/**
  Implements a Bank Of Tuners.

  @author Jeff Schoen
  @version $Id: BoT.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME BoT
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 nfft;
  int_4 nout;
  int_4 nppt;
  int_4 nrpt;
  int_4 gain;
  int_4 flags;
  double foff;
  double ratio;
} BoTPlan;

#include "CoreProtos.h"

#define flagsList "Real,Pkt"
#define REAL 0x1
#define PKT  0x2

/* CORE Code */

void* BoT_alloc (char *config) {
  BoTPlan *plan = BoT_plan();
  HW_open(config);
  plan->nfft = 16;
  plan->nout = 16;
  plan->nppt = 4;
  plan->nrpt = 16;
  plan->foff = 0;
  plan->gain = 0;
  plan->flags = 0;
  return (void*)plan;
}

int_4 BoT_setkey (BoTPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
  if      (keyMatch("L:NFFT")) { plan->nfft  = ldata[0]; }
  else if (keyMatch("L:NOUT")) { plan->nout  = ldata[0]; }
  else if (keyMatch("L:NPPT")) { plan->nppt  = ldata[0]; }
  else if (keyMatch("L:NRPT")) { plan->nrpt  = ldata[0]; }
  else if (keyMatch("L:GAIN")) { plan->gain  = ldata[0]; }
  else if (keyMatch("L:FLAGS")){ plan->flags = ldata[0]; }
  else if (keyMatch("D:FOFF")) { plan->foff  = *(double *)data; }
  else if (keyMatch("D:RATIO")){ plan->ratio = *(double *)data; }
  else if (keyMatch("F:FPPF")) { HW_write(COREG_USER1,plan->nppt-1); 
				 HW_loadPPF(COREG_USER2,plan->nppt,len/4,(real_4*)data); }
  else if (keyMatch("F:RFIR")) { HW_loadTaps(COREG_USER3,len/4,(real_4*)data); }
  else return 0;
  return len;
}

int_4 BoT_getkey (BoTPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:NFFT"))  { *(int_4*)data  = plan->nfft; return 4; }
  if (keyMatch("L:NOUT"))  { *(int_4*)data  = plan->nout; return 4; }
  if (keyMatch("L:NPPT"))  { *(int_4*)data  = plan->nppt; return 4; }
  if (keyMatch("L:NRPT"))  { *(int_4*)data  = plan->nrpt; return 4; }
  if (keyMatch("L:GAIN"))  { *(int_4*)data  = plan->gain; return 4; }
  if (keyMatch("D:FOFF"))  { *(double*)data = plan->foff; return 8; }
  if (keyMatch("D:RATIO")) { *(double*)data = plan->ratio; return 8; }
  if (keyMatch("L:FLAGS")) { *(int_4*)data  = plan->flags; return 4; }
  if (keyMatch("S:FLAGSLIST")) { strcpy(data,flagsList); return strlen(flagsList); }
  return 0;
}

int_4 BoT_init (BoTPlan *plan) {
  int_4 pow2s=pow2sizem1(plan->nfft), pow2=pow2s>>16, oreal=(plan->flags&REAL);
#ifdef CORE_IS_FPGA
  HW_write(COREG_SYS,0x00008800);	
  HW_write(COREG_DEC,plan->nout-1);	
  HW_write(COREG_GAIN,fftscale(plan->gain,pow2)); 
  HW_write(COREG_RATIO,R2L((oreal?0.5:1.0)/plan->ratio));	
  HW_write(COREG_FRAME,pow2s);
  HW_write(COREG_FREQ,R2L(-2*plan->foff-1.0/plan->nfft));	// odd bins
  HW_write(COREG_FREQ,R2L(-2*plan->foff));			// even bins
  HW_write(COREG_FLAG,plan->flags);	
  HW_write(COREG_SYS,0x00008801);	
#endif
  return 0;
}

int_4 BoT_work (BoTPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#ifdef CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 BoT_free (BoTPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}
