/**
  Implements Envelope Measurement and Track functions 

  @author Jeff Schoen
  @version $Id: EMT.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME EMT
#define CORE_AREA nxm_ice_core

/* CORE defines */
#include "CoreDefs.h"

/* CORE Plan handle */
typedef struct {
  HALO halo;
  int_4 nmax;
  int_4 nexp;
} EMTPlan;

#include "CoreProtos.h"

/* CORE Code */

void* EMT_alloc (char *config) {
  EMTPlan *plan = EMT_plan();
  plan->nmax = 1;
  plan->nexp = 1;
  HW_open(config);
  return (void*)plan;
}

int_4 EMT_setkey (EMTPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
       if (keyMatch("L:NMAX")) { plan->nmax  = ldata[0]; HW_write(COREG_DEC,plan->nmax-1); }
  else if (keyMatch("L:NEXP")) { plan->nexp  = ldata[0]; HW_write(COREG_FRAME,plan->nexp-1); }
  else return 0;
  return len;
}

int_4 EMT_getkey (EMTPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
  if (keyMatch("L:NMAX")) { ldata[0] = plan->nmax;  return 4; }
  if (keyMatch("L:NEXP")) { ldata[0] = plan->nexp;  return 4; }
  return 0;
}

int_4 EMT_init (EMTPlan *plan) {
#if CORE_IS_FPGA
  int_4 sys = getFmts(FMTS);
  HW_write(COREG_SYS,sys);
  HW_write(COREG_DEC,plan->nmax-1);
  HW_write(COREG_RATE,pow2sizem1(plan->nexp));
  HW_write(COREG_SYS,sys|CORE_ENA); 
#endif
  return 0;
}

int_4 EMT_work (EMTPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#if CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 EMT_free (EMTPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

