package nxm.ice.core;

import nxm.ice.lib.*;
import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;

/**
  Implements an Envelope Measurement and Track algorithm with Clip/Blank control.
  Uses a sliding data max with exponential averaging.

  @author Jeff Schoen
  @version $Id: EMT.java,v 1.8 2011/05/24 13:21:13 jgs Exp $
*/
public class EMT extends CoreFactory {

// native implementation wrappers
 public static class VHS extends CoreNative { public native long alloc(); }
 public static class ICE extends CoreNative { public native long alloc(); }

 // java implementation
 public static class JVM extends Core {

  private int nmax=1,nexp=1,ibps,obps;
  private float value;

  /** set parameters */
  public void set (String key, Data value) {
         if (key.equals("NMAX")) nmax  = value.toL();
    else if (key.equals("NEXP")) nexp  = value.toL();
    else super.set(key,value);
  }

  /** get parameters */
  public Data get (String key, byte type) {
    if (key.equals("NMAX"))  return new Data(nmax);
    if (key.equals("NEXP"))  return new Data(nexp);
    return super.get(key,type);
  }

  /** Ready this engine with current parameters */
  public int open() {
    ibps = getFormatBytes(1);
    obps = getFormatBytes(2);
    return 0;
  }

  /** process a buffer of data */
  public int process (byte[] a, int na, byte[] b, int nb) {
    int i,j,n=0,n1=na/ibps;
    float p0=(float)(1.0/nexp), p1=(float)(1.0-p0);
    if (ibps==1) {
      byte vm=0;
      for (i=j=0; i<n1; i++) {
        byte v = a[i]; if (v<0) v=(byte)(-v);
	if (j==0) vm=v; else if (v>vm) vm=v;
        if (++j<nmax) continue; else j=0;
        value = value*p1 + vm*p0;
        b[n++] = (byte)value;
      }
    }
    else if (ibps==2) {
      short[] ai = Convert.castI(a,true);
      short[] bi = Convert.castI(b,false);
      short vm=0,vv;
      for (i=j=0; i<n1; i++) {
        short v = (short)Math.abs(ai[i]);
	if (j==0) vm=v; else if (v>vm) vm=v;
        if (++j<nmax) continue; else j=0;
        value = value*p1 + vm*p0;
        bi[n++] = (short)value;
      }
      Convert.uncast(ai,a,false);
      Convert.uncast(bi,b,true);
    }
    else if (ibps==4) {
      short[] ai = Convert.castI(a,true);
      short[] bi = Convert.castI(b,false);
      float vm=0;
      for (i=j=0; i<n1; i++) {
        float vr = (float)ai[i+i], vi = (float)ai[i+i+1]; 
        float v = Math.max(Math.abs(vr),Math.abs(vi));
	if (j==0) vm=v; else if (v>vm) vm=v;
        if (++j<nmax) continue; else j=0;
        value = value*p1 + vm*p0;
        bi[n++] = (short)value;
      }
      Convert.uncast(ai,a,false);
      Convert.uncast(bi,b,true);
    }
    return n*obps;
  }

  }
}
