/**
  Implements a FIR Filter

  @author Jeff Schoen
  @version $Id: FIR.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Define Signatures */
#define CORE_NAME FIR
#define CORE_AREA nxm_ice_core

/* Generic defines */
#include "CoreDefs.h"

typedef struct {
  HALO halo;
  int_4 taps;
  int_4 order;
  int_4 dec;
  int_4 gain;
} FIRPlan;

#include "CoreProtos.h"

/* CORE Code */

void* FIR_alloc (char *config) {
  FIRPlan *plan = FIR_plan();
  HW_open(config);
  plan->dec = 1;
  plan->gain = 0;
  return (void*)plan;
}

int_4 FIR_setkey (FIRPlan *plan, char *key, void *data, int_4 len) {
  int_4 *ldata = (int_4*)data;
       if (keyMatch("L:DECIMATION")) { plan->dec = ldata[0]; }
  else if (keyMatch("L:ORDER"))      { plan->order = ldata[0]; }
#ifdef CORE_IS_FPGA
  else if (keyMatch("F:TAPS"))       { plan->taps = len/4; HW_loadFIR(COREG_USER4,plan->order,plan->dec,plan->taps,(real_4*)data); }
  else if (keyMatch("L:GAIN"))       { plan->gain = ldata[0]; HW_write(COREG_USER1,(int_4)(0x1000*pow(2.0,plan->gain/6))); }
#endif
  else return 0;
  return len;
}

int_4 FIR_getkey (FIRPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:DECIMATION")) { *(int_4*)data = plan->dec; return 4; }
  if (keyMatch("L:GAIN"))       { *(int_4*)data = plan->gain; return 4; }
  if (keyMatch("L:ORDER"))      { *(int_4*)data = plan->order; return 4; }
  return 0;
}

int_4 FIR_init (FIRPlan *plan) {
#ifdef CORE_IS_FPGA
  HW_write(COREG_SYS,CORE_ENA);	// enable
#endif
  return 0;
}

int_4 FIR_work (FIRPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#ifdef CORE_IS_FPGA
  n = HW_proc (buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 FIR_free (FIRPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

