package nxm.ice.core;

import nxm.ice.lib.*;
import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;

/**
  Implements data copy with optional decimation and reformat.

  @author Jeff Schoen
  @version $Id: Noop.java,v 1.8 2011/05/24 13:21:13 jgs Exp $
*/
public class Noop extends CoreFactory {

 // native implementation wrappers
 public static class CPU extends CoreNative { public native long alloc(); }
 public static class VHS extends CoreNative { public native long alloc(); }
 public static class ICE extends CoreNative { public native long alloc(); }

 // java implementation
 public static class JVM extends Core {

  private int dec,ioff,ibps,obps;
  private Task.Stream tsi,tso;
  private Task.Decimate tdec;

  /** set parameters */
  public void set (String key, Data value) {
    if (key.equals("DECIMATION")) dec  = value.toL();
    else super.set(key,value);
  }

  /** get parameters */
  public Data get (String key, byte type) {
    if (key.equals("DECIMATION"))  return new Data(dec);
    else return super.get(key,type);
  }

  /** Ready this engine with current parameters */
  public int open() {
    ioff = 0;
    ibps = getFormatBytes(1);
    obps = getFormatBytes(2);
    tsi  = new Task.Stream();
    tdec = new Task.Decimate(tsi,dec);
    tso  = new Task.Stream(tdec);
    return 0;
  }

  /** process stream approach */
  public int process () {
    tsi.set(bi1,ni1);
    tso.set(bo1,no1);
    tsi.push();
    return tso.length();
  }

  /** process non-stream approach */
  public int processX () {
    int n=0,i=ioff,n1=ni1/ibps;
    if (ibps==1) {
      byte[] a=bi1, b=bo1;
      for (; i<n1; i+=dec) b[n++]=a[i];
    }
    else if (ibps==2) {
      short[] ai = Convert.castI(bi1,true);
      short[] bi = Convert.castI(bo1,false);
      for (; i<n1; i+=dec) bi[n++]=ai[i];
      Convert.uncast(ai,bi1,false);
      Convert.uncast(bi,bo1,true);
    }
    else if (ibps==4) {
      int[] ai = Convert.castL(bi1,true);
      int[] bi = Convert.castL(bo1,false);
      for (; i<n1; i+=dec) bi[n++]=ai[i];
      Convert.uncast(ai,bi1,false);
      Convert.uncast(bi,bo1,true);
    }
    ioff = i%n1;
    return n*obps;
  }

 }
}
