/**
  Implements Real to Complex conversion

  @author Jeff Schoen
  @version $Id: PCIe.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Signatures */
#define CORE_NAME PCIe
#define CORE_AREA nxm_ice_core

#include "CoreDefs.h"

/* CORE Plan handle */
typedef struct {
  HALO halo;
  int_4 lanes;
  int_4 flags;
} PCIePlan;

#include "CoreProtos.h"

/* CORE Code */

void* PCIe_alloc (char *config) {
  PCIePlan *plan = PCIe_plan();
  HW_open(config);
  plan->flags = 0;
  plan->lanes = 8;
  return (void*)plan;
}

int_4 PCIe_setkey (PCIePlan *plan, char *key, void *data, int_4 len) {
       if (keyMatch("L:FLAGS")) { plan->flags = *(int_4 *)data; }
  else if (keyMatch("L:LANES")) { plan->lanes = *(int_4 *)data; }
  else return 0;
  return len;
}

int_4 PCIe_getkey (PCIePlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:FLAGS")) { *(int_4*)data = plan->flags; return 4; }
  if (keyMatch("L:LANES")) { *(int_4*)data = plan->lanes; return 4; }
  return 0;
}

int_4 PCIe_init (PCIePlan *plan) {
  int_4 sysreg = 0x0001;
  if (plan->lanes==8) sysreg |= (0x3<<8);
  if (plan->lanes==4) sysreg |= (0x2<<8);
  if (plan->lanes==2) sysreg |= (0x1<<8);
  HW_write(COREG_FLAG,plan->flags);
  HW_write(COREG_SYS,sysreg);	// enable PCIe with specified lanes
  return 0;
}

int_4 PCIe_work (PCIePlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
  n = HW_procw(buf1,n1, buf2,n2);
  return n;
}

int_4 PCIe_free (PCIePlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

