/**
  Implements Real to Complex conversion

  @author Jeff Schoen
  @version $Id: R2C.c,v 1.3 2005/03/29 14:56:00 jgs Exp $
*/

/* Signatures */
#define CORE_NAME R2C
#define CORE_AREA nxm_ice_core

#include "CoreDefs.h"

/* CORE Plan handle */
typedef struct {
  HALO halo;
  int_4 state;
  int_4 flags;
  double foff;
} R2CPlan;

#include "CoreProtos.h"

/* CORE Code */

void* R2C_alloc (char *config) {
  R2CPlan *plan = R2C_plan();
  HW_open(config);
  plan->state = 0;
  plan->flags = 0;
  plan->foff = 0;
  return (void*)plan;
}

int_4 R2C_setkey (R2CPlan *plan, char *key, void *data, int_4 len) {
       if (keyMatch("L:FLAGS"))     { plan->flags = *(int_4 *)data; }
  else if (keyMatch("D:FREQUENCY")) { plan->foff = *(double *)data; }
  else return 0;
  return len;
}

int_4 R2C_getkey (R2CPlan *plan, char *key, void *data, int_4 len) {
  if (keyMatch("L:FLAGS"))      { *(int_4*)data = plan->flags; return 4; }
  if (keyMatch("D:FREQUENCY"))  { *(double*)data = plan->foff; return 8; }
  return 0;
}

int_4 R2C_init (R2CPlan *plan) {
#ifdef CORE_IS_FPGA
  HW_write(COREG_FLAG,plan->flags);
  HW_write(COREG_FREQ,R2L(plan->foff));
  HW_write(COREG_SYS,0x00008001);	// enable R2C
#endif
  return 0;
}

int_4 R2C_work (R2CPlan *plan) {
  int_4 n=0;
  VBUFI(buf1,n1);
  VBUFO(buf2,n2);
#ifdef CORE_IS_FPGA
  n = HW_proc(buf1,n1, buf2,n2);
#endif
  return n;
}

int_4 R2C_free (R2CPlan *plan) {
  HW_close();
  free(plan);
  return 0;
}

