//************
//  Description: Generic midas driver for an ICECORE library
//
//  Author:	Jeff Schoen	3/13
//
//  $Revision:	$
//  $Modtime:	$
//
//************
#include <stdio.h>
#include <primitive.h>
#include <fintrinsics.h>
#include "corelib.h"
#define MARGS 64

void mainroutine () 
{

  CPHEADER hi, ho;

  int_4 i,n,id,lval,xfer,ngot,nalloc,stat,np,dec=1;
  int_4 *buf,wid[MARGS];
  real_8 dval,wdval[MARGS];
  string iform,oform,wtkey[MARGS];

  // open input file
  m_init (hi, m_apick(1), "1000,2000","S#,C#", 0);
  m_open (hi, HCBF_INPUT+HCBF_OPTIONAL);

  // create an instance of the core by name
  CORE *core = core_alloc (m_apick(3).c_str());

  // determine formats
  i = m_get_uswitch("IFORM",iform);
  if (i<0) iform = hi.format;
  i = m_get_uswitch("OFORM",oform);
  if (i<0) oform = iform;
  core_setfmt(core,iform.c_str(),oform.c_str());

  // parse and set the parameters 
  for (np=0; np<MARGS && m_ppick(np+4); np++) {
    string arg = m_upick(np+4);
    if ((i=arg.find('='))<0) m_error("All extra parameters must be key=value not "+arg);
    string key=arg.substr(0,i), val=arg.substr(i+1), kt="L";
    if (key[1]==':') { kt=key[0]; key=key.substr(2); }
    string tkey = kt+":"+key;
    if (kt=="D") {
      m_a2d(val.c_str(),dval); id=m_dwinit(key,dval,1.0,-1.0,1.0);
      if (core_setkey(core,tkey.c_str(),(void*)&dval,8)!=8) m_error("Setting "+arg);
    }
    else if (kt=="L") {
      m_a2d(val.c_str(),dval); lval=dval; id=m_lwinit(key,lval,1,-1,1);
      if (core_setkey(core,tkey.c_str(),(void*)&lval,4)!=4) m_error("Setting "+arg);
      if (key.substr(0,3)=="DEC") dec = max(1,lval); // special param
    } 
    else {
      m_error ("Unsupported parameter type="+kt);
    }
    wid  [np] = id;
    wtkey[np] = tkey;
    wdval[np] = dval;
  }

  // open the core
  stat = core_open (core);

  // override decimation 
  dec = m_get_switch_def("ODEC",dec);

  // open output file
  m_propagate (hi, ho);
  ho.file_name = m_apick(2);
  ho.xdelta = hi.xdelta*dec;
  m_open (ho, HCBF_OUTPUT+HCBF_OPTIONAL);

  // prepare the data buffer
  xfer = m_get_switch_def("TL",4096);
  hi.xfer_len = xfer;
  nalloc = xfer*max(hi.bpa,ho.bpa);
  buf = (int_4*) malloc(nalloc);

  m_sync();

  while (!Mc->break_) {

    // check for real-time parameter changes
    for (n=0; n<np; n++) {
      id = wid[n];
      if (wtkey[n][0]=='D') {
        dval = wdval[n];
        if (m_dwget(id,dval)>0) core_setkey (core,wtkey[n].c_str(),(void*)&dval,8);
        wdval[n] = dval;
      } else {
        lval = wdval[n];
        if (m_lwget(id,lval)>0) core_setkey (core,wtkey[n].c_str(),(void*)&lval,4);
        wdval[n] = lval;
      }
    }

    // input data 
    ngot = 0;
    if (hi.open) {
      m_grabx (hi,buf,ngot);
      if (ngot<=0) break;
    } 

    // process and output data
    if (ngot>0) {
      stat = core_process (core, (int_1*)buf,ngot*hi.bpa, (int_1*)buf,nalloc);
      m_filad (ho,buf,stat/ho.bpa);
    } else {
      m_pause(Mc->pause);
    }

  }

  // close all resources
  stat = core_close (core);
  m_close(hi);
  m_close(ho);
  free(buf);

}
