c************
c  Description:	Generic midas driver for an ICECORE library
c
c  Author:	Jeff Schoen	2/2013
c
c  $Revision:	$
c  $Modtime:	$
c
c************
	program icecore

	implicit none
	include 'primitive.inc'
	include 'corelib.inc'
	record /HEADER/ hi, ho
	integer*4 i,n,np,id,lval,ngot,xfer,nalloc,stat,dec
	integer*4 buf(1),core(1),wid(64)
	real*8    dval,wdval(64)
	character iform*2,oform*2,styp*8,kt,wtype(64)
	character*40 arg,val,sdev
	character*16 func,key,wname(64)
	pointer (pbuf,buf),(pcore,core)

	call M$CHECKIN

	! open input file
	call M$INIT (hi, M$APICK(1), '1000,2000','S#,C#', 0)
	call M$OPEN (hi, HCBF_INPUT+HCBF_OPTIONAL)

	! create an instance of the core by name
	func  = M$APICK(3)
	pcore = CORE_ALLOC (func)

	! determine formats
	i = M$GET_USWITCH('IFORM',iform)
	if (i.le.0) iform = hi.format
	i = M$GET_USWITCH('OFORM',oform)
	if (i.le.0) oform = iform
	i = CORE_SETFMT(core,iform,oform)

	! parse and set the parameters 
	dec = 1
	do n = 1,64
	  arg = M$UPICK(n+3)
	  if (arg.eq.' ') goto 9
	  i = INDEX(arg,'=')
	  if (i.eq.0) call M$ERROR('All extra parameters must be key=value not '//arg)
	  key = arg(1:i-1)
	  val = arg(i+1:)
	  if (key(2:2).eq.':') then
	    kt = key(1:1) 
	    key = key(3:)
	  else
	    kt = 'L'
	  endif
	  if (kt.eq.'D') then
	    stat = M$A2D(val,dval)
	    id = M$DWINIT(key,dval,1.0,-1.0,1.0)
	    if (CORE_SETKEY(core,'D:'//key,dval,8).ne.8) call M$ERROR('Setting '//arg)
	  elseif (kt.eq.'L') then
	    stat = M$A2D(val,dval)
	    lval = NINT(dval)
	    id = M$LWINIT(key,lval,1,-1,1)
	    if (CORE_SETKEY(core,'L:'//key,lval,4).ne.4) call M$ERROR('Setting '//arg)
	    if (key(1:3).eq.'DEC') dec = max(1,lval)
	  else
	    call M$ERROR ('Unsupported parameter type='//kt)
	  endif
	  wid(n)   = id
	  wname(n) = key
	  wtype(n) = kt
	  wdval(n) = dval
	enddo
 9	np = n-1

	! open the core
	stat = CORE_OPEN (core)

	! override decimation
	dec = M$GET_SWITCH_DEF('ODEC',dec)

	! open output file
	call M$PROPAGATE (hi, ho)
	ho.file_name = M$APICK(2)
	ho.format = oform
	ho.xdelta = hi.xdelta*dec
	call M$OPEN (ho, HCBF_OUTPUT+HCBF_OPTIONAL)

	! prepare the data buffer
	xfer = M$GET_SWITCH('TL')
	if (xfer.lt.0) xfer = 4096
	hi.xfer_len = xfer
	nalloc = xfer*max(hi.bpa,ho.bpa)
	call M$MALLOC(nalloc,pbuf)

	call M$SYNC

	do while (.not.Mc.break)

	  ! check for real-time parameter changes
	  do n = 1,np
	    if (wtype(n).eq.'D') then
	      dval = wdval(n)
	      i = M$DWGET(wid(n),dval)
	      if (i.gt.0) stat = CORE_SETKEY (core,'D:'//wname(n),dval,8)
	      wdval(n) = dval
	    else
	      lval = wdval(n)
	      i = M$LWGET(wid(n),lval)
	      if (i.gt.0) stat = CORE_SETKEY (core,'L:'//wname(n),lval,4)
	      wdval(n) = lval
	    endif
	  enddo

	  ! input data or pause
	  ngot = 0
	  if (hi.open) then
	    call M$GRABX (hi,buf,ngot)
	    if (ngot.le.0) goto 99
	  else
	    call M$PAUSE(Mc.pause)
	  endif

	  ! process and output data
	  if (ngot.gt.0) then
	    stat = CORE_PROCESS (core, buf,ngot*hi.bpa, buf,nalloc)
	    call M$FILAD (ho,buf,stat/ho.bpa)
	  endif

	enddo

 99	continue
	! close all resources
	stat = CORE_CLOSE (core)
	call M$MFREE(nalloc,pbuf)
	call M$CLOSE(hi)
	call M$CLOSE(ho)

 	call M$RETURN

	end

