/**
  Common prototypes for processing COREs

  CORE_AREA is the java native package location
  CORE_NAME is the Library Function Name
  CORE_TYPE is the Processor Type Java,CPU,GPU,MLS,VHS,ICE

  @author Jeff Schoen
  @version $Id: CoreProtos.h,v 1.1 2011/04/27 13:09:47 jgs Exp $
*/

#ifdef CORE_AREA
#define CEXPORT static
#else
#define CEXPORT
#endif

#ifdef CORE_NAME
#define BHANDLE(A) A##Plan
#define CHANDLE(A) BHANDLE(A)
#define BMETHOD(A,B) A##_##B
#define CMETHOD(A,B) BMETHOD(A,B)
#define BGLOBAL(A,B) Core##B##_##A##_alloc
#define CGLOBAL(A,B) BGLOBAL(A,B)

CEXPORT void* CMETHOD(CORE_NAME,alloc)  (char* config);
CEXPORT int_4 CMETHOD(CORE_NAME,setkey) (CHANDLE(CORE_NAME)*handle, char* key, void* data, int_4 len);
CEXPORT int_4 CMETHOD(CORE_NAME,getkey) (CHANDLE(CORE_NAME)*handle, char* key, void* data, int_4 len);
CEXPORT int_4 CMETHOD(CORE_NAME,init)   (CHANDLE(CORE_NAME)*handle);
CEXPORT int_4 CMETHOD(CORE_NAME,work)   (CHANDLE(CORE_NAME)*handle);
CEXPORT int_4 CMETHOD(CORE_NAME,free)   (CHANDLE(CORE_NAME)*handle);

// sets configuration values
CEXPORT int_4 CMETHOD(CORE_NAME,config)(CHANDLE(CORE_NAME)*handle, int_4 key, int_4 value) {
  HALO *pt = (HALO*)handle;
  pt->fmts = value;
  return 0;
}

// allocates and initializes HALO
CEXPORT CHANDLE(CORE_NAME)* CMETHOD(CORE_NAME,plan) () {
  int_4 size = sizeof(CHANDLE(CORE_NAME));
  HALO *pt = (HALO*)malloc(size);
  memset (pt,0,size);
  pt->init   = &CMETHOD(CORE_NAME,init);
  pt->config = &CMETHOD(CORE_NAME,config);
  pt->setkey = &CMETHOD(CORE_NAME,setkey);
  pt->getkey = &CMETHOD(CORE_NAME,getkey);
  pt->work   = &CMETHOD(CORE_NAME,work);
  pt->free   = &CMETHOD(CORE_NAME,free);
  pt->psize  = size - sizeof(HALO);
  return (CHANDLE(CORE_NAME)*)pt;
}
void* CGLOBAL(CORE_NAME,CORE_SIG) (char* config) {
  return CMETHOD(CORE_NAME,alloc)(config);
}
#endif

#ifdef CORE_AREA
#define IGLOBAL(A,B,C) Java_##A##_##B##_00024##C##_alloc
#define JGLOBAL(A,B,C) IGLOBAL(A,B,C)
#ifndef NONEXTMIDAS
#include "native.h"
JNIEXPORT jlong JNICALL JGLOBAL(CORE_AREA,CORE_NAME,CORE_SIG) (JNIEnv *env, jobject obj) {
  jobject config =  GetStringObject(obj,"config");
  char *cconfig = (config==NULL)? 0 : (char *)GetString(config);
  jpointer handle = (jpointer) CMETHOD(CORE_NAME,alloc)((char *)cconfig);
  if (config!=NULL) ReleaseString(config,cconfig);
  return (jlong)handle;
}
#endif
#endif

