/**
  Description:	Header file for data packet headers 

  Author:	Jeff Schoen

  $Revision:	$
  $Modtime:	$

*/

typedef struct {
  real_8 start,delta;
} ICEPKTX;

typedef struct {
  int_4 ramphys,ramsize,dataoffset,datasize;
} ICEPKTR;

typedef struct {
  int_1	 keys[4];	/* barker syncs[2:0] and packet type[3] */
  int_4	 count;		/* packet number since start of xfer */
  int_4	 elem;		/* data elements in packet */
  int_2	 user;		/* user defined field */
  int_2	 channel;	/* port channel number */
  int_1	 id1,id2;	/* subsystem IDs */
  int_1	 rep;		/* data rep (Vax,Ieee,Eeei) */
  int_1	 bpa;		/* bytes per atom=(mode,type) */
  int_1	 mode;		/* data format mode (Complex, Scalar) */
  int_1	 type;		/* data format type (Pack,Byte,Int,Long) */
  int_1	 tcmode;	/* timecode mode */
  int_1	 tcstatus;	/* timecode status */
  double tcoff;		/* sample offset from data[0] to TC */
  double tcwsec;	/* timecode 1950 whole seconds */
  double tcfsec;	/* timecode fractional seconds */
 union {
  real_8 dval[2];	/* user fields */
  real_4 fval[4];	/* user fields */
  int_4	 lval[4];	/* user fields */
  int_2	 ival[8];	/* user fields */
  int_1	 bval[16];	/* user fields */
  ICEPKTX X;
  ICEPKTR R;
 } u;
} ICEPACKETSTRUCT;

#define ICE_HDR_SZ 64

typedef struct {
  int_1 func; 		/* defines the method to call */
  int_1 flag; 		/* is the protocol version number and options mask */
  int_1 info; 		/* channel number for fast routing of monitored properties */
  int_1 rep;  		/* packet data representation (VAX|IEEE|EEEI) */
  int_1 seq;  		/* sequence number (0-127) for Reliable packets */
  int_1 try_;  		/* number of resends for Reliable packets */
  int_1 rpt;  		/* report sequence number */
  int_1 adj;  		/* number of bytes in adjunct header */
} RMIFPACKETSTRUCT;

/* 
  Network overhead per packet
  12by - dead time between packets
   8by - ethernet preamble (0x1FD,0x55,0x55,0x55,0x55,0x55,0x55,0x5D)
  14by - ethernet header
  20by - IP header
   8by - UDP header
  56by - SDDS header
  1Kby - SDDS data
   4by - ethernet trailer FCS

  Total of 110by/1024by = 12% overhead
  Max 1G throughput = 125Mby / 1.12 = 111Mby/sec data rate
  Max 10G throughput = 1250Mby / 1.12 = 1110Mby/sec data rate
*/

typedef struct {
  int_2 keys;		/* barker packet sync 0x55AA */
  int_2 port;		/* UDP source|destination port (0|29495 or 0x0|0x7337) */
  int_4 addr;		/* UDP source|destination address */
			/* now starts the actual SDDS UDP payload */
  int_1 fid;		/* format identifier */
  int_1 bits;		/* bits per data word 4|8|16 */
  int_2	frame;		/* frame sequence number (0x1F=parity packet) */
  int_2 tcinfo;		/* Time Tag info {1msV,ttV,sscV,00,1msPtr[10:0]} */
  int_2	tccorr;		/* Time from sample to 1msec hack in 250psec tics (min SR=60kHz) */ 
  int_8	tctics;		/* Number of 250psec tics since jan 1 to timecode hack */
  int_4	tcext;		/* Timecode extended precision (250psec/2^32) */
  int_1	sscinfo[12];	/* SSC information */
  int_1	ssdinfo[4];	/* SSD information | IP address */
  int_1	aadinfo[20];	/* AAD information */
  int_1 data[1024];	/* packet data */
} SDDSPACKETSTRUCT;

#define SDDS_HDR_SZ 56
#define SDDS_PKT_SZ 1080

/* 
  Network overhead per packet
  12by - dead time between packets
   8by - ethernet preamble (0x1FD,0x55,0x55,0x55,0x55,0x55,0x55,0x5D)
  14by - ethernet header
  20by - IP header
   8by - UDP header
  28by - VRT header
1440by - VRT data
   4by - VRT trailer
   4by - ethernet trailer FCS

  Total of 86by/1440by = 6% overhead
  Max 1G throughput     1.250 (8B10B) = 1G/8 = 125Mby / 1.06 = 118Mby/sec data rate
  Max 10G throughput 10.3125 (64B66B) = 10G/8 = 1250Mby / 1.06 = 1180Mby/sec data rate 
*/

typedef struct {
  int_4 header;		/* Pkt type, fields, size */
  int_4 streamID;	/* Stream Identifier */
  int_4 orgID;		/* Organizational Unique ID */
  int_4 classID;	/* Information:Packet Class ID */
  int_4 tsis;		/* Time Stamp Integer Seconds */
  int_8 tsfs;		/* Time Stamp Fractional Seconds */
  int_1 data[1440];	/* Packet data */
  int_4 trailer;	/* Event trailer */
} VRTPACKETSTRUCT;

#define VRT_HDR_SZ 32
#define VRT_PKT_SZ 1472

typedef struct {
  int_4 header;		/* Pkt type, fields, size */
  int_4 streamID;	/* Stream Identifier */
  int_4 orgID;		/* Organizational Unique ID */
  int_4 classID;	/* Information:Packet Class ID */
  int_4 tsis;		/* Time Stamp Integer Seconds */
  int_8 tsfs;		/* Time Stamp Fractional Seconds */
  int_4 trailer;	/* Event trailer */
  int_4 bits;		/* number of bits for internal use */
} VRTPKTHDRSTRUCT;

typedef struct {
  int_1 destAddr[6];	/* MAC address of dest NIC 01-00-5E-XX-XX-XX */
  int_1 srcAddr[6];	/* MAC address of source NIC 01-00-5E-XX-XX-XX */
  int_2 type;		/* 0x0800 IP */
} EHDRSTRUCT;

typedef struct {
  int_1 versionLength;	/* 0x45   IP version = 4, header length in 32b words = 5 */
  int_1 typeOfService;	/* 0x00   precedence, throughput, reliability */
  int_2 totalLength;	/* 0x0454 for SDDS packet + IPhdr + UDPhdr  or 1080 + 20 + 8 = 1108 */
  int_2 identification;	/* 0x0000 unique identifier */
  int_2 fragmentOffset;	/* 0x4000 do not fragment bit + frag offset = 0 */
  int_1 timeToLive;	/* 32     max hops before discard */
  int_1 protocol;	/* transport layer protocol 1=ICMP 2=IGMPv2 6=TCP 8=EGP 17=UDP 89=OSPF */
  int_2 headerChecksum;	/* calculate from IP header only */
  int_1 srcAddr[4];	/* IP addr in range 224.0.0.1 to 239.255.255.255 */
  int_1 destAddr[4];	/* IP addr in range 224.0.0.1 to 239.255.255.255 */
} IPHDRSTRUCT;

typedef struct {
  int_2 srcPort;	/* source port = 0 */
  int_2 destPort;	/* destination port = 29495 or 0x7337 */
  int_2 length;		/* length of UDP datagram = 1088 or 0x440 */
  int_2 checksum;	/* 0 = not computed */
} UDPHDRSTRUCT;

/*
  Processor/IO Module Hypertransport packet
*/
typedef struct {
  int_1 key; 		/* Start Key = 0xA5 */
  int_1 cmd; 		/* Command = 0x00 for data */
  int_1 len; 		/* Data len in number of 64by packets */
  int_1 chn;  		/* Channel Number */
} HYPERPACKETSTRUCT;

/*
  File Exchange Packet
*/
typedef struct {
  int_4 key;		/* Header Sync word */
  int_4 state;		/* Current state */
  int_4 offset;		/* Current offset */
  int_4 avail;		/* available buffer space */
  int_4 length;		/* available buffer space */
  int_4 sequence;	/* packet sequence number */
  double time;		/* time of last transmission */
  char fromfile[128];	/* filename path root */
  char tofile[128];	/* filename path root */
  char chksum[128];	/* file checksum */
  char data[4096-384-4*12];	/* fill a 4K page */
} FEXPACKETSTRUCT;
