/***************
*
*  tiolib.h: TBD
*  
*  Revisions:
*  tbd  Initial Release

*
****************/

#ifndef _TIOLIB_H_
#define _TIOLIB_H_

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#define TIO_MAGIC 0xf8b1c0ad
#define QSIZE 2
#define IODIRN -1
#define IODIRW 0
#define IODIRR 1
#define MEG 1048576

/* disk map struct */
typedef struct {
  off_t off;
  char path[FILENAME_MAX];
} s_diskmap;

/* file system struct */
typedef struct {
  off_t tsize, off, extend;
  int ssize, ndisks, nthreads;
  char type[10];
  char iceloc[FILENAME_MAX];
  char fuseloc[FILENAME_MAX];
  char dmfile[FILENAME_MAX];
  char dev[FILENAME_MAX];
  s_diskmap *disks;
} s_fsprops;

/* queue struct */
typedef struct {
  void *buf;
  off_t rsize;
  pthread_mutex_t rlock;
  pthread_cond_t rcond;
  pthread_mutex_t wlock;
  pthread_cond_t wcond;
} s_q;

/* device struct */
typedef struct {
  FILE *fd;
  pthread_mutex_t dlock;
  off_t seek;
  int dnum;
  s_q q[QSIZE];
  int rind;
  int wind;
  int last_read;
  int extra_read;
  off_t off;
} s_dev;

/* thread struct */
typedef struct {
  s_dev *devs;
  void *buf;
  int num_reads;
  int tot_reads;
  off_t tot_tot;
  off_t tot_proc;
  int ndevs;
  int tnum;
  int RFEOF;
  int WFEOF;
  int stripe_size;
  pthread_t tid;
} s_t;

/* global struct */
typedef struct s_tio {
  s_t *threads;
  s_fsprops *fs;
  int init; 
  int dir;
  int magic;
  int dev_beg;
} s_tio;

/* private functions */
int join_unlock ( void* );
int mapReads (void*, off_t );
void *read_disk ( void* );
void *read_buff ( void* );
void *write_disk ( void* );
void *write_buff ( void* );
void perform_read_disk ( void*, off_t );
void perform_read_buff ( void*, void*, off_t );
void perform_write_disk ( void* );
void perform_write_buff ( void*, void* );
int launch_write_threads ( void* );

/* public functions */
void *tio_malloc ( int );
int tio_free ( void* );
void *tio_open ();
int tio_set_config ( void*, char* );
int tio_set_dir ( void*, int );
int tio_init ( void* );
int tio_launch_read_threads ( void*, off_t );
int tio_join_threads ( void* );
int tio_wait_threads ( void* );
int tio_seek ( void*, off_t );
int tio_set_offset ( void*, off_t );
int tio_read ( void*, void*, off_t );
int tio_write ( void*, void*, off_t );
int tio_close ( void* );
off_t tio_write_det ( void*, char*, off_t );
off_t tio_update_det ( void*, char*, off_t );
off_t tio_finalize_det ( void*, char*, off_t );

#endif
