/**
  Template Java/C wrapper code for an CoreNative Library.

  @author Jeff Schoen
  @version $Id: CoreNative.h,v 1.5 2011/07/27 18:45:15 ntn Exp $
*/
#include "native.h"
#include "CoreDefs.h"

#define PTR (void*)
#define CPTR (void*)(jpointer)
#define PPTR (HALO*)(jpointer)

FILE *fpi,*fpo,*fdi,*fdo;

JNIEXPORT jlong JNICALL Java_nxm_ice_lib_CoreNative_alloc (JNIEnv *env, jobject obj, jstring config) {
  return min(0,max(0,0)); /* must be over-ridden */
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_config (JNIEnv *env, jobject obj, jlong handle,
		jint key, jint value) {
  HALO *pt = PPTR handle;
  return (*pt->config) (CPTR handle,key,value);
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_setkey (JNIEnv *env, jobject obj, jlong handle,
		jstring key, jobject buffer, jint len) {
  HALO *pt = PPTR handle;
  char *ckey = (char*)GetString(key);
  jbyte *buf = GetArray(buffer), *bufp = buf;
  jint status, i, needpad = (ckey[0]=='S' && ckey[1]==':');
  if (needpad) { bufp=malloc(len+1); for (i=0;i<len&&buf[i]!=' ';i++) bufp[i]=buf[i]; bufp[i]=0; len=i; }
  status = (*pt->setkey) (CPTR handle,ckey,(void*)bufp,len);
  if (needpad) free(bufp);
  ReleaseArray(buffer,buf);
  ReleaseString(key,ckey);
  return status;
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_getkey (JNIEnv *env, jobject obj, jlong handle,
		jstring key, jobject buffer, jint len) {
  HALO *pt = PPTR handle;
  char *ckey = (char*)GetString(key);
  jbyte *buf = GetArray(buffer);
  jint status = (*pt->getkey) (CPTR handle,ckey,(void*)buf,len);
  ReleaseArray(buffer,buf);
  ReleaseString(key,ckey);
  return status;
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_init (JNIEnv *env, jobject obj, jlong handle) {
  HALO *pt = PPTR handle;
  char *tfn=getenv("ICECORE_TFN"), tmp[256];
  jint status = (*pt->init) (CPTR handle);
  if (tfn!=NULL) {
    sprintf(tmp,"%s_pin" ,tfn); fpi = fopen(tmp,"wb");
    sprintf(tmp,"%s_pout",tfn); fpo = fopen(tmp,"wb");
    sprintf(tmp,"%s_din" ,tfn); fdi = fopen(tmp,"wb");
    sprintf(tmp,"%s_dout",tfn); fdo = fopen(tmp,"wb");
  } else {
    fpi = fpo = fdi = fdo = NULL;
  }
  return status;
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_work (JNIEnv *env, jobject obj, jlong handle,
		jobject bi1, jint ni1, jobject bi2, jint ni2, jobject bo1, jint no1, jobject bo2, jint no2) {
  HALO *pt = PPTR handle;
  int_4 ptvs = sizeof(HALO) - sizeof(pt->config)*6 + pt->psize; /* plan test vector size */
  if (ni1>0) pt->bi1 = GetArray(bi1); pt->ni1=ni1;
  if (ni2>0) pt->bi2 = GetArray(bi2); pt->ni2=ni2;
  if (no1>0) pt->bo1 = GetArray(bo1); pt->no1=no1;
  if (no2>0) pt->bo2 = GetArray(bo2); pt->no2=no2;
  if (fpi!=NULL) fwrite((void*)&pt->func,ptvs,1,fpi);
  if (fdi!=NULL && ni1>0) fwrite((void*)pt->bi1,ni1,1,fdi);
  jint status = (*pt->work) (CPTR handle);
  if (fpo!=NULL) fwrite((void*)&pt->func,ptvs,1,fpo);
  if (fdo!=NULL && status>0) fwrite((void*)pt->bo1,status,1,fdo);
  if (ni1>0) ReleaseArray(bi1,pt->bi1);
  if (ni2>0) ReleaseArray(bi2,pt->bi2);
  if (no1>0) ReleaseArray(bo1,pt->bo1);
  if (no2>0) ReleaseArray(bo2,pt->bo2);
  return status;
}

JNIEXPORT jint  JNICALL Java_nxm_ice_lib_CoreNative_free (JNIEnv *env, jobject obj, jlong handle) {
  HALO *pt = PPTR handle;
  jint status = (*pt->free) (CPTR handle);
  if (fpi!=NULL) fclose(fpi);
  if (fpo!=NULL) fclose(fpo);
  if (fdi!=NULL) fclose(fdi);
  if (fdo!=NULL) fclose(fdo);
  return status;
}
