package nxm.ice.lib;

import nxm.sys.lib.*;
import java.io.InputStream;
import java.io.FileOutputStream;

/**
  Extracts information and data from an ICE Archive file of packetized data.

  @author Jeff Schoen
  @version $Id: ICEExtract.java,v 1.10 2011/02/02 22:04:13 jgs Exp $
*/
public class ICEExtract extends InputStream {

  static int XFER=0x10000;

  DataFile dfi,dfo;
  Keywords kw;
  long pos_cur=0;
  long pos_dat=512;
  long pos_ext=0;
  long pos_end=0;

  public ICEExtract () { }

  public ICEExtract (Object ref, String ifname, double start, double duration) {
    open(ref,ifname,start,duration);
  }

  public boolean open (Object ref, String ifname, double start, double duration) {
    // open the circular packetized input file
    dfi = new DataFile(ref,ifname);
    if (!dfi.open(dfi.INPUT)) return false;
    double delta  = dfi.getXDelta();
    double dbpe   = dfi.getBPE();
    double ospblk = Math.round(512.0/dbpe);
    double ostart = 0.0;
    double osize  = Math.floor((duration/delta)/ospblk) * ospblk;
    // create a flattened internal file
    dfo = new DataFile(ref,"NULL",dfi,0);
    dfo.setPacket("NONE");
    dfo.setDetached(0);
    dfo.setDataStart(pos_dat);
    dfo.setDataSize(osize*dbpe);
    pos_ext = pos_dat + (long)dfo.getDataSize();
    kw = dfo.getKeywordsObject();
    dfo.setExtStart((int)(pos_ext/512));
    dfo.setExtSize(kw.size);
    pos_end = pos_ext + (long)kw.size;
    return true;
  }

  // Input Stream methods

  public int available() {
    long avail = pos_end-pos_cur;
    return (int)Math.min(avail,1024*1024);
  }

  public void close() {
  }

  public int read() {
    byte[] b = new byte[1];
    read(b,0,1);
    return (int)b[0];
  }

  public int read (byte[] b) {
    return read(b,0,b.length);
  }

  public int read (byte[] b, int off, int len) {
    int ioff = off;
    if (pos_cur<pos_dat && len>0) {
      int xfer = (int)Math.min(len,pos_dat-pos_cur);
      System.arraycopy(dfo.hb,(int)pos_cur, b,off, xfer);
      off+=xfer; pos_cur+=xfer; len-=xfer;
    }
    if (pos_cur<pos_ext && len>0) {
      int xfer = (int)Math.min(len,pos_ext-pos_cur);
      xfer = dfi.read(b,off,xfer);
      off+=xfer; pos_cur+=xfer; len-=xfer;
    }
    if (pos_cur<pos_end && len>0) {
      int xfer = (int)Math.min(len,pos_end-pos_cur);
      System.arraycopy(kw.buf,(int)(pos_cur-pos_ext), b,off, xfer);
      off+=xfer; pos_cur+=xfer; len-=xfer;
    }
    return off-ioff;
  }

  public long skip (long n) {
    pos_cur += n;
    return n;
  }

  // static interface methods

  public static InputStream asInputStream (Object ref, String ifname, double start, double duration) {
    ICEExtract fis = new ICEExtract();
    boolean ok = fis.open(ref,ifname,start,duration);
    return ok? fis : null;
  }

  public static int toFile (String ifname, double start, double duration, String ofname) {
    return toFile(Shell.getMidasContext(),ifname,start,duration,ofname);
  }

  public static int toFile (Object ref, String ifname, double start, double duration, String ofname) {
    try {
      InputStream fis = asInputStream(ref,ifname,start,duration);
      FileOutputStream fos = new FileOutputStream(ofname);
      byte[] buffer = new byte[XFER];
      for (;;) {
        int xfer = fis.read(buffer,0,XFER);
        if (xfer>0) fos.write(buffer,0,xfer);
        if (xfer<=0) break;
      }
      fos.close();
      fis.close();
    } catch (Exception e) {
      System.out.println("Problem in ICEExtract.toFile: "+e);
      return -1;
    }
    return 0;
  }

}
