package nxm.ice.lib;

import nxm.ice.lib.IceHW;

public class IceMC implements IceHW {

  // special static vars recognized by Ice JVM loader
  public static int saddr,sdata; // system bus
  public static int maddr,mdata; // memory bus
  public static int ioadr,ioctl; // IO bus

  public native static Object getObjectAt (int addr);

  // normal static vars
  public static int runMode;
  public static int tick;
  public static int CRC;

  public static void main () {
    run();
  }

  public static void run() {
    reset();
    for (int beat=1;;beat++) {
      processMessages();
      processMsg();
      processDMA();
      awMem(HBT,beat);
    }
  }

  public static int isBus (int addr) {
    if ((addr>>28) == 15) return 0;
    if ((addr>>24) != 0) return 1;
    return 0;
  }

  public static void processMsg() {

    int stat = arBus(REG_AMCSR);
    if ((stat&MCSR_OMB1F)==0) return;

    int r1 = arBus(REG_AIMB1);
    int r2 = arBus(REG_AIMB2);
    int r3 = arBus(REG_AIMB3);
    int r4 = arBus(REG_AIMB4);
//    int r5 = arBus(REG_AOMB1);
    int r0 = 0; //(((r1^r2)^r3)^r4);
    int i2 = r2<<2;

    r0 = r3;
    if (r1==0) {	// adon_null
      r0=CRC;
      CRC=0;
    }
    else if (r1==1) {	// adon_wr
      if (isBus(i2)!=0) awBus(i2,r3);
      else              awMem(i2,r3);
    }
    else if (r1==2) {	// adon_rd
      if (isBus(i2)!=0) r0=arBus(i2);
      else              r0=arMem(i2);
    }
    else if (r1==4) {	// adon_dma
      r0 = r3&0xFFFF;
      awBus(REG_AMCSR,MCSR_RFIFO);
      awBus(REG_AOMB1,r0);
      if ((r4&0x4)!=0) wfifo (i2,r0);
      else             rfifo (i2,r0);
      return;
    }
    else if (r1==9) {	// adon_init
      if (r2>=0) {
        initMC();
        runMode=1;
      } else {
        runMode=0;
      }
    }
    else if (r1==14) {	// adon_wrioc
      r2 = ((r2&0xC0)<<(HYP_SHF-6)) | IOC_ADDR | (r2&0x3F);
      awBus(r2,r3);
    }
    else if (r1==15) {	// adon_rdioc
      r2 = ((r2&0xC0)<<(HYP_SHF-6)) | IOC_ADDR | (r2&0x3F);
      r0 = arBus(r2);
    }
    else if (r1==16) {	// adon_wriob
    }
    else if (r1==17) {	// adon_rdiob
    }
    else if (r1==20) {	// adon_jtag
      if (r2 != -1) {
        r0 = dojtag(r2,r3,r4);
        //if (r4==0x1040000F) CRC = CRC32(CRC,r3,32);
      }
      else for (int i=0; i<(r4&0xFFFF); i++) {
        int j = arMem(r3<<2); r3++;
        dojtag (0,j,(r4&0xFFFF0000)|0x0F);
      }
    }
    awBus(REG_AOMB1,r0);
  }

  public static void processDMA() {
    awBus(REG_AOMB4,tick);
    tick = tick+1;
  }

  public static void processMessages() {
    maddr = MSG;
    int key = mdata;
    if ((key>>24)!=0x55) return;
    int node = mdata;
    int head = mdata;
    int addr = mdata;
    int func = head & 0x3F;
    int olen = head >> 16;
    int size = olen >> 2;
    int data = MSG+16;
         if (func==1) { busr (addr,data,size); }
    else if (func==2) { busw (addr,data,size); olen=0; }
    else if (func==3) { memr (addr,data,size); }
    else if (func==4) { memw (addr,data,size); olen=0; }
    else if (func==6) { initAlgorithm(); }
    else if (func==7) { modifyAlgorithm(data,size); }
    else if (func==14) { busrwm (addr,data,size); }
    else if (func==15) { memi (data,size); olen=0; }
    else {
      int port = addr & 0xFFFF;
      int type = addr>>16;
      int route = mdata;
      if (func==8) openAlgorithm(port,type,route);
      if (func==9) closeAlgorithm(port,type,route);
    }
    maddr = MSG;
    mdata = 0xAA000000 | olen;
  }

  public static void initAlgorithm () {
    resetDMA();
    runMode = ALG_INIT;
  }

  public static void openAlgorithm (int port, int type, int route) {
    awBus(DMAC_ROUTE_SET,route);
  }

  public static void processAlgorithm () {
  }

  public static void closeAlgorithm (int port, int type, int route) {
    awBus(DMAC_ROUTE_CLR,route);
    if (arBus(DMAC_CTL)==0) resetDMA();
  }

  public static void modifyAlgorithm (int data, int size) {
  }

  public static void  initMC () {
    maddr = 0x27000<<2;
    for (int i=0; i<0x1000; i++) mdata = 0;
    awBus(DMAC_ROUTE_CLR,-1);
  }
  public static void  wfifo (int addr, int count) {
    maddr = addr;
    while (count>0) {
      int stat = arBus(REG_AMCSR);
      if ((stat&MCSR_IFIFO_FULL)==0) {
        int data = mdata;
        awBus(REG_AFIFO,data);
        count--;
      }
    }
  }
  public static void  rfifo (int addr, int count) {
    maddr = addr;
    while (count>0) {
      int stat = arBus(REG_AMCSR);
      if ((stat&MCSR_OFIFO_EMPTY)==0) {
        int data = arBus(REG_AFIFO);
        mdata = data;
        count--;
      }
    }
  }

  public static int  dojtag (int tms, int tdi, int ena) {
    int i, tdo=tdi;
    awBus(JTAG_BASE+0,tms);
    awBus(JTAG_BASE+4,tdi);
    awBus(JTAG_BASE+8,ena);
    while( (arBus(JTAG_BASE+12)&0x1) != 0); // wait for scan to finish
    if ((ena&0x0F00)!=0) tdo = arBus(JTAG_BASE+0);
    return tdo;
  }

  // compare of synchronous current and config DMA registers
  protected static int dmaSync (int addr) {
    maddr = addr;
    int data = mdata;
    data -= mdata;
    return data;
  }

  protected static void reset () {
    awBus(DMAC_ROUTE_RST,-1);
    runMode=0;
  }

  protected static void resetDMA () {
    awBus(DMAC_ROUTE_RST,-1);
    udelay(20);
    awBus(DMAC_ROUTE_RST,-1);
    for (int i=0; i<8;  i++) setupDMAi(i, 0, 0);
    for (int i=0; i<8;  i++) setupDMAo(i, 0, 0);
    for (int i=0; i<32; i++) setupDMAt(i, 0);
    awBus(DMAC_ROUTE_FLG,0);
  }
    
  // sets up DMA out of block into RAM
  static void setupDMAo (int i, int cur_addr, int cfg_addr) {
    maddr = DMA_XO + (i<<3);
    mdata = cur_addr; 
    mdata = cfg_addr; 
  }

  // sets up DMA into block from RAM
  static void setupDMAi (int i, int cur_addr, int cfg_addr) {
    maddr = DMA_XI + (i<<3);
    mdata = cur_addr; 
    mdata = cfg_addr; 
  }
  
  // sets up DMA out of tuner/mcore into RAM
  static void setupDMAt (int i, int cur_addr) {
    maddr = DMA_TO + (i<<2);
    mdata = cur_addr; 
  }
  
  static void busr (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    for (int i=0; i<size; i++) mdata = sdata;
  }

  static void busw (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    for (int i=0; i<size; i++) sdata = mdata;
  }

  static void busrwm (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    int waddr = mdata;
    int wdata = mdata;
    maddr = madr;
    for (int i=0; i<size; i++) {
      saddr = waddr;
      sdata = wdata;
      udelay(20);
      saddr = addr;
      mdata = sdata;
    }
  }

  protected static void udelay (int usec) {
    for (int i=0; i<usec; i++) {
      for (int j=0; j<16; j++);
    }
  }

  static int divide (int num, int denom) {
    boolean neg = (num<0);
    if (num<0) num = -num;
    if (denom<0) { denom = -denom; neg = !neg; }
    int a=num, b=0, i=0;
    while (a!=0) { a=a>>4; i=i+4; }		// find leftmost non-zero nibble
    while (i-- > 0) {				// appends one bit from numerator to a
      a = (a << 1) + ((num & (1 << i)) >> i);
      b = b << 1;
      if (a >= denom) { a=a-denom; b=b+1; }
    }
    return neg? -b : b;
  }

  final static int rmax=64;
  final static int wmax=64;

  static void memr (int addr, int madr, int size) {
    int data; 
    if ((addr&0xFFF00000) > 0) {	// DRAM
      for (size<<=2; size>rmax; size-=rmax) {
        maddr = madr;
        ioadr = addr;
        ioctl = IOCTL_RD|rmax;
	madr+=rmax; addr+=rmax;
      }
      maddr = madr;
      ioadr = addr;
      ioctl = IOCTL_RD|size;
    } else {				// BRAM
      for (int i=0; i<size; i++) {
        maddr = addr; data = mdata; addr+=4;
        maddr = madr; mdata = data; madr+=4;
      }
    }
  }

  static void memw (int addr, int madr, int size) {
    int data;
    if ((addr&0xFFF00000) > 0) {	// DRAM
      for (size<<=2; size>wmax; size-=wmax) {
        maddr = madr;
        ioadr = addr;
        ioctl = IOCTL_WR|wmax;
	madr+=wmax; addr+=wmax;
      }
      maddr = madr;
      ioadr = addr;
      ioctl = IOCTL_WR|size;
    } else {				// BRAM
      for (int i=0; i<size; i++) {
        maddr = madr; data = mdata; madr+=4;
        maddr = addr; mdata = data; addr+=4;
      }
    }
  }

  static void memi (int madr, int size) {
    for (int i=0; i<size; i++) {
      maddr = madr;
      ioadr = mdata;
      ioctl = IOCTL_WR|0x4;
      madr += 4;
      udelay(50);
    }
  }

  protected static void awBus (int addr, int data) {
    saddr = addr;
    sdata = data;
  }
  protected static int arBus (int addr) {
    saddr = addr;
    saddr = addr; // make sure address is presented for ample clocks
    return sdata;
  }

  protected static void awMem (int addr, int data) {
    maddr = addr;
    mdata = data;
  }
  protected static int arMem (int addr) {
    maddr = addr;
    return mdata;
  }

}
