package nxm.ice.lib;

import nxm.ice.lib.IceHW;

public class IcePM implements IceHW {

  // special static vars recognized by Ice JVM loader
  public static int saddr,sdata; // system bus
  public static int maddr,mdata; // memory bus
  public static int ioadr,ioctl; // IO bus

  public native static Object getObjectAt (int addr);

  // normal static vars
  public static int runMode;

  public static void main () {
    run();
  }

  public static void run() {
    reset();
    for (int beat=1;;beat++) {
      processMessages();
      if (runMode==ALG_PROCESS) processAlgorithm();
      awMem(HBT,beat);
    }
  }

  public static void processMessages() {
    maddr = MSG;
    int key = mdata;
    if ((key>>24)!=0x55) return;
    int node = mdata;
    int head = mdata;
    int addr = mdata;
    int func = head & 0x3F;
    int olen = head >> 16;
    int size = olen >> 2;
    int data = MSG+16;
         if (func==1) { busr (addr,data,size); }
    else if (func==2) { busw (addr,data,size); olen=0; }
    else if (func==3) { memr (addr,data,size); }
    else if (func==4) { memw (addr,data,size); olen=0; }
    else if (func==6) { initAlgorithm(); }
    else if (func==7) { modifyAlgorithm(data,size); }
    else if (func==14) { busrwm (addr,data,size); }
    else if (func==15) { memi (data,size); olen=0; }
    else {
      int port = addr & 0xFFFF;
      int type = addr>>16;
      int route = mdata;
      if (func==8) openAlgorithm(port,type,route);
      if (func==9) closeAlgorithm(port,type,route);
    }
    maddr = MSG;
    mdata = 0xAA000000 | olen;
  }

  public static void initAlgorithm () {
    resetDMA();
    runMode = ALG_INIT;
  }

  public static void openAlgorithm (int port, int type, int route) {
    awBus(DMAC_ROUTE_SET,route);
    if (type==IOPT_TBANK) {
      initTBank(port);
      runMode = ALG_PROCESS;
    }
  }

  public static void processAlgorithm () {
    FTTStruct ftt = (FTTStruct)getObjectAt(FTT_LOC);
    int stages = ftt.stages;
    int cstage = ftt.cstage;
    if (stages>1) doTBank(cstage,true);
    if (stages>0) doTBank(stages-1,false);
  }

  public static void closeAlgorithm (int port, int type, int route) {
    awBus(DMAC_ROUTE_CLR,route);
    if (type==IOPT_TBANK) {
      exitTBank(port);
      udelay(1000);
      runMode = ALG_DONE;
    }
    if (arBus(DMAC_CTL)==0) resetDMA();
  }

  public static void modifyAlgorithm (int data, int size) {
    
  }

  protected static void initTBank (int port) {

    FTTStruct ftt = (FTTStruct)getObjectAt(FTT_LOC);
    TBStruct tb = null;

    int oact = ftt.active;
    if (port>2) port=7;
    int act = oact|port;
    int top = (act==2)?1:0;
    int stages = ftt.stages;
    ftt.chantop = top;
    ftt.active = act;
    for (int i=0; i<3; i++) {
      int tbi = (i>1)? FTT_TB2 : (i>0)? FTT_TB1 : FTT_TB0;
      tb = (TBStruct)getObjectAt(tbi);
      tb.chan = top; // needs to be before return if support A&B
      tb.pindx = 0;
      tb.tphase = 0;
      tb.tdphase = 0;
    }
    if (oact!=0) return;
    ftt.cstage = 0;
    ftt.lstage = -1;
    ftt.busy = -1;
    ftt.dca0 = null;
    ftt.dca1 = null;
    // set up the DMA channels
    tb = (TBStruct)getObjectAt(FTT_TB0);
    int nchan = tb.nchan;
    int pcfg  = tb.pcfg;
    int pstrt = tb.pstrt;
    int psize = tb.psize;
    int inc   = (stages>1)? 2:1;
    for (int i=0,j=0,adr=pcfg|pstrt; i<nchan; i++,j+=inc,adr+=psize) setupDMAt (j,adr);
    ftt.tbctl = arBus(TUNC_CTL); // store BPASS tuner reg
    if (stages>1) ssBank(0);
//    if (stages>1) awBus(DMAC_ROUTE_FLG,ROUTF_TXAOB);
  }

  protected static void doTBank (int stage, boolean bpass) {

    FTTStruct ftt = (FTTStruct)getObjectAt(FTT_LOC);
    TBStruct tb = null, tbn = null;

    // check busy
    int dcaa = bpass? FTT_DCA1 : FTT_DCA0;
    int dci  = arMem(dcaa);
    if (dci!=0 && dmaSync(dci)!=0) return;

    // check config
    int act  = ftt.active;
    int tbi  = (stage>0)? FTT_TB1 : FTT_TB0;
    tb = (TBStruct)getObjectAt(tbi);
    int tbni  = tbi+TB_SIZEOF; // 20elem*4by
    tbn = (TBStruct)getObjectAt(tbni);
    int chan = tb.chan;
    int ctop = ftt.chantop;
    boolean odd = (chan&1)!=0 && (act!=7);
    int io = bpass? PM_TUNB : odd? ftt.omodb : ftt.omoda;
    int route = 0x01000000 << io;
    dci = DMA_XI + (io<<3);

    // stop last tuner bank pass
    if (stage == ftt.busy) {
      awBus(DMAC_ROUTE_CLR,route);
      ftt.busy = -1;
      // just finished with last bank, now update pointers
      if (chan==ctop) tbn.pindx = (tbn.pindx+tb.prod) & tbn.pmask;
      ssBank(0);
    }
//    else if (bpass && ftt.busy>=0) return;  // busy on another stage

    // get current address
    int addr;
    if (stage==0) {
      int chani = chan;
      if (bpass) chani<<=1;
      addr = arMem(DMA_TO+(chani<<2));
    } 
    else if (chan==ctop) addr = getBase(tb,chan);
    else addr = ftt.addr;

    // check availability
    int mask = tb.pmask;
    int indx = tb.index;
    int xfer = tb.xfer;
    int size = tb.psize;
    int diff = (addr&mask) - indx;
    if (diff<0) diff += size;
    if (diff<xfer) return;

    // setup tuner bank pass
    if (bpass) {
      ssBank(0);
      if (chan==ctop) ftt.base = getBase (tbn,chan);
      int base = ftt.base;
      int incr   = tbn.psize;
      if (act!=7) incr <<= 1;
      int lstage = ftt.lstage;
      int phinc  = tbn.phinc;
      int dphinc = tbn.dphinc;
      int tphase = tbn.tphase;
      int tdphase= tbn.tdphase;
      int start  = tphase;
      int taddr  = TUNB_CTL;
      int tinc   = 0x100;
      int cpb2   = tbn.mchan<<1;
      for (int i=1; i<cpb2; i+=2) {
        if (stage!=lstage) awBus(taddr+GCFPGA_FREQ,-phinc);
        if (chan==ctop) awBus(taddr+GCFPGA_POFF,-start);
        start += tdphase;
        setupDMAt(i,base);
        phinc += dphinc;
        base  += incr;
        taddr += tinc;
      }
      if (chan==ctop) {
       tbn.tphase = tphase + tbn.tphinc;
       tbn.tdphase = tdphase + tbn.tdphinc;
      }
      if (act==3) {
        if (odd) base =     base - tbn.psize;
        else     base = ftt.base + tbn.psize;
      }
      // start the tuner bank core
      ssBank(1);
      ftt.lstage = stage;
      ftt.busy = stage;
      ftt.base = base;
    }

    // start the transfer
    int start = indx;
    int stop  = (indx+xfer) & mask;
    int base  = addr & (~mask);
    awMem(dci+0,base+start);
    awMem(dci+4,base+stop);
    awBus(DMAC_ROUTE_SET,route);

    // setup for next pass
    if ((act&3)!=3) chan++;
    if (++chan >= tb.nchan) {
      chan = ctop;
      indx = (indx+tb.cons) & mask;
      tb.index = indx;
    }
    if (stage>0) {
      addr += size;
      //if (act!=7) addr += size;
      if ((act&3)!=3) addr += size;
      ftt.addr = addr; 
    }
    tb.chan = chan;
    awMem(dcaa,dci);
  }

  protected static void exitTBank (int port) {
    FTTStruct ftt = (FTTStruct)getObjectAt(FTT_LOC);
    int oact = ftt.active;
    if (port>2) port=7;
    int act = oact & ~port;
    ftt.active = act;
    if (act!=0) return;
    int stages = ftt.stages;
    if (stages>1) ssBank(0);
  }

  static int getBase (TBStruct tb, int chan) {
    return tb.pcfg | (tb.pstrt + tb.psize*chan + tb.pindx);
  }

  static void ssBank (int mode) {
    int ctl = arMem(FTT_TBCTL);
    if (mode<=0) ctl &= ~0x1;
    awBus(TUNC_CTL, ctl);
  }

  // compare of synchronous current and config DMA registers
  protected static int dmaSync (int addr) {
    maddr = addr;
    int data = mdata;
    data -= mdata;
    return data;
  }

  protected static void reset () {
    awBus(DMAC_ROUTE_RST,-1);
    runMode=0;
  }

  protected static void resetDMA () {
    awBus(DMAC_ROUTE_RST,-1);
    udelay(20);
    awBus(DMAC_ROUTE_RST,-1);
    for (int i=0; i<8;  i++) setupDMAi(i, 0, 0);
    for (int i=0; i<8;  i++) setupDMAo(i, 0, 0);
    for (int i=0; i<32; i++) setupDMAt(i, 0);
    awBus(DMAC_ROUTE_FLG,0);
  }
    
  // sets up DMA out of block into RAM
  static void setupDMAo (int i, int cur_addr, int cfg_addr) {
    maddr = DMA_XO + (i<<3);
    mdata = cur_addr; 
    mdata = cfg_addr; 
  }

  // sets up DMA into block from RAM
  static void setupDMAi (int i, int cur_addr, int cfg_addr) {
    maddr = DMA_XI + (i<<3);
    mdata = cur_addr; 
    mdata = cfg_addr; 
  }
  
  // sets up DMA out of tuner/mcore into RAM
  static void setupDMAt (int i, int cur_addr) {
    maddr = DMA_TO + (i<<2);
    mdata = cur_addr; 
  }
  
  static void busr (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    for (int i=0; i<size; i++) mdata = sdata;
  }

  static void busw (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    for (int i=0; i<size; i++) sdata = mdata;
  }

  static void busrwm (int addr, int madr, int size) {
    saddr = addr;
    maddr = madr;
    int waddr = mdata;
    int wdata = mdata;
    maddr = madr;
    for (int i=0; i<size; i++) {
      saddr = waddr;
      sdata = wdata;
      udelay(20);
      saddr = addr;
      mdata = sdata;
    }
  }

  protected static void udelay (int usec) {
    for (int i=0; i<usec; i++) {
      for (int j=0; j<16; j++);
    }
  }

  static int divide (int num, int denom) {
    boolean neg = (num<0);
    if (num<0) num = -num;
    if (denom<0) { denom = -denom; neg = !neg; }
    int a=num, b=0, i=0;
    while (a!=0) { a=a>>4; i=i+4; }		// find leftmost non-zero nibble
    while (i-- > 0) {				// appends one bit from numerator to a
      a = (a << 1) + ((num & (1 << i)) >> i);
      b = b << 1;
      if (a >= denom) { a=a-denom; b=b+1; }
    }
    return neg? -b : b;
  }

  final static int rmax=64;
  final static int wmax=64;

  static void memr (int addr, int madr, int size) {
    int data; 
    if ((addr&0xFFF00000) > 0) {	// DRAM
      for (size<<=2; size>rmax; size-=rmax) {
        maddr = madr;
        ioadr = addr;
        ioctl = IOCTL_RD|rmax;
	madr+=rmax; addr+=rmax;
      }
      maddr = madr;
      ioadr = addr;
      ioctl = IOCTL_RD|size;
    } else {				// BRAM
      for (int i=0; i<size; i++) {
        maddr = addr; data = mdata; addr+=4;
        maddr = madr; mdata = data; madr+=4;
      }
    }
  }

  static void memw (int addr, int madr, int size) {
    int data;
    if ((addr&0xFFF00000) > 0) {	// DRAM
      for (size<<=2; size>wmax; size-=wmax) {
        maddr = madr;
        ioadr = addr;
        ioctl = IOCTL_WR|wmax;
	madr+=wmax; addr+=wmax;
      }
      maddr = madr;
      ioadr = addr;
      ioctl = IOCTL_WR|size;
    } else {				// BRAM
      for (int i=0; i<size; i++) {
        maddr = madr; data = mdata; madr+=4;
        maddr = addr; mdata = data; addr+=4;
      }
    }
  }

  static void memi (int madr, int size) {
    for (int i=0; i<size; i++) {
      maddr = madr;
      ioadr = mdata;
      ioctl = IOCTL_WR|0x4;
      madr += 4;
      udelay(50);
    }
  }

  protected static void awBus (int addr, int data) {
    saddr = addr;
    sdata = data;
  }
  protected static int arBus (int addr) {
    saddr = addr;
    saddr = addr; // make sure address is presented for ample clocks
    return sdata;
  }

  protected static void awMem (int addr, int data) {
    maddr = addr;
    mdata = data;
  }
  protected static int arMem (int addr) {
    maddr = addr;
    return mdata;
  }

}
