/* Include code signatures */
#include "tiolib.h"

/* Create Wrappers */
#include "cdefs.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "native.h"
#define CMETHOD(M) IceResource_## M
#define JMETHOD(M) Java_nxm_ice_lib_IceResource_## M

#define PTR (void*)
#define CPTR (void*)(long)
#define JPTR (jlong)(long)

JNIEXPORT jlong JNICALL JMETHOD (alloc) (JNIEnv *env, jobject obj) {
  return JPTR tio_open ();
}

JNIEXPORT jlong JNICALL JMETHOD (setConfig) (JNIEnv *env, jobject obj, jlong handle,
               jstring path) {
  char *cpath = (char*)GetString(path);
  jlong status = (jlong) tio_set_config (CPTR handle, cpath);
  ReleaseString(path,cpath);
  return status;
}

JNIEXPORT jint  JNICALL JMETHOD (setDirection) (JNIEnv *env, jobject obj, jlong handle,
                jint dir) {
  return (jint) tio_set_dir (CPTR handle,dir);
}

JNIEXPORT jint  JNICALL JMETHOD(init) (JNIEnv *env, jobject obj, jlong handle) {
  return (jint) tio_init (CPTR handle);
}

JNIEXPORT jlong JNICALL JMETHOD(writeDet) (JNIEnv *env, jobject obj, jlong handle,
               jstring path, jlong len) {
  char *cpath = (char*)GetString(path);
  jlong status = (jlong) tio_write_det (CPTR handle, cpath, (off_t)len);
  ReleaseString(path,cpath);
  return status;
}

JNIEXPORT jlong JNICALL JMETHOD(updateDet) (JNIEnv *env, jobject obj, jlong handle,
               jstring path, jlong len) {
  char *cpath = (char*)GetString(path);
  jlong status = (jlong) tio_update_det (CPTR handle, cpath, (off_t)len);
  ReleaseString(path,cpath);
  return status;
}

JNIEXPORT jlong JNICALL JMETHOD(finalizeDet) (JNIEnv *env, jobject obj, jlong handle,
               jstring path, jlong len) {
  char *cpath = (char*)GetString(path);
  jlong status = (jlong) tio_finalize_det (CPTR handle, cpath, (off_t)len);
  ReleaseString(path,cpath);
  return status;
}

JNIEXPORT jint  JNICALL JMETHOD(seek) (JNIEnv *env, jobject obj, jlong handle,
                jlong off) {
  return (jint) tio_seek (CPTR handle, (off_t)off);
}

JNIEXPORT jint  JNICALL JMETHOD(setOffset) (JNIEnv *env, jobject obj, jlong handle,
                jlong off) {
  return (jint) tio_set_offset (CPTR handle, (off_t)off);
}

JNIEXPORT jint  JNICALL JMETHOD(nativeRead) (JNIEnv *env, jobject obj, jlong handle,
                jlong buf, jint n1) {
  return (jint) tio_read (CPTR handle,CPTR buf,n1);
}

JNIEXPORT jint  JNICALL JMETHOD(nativeReadB) (JNIEnv *env, jobject obj, jlong handle, 
		jobject buffer1, jint n1) {
  jbyte *buf1 = GetArray(buffer1);
  jint status = (jint) tio_read (CPTR handle,PTR buf1,n1);
  ReleaseArray(buffer1,buf1);
  return status;
}

JNIEXPORT jint  JNICALL JMETHOD(nativeWrite) (JNIEnv *env, jobject obj, jlong handle,
                jlong buf, jint n1) {
  return (jint) tio_write (CPTR handle,CPTR buf,n1);
}

JNIEXPORT jint  JNICALL JMETHOD(nativeWriteB) (JNIEnv *env, jobject obj, jlong handle, 
		jobject buffer1, jint n1) {
  jbyte *buf1 = GetArray(buffer1);
  jint status = (jint) tio_write (CPTR handle,PTR buf1,n1);
  ReleaseArray(buffer1,buf1);
  return status;
}

JNIEXPORT jint  JNICALL JMETHOD(free) (JNIEnv *env, jobject obj, jlong handle) {
  return (jint) tio_close (CPTR handle);
}

JNIEXPORT jint  JNICALL JMETHOD(launchReads) (JNIEnv *env, jobject obj, jlong handle,
                jlong len) {
  return (jint) tio_launch_read_threads (CPTR handle, (off_t)len);
}

JNIEXPORT jint  JNICALL JMETHOD(joinReads) (JNIEnv *env, jobject obj, jlong handle) {
  return (jint) tio_join_threads (CPTR handle);
}

JNIEXPORT jint  JNICALL JMETHOD(joinWrites) (JNIEnv *env, jobject obj, jlong handle) {
  return (jint) tio_wait_threads (CPTR handle);
}

JNIEXPORT jlong JNICALL JMETHOD (malloc_1buf) (JNIEnv *env, jobject obj, jlong size) {
  jlong rv;
#ifdef _LINX
  rv = JPTR tio_malloc (size);
#else
  rv = JPTR malloc(size);
#endif 
  return rv;
}

JNIEXPORT jint  JNICALL JMETHOD(free_1buf) (JNIEnv *env, jobject obj, jlong handle) {
#ifdef _LINX
  tio_free(CPTR handle);
#else
  free(CPTR handle);
#endif
  return 0;
}

int_4 dummy () {
  return min(1,max(3,4));
}
