package nxm.ice.lib;

import nxm.sys.inc.Constants;
import nxm.sys.inc.DataTypes;

import nxm.sys.lib.Data;
import nxm.sys.lib.Convert;
import nxm.sys.lib.MidasException;

/**
  Task is an element in a chain of sequential parallel operations.

  @author Jeff Schoen
  @version $Id: Task.java,v 1.2 2011/05/24 13:20:30 jgs Exp $
*/
public class Task extends Object {

  public Task taskOut;

  public Task() { }
  public void connect (Task t) { taskOut=t; }
  public void proc (int v) { }

  public static class Copy extends Task {
    public Copy() { }
    public void proc (int v) { taskOut.proc(v); }
  }

  public static class Stream extends Task {
    int ioff,nbuf;
    byte[] buf;
    public Stream () { }
    public Stream (Task src) { 
      src.connect(this);
    }
    public void set (byte[] b, int n) { ioff=0; nbuf=n; buf=b; }
    public void proc (int v) { Convert.packL(buf,ioff,v); ioff+=4; }
    public void push () { 
      for (; ioff<nbuf; ioff+=4) taskOut.proc( Convert.unpackL(buf,ioff) );
    }
    public int length() { return ioff; }
  }

  public static class Decimate extends Task {
    int i,n;
    public Decimate (Task src, int dec) { 
      src.connect(this);
      n=dec;
    }
    public void proc (int v) { 
      if (i==0) taskOut.proc(v);
      if (++i==n) i=0;
    }
  }

  public static class Filter extends Task {
    float[] taps,data;
    int i,j,k,n;
    public Filter (Task src, float[] taps) { 
      src.connect(this);
      this.taps=taps; 
      n=taps.length;
      data=new float[n];
    }
    public void proc (int v) { 
      data[i++]=v;
      v=0; 
      for (j=0,k=i; j<n; j++,k++) {
        if (k>=n) k=0;
        v += taps[j]*data[k];
      }
      taskOut.proc(v);
    }
  }

}
