/*
  routines for calling FPGA functions as an attached processor 
*/
#include <dlfcn.h>
#include <stdlib.h>
#include "icelib.h"
#include "corelib.h"
#include "headers.h"

/* C/C++ Interface */
CORE *core_alloc (const char *cname) {
  void *libhandle, *(*func)(), *corehandle;
  char path[80],sig[8],name[40],function[40],dev[40],config[128],opt[16],typ;
  int i,j,lc,cindex,pmi,ci,side,coreio;
  HEADER hwf;

  i = (int)((long)index(cname,';')-(long)cname);
  if (i>0) {
    strncpy(name,cname,i); name[i]=0;
    strcpy(opt,"XMAREA_");
    strcat(opt,cname+i+1);
  } else {
    strcpy(name,cname);
    strcpy(opt,"ICEROOT");
  }

  i = m_get_uswitch("CORE",sig,8);
  if (i>0) typ=sig[0]; else typ='C';
  coreio = (sig[3]=='X')?0:1;

  i = m_get_uswitch("COREDEV",dev,40);
  if (i<=0) strcpy(dev,"PIC1AUTO:11");
  if (strstr(dev,":")==NULL) strcat(dev,":11");

  /* parse the core type */
       if (typ=='C') strcpy(sig,"CPU");
  else if (typ=='V') strcpy(sig,"VHS");
  else if (typ=='I') strcpy(sig,"ICE");
  else if (typ=='G') strcpy(sig,"GPU");
  else return 0;

  /* parse for special implementation configuration */
  strcpy(config,"");
  if (typ=='I') {
//    i = (int)((void*)index(dev,':')-(void*)dev);
    i = (int)((long)index(dev,':')-(long)dev);
    j=i+1; if (strncmp(dev+j,"CORE",4)==0) j+=4;
    cindex = atoi(dev+j);
    if (cindex<10) cindex+=10;
    pmi = cindex/10; ci = cindex%10; side = (ci+1)%2+1;
    dev[i]=0; /* remove :port */
    i  = m_hwf_open(&hwf);
    lc = m_hwf_alias(&hwf,dev,config,40,128);
    i  = m_hwf_close(&hwf);
    sprintf(config+lc,",IPORT=PM0CORE%d,OPORT=PM%dCORE%d,COREIO=%d",side,pmi,ci,coreio);
#if DEBUG
    printf("Alias=%s Port=%d Config=%s\n",dev,cindex,config);
#endif
  }

  /* open the core library */
  sprintf(path,"%s/core/Core%s_%s.so",getenv(opt),sig,name);
  sprintf(function,"Core%s_%s_alloc",sig,name);
#if DEBUG
  printf("Alloc name=%s path=%s routine=%s\n",name,path,function);
#endif
  libhandle = dlopen(path,  RTLD_LAZY|RTLD_GLOBAL);
  if (!libhandle) { printf("Err opening library %s\n",path); return 0; }
  func = (void *(*) ()) dlsym(libhandle, function);
  if (!func) { printf("Err opening function %s\n",function); return 0; }
  corehandle = (*func)(config);
  if (!corehandle) { printf("Err running function %s\n",function); return 0; }
  return corehandle;
}
int_4 core_open (CORE *handle) {
  return (*handle->init) (handle);
}
int_4 core_process (CORE *handle, int_1 *buf1,int_4 n1, int_1 *buf2,int_4 n2) {
  return (*handle->work) (handle,buf1,n1,buf2,n2);
}
int_4 core_close (CORE *handle) {
  return (*handle->free) (handle);
}
int_4 core_setkey (CORE *handle, const char *name, void *value, int_4 size) {
  return (*handle->setkey) (handle,name,value,size);
}
int_4 core_getkey (CORE *handle, const char *name, void *value, int_4 size) {
  return (*handle->getkey) (handle,name,value,size);
}
int_4 core_setfmt (CORE *handle, const char *iform, const char *oform) {
  int fmts = (getFormat(iform)<<0) | (getFormat(oform)<<8);
  return (*handle->config) (handle,1,fmts);
}

char *f2c (char *fstr, int flen) {
  int i; 
  for (i=0; fstr[i]>' ' && i<flen; i++);
  fstr[i]=0;
  return fstr;
}

/* Fortran Interface */
int_8 core_alloc_   (char *name, int ln) { return (int_8)core_alloc(f2c(name,ln)); }
int_4 core_open_    (CORE *handle) { return core_open(handle); }
int_4 core_close_   (CORE *handle) { return core_close(handle); }
int_4 core_process_ (CORE *handle, int_1 *buf1,int_4 *n1, int_1 *buf2,int_4 *n2){ return core_process(handle,buf1,*n1,buf2,*n2); }
int_4 core_setkey_  (CORE *handle, char *name,void *value,int_4 *size, int len) { return core_setkey(handle,f2c(name,len),value,*size); }
int_4 core_getkey_  (CORE *handle, char *name,void *value,int_4 *size, int len) { return core_getkey(handle,f2c(name,len),value,*size); }
int_4 core_setfmt_  (CORE *handle, char *iform, char *oform, int lif, int lof)  { return core_setfmt(handle,f2c(iform,lif),f2c(oform,lof)); }
