/*
  routines for calling FPGA functions as an attached processor 
*/

#define FPGA_DMASZ 1048576

int_4 FPGA_open (PICSTRUCT **pp, char *config) {
  int_4 status,ipm,opm,ipo,opo,icn,ocn,coreio;
  PICSTRUCT *p = (PICSTRUCT *)malloc( sizeof(PICSTRUCT) ); *pp = p;

  status = pic_open (p, config, 0, 0);
  if (status<0) { printf("Err opening device=%s\n",config); return -1; }

  /* config has the form IPORT=PM0CORE1,OPORT=PM1CORE1 */
  ipo=findflag("IPORT",p->config); ipm=p->config[ipo+8]-'0'; icn=p->config[ipo+13]-'0';
  opo=findflag("OPORT",p->config); opm=p->config[opo+8]-'0'; ocn=p->config[opo+13]-'0';

  coreio = findintflagdef("COREIO",p->config,1);
  if (coreio==0) {
    p->pmi    = opm;
    p->pindex = ocn;
    p->dmaci  = p->dmaco = -(opm*10+ocn);
    return 0;
  }

  p->pmi=ipm; p->config[ipo]='X';
  status = pic_ioport (p, IOPT_CORE, icn, -1, 1, -16, 100000000, 0.0, 1, 0, 0);
  if (status<0) { printf("Err opening output channel dmao=%d\n",status); return -1; }
  p->dmaco = status; p->config[ipo]='I';
    status = pic_setkeyl (p, p->dmaco,KEY_CORE+0,0x8803);

  p->pmi=opm; p->config[opo]='X';
  status = pic_ioport (p, IOPT_CORE, ocn, -1, -1, -16, 100000000, 0.0, 1, 0, 0); 
  if (status<0) { printf("Err opening input channel dmai=%d\n",status); return -1; }
  p->dmaci = status; p->config[opo]='O';

  status = pic_mapmem (p, &p->mapi, FPGA_DMASZ, 1);
  status = pic_mapmem (p, &p->mapo, FPGA_DMASZ, 1);

  status = pic_dmasetup (p, p->dmaci, -1, &p->mapi, -1, 0);
  status = pic_dmasetup (p, p->dmaco,  1, &p->mapo, -1, 0);

  status = pic_dmafunc (p, p->dmaci, DMA_CONTINUOUS);
  status = pic_dmafunc (p, p->dmaco, DMA_ONDEMAND);

  /* for sys_wr and wpb routines */
  p->pmi    = opm;
  p->pindex = ocn;

  vprint("FPGA_open config=%s dmaco=%d:%d:%d dmaci=%d:%d:%d\n",p->config,ipm,icn,p->dmaco,opm,ocn,p->dmaci);
  return 0; 
}

int_4 FPGA_syswr (PICSTRUCT *p, int_4 addr, int_4 value) { 
  vprint("FPGA_syswr dmac=%d addr=%08x value=%08x\n",p->dmaci,addr,value);
  return pic_setkeyl (p,p->dmaci,KEY_CORE+addr,value);
}

int_4 FPGA_sysrd (PICSTRUCT *p, int_4 addr) { 
  int_4 value = pic_getkeyl (p,p->dmaci,KEY_CORE+addr);
  vprint("FPGA_syswr dmac=%d addr=%08x value=%08x\n",p->dmaci,addr,value);
  return value;
}

int_4 FPGA_process (PICSTRUCT *p, int_4 *in, int_4 nin, int_4 *out, int_4 nout) { 
  int_4 status=0;
  if (p->dmaco>0) status = pic_dmaxfer (p, p->dmaco, in, nin, 0);
  if (p->dmaci>0) status = pic_dmaxfer (p, p->dmaci, out, nout, FLG_NOWAIT);
  return status;
}

int_4 FPGA_close (PICSTRUCT *p) {
  int_4 status;
  if (p->dmaci>0) status = pic_dmafunc (p, p->dmaci, DMA_CANCEL);
  if (p->dmaco>0) status = pic_dmafunc (p, p->dmaco, DMA_CANCEL);
  if (p->dmaci>0) status = pic_mapmem (p, &p->mapi, FPGA_DMASZ, -1);
  if (p->dmaco>0) status = pic_mapmem (p, &p->mapo, FPGA_DMASZ, -1);
  vprint("FPGA_close config=%s\n",p->config);
  pic_close (p);
  free((void*)p);
  return 0;
}

int_4 FPGA_wpb (PICSTRUCT *p, int_4 node, int_4 addr, int_4 data) { 
  int csel = (p->pindex==4)? 0x80 : (p->pindex==3)? 0x40 : (p->pindex==2)? 0x08 : 0x04;
  addr = (addr&0x00FFFFFF) | (csel<<24);
  return pic_wpb(p,node,addr,data);
}

