/****************************************************************************/
/*                                                                          */
/*   A2DM1X module interface routines                                       */
/*                                                                          */
/*   Note: all jpmem access must be to port 2 as this dual site module      */
/*         is only controlled from the B sides JTAG connection              */
/*                                                                          */
/****************************************************************************/

#include "iomlib_a2dm1x.h"

int_4 a2dm1x_debug(A2DM1X_State_Obj *A)
{
  int_4 i;
  if (A->id != A2DM1X_ID) {
    printf(" \n");
    printf("WARNING: A2DM1X object not initialized.\n\n");
    return (-1);
  }
  printf(" \n");
  printf("A2DM1X Debug ...\n");
  printf("  chksum:      0x%8.8X  \n",A->chksum);
  printf("  id:          0x%8.8X  \n",A->id);     
  printf("  mtrev:       0x%8.8X  \n",A->mtrev);  
  printf("  sysreg:      0x%8.8X  \n",A->sysreg);  
  printf("  refclk:      %i       \n",A->refclk); 
  printf("  ovsr:        %i       \n",A->ovsr);   
  printf("  samprate:    %18.6f   \n",A->samprate); 
  printf("  foffset:     %i       \n",A->foffset); 
  printf("  debuglevel:  %i       \n",A->debuglevel);
  printf("  bits:        %i       \n",A->bits);  
  printf("  ch1pga:      %i       \n",A->ch1pga);   
  printf("  ch2pga:      %i       \n",A->ch2pga);
  printf("  test:        0x%2.2X  \n",A->test);  
  printf("  opts:        0x%2.2X  \n",A->opts);  
  printf(" \n");
#ifdef BOGUS
  printf("  pinval     = 0x%8.8X  \n",A->pinval);
  printf("  addr[PTCFG_PIN_BASE] = 0x%8.8X \n",A->addr[PTCFG_PIN_BASE]);
  printf("  addr[PTCFG_RG_BASE]  = 0x%8.8X \n",A->addr[PTCFG_RG_BASE]);
  printf("  addr[PTCFG_RG_WROFF] = 0x%8.8X \n",A->addr[PTCFG_RG_WROFF]);
  printf("  addr[PTCFG_RG_RDOFF] = 0x%8.8X \n",A->addr[PTCFG_RG_RDOFF]);
  if (A->debuglevel > 1) {
      for (i=0; i<A2DM1X_RG_MAXCNT; i++)
        printf("  A2DM1X_RG  REG: %3i   ADD: 0x%8.8X   VAL: 0x%2.2X\n",
                i, A->rgadd[i], A->rgval[i]);
  }
#endif
  if (A->debuglevel > 2) {
    ptcfg_debug(ADF4350_ID, &A->SYN);
#ifdef BOGUS
    ptcfg_debug(ADC12D_ID, &A->ADC);
#endif
    ptcfg_debug(PE4302_ID, &A->ATN1);
    ptcfg_debug(PE4302_ID, &A->ATN2);
    ptcfg_debug(PE4302_ID, &A->ATN3);
    ptcfg_debug(PE4302_ID, &A->ATN4);
  }
  return (0);  
}


int_4 a2dm1x_get_state(PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST)
{ 
  int_4 st;
  st = iom_get_state(p,mport,A2DM1X_STATE,(void*)ADST,sizeof(A2DM1X_State_Obj));
  if (p->verbose >= 3) a2dm1x_debug(ADST);
  return (st);
}


int_4 a2dm1x_set_state(PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST) 
{ 
  int_4 st;
  if (p->verbose >= 2) a2dm1x_debug(ADST);
  st = iom_set_state(p,mport,A2DM1X_STATE,(void*)ADST,sizeof(A2DM1X_State_Obj));
  return (st);
}


int_4 a2dm1x_init_state(PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST)
{
  int_4 i;

  /* Local handles for clarity */
  ADF4350_Reg_Obj *SY = &ADST->SYN;
#ifdef BOGUS
  ADC12D_Reg_Obj  *AD = &ADST->ADC;
#endif
  PE4302_Reg_Obj  *A1 = &ADST->ATN1;
  PE4302_Reg_Obj  *A2 = &ADST->ATN2;
  PE4302_Reg_Obj  *A3 = &ADST->ATN3;
  PE4302_Reg_Obj  *A4 = &ADST->ATN4;

  mport = A2DM1X_PORT_DSM;  /* Override port for dual-site module */

  /* Zero the state buffer */
  vfill (0, (int_4*)ADST, sizeof(A2DM1X_State_Obj));

  /* Check for invalid size of state object */
  if (sizeof(A2DM1X_State_Obj) > A2DM1X_STATE_MAX_BYTES) {
    printf("ERROR: A2DM1X state obj: %lu   Max: %i \n",
             sizeof(A2DM1X_State_Obj), A2DM1X_STATE_MAX_BYTES);
    printf("ERROR: A2DM1X invalid state. This is bad.\n");
    return (-1);
  }
  i = (int_4) (sizeof(A2DM1X_State_Obj)/4);
  if ((i*4) != ((int_4)sizeof(A2DM1X_State_Obj))) {
    printf("ERROR: A2DM1X state obj: %lu   Max: %i \n",
             sizeof(A2DM1X_State_Obj), A2DM1X_STATE_MAX_BYTES);
    printf("ERROR: A2DM1X object size not multiple of 4 bytes. \n");
    return (-1);
  }
  printf("A2DM1X Object Size: %i of %i available bytes\n",
    (int_4)sizeof(A2DM1X_State_Obj),A2DM1X_STATE_MAX_BYTES);

  /* Default the fields    */
  ADST->chksum    = 0xFFFFFFFF;
  ADST->id        = A2DM1X_ID; 
  ADST->refclk    = 10000000;
  ADST->ovsr      = 1;            
  ADST->samprate  = 100000000.0;
  ADST->foffset   = 0x80000000;
  ADST->bits      = 16;
  ADST->ch1pga    = -100;
  ADST->ch2pga    = -100;
  ADST->opts      = 0;
#ifdef BOGUS
  ADST->pinval    = 0x00000000;

  /* Initialize base and offset addresses */
  for (i=0; i<PTCFG_ADDR_MAX; i++)  ADST->addr[i] = 0x00000000;
  for (i=0; i<A2DM1X_RG_MAXCNT; i++) {
    ADST->rgadd[i] = 0x00000000;
    ADST->rgval[i] = 0x00;
  }
#endif

  /* Initalize the hardware objects (not the actual hardware) */
  ptcfg_init(ADF4350_ID, SY, 0);
  ptcfg_addr(ADF4350_ID, SY, PTCFG_PIN_BASE, A2DM1X_CLK_PINS);
  ptcfg_addr(ADF4350_ID, SY, PTCFG_RG_BASE,  A2DM1X_CLK_RG_BASE);

#ifdef BOGUS
  ptcfg_init(ADC12D_ID, AD, 0);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_PIN_BASE, A2DM1X_ADC_PINS);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_BASE,  A2DM1X_ADC_RG_BASE); 
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_WROFF, A2DM1X_ADC_RG_WROFF);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_RDOFF, A2DM1X_ADC_RG_RDOFF); 
#endif

  ptcfg_init(PE4302_ID, A1, 0);
  ptcfg_addr(PE4302_ID, A1, PTCFG_RG_BASE,  A2DM1X_ATN1_RG_BASE);
  ptcfg_init(PE4302_ID, A2, 0);
  ptcfg_addr(PE4302_ID, A2, PTCFG_RG_BASE,  A2DM1X_ATN2_RG_BASE);
  ptcfg_init(PE4302_ID, A3, 0);
  ptcfg_addr(PE4302_ID, A3, PTCFG_RG_BASE,  A2DM1X_ATN3_RG_BASE);
  ptcfg_init(PE4302_ID, A4, 0);
  ptcfg_addr(PE4302_ID, A4, PTCFG_RG_BASE,  A2DM1X_ATN4_RG_BASE);

  /* Push state info onto module */
  return (a2dm1x_set_state(p,A2DM1X_PORT_DSM,ADST));
}   


int_4 a2dm1x_cfg_clk (PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST)
{
  int_4  i;
  int_u4 rgval;
  ADF4350_Reg_Obj *A;

  /* Local handle to the ADF4350 configuration */
  A = &ADST->SYN;

  /* Set desired reference and frequency output and find a solution */
  A->fref = (real_8) ADST->refclk;
  A->fdes = ((real_8)ADST->samprate)*((real_8)ADST->ovsr);
  i = adf4350_pt_findsoln(A);
  A->debuglevel = 5;
  if (i > 0)
    printf("WARNING: A2DM1X clock -> Req: %20.6f  Act: %20.6f\n",
            A->fdes, A->fsynth);
  if (i < 0) return (-1);

  ptcfg_set(ADF4350_ID, A, ADF4350_RG_4h, ADF4350_RG_4h_RF_PWR,     0x03);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_4h, ADF4350_RG_4h_RF_ENABLE,  0x01);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_4h, ADF4350_RG_4h_AUX_PWR,    0x03);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_4h, ADF4350_RG_4h_AUX_ENABLE, 0x01);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_4h, ADF4350_RG_4h_AUX_SELECT, 0x00);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_3h, ADF4350_RG_3h_FL_CLKDIV,  0x0800);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_3h, ADF4350_RG_3h_FL_CLKMODE, 0x00);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_3h, ADF4350_RG_3h_CSR,        0x00);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_2h, ADF4350_RG_2h_CP_CURR,    0x04);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_2h, ADF4350_RG_2h_MUXOUT,     0x03);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_1h, ADF4350_RG_1h_PHASE,      0x0000);
  ptcfg_set(ADF4350_ID, A, ADF4350_RG_1h, ADF4350_RG_1h_PRESCALER,  0x00);

  /* Implement valid configuration */
  if (A->id != ADF4350_ID) return (-1);
  
  /* Enable chip */
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_PIN, 0x0001, 0x01);

  /* Configure chip registers in reverse order */
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_5h, 0, 0);
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_4h, 0, 0);
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_3h, 0, 0);
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_2h, 0, 0);
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_1h, 0, 0);
  ptcfg_sethw(p, mport, ADF4350_ID, A, ADF4350_RG_0h, 0, 0);

  /* Debug output */
  A->debuglevel = 5;
  ptcfg_debug(ADF4350_ID,A);

  /* Wait for synthesizer phase lock */
  for (i=0; i<100; i++) {
    rgval = ptcfg_gethw(p, mport, ADF4350_ID, A, ADF4350_RG_PIN, 0xFFFFFFFF, 0);
    if ((rgval & ADF4350_RG_PIN_LD) == ADF4350_RG_PIN_LD) break;   
    udelay(1000);
  }
  if ((rgval & ADF4350_RG_PIN_LD) != ADF4350_RG_PIN_LD) {
    printf("WARNING: A2DM1X clock synth failed to lock. Check reference.\n");
    return (-1);
  }

  return (0);
}

int_4 a2dm1x_cfg_pga(PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST)
{ 
  int_4  idx;
  int_1  pga_des;
  int_u1 rgvalA, rgvalB;
  PE4302_Reg_Obj *ATNA, *ATNB;
  real_8 pga_gain_max, pga_gain_min, total_attn;
  const real_8 max_pga_gain = 28.0; /* 15.0 + 15.0 - 2.0 */
  const real_8 max_pga_attn = 63.0; /* 31.5 + 31.5       */

  const int_u1 pga_gain_tbl[64][2] = {
   {0x00,0x00}, {0x00,0x02}, {0x00,0x04}, {0x00,0x06}, {0x00,0x08}, {0x00,0x0A},
   {0x00,0x0C}, {0x00,0x0E}, {0x00,0x10}, {0x00,0x12}, {0x00,0x14}, {0x00,0x16},
   {0x00,0x18}, {0x00,0x1A}, {0x00,0x1C}, {0x00,0x1E}, {0x00,0x20}, {0x02,0x20},
   {0x04,0x20}, {0x06,0x20}, {0x08,0x20}, {0x0A,0x20}, {0x0C,0x20}, {0x0E,0x20},
   {0x10,0x20}, {0x12,0x20}, {0x14,0x20}, {0x16,0x20}, {0x18,0x20}, {0x1A,0x20},
   {0x1C,0x20}, {0x1E,0x20}, {0x20,0x20}, {0x22,0x20}, {0x24,0x20}, {0x28,0x20},
   {0x2A,0x20}, {0x2C,0x20}, {0x2E,0x20}, {0x30,0x20}, {0x32,0x20}, {0x34,0x20},
   {0x36,0x20}, {0x38,0x20}, {0x3A,0x20}, {0x3C,0x20}, {0x3E,0x20}, {0x3F,0x21},
   {0x3F,0x23}, {0x3F,0x25}, {0x3F,0x27}, {0x3F,0x29}, {0x3F,0x2B}, {0x3F,0x2D},
   {0x3F,0x2F}, {0x3F,0x31}, {0x3F,0x33}, {0x3F,0x35}, {0x3F,0x37}, {0x3F,0x39},
   {0x3F,0x3B}, {0x3F,0x3D}, {0x3F,0x3F}, {0x3F,0x3F} };

  /* First, see if max attenuation is requested (pga <= -100dB) */
  pga_des = (mport == 2) ? (ADST->ch2pga) : (ADST->ch1pga);
  if (pga_des <= -100) {
    total_attn = -100.0;
    rgvalA = pga_gain_tbl[63][0];
    rgvalB = pga_gain_tbl[63][1];
  }
  else {
    pga_gain_max = max_pga_gain + 0.5;
    pga_gain_min = max_pga_gain - max_pga_attn;
    total_attn   = pga_gain_max - ((real_8) pga_des);
    total_attn   = (total_attn >= 0.0) ? (total_attn) : (0.0);
    total_attn   = (total_attn <= max_pga_attn) ? (total_attn) : (max_pga_attn);
    idx          = (int_4) total_attn;
    idx = (idx >=  0) ? (idx) : (0);
    idx = (idx <= 63) ? (idx) : (63);
    rgvalA = pga_gain_tbl[idx][0];
    rgvalB = pga_gain_tbl[idx][1];
  }

  printf("PGA-> PORT:%i GAIN: %i ATN: %6.2f ATN-A: 0x%2.2X ATN-B: 0x%2.2X \n",
    mport,pga_des,total_attn,rgvalA,rgvalB);

  if (mport == 1) {
    ADST->ch1pga = pga_gain_max - total_attn;
    ATNA = &ADST->ATN1;
    ATNB = &ADST->ATN2;
  }
  else {
    ADST->ch2pga = pga_gain_max - total_attn;
    ATNA = &ADST->ATN3;
    ATNB = &ADST->ATN4;
  }
  mport = A2DM1X_PORT_DSM;  /* Override port for dual-site module */
  ptcfg_sethw(p,mport,PE4302_ID,ATNA,PE4302_RG_0h,PE4302_RG_0h_ATTN,rgvalA);
  ptcfg_sethw(p,mport,PE4302_ID,ATNB,PE4302_RG_0h,PE4302_RG_0h_ATTN,rgvalB);
  return (0);
}


int_4 a2dm1x_cfg_adc(PICSTRUCT *P, int_4 mport, A2DM1X_State_Obj *ADST)
{
  int_4 ical;
  ADC12D_Reg_Obj *AD;

  /* Local handle for ADC12D */
  /* AD = &ADST->ADC; */
  AD = (ADC12D_Reg_Obj *)malloc(sizeof(ADC12D_Reg_Obj));
  if (AD == (ADC12D_Reg_Obj *)NULL) {
    printf("ERROR: A2DM1X memory allocation error in ADC12D.\n");
    return (-1);
  }

  /* Check for valid sample rate */
  if ((ADST->samprate*((real_8)ADST->ovsr)) > 1600000000.0){
    printf("ERROR: A2DM1X invalid sample rate: %20.6f \n",
        (ADST->samprate*((real_8)ADST->ovsr)));
    return (-1);
  }

  /* Initialize the ADC12D read/write interfaces */
  ptcfg_init(ADC12D_ID, AD, 0);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_PIN_BASE, A2DM1X_ADC_PINS);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_BASE,  A2DM1X_ADC_RG_BASE); 
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_WROFF, A2DM1X_ADC_RG_WROFF);
  ptcfg_addr(ADC12D_ID, AD, PTCFG_RG_RDOFF, A2DM1X_ADC_RG_RDOFF); 


  /* Turn on 90 degree clock mode */
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_0h_CFG, ADC12D_RG_0h_DPS, 0x0000);

  /* Turn on / off test pattern mode */
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_0h_CFG, ADC12D_RG_0h_TPM, 0x0000);

  /* Set output format to 2's compliment */
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_0h_CFG, ADC12D_RG_0h_2SC, 0x0001);

  /* Set FSR for both channels */
#ifdef BOGUS
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_3h_IFSRA, ADC12D_RG_3h_IFM, 0x7FFF);
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_Bh_QFSRA, ADC12D_RG_Bh_IFM, 0x7FFF);
#endif
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_3h_IFSRA, ADC12D_RG_3h_IFM, 0x0000);
  ptcfg_set(ADC12D_ID, AD, ADC12D_RG_Bh_QFSRA, ADC12D_RG_Bh_IFM, 0x0000);

  /* Set LFS bit based on sample rate */
  if ((ADST->samprate*((real_8)ADST->ovsr)) > 300000000.0) 
    ptcfg_set(ADC12D_ID, AD, ADC12D_RG_0h_CFG, ADC12D_RG_0h_LFS, 0x0000);
  else
    ptcfg_set(ADC12D_ID, AD, ADC12D_RG_0h_CFG, ADC12D_RG_0h_LFS, 0x0001);

  /* Configure for Non-Demux Mode {0=Demux, 1=No Demux } */ 
  /* ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_PIN, 
                 ADC12D_RG_PIN_NDM, 0x0001); */
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_PIN, 
                 ADC12D_RG_PIN_NDM, 0x0000); 

  /* Initialize the ADC12D */
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_0h_CFG,    0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_1h_RESV,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_2h_IOFF,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_3h_IFSRA,  0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_4h_CALADJ, 0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_5h_CALVAL, 0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_6h_RESV,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_7h_DESADJ, 0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_8h_RESV,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_9h_RESV,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Ah_QOFF,   0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Bh_QFSRA,  0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Ch_CAPDLY, 0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Dh_FAPDLY, 0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Eh_ASYNC,  0, 0);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_Fh_RESV,   0, 0);

  /* Dump debug information */
  AD->debuglevel = 5;
  ptcfg_debug(ADC12D_ID,AD);

  /* Run a calibration cycle and wait for it to complete */
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_4h_CALADJ, 
                ADC12D_RG_4h_CSS, 0x0001);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_0h_CFG, 
                ADC12D_RG_0h_CAL, 0x0000);
  udelay(1000);
  ptcfg_sethw(P, mport, ADC12D_ID, AD, ADC12D_RG_0h_CFG, 
                ADC12D_RG_0h_CAL, 0x0001);
  udelay(100);

  for (ical=0; ical<1000; ical++) {
    if (0 == ptcfg_gethw(P, mport, ADC12D_ID, AD, ADC12D_RG_PIN, 
                          ADC12D_RG_PIN_CAL, 0)) break;      
    udelay(1000);
  }
  printf("*** ICAL: %i \n",ical);

  /* Enable the outputs */


  if (AD != NULL) free(AD);

  return (0);
}


int_4 a2dm1x_implement(PICSTRUCT *p, int_4 mport, A2DM1X_State_Obj *ADST, 
                         int_4 flags)
{ 
  A2DM1X_State_Obj ADSTCur;     
  int_u1 sysreg; 
  int_1  tcxo;
  int_4  rtn;
  int_4  pga1,pga2;

  /* Turn off TCXO if external reference or external clock */
  if (ADST->opts & RFOPT_XADCLK) ADST->opts &= (~RFOPT_XREF);
  tcxo = (ADST->opts&(RFOPT_XREF|RFOPT_XADCLK)) ? (0) : (1);

  /* System register configuration */
  sysreg = 0x00;
  if (tcxo == 0)       sysreg |= A2DM1X_SYSREG_PDTCXO; 
  if (ADST->test != 0) sysreg |= A2DM1X_SYSREG_TESTRMP;
  if (mport == 2)      sysreg |= A2DM1X_SYSREG_TESTCLK;
  if (ADST->bits < 0)  sysreg |= A2DM1X_SYSREG_CMPLX;
  if (mport == 3)      sysreg |= A2DM1X_SYSREG_INTREF;
  if (ADST->opts & RFOPT_XREF) sysreg |= A2DM1X_SYSREG_EXTREF;
  if (findintflag("MTGO",p->config) > 0) sysreg |= A2DM1X_SYSREG_MTGO;

  /* Disable output during configuration */
  sysreg &= (~A2DM1X_SYSREG_ENABLE); 
  pic_jpmemwr(p,A2DM1X_PORT_DSM,A2DM1X_SYSREG,sysreg,1);

  /* Output complex data with center frequency offset */
  /* pic_jpmemwr(p,A2DM1X_PORT_DSM,A2DM1X_FOFF+(mport-1)*4,ADST->foffset,4); */

  /* First, set PGA's to maximum attenuation for ADC calibration */
  pga1 = ADST->ch1pga;
  pga2 = ADST->ch2pga;
  ADST->ch1pga = -100;
  ADST->ch2pga = -100;
  rtn = a2dm1x_cfg_pga(p,mport,ADST);
#ifdef BOGUS
  ADST->ch1pga = pga1;
  ADST->ch2pga = pga2;
  if (rtn != 0) return (-1);
#endif

  /* A/D clock configuration */   
  rtn = a2dm1x_cfg_clk(p,A2DM1X_PORT_DSM,ADST);
  if (rtn != 0) return (-2);

  /* ADC12D1600 configuration */   
  rtn = a2dm1x_cfg_adc(p,A2DM1X_PORT_DSM,ADST);
  if (rtn != 0) return (-3);

  /* Now, set PGA's to requested gains */
  rtn = a2dm1x_cfg_pga(p,mport,ADST);
  if (rtn != 0) return (-1);

  /* Enable module output */
  sysreg |= A2DM1X_SYSREG_ENABLE; 
  pic_jpmemwr(p,A2DM1X_PORT_DSM,A2DM1X_SYSREG,sysreg,1);
  ADST->sysreg = sysreg;

  /* update state */
  a2dm1x_set_state(p,A2DM1X_PORT_DSM,ADST);

  return (0);
}

 
/*---------------------------------------------------------------------------*/
/* A2DM1X - Public Interfaces for Module Initialization, Setup, and Adjust   */
/*---------------------------------------------------------------------------*/
#ifdef BOGUS
int_4 a2dm1x_enable(PICSTRUCT *p, int_4 mport, int_4 ena) {
  A2DM1X_State_Obj ADST; 
  printf("ENABLE: %i\n",ena);
  a2dm1x_get_state(p,A2DM1X_PORT_DSM,&ADST);
  ADST.enable = (ena > 0) ? (1) : (0);
  if (ADST.enable > 0) 
    ADST.sysreg |= A2DM1X_SYSREG_ENABLE;
  else
    ADST.sysreg &= (~A2DM1X_SYSREG_ENABLE);
  pic_jpmemwr (p,mport,A2DM1X_SYSREG,ADST.sysreg,1);
  a2dm1x_set_state(p,A2DM1X_PORT_DSM,&ADST);
  return (0);
}
#endif
  
int_4 a2dm1x_get_gain(PICSTRUCT *p, int_4 mport)
{ 
  int_4 pga;
  A2DM1X_State_Obj ADST;     
  a2dm1x_get_state(p,A2DM1X_PORT_DSM,&ADST);
  pga = (mport == 2) ? (ADST.ch2pga) : (ADST.ch1pga);
  return (pga);
}

int_4 a2dm1x_set_gain(PICSTRUCT *p, int_4 mport, int_4 gain)
{ 
  int_4 stat = 0;
  A2DM1X_State_Obj ADST;     
  a2dm1x_get_state(p,A2DM1X_PORT_DSM,&ADST);
  if (mport == 2)
    ADST.ch2pga = gain;
  else
    ADST.ch1pga = gain;
  stat = a2dm1x_cfg_pga(p,mport,&ADST);
  a2dm1x_set_state(p,A2DM1X_PORT_DSM,&ADST);
  return (stat);
}

int_4 a2dm1x_get_opts (PICSTRUCT *p, int_4 mport)
{
  A2DM1X_State_Obj ADST;     
  a2dm1x_get_state(p,A2DM1X_PORT_DSM,&ADST);
  return ((int_4)ADST.opts);
}

int_4 a2dm1x_set_opts (PICSTRUCT *p, int_4 mport, int_4 flags)
{
  A2DM1X_State_Obj ADST;     
  a2dm1x_get_state(p,A2DM1X_PORT_DSM,&ADST);
  if (!(flags & RFOPT_ENABLE)) return (0); /* Options already enabled */
  a2dm1x_implement(p,mport,&ADST,0);
  return (0);
}


int_4 a2dm1x_setup(PICSTRUCT *p, int_4 mport, int_4 dir, int_4 bits, 
                     real_8 rate, int_4 gain, int_4 flags)
{ 
  int_4 i,iflag; 
  real_8 d;
  A2DM1X_State_Obj ADST; 

  mport = A2DM1X_PORT_DSM;  /* Override port for dual-site module */

  /* Initialize state information on module */
  a2dm1x_init_state(p,mport,&ADST);

  ADST.bits = bits; 
  if (mport == 2) 
    ADST.ch2pga = gain;
  else 
    ADST.ch1pga = gain;
  ADST.test = findintflagdef("A2DTEST",p->config,0);
  ADST.samprate = (bits<0)? rate*2 : rate;
  ADST.opts = findmaskflagdef("A2DOPTS",p->config,RFOPTLIST,0);

  /* Check for oversampling ratio */
  iflag = findintflag("MOVSR",p->config);
  iflag = (iflag > 0) ? (iflag) : (1);
  iflag = (iflag < 127) ? (iflag) : (127);
  ADST.ovsr = iflag;

  /* adjust center freq of hilbert output ? */
  d = finddblflag("MFREQ",p->config);
  if (d != -1.0) { 
    d /= (-ADST.samprate/4); 
    if (d> 1.0) d-=2.0;
    if (d<-1.0) d+=2.0;
    ADST.foffset = R2L(d);
   }

  /* setup base register modes - disable module output */
  pic_jpmemwr(p,mport,A2DM1X_SYSREG,0,1);

  /* Determine if external or internal reference is to be used */
  ADST.refclk = 10000000; 
  ADST.opts &= (~RFOPT_XREF);
  i = findintflag("PREFX",p->config);
  if (i > 0) ADST.opts |= RFOPT_XREF;
  if (i > 100) ADST.refclk = i; /* in Hz */
  else if (i > 1) ADST.refclk = i*1000000; /* in MHz */

  /* Determine if external clock is to be applied */
  if ((findflag("MUXCLK=A",p->config) >= 0) ||
      (findflag("MUXCLK=B",p->config) >= 0) ||
      (findflag("MUXCLK=N",p->config) >= 0))
  {
     printf ("A2DM1X: Error only MUXCLK=P allowed\n");
  }

  /* implement with enable */
  /* ADST.enable = 0; */
  a2dm1x_implement(p,mport,&ADST,-1);

  return (0);
}

