/*  
   Native network utilites for the ICE libraries


   Address specification format is    local IF:port / MCast Address:port

   send   udp:192.168.201.105:7777/224.1.40.87:7777

   recv   udp:192.168.201.105:7777/224.1.40.87:7777

*/
#if _LINX

#define DEBUG 0

#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#define MAXCLIENT 8
#define TXBUFSIZE 262080
#define RXBUFSIZE 8388608

int clients=0;
int clientsocks[MAXCLIENT];
struct sockaddr_in clientaddrs[MAXCLIENT];

void nudelay (int usec);
int_4 getaddr (int_4 sock, int_4 chan, struct sockaddr_in *sa);
int_4 addr2str (const struct sockaddr_in *sa, char *str, int_4 len);
int_4 formAddr (const char *host, int port, struct sockaddr_in *address);

int_4 nioOpen (char *url, int_4 chans, int_4 flags)
{
  int_4 i,status,reuse=1,unblock=0,sock,net,port,oport;
  unsigned int bufsize;
  char murl[64],name[64],oname[64];
  struct sockaddr_in clientaddr,serveraddr;
  struct ip_mreq group;
  socklen_t rlen,len;

  strcpy(murl,url); for (i=0; i<strlen(murl); i++) if (murl[i]==':') murl[i]=' ';
  sscanf(murl,"udp %s %d/%s %d",name,&port,oname,&oport);
  if (strlen(oname)==0) strcpy(oname,"localhost");

#if DEBUG
  printf("nioOpen url=%s chans=%d flags=%x addr=%s port=%d oaddr=%s oport=%d \n",url,chans,flags,name,port,oname,oport);
#endif

  sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sock<0) { perror("Err opening socket\n"); return -1; }

  serveraddr.sin_family = AF_INET;
  serveraddr.sin_port = htons(port);
  serveraddr.sin_addr.s_addr = (strlen(name)==0)? INADDR_ANY : inet_addr(name);

  inet_aton(oname,&group.imr_multiaddr);
  inet_aton( name,&group.imr_interface);
  net = group.imr_multiaddr.s_addr & 0xFF;

  /* join multicast group */
  if (flags < 0) {
    if (net>=224 && net<=240) {
      status = setsockopt (sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &group, sizeof(group));
#if DEBUG
      printf("Joining multicast group=%s on %s status=%d\n",oname,name,status); 
#endif
      if (status<0) { perror("Err joining multicast group\n"); return -5; }
      serveraddr.sin_addr.s_addr = group.imr_multiaddr.s_addr; /* pull interface from here strangely enough */
    }
    unblock = 1;
  } 

  status = bind (sock, (struct sockaddr *)&serveraddr, sizeof(serveraddr));
  if (status<0) { perror("Err binding socket\n"); return -4; }

  status = ioctl (sock, FIONBIO, &unblock);
  if (status<0) { perror("Err unblocking socket\n"); return -2; }

  len = sizeof(bufsize);
  status = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &reuse, len);
  if (status<0) { perror("Err setting port reuse\n"); return -3; }

  bufsize=RXBUFSIZE;
  status = setsockopt (sock, SOL_SOCKET, SO_RCVBUF, &bufsize, len);
  if (status<0) { perror("Err sizing receive socket buffer\n"); return -3; }

  bufsize=TXBUFSIZE;
  status = setsockopt (sock, SOL_SOCKET, SO_SNDBUF, &bufsize, len);
  if (status<0) { perror("Err sizing send socket buffer\n"); return -3; }

  rlen=len; status = getsockopt (sock, SOL_SOCKET, SO_RCVBUF, &bufsize, &rlen);
  if (bufsize<RXBUFSIZE) printf ("RCV buffer is only %d bytes\n",bufsize);

  rlen=len; status = getsockopt (sock, SOL_SOCKET, SO_SNDBUF, &bufsize, &rlen);
  if (bufsize<TXBUFSIZE) printf ("SND buffer is only %d bytes\n",bufsize);

  /* store output address */
  if (flags>=0) {
    formAddr (oname,oport,&clientaddr);
    addr2str(&clientaddr,oname,64);
#if DEBUG
    printf("Forming output address=%s\n",oname); 
#endif
    for (i=0; i<clients && clientsocks[i]!=sock; i++);
    clientaddrs[i] = clientaddr;
    clientsocks[i] = sock;
    if (i>=clients) clients++;
  }

  return sock;
}

int_4 nioClose (int_4 sock)
{
  int_4 status = close (sock);
#if DEBUG
  printf("Closing sock=%d status=%d\n",sock,status); 
#endif
  return status;
}

void* nioAlloc (int_4 bytes)
{
  void* pdata = malloc(bytes);
  return pdata;
}

int_4 nioFree (void* pdata)
{
  free(pdata);
  return 0;
}

int_4 nioSend (int_4 sock, int_4 chan, void* vaddr, int_4 bytes, int_4 flags)
{
  int_4 status;
  struct sockaddr_in address;
  socklen_t addrlen = getaddr(sock,chan,&address);
  status = sendto (sock, vaddr, bytes, 0,  (struct sockaddr *)&address, addrlen);
#if DEBUG
  printf("nioSend sock=%d chan=%d vaddr=%p bytes=%d flags=%x status=%d\n",sock,chan,vaddr,bytes,flags,status);
#endif
  if (status<=-1) return 0;
  if (flags>0) nudelay (flags);
  return status;
}

int_4 nioRecv (int_4 sock, int_4 chan, void* vaddr, int_4 bytes, int_4 flags)
{
  int_4 status;
  socklen_t addrlen;
  struct sockaddr_in address;
  status = recvfrom (sock, vaddr, bytes, 0,  (struct sockaddr *)&address, &addrlen);
#if DEBUG
  printf("nioRecv sock=%d chan=%d vaddr=%p bytes=%d flags=%x status=%d\n",sock,chan,vaddr,bytes,flags,status);
#endif
  if (status<=-1) {
    if (errno==EAGAIN) return 0;
    perror("nioRecv ");
  }
  return status;
}

int_4 formAddr (const char *host, int_4 port, struct sockaddr_in *address) 
{
  struct hostent *hostent;
  hostent = gethostbyname(host);
  if (hostent==NULL) {
    printf("formAddr(): gethostbyname host=%s returned null\n",host);
    return -1;
  }
  address->sin_family = hostent->h_addrtype;
  address->sin_port = htons(port);
  memcpy(&address->sin_addr, hostent->h_addr, hostent->h_length);
  return 0;
}

int_4 addr2str (const struct sockaddr_in *sa, char *str, int_4 lstr) 
{
  int_4 retstrlen = 0;
  const struct in_addr *sin_addr = &sa->sin_addr; /* in network byte-order */
  const int_u1 *addr = (void*) sin_addr; /* cast to byte array to easily get IP address */
  u_int16_t port = ntohs(sa->sin_port);  /* convert from network byte-order to host */
  retstrlen = (int_4) snprintf(str, (size_t)lstr, "%u.%u.%u.%u:%d", addr[0], addr[1], addr[2], addr[3], port);
  return retstrlen;
}

int_4 getaddr (int_4 sock, int_4 chan, struct sockaddr_in *sa) 
{
  int_4 i, addrlen=sizeof(*sa);
  int_u1 *addr = (int_u1 *)&sa->sin_addr;
  for (i=0; i<clients && clientsocks[i]!=sock; i++);
  if (i>=clients) return 0;
  memcpy (sa,clientaddrs+i,addrlen); 
  if (chan+addr[3]>255) addr[2]++;
  addr[3] += chan; /* adjust addr for channel number */
  return addrlen;
}

void nudelay (int usec) {
  int ksec,udelay; struct timeval tv,tw;
  if (usec>=10000) { ksec = usec/10000*10000; usleep(ksec); usec-=ksec; }
  if (usec<=0) return;
  gettimeofday(&tv,0);
  TRYAGAIN:
  gettimeofday(&tw,0);
  udelay = (tw.tv_sec-tv.tv_sec)*1000000 + tw.tv_usec - tv.tv_usec;
  if (udelay < usec) goto TRYAGAIN;
}

#else

int_4 nioOpen (char *url, int_4 chans, int_4 flags)
{
  printf("nioOpen stub url=%s chans=%d flags=%x\n",url,chans,flags);
  return 1;
}

int_4 nioClose (int_4 sock)
{
  printf("nioClose stub sock=%d\n",sock);
  return 0;
}

void* nioAlloc (int_4 bytes)
{
  printf("nioAlloc stub bytes=%d\n",bytes);
  return 0;
}

int_4 nioFree (void* vaddr)
{
  printf("nioFree stub vaddr=%p\n",vaddr);
  return 0;
}

int_4 nioSend (int_4 sock, int_4 chan, void* vaddr, int_4 bytes, int_4 flags)
{
  printf("nioSend stub sock=%d chan=%d vaddr=%p bytes=%d flags=%x\n",sock,chan,vaddr,bytes,flags);
  return 0;
}

int_4 nioRecv (int_4 sock, int_4 chan, void* vaddr, int_4 bytes, int_4 flags)
{
  printf("nioRecv stub sock=%d chan=%d vaddr=%p bytes=%d flags=%x\n",sock,chan,vaddr,bytes,flags);
  return 0;
}

#endif
