/****************************************************************************/
/*                                                                          */
/*   ADC12D interface routines                                              */
/*                                                                          */
/****************************************************************************/

#include "part_adc12d.h"

/*--------------------------------------------------------------------------*/
uint32_t adc12d_pt_get(ADC12D_Reg_Obj *P, int32_t idreg, uint32_t keymask)
{
  uint16_t km;
  uint16_t urg16;         
  uint32_t urg32, ureg;
  if (idreg == ADC12D_RG_PIN) {
    urg32 = P->pinval;
    ureg  = u32_getbits(keymask,urg32);
  }
  else {
    if ((idreg < 0)||(idreg >= ADC12D_RG_MAXCNT)) return (0);
    urg16 = P->rgval[idreg];
    km = (uint16_t) (keymask & 0x0000FFFF);
    ureg = u16_getbits(km,urg16);
  }
  return (ureg);
}


uint32_t adc12d_pt_gethw(PICSTRUCT *PS, int32_t port, ADC12D_Reg_Obj *P, 
                           int32_t idreg, uint32_t keymask)
{
  uint32_t rgadd, rgval, ureg;
  if (idreg == ADC12D_RG_PIN) {
    rgadd = P->addr[PTCFG_PIN_BASE];
    rgval = (0x000000FF)&(pic_jpmemrd(PS, port, rgadd, 1));
    P->pinval = rgval;
  }
  else {
    if ((idreg < 0)||(idreg >= ADC12D_RG_MAXCNT)) return (0);
    rgadd = (P->addr[PTCFG_RG_BASE])|(P->addr[PTCFG_RG_RDOFF])|(P->rgadd[idreg])
;
    rgval = (0x0000FFFF)&(pic_jpmemrd(PS, port, rgadd, 2));
    P->rgadd[idreg] = rgval;
  }
  ureg = adc12d_pt_get(P,idreg,keymask);
  return (ureg);
}

 
int32_t adc12d_pt_set(ADC12D_Reg_Obj *P, int32_t idreg, 
                           uint32_t keymask, uint32_t keyval)
{
  uint16_t km = (uint16_t) (keymask & 0x0000FFFF);
  uint16_t kv = (uint16_t) (keyval  & 0x0000FFFF);
  if (idreg == ADC12D_RG_PIN) {
    P->pinval = (P->pinval & (~keymask))|u32_setbits(keymask,keyval);
  }
  else {
    if ((idreg < 0)||(idreg >= ADC12D_RG_MAXCNT)) return (-1);
    P->rgval[idreg] = (P->rgval[idreg] & (~km))|u16_setbits(km,kv);
  }
  return(0);
}

 
int32_t adc12d_pt_sethw(PICSTRUCT *PS, int32_t port, ADC12D_Reg_Obj *P, 
                          int32_t idreg, uint32_t keymask, uint32_t keyval)
{
  int32_t  rtn;
  uint32_t rgadd, rgval;
  if (keymask != 0x00000000) {
    adc12d_pt_set(P, idreg, keymask, keyval);
  }
  if (idreg == ADC12D_RG_PIN) {
    rgadd = P->addr[PTCFG_PIN_BASE];
    rgval = P->pinval;
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 2);
  }
  else {
    if ((idreg < 0)||(idreg >= ADC12D_RG_MAXCNT)) return (-1);
    rgadd = (P->addr[PTCFG_RG_BASE])|(P->addr[PTCFG_RG_WROFF])|(P->rgadd[idreg]);
    rgval = P->rgval[idreg];
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 2);
  }
  udelay(100);
  return(rtn);
}
 

int32_t adc12d_pt_addr(ADC12D_Reg_Obj *P, uint32_t addr_key, uint32_t addr)
{
  int32_t rtn = -1;
  if (addr_key < PTCFG_ADDR_MAX) {
    P->addr[addr_key] = addr; 
    rtn = 0;
  }
  return (rtn);
}


int32_t adc12d_pt_init(ADC12D_Reg_Obj *P, int32_t module_type)
{
  int32_t i;

  uint16_t adc12d_a2dm18_reginit[ADC12D_RG_MAXCNT][2] = {
         {(0x0<<1),  0x2000},
         {(0x1<<1),  0x2907}, 
         {(0x2<<1),  0x0000},  
         {(0x3<<1),  0x4000},
         {(0x4<<1),  0xDB4B}, 
         {(0x5<<1),  0x0000},
         {(0x6<<1),  0x1C2E},   
         {(0x7<<1),  0x8142},
         {(0x8<<1),  0x0F0F}, 
         {(0x9<<1),  0x0000},
         {(0xA<<1),  0x0000}, 
         {(0xB<<1),  0x4000},
         {(0xC<<1),  0x0004}, 
         {(0xD<<1),  0x0000},
         {(0xE<<1),  0x0003}, 
         {(0xF<<1),  0x001D} };

  /* Default structure contents */
  P->id         = ADC12D_ID;
  P->mtype      = module_type;
  P->debuglevel = 0;
  P->pinval     = 0x00000000;

  /* Initialize base and offset addresses */ 
  for (i=0; i<PTCFG_ADDR_MAX; i++)  P->addr[i] = 0x00000000;
  
  /* Future case logic for different module types would go here */
  /* Assign all of the address space and default values */
  for (i=0; i<ADC12D_RG_MAXCNT; i++) {
    P->rgadd[i] = adc12d_a2dm18_reginit[i][0];
    P->rgval[i] = adc12d_a2dm18_reginit[i][1];
  }

  return (0);
}


int32_t adc12d_pt_debug(ADC12D_Reg_Obj *P)
{
  int32_t i;

  if (P->id != ADC12D_ID) {
    printf("\n");
    printf("WARNING: ADC12D object not initialized.\n");
    printf("\n");
    return (-1);
  }
  else {
    printf("\n");
    printf("ADC12D Debug ...\n");
    printf("  id         = 0x%8.8X \n",P->id);        
    printf("  mtype      = %i      \n",P->mtype);     
    printf("  debuglevel = %i      \n",P->debuglevel);
    printf("  pinval     = 0x%8.8X \n",P->pinval);    
    printf("  addr[PTCFG_PIN_BASE] = 0x%8.8X \n",P->addr[PTCFG_PIN_BASE]);
    printf("  addr[PTCFG_RG_BASE]  = 0x%8.8X \n",P->addr[PTCFG_RG_BASE]);
    printf("  addr[PTCFG_RG_WROFF] = 0x%8.8X \n",P->addr[PTCFG_RG_WROFF]);
    printf("  addr[PTCFG_RG_RDOFF] = 0x%8.8X \n",P->addr[PTCFG_RG_RDOFF]);
    if (P->debuglevel > 1) {
      for (i=0; i<ADC12D_RG_MAXCNT; i++)
        printf("  ADF12D_RG  REG: %3i   ADD: 0x%8.8X   VAL: 0x%4.4X\n",
                i, P->rgadd[i], P->rgval[i]);
    }
    printf("\n");
  }
  return (0);  
}
