#ifndef _PART_ADC12D_H
#define _PART_ADC12D_H

#include "part_common.h"

/*---------------------------------------------------------------------------*/
/* ADC12D Object */
#define ADC12D_ID        0x2748  /* 16-bit unique part identifier */
#define ADC12D_RG_MAXCNT 16      /* Total ADC12D config/status registers */
typedef struct
{
  uint16_t id;         /* Unique part ID (validates structure contents)      */
  int8_t   mtype;      /* Module type                                        */
  int8_t   debuglevel; /* Diagnostic output level                            */
  uint16_t pinval;                   /* ADC12D Pin I/O Register        */
  uint16_t addr[PTCFG_ADDR_MAX];     /* Base and offset addresses      */
  uint16_t rgadd[ADC12D_RG_MAXCNT];  /* ADC12D Register Address offset */
  uint16_t rgval[ADC12D_RG_MAXCNT];  /* ADC12D Register Value          */

} ADC12D_Reg_Obj;

/* ADC12D Register ID's */
#define ADC12D_RG_PIN        16   /* User Control of Device Pins via FPGA */

/* ADC12D Register ID's */
#define ADC12D_RG_0h_CFG      0 
#define ADC12D_RG_1h_RESV     1
#define ADC12D_RG_2h_IOFF     2
#define ADC12D_RG_3h_IFSRA    3 
#define ADC12D_RG_4h_CALADJ   4
#define ADC12D_RG_5h_CALVAL   5
#define ADC12D_RG_6h_RESV     6
#define ADC12D_RG_7h_DESADJ   7
#define ADC12D_RG_8h_RESV     8
#define ADC12D_RG_9h_RESV     9
#define ADC12D_RG_Ah_QOFF    10 
#define ADC12D_RG_Bh_QFSRA   11 
#define ADC12D_RG_Ch_CAPDLY  12 
#define ADC12D_RG_Dh_FAPDLY  13 
#define ADC12D_RG_Eh_ASYNC   14 
#define ADC12D_RG_Fh_RESV    15 

/* ADC12D Pin Keys      */
#define ADC12D_RG_PIN_NDM 0x0001  /* Non-Demux Mode {0=Demux, 1=No Demux  */
#define ADC12D_RG_PIN_CAL 0x0002  /* CALRUN calibration start/stop status */

/* ADC12D Register Keys */
#define ADC12D_RG_0h_CAL  0x8000  /* Calibration enable / disable / reset  */
#define ADC12D_RG_0h_DPS  0x4000  /* DCLK Phase Select {0=0deg, 1=90deg}   */
#define ADC12D_RG_0h_OVS  0x2000  /* LVDS Output Voltage Select            */
#define ADC12D_RG_0h_TPM  0x1000  /* Test Pattern Mode {0=normal, 1=test}  */
#define ADC12D_RG_0h_PDI  0x0800  /* Power down I-channel                  */
#define ADC12D_RG_0h_PDQ  0x0400  /* Power down Q-channel                  */
#define ADC12D_RG_0h_LFS  0x0100  /* LowFreq Select {0=300MHz, 1=>300MHz}  */
#define ADC12D_RG_0h_DES  0x0080  /* Dual Edge Sampling {0=NonDES, 1=DES}  */
#define ADC12D_RG_0h_DEQ  0x0040  /* DES mode input select {0=I, 1=Q}      */
#define ADC12D_RG_0h_DIQ  0x0020  /* DESIQ mode {0=normal, 1=I&Q Shorted}  */
#define ADC12D_RG_0h_2SC  0x0010  /* Two's compliment {0=OB, 1=2Comp}      */
#define ADC12D_RG_0h_TSE  0x0008  /* Time stamp enable                     */
#define ADC12D_RG_0h_SDR  0x0004  /* Single data rate force {0=DDR, 1=SDR} */
#define ADC12D_RG_2h_IOS  0x1000  /* I-Channel offset sign {0=pos, 1=neg}  */
#define ADC12D_RG_2h_IOM  0x0FFF  /* I-Channel offset magnitude (2048 max) */
#define ADC12D_RG_3h_IFM  0x7FFF  /* I-Channel full scale adjust magnitude */
#define ADC12D_RG_4h_CSS  0x4000  /* Cal Seq Select {1=normal, 0=no reset} */
#define ADC12D_RG_4h_SSC  0x0080  /* SPI Scan Ctl {0=default, 1=read 5h}   */
#define ADC12D_RG_5h_SS   0xFFFF  /* Cal Seq results                       */
#define ADC12D_RG_7h_DTA  0xFE00  /* DES Timing Adjust (30fs increments)   */
#define ADC12D_RG_Ah_QOS  0x1000  /* Q-Channel offset sign {0=pos, 1=neg}  */
#define ADC12D_RG_Ah_QOM  0x0FFF  /* Q-Channel offset magnitude (2048 max) */
#define ADC12D_RG_Bh_IFM  0x7FFF  /* Q-Channel full scale adjust magnitude */
#define ADC12D_RG_Ch_CAM  0xFFF0  /* Coarse Adjust Delay Magnitude         */
#define ADC12D_RG_Ch_STA  0x0008  /* Tadj feature {0=off, 1=on}            */
#define ADC12D_RG_Ch_DCC  0x0004  /* Duty cycle correct {1=on, 0=disable}  */
#define ADC12D_RG_Dh_FAM  0xFC00  /* Fine aperature delay adjust magnitude */
#define ADC12D_RG_Eh_DRC  0xFF00  /* Delay reference clock magnitude       */
#define ADC12D_RG_Eh_DCK  0x0040  /* DESCLK-IQ mode                        */
#define ADC12D_RG_Eh_SP   0x0018  /* Refclk phase latch {0,90,180,270}     */
#define ADC12D_RG_Eh_ES   0x0004  /* Enable slave {1=enable, 0=disable}    */
#define ADC12D_RG_Eh_DOC  0x0002  /* Disable output ref clocks             */
#define ADC12D_RG_Eh_DR   0x0001  /* Disable reset pin                     */


#define ADC12D_FCLK_MAX 1600000000.0   
#define ADC12D_FCLK_MIN  137500000.0 

#endif  /* _PART_ADC12D_H */
