/****************************************************************************/
/*                                                                          */
/*   ADF4350 interface routines                                             */
/*                                                                          */
/****************************************************************************/

#include "part_adf4350.h"

/*--------------------------------------------------------------------------*/
uint32_t adf4350_pt_get(ADF4350_Reg_Obj *P, int32_t idreg, uint32_t keymask)
{
  uint32_t urg32, ureg;
  if (idreg == ADF4350_RG_PIN) {
    urg32 = P->pinval;         
    ureg  = u32_getbits(keymask,urg32);
  }
  else {
    if ((idreg < 0)||(idreg >= ADF4350_RG_MAXCNT)) return (0);
    urg32 = P->rgval[idreg];
    ureg  = u32_getbits(keymask,urg32);
  }
  return (ureg);
}


uint32_t adf4350_pt_gethw(PICSTRUCT *PS, int32_t port, ADF4350_Reg_Obj *P,
                            int32_t idreg, uint32_t keymask)
{
  uint32_t rgadd, rgval, ureg;
  if (idreg == ADF4350_RG_PIN) {
    rgadd = P->addr[PTCFG_PIN_BASE];
    rgval = (0x000000FF)&(pic_jpmemrd(PS, port, rgadd, 1));
    P->pinval = rgval;
  }
  else {
    if ((idreg < 0)||(idreg >= ADF4350_RG_MAXCNT)) return (0);
    rgadd = (P->addr[PTCFG_RG_BASE])|(P->addr[PTCFG_RG_RDOFF])|(P->rgadd[idreg]);
    rgval = (0xFFFFFFFF)&(pic_jpmemrd(PS, port, rgadd, 4));
    P->rgadd[idreg] = rgval;
  }
  ureg = u32_getbits(keymask,rgval);
  return (ureg);
}


int32_t adf4350_pt_set(ADF4350_Reg_Obj *P, int32_t idreg, uint32_t keymask, 
                       uint32_t keyval)
{
  if (idreg == ADF4350_RG_PIN) {
    P->pinval = (P->pinval & (~keymask))|u32_setbits(keymask,keyval);
  }
  else {
    if ((idreg < 0)||(idreg >= ADF4350_RG_MAXCNT)) return (-1);
    P->rgval[idreg] = (P->rgval[idreg] & (~keymask))|u32_setbits(keymask,keyval);
  }
  return(0);
}

 
int32_t adf4350_pt_sethw(PICSTRUCT *PS, int32_t port, ADF4350_Reg_Obj *P, 
                          int32_t idreg, uint32_t keymask, uint32_t keyval)
{
  int32_t  rtn;
  uint32_t rgadd, rgval;
  if (keymask != 0x00000000) {
    adf4350_pt_set(P, idreg, keymask, keyval);
  }
  if (idreg == ADF4350_RG_PIN) {
    rgadd = P->addr[PTCFG_PIN_BASE];
    rgval = P->pinval;
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 1);
  }
  else {
    if ((idreg < 0)||(idreg >= ADF4350_RG_MAXCNT)) return (-1);
    rgadd = (P->addr[PTCFG_RG_BASE])|(P->addr[PTCFG_RG_WROFF])|(P->rgadd[idreg]);
    rgval = P->rgval[idreg];
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 4);
  }
  udelay(100);
  return(rtn);
}


int32_t adf4350_pt_addr(ADF4350_Reg_Obj *P, uint32_t addr_key, uint32_t addr)
{
  int32_t rtn = -1;
  if (addr_key < PTCFG_ADDR_MAX) {
    P->addr[addr_key] = addr; 
    rtn = 0;
  }
  return (rtn);
}


int32_t adf4350_pt_init(ADF4350_Reg_Obj *P, int32_t module_type)
{
  int32_t  i;

  uint32_t adf4350_a2dm18_reginit[ADF4350_RG_MAXCNT][2] = {
           {0x00000000, 0x04B00000},
           {0x00000000, 0x00004001}, 
           {0x00000000, 0x0F02B142},  
           {0x00000000, 0x00001003},
           {0x00000000, 0x00C205FC}, 
           {0x00000000, 0x00580005} };

  /* Configure the fields */
  P->id         = ADF4350_ID;
  P->debuglevel = 0;
  P->mtype      = 0;
  P->validflag  = -1;
  P->fdes       = 150000000.0;
  P->fsynth     = 150000000.0;
  P->fref       = 10000000.0;
  P->fpd        = 10000000.0;
  P->fvco       = ADF4350_VCO_FMAX;
  P->bscd       = 1;
  P->R          = 1;
  P->D          = 1;
  P->T          = 1;
  P->U          = 1;
  P->N          = 1;
  P->F          = 1;
  P->M          = 1;
  P->pinval     = 0x00000000;
 
  /* Initialize base and offset addresses */ 
  for (i=0; i<PTCFG_ADDR_MAX; i++)  P->addr[i] = 0x00000000;
  
  /* Future case logic for different module types would go here */
  /* Assign all of the address space and default values */
  for (i=0; i<ADF4350_RG_MAXCNT; i++) {
    P->rgadd[i] = adf4350_a2dm18_reginit[i][0];
    P->rgval[i] = adf4350_a2dm18_reginit[i][1];
  }

  return (0);
}


int32_t adf4350_pt_debug(ADF4350_Reg_Obj *P)
{
  int32_t i;

  if (P->id != ADF4350_ID) {
    printf("\n");
    printf("WARNING: ADF4350 object not initialized.\n");
    printf("\n");
    return (-1);
  }
  else {
    printf("\n");
    printf("ADF4350 Debug ...\n");
    printf("  id         = 0x%8.8X \n",P->id);        
    printf("  debuglevel = %i      \n",P->debuglevel);
    printf("  mtype      = %i      \n",P->mtype);     
    printf("  validflag  = %i      \n",P->validflag); 
    printf("  fdes       = %20.6f  \n",P->fdes); 
    printf("  fsynth     = %20.6f  \n",P->fsynth);
    printf("  fref       = %20.6f  \n",P->fref);
    printf("  fpd        = %20.6f  \n",P->fpd); 
    printf("  fvco       = %20.6f  \n",P->fvco);
    printf("  bscd       = %u      \n",P->bscd);
    printf("  R          = %u      \n",P->R);
    printf("  D          = %u      \n",P->D);
    printf("  T          = %u      \n",P->T);
    printf("  U          = %u      \n",P->U);
    printf("  N          = %u      \n",P->N);
    printf("  F          = %u      \n",P->F);
    printf("  M          = %u      \n",P->M);
    printf("  pinval     = 0x%8.8X \n",P->pinval);    
    printf("  addr[PTCFG_PIN_BASE] = 0x%8.8X \n",P->addr[PTCFG_PIN_BASE]);
    printf("  addr[PTCFG_RG_BASE]  = 0x%8.8X \n",P->addr[PTCFG_RG_BASE]);
    printf("  addr[PTCFG_RG_WROFF] = 0x%8.8X \n",P->addr[PTCFG_RG_WROFF]);
    printf("  addr[PTCFG_RG_RDOFF] = 0x%8.8X \n",P->addr[PTCFG_RG_RDOFF]);
    if (P->debuglevel > 1) {
      for (i=0; i<ADF4350_RG_MAXCNT; i++)
        printf("  ADF4350_RG  REG: %3i   ADD: 0x%8.8X   VAL: 0x%8.8X\n",
                i, P->rgadd[i], P->rgval[i]);
    }
    printf("\n");
  }
  return (0);
}


int32_t adf4350_pt_findsoln(ADF4350_Reg_Obj *P)
{
  float64_t pdratio, dfrac;
  int32_t  nratio;
  uint32_t U_idx;

  /* First, test reference to make certain it is in valid range */
  if ((P->fref < 10000000.0)||(P->fref > 250000000.0)) return (-2);

  /* Always disable the reference doubler to reduce noise and jitter */
  P->D = 0;

  /* Always enable reference half to force 50% duty cycle */
  P->T = 1;
  P->fpd = 0.5*(P->fref);

  /* Now select the highest reference frequency below 32MHz */
  P->R = (uint32_t) ((P->fpd/32000000.0)+1.0);
  P->R = (P->R > 0) ? (P->R) : (1);
  P->fpd = (P->fpd)/(P->R);

  /* Now compute the dividers */
  P->fsynth = (P->fdes >= ADF4350_VCO_FSYNTH_MIN) ? (P->fdes) : (ADF4350_VCO_FSYNTH_MIN); 
  P->fsynth = (P->fsynth <= ADF4350_VCO_FMAX) ? (P->fsynth) : (ADF4350_VCO_FMAX);

  /* Compute RF divider */ 
  { P->U = 1; U_idx = 0x00; }
  if (P->fsynth <= 2200000000.0) { P->U = 2;  U_idx = 0x01; }
  if (P->fsynth <= 1100000000.0) { P->U = 4;  U_idx = 0x02; }
  if (P->fsynth <=  550000000.0) { P->U = 8;  U_idx = 0x03; }
  if (P->fsynth <=  275000000.0) { P->U = 16; U_idx = 0x04; }

  /* Compute nominal VCO frequency */
  P->fvco = (P->fsynth)*(P->U);

  /* Fix modulus at 3000 */
  P->M = 3000;

  /* Solve for nearest valid VCO frequency */
  pdratio = (P->fvco)/(P->fpd);
  nratio  = (int32_t) (pdratio);
  nratio  = (nratio > 23) ? (nratio) : (23);
  nratio  = (nratio < 65535) ? (nratio) : (65535);
  P->N    = nratio;
  dfrac   = pdratio - ((float64_t) nratio);
  dfrac   = (dfrac > 0.0) ? (dfrac) : (0.0);
  dfrac   = (dfrac < 1.0) ? (dfrac) : (0.0);
  P->F    = (uint32_t) (0.5 + dfrac*((float64_t) P->M));
  if (P->F == P->M) {
    P->F = 0;
    P->N = P->N + 1;
  }

  /* Compute the exact synthesizer frequency */
  pdratio = ((float64_t)P->N) + (((float64_t)P->F)/((float64_t)P->M));
  P->fsynth = (1.0/((float64_t)P->U))*(P->fpd*pdratio);

  /* Band select clock divider (62.5kHz is target) */
  nratio = (int32_t) (0.5+(2.0*(P->fpd)/62500.0));
  nratio = (nratio >=   1) ? (nratio) : (1);
  P->bscd = (nratio <= 255) ? (nratio) : (255);

  /* Set valid / invalid solution flag */
  dfrac = P->fdes - P->fsynth;
  dfrac = (dfrac > 0.0) ? (dfrac) : (-1.0*dfrac);
  if (dfrac < 0.001) 
    P->validflag = 0;
  else if (dfrac < 500.0)
    P->validflag = 1;
  else
    P->validflag = -1;

  /* Setup the configuration registers */
  adf4350_pt_set(P, 4, ADF4350_RG_4h_BS_CLKDIV,  P->bscd);
  adf4350_pt_set(P, 4, ADF4350_RG_4h_U,          U_idx);
  adf4350_pt_set(P, 3, ADF4350_RG_3h_FL_CLKDIV,  0x0800);
  adf4350_pt_set(P, 3, ADF4350_RG_3h_FL_CLKMODE, 0x00);
  adf4350_pt_set(P, 3, ADF4350_RG_3h_CSR,        0x00);
  adf4350_pt_set(P, 2, ADF4350_RG_2h_LDF,        0x00);
  if (P->F == 0) adf4350_pt_set(P, 2, ADF4350_RG_2h_LDF, 0x01);
  adf4350_pt_set(P, 2, ADF4350_RG_2h_R,          P->R);
  adf4350_pt_set(P, 2, ADF4350_RG_2h_D,          P->D);
  adf4350_pt_set(P, 2, ADF4350_RG_2h_T,          P->T);
  adf4350_pt_set(P, 1, ADF4350_RG_1h_M,          P->M);
  adf4350_pt_set(P, 1, ADF4350_RG_1h_PHASE,      0x0000);
  adf4350_pt_set(P, 1, ADF4350_RG_1h_PRESCALER,  0x00);
  adf4350_pt_set(P, 0, ADF4350_RG_0h_F,          P->F);
  adf4350_pt_set(P, 0, ADF4350_RG_0h_N,          P->N);

  return (P->validflag);
}
