#ifndef _PART_ADF4350_H
#define __PART_ADF4350_H

#include "part_common.h"

/*---------------------------------------------------------------------------*/
/* ADF4350 Object */
#define ADF4350_ID        0xD7C6 /* 16-bit unique part identifier      */
#define ADF4350_RG_MAXCNT 6      /* Maximum of 6 configuration registers */
typedef struct
{
  int32_t id;         /* Part identifier (validates structure contents)     */
  int8_t  debuglevel; /* Diagnostic output level                            */
  int8_t  mtype;      /* Module type                                        */
  int8_t  validflag;  /* -1 = no solution, 0 = solution, 1 = imperfect soln */
  int8_t  pad_align;  /* Align structure to boundary                        */
  float64_t fdes;      /* Desired output frequency (Hz)                      */
  float64_t fsynth;    /* Exact output frequency of synthesizer (Hz)         */
  float64_t fref;      /* Input reference frequency (Hz)                     */
  float64_t fpd;       /* PLL comparison frequency (Hz)                      */
  float64_t fvco;      /* VCO center frequency (Hz)                          */
  uint32_t bscd;      /* VCO band select clock divider {1,255}              */
  uint32_t R;         /* R reference divider {1,1023}                       */
  uint32_t D;         /* D reference doubler enable / disable {0,1}         */
  uint32_t T;         /* T reference half divider enable / disable {0,1}    */
  uint32_t U;         /* U VCO RF output divider {1,16}                     */
  uint32_t N;         /* N VCO integer divider {23,65535}                   */
  uint32_t F;         /* F VCO fractional divider numerator {0,4095}        */
  uint32_t M;         /* M VCO fractional divider denomenator {0,4095}      */
  uint16_t pinval;                   /* ADF4350 Pin I/O Register         */
  uint16_t addr[PTCFG_ADDR_MAX];     /* Base and offset addresses        */
  uint16_t rgadd[ADF4350_RG_MAXCNT]; /* ADF4350 Register Address Offset  */
  uint32_t rgval[ADF4350_RG_MAXCNT]; /* ADF4350 Register Value           */

} ADF4350_Reg_Obj;

/* ADF4350 PIN Register ID */
#define ADF4350_RG_PIN  6         /* User Control of Device Pins via FPGA */

/* ADF4350 Register ID's */
#define ADF4350_RG_0h   0 
#define ADF4350_RG_1h   1
#define ADF4350_RG_2h   2
#define ADF4350_RG_3h   3 
#define ADF4350_RG_4h   4
#define ADF4350_RG_5h   5

/* ADF4350 Pin Keys      */
#define ADF4350_RG_PIN_LD        0x00000010 /* Lock detect pin               */

/* ADF4350 Register Keys */
#define ADF4350_RG_4h_RF_PWR     0x00000018 /* RF output power {0,3}          */
#define ADF4350_RG_4h_RF_ENABLE  0x00000020 /* RF output enable / disable     */
#define ADF4350_RG_4h_AUX_PWR    0x000000C0 /* AUX output power {0,3}         */
#define ADF4350_RG_4h_AUX_ENABLE 0x00000100 /* AUX output enable / disable    */
#define ADF4350_RG_4h_AUX_SELECT 0x00000200 /* AUX from RF divider or VCO     */
#define ADF4350_RG_4h_BS_CLKDIV  0x000FF000 /* Band select clock div {1,255}  */
#define ADF4350_RG_4h_U          0x00700000 /* RF output divider {1,16}       */
#define ADF4350_RG_3h_FL_CLKDIV  0x00007FF8 /* Fast lock clock divide {0,4095}*/
#define ADF4350_RG_3h_FL_CLKMODE 0x00018000 /* FL Clock mode {0,3}            */
#define ADF4350_RG_3h_CSR        0x00020000 /* Cycle slip enable / disable    */
#define ADF4350_RG_2h_LDF        0x00000100 /* Lock detect Integer / Fraction */
#define ADF4350_RG_2h_CP_CURR    0x00001E00 /* Charge Pump Current {0,15}     */
#define ADF4350_RG_2h_R          0x00FFC000 /* R reference divider            */
#define ADF4350_RG_2h_D          0x02000000 /* D reference doubler            */
#define ADF4350_RG_2h_T          0x01000000 /* T reference half               */
#define ADF4350_RG_2h_MUXOUT     0x1C000000 /* MUX output selection {0,7}     */
#define ADF4350_RG_1h_M          0x00007FF8 /* M fractional denominator       */
#define ADF4350_RG_1h_PHASE      0x07FF8000 /* M fractional denominator       */
#define ADF4350_RG_1h_PRESCALER  0x08000000 /* Prescaler select {0,1}         */
#define ADF4350_RG_0h_F          0x00007FF8 /* F fractional divider numerator */
#define ADF4350_RG_0h_N          0x7FFF8000 /* N integer divider              */

/* ADF4350 Parameters */
#define ADF4350_VCO_FMAX 4400000000.0
#define ADF4350_VCO_FMIN 2200000000.0
#define ADF4350_VCO_FSYNTH_MAX 4400000000.0
#define ADF4350_VCO_FSYNTH_MIN  137500000.0 

#endif  /* _PART_ADF4350_H */
