/****************************************************************************/
/*                                                                          */
/*   iomlib_common - libraries common to parts and modules                  */
/*                                                                          */
/****************************************************************************/

#include "part_common.h"


/*****************************************************************************/
/* Setbit/Getbit routines for register content manipulation                  */
/*  - Setbit writes a "uval" value into the "umask" position of the register */
/*  - Getbit reads a "uval" value from the "umask" position in the register  */
/*  - "umask" position does not need to be contiguous (helpful for SiLabs)   */
/*  - Warning: these are little endian routines, but at least it isolates    */
/*             the endian issues in one place for future updates             */
/*****************************************************************************/
/*----------------------------------------------------------------------------*/
uint8_t u8_setbits(uint8_t umask, uint8_t uval)
{
  uint8_t i;
  uint8_t im = 0x01;
  uint8_t om = 0x01;
  uint8_t rtnval = 0x00;
  for (i=0; i<(8*sizeof(uint8_t)); i++) {
    if (umask & om) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      im = (im << 1);
    } 
    om = (om << 1);
  }
  return (rtnval);
}

uint8_t u8_getbits(uint8_t umask, uint8_t uval)
{
  uint8_t i;
  uint8_t im = 0x01;
  uint8_t om = 0x01;
  uint8_t rtnval = 0x00;
  for (i=0; i<(8*sizeof(uint8_t)); i++) {
    if (umask & im) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      om = (om << 1);
    } 
    im = (im << 1);
  }
  return (rtnval);
}

uint16_t u16_setbits(uint16_t umask, uint16_t uval)
{
  uint16_t i;
  uint16_t im = 0x0001;
  uint16_t om = 0x0001;
  uint16_t rtnval = 0x0000;
  for (i=0; i<(8*sizeof(uint16_t)); i++) {
    if (umask & om) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      im = (im << 1);
    } 
    om = (om << 1);
  }
  return (rtnval);
}

uint16_t u16_getbits(uint16_t umask, uint16_t uval)
{
  uint16_t i;
  uint16_t im = 0x0001;
  uint16_t om = 0x0001;
  uint16_t rtnval = 0x0000;
  for (i=0; i<(8*sizeof(uint16_t)); i++) {
    if (umask & im) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      om = (om << 1);
    } 
    im = (im << 1);
  }
  return (rtnval);
}

uint32_t u32_setbits(uint32_t umask, uint32_t uval)
{
  uint32_t i;
  uint32_t im = 0x00000001;
  uint32_t om = 0x00000001;
  uint32_t rtnval = 0x00000000;
  for (i=0; i<(8*sizeof(uint32_t)); i++) {
    if (umask & om) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      im = (im << 1);
    } 
    om = (om << 1);
  }
  return (rtnval);
}

uint32_t u32_getbits(uint32_t umask, uint32_t uval)
{
  uint32_t i;
  uint32_t im = 0x00000001;
  uint32_t om = 0x00000001;
  uint32_t rtnval = 0x00000000;
  for (i=0; i<(8*sizeof(uint32_t)); i++) {
    if (umask & im) {
      rtnval = (uval & im) ? (rtnval | om) : (rtnval & (~om));
      om = (om << 1);
    } 
    im = (im << 1);
  }
  return (rtnval);
}

