/****************************************************************************/
/*                                                                          */
/*   PE4302 interface routines                                             */
/*                                                                          */
/****************************************************************************/

#include "part_pe4302.h"

/*--------------------------------------------------------------------------*/
uint32_t pe4302_pt_get(PE4302_Reg_Obj *P, int32_t idreg, uint32_t keymask)
{
  uint32_t urg32, ureg;
  if (idreg == PE4302_RG_PIN) {
    urg32 = P->pinval;         
    ureg  = u32_getbits(keymask,urg32);
  }
  else {
    if ((idreg < 0)||(idreg >= PE4302_RG_MAXCNT)) return (0);
    urg32 = P->rgval[idreg];
    ureg  = u32_getbits(keymask,urg32);
  }
  return (ureg);
}


uint32_t pe4302_pt_gethw(PICSTRUCT *PS, int32_t port, PE4302_Reg_Obj *P,
                            int32_t idreg, uint32_t keymask)
{
  /* There are no pins or registers to read on this part - write only */
  return (0x00000000);
}


int32_t pe4302_pt_set(PE4302_Reg_Obj *P, int32_t idreg, uint32_t keymask, 
                       uint32_t keyval)
{
  if (idreg == PE4302_RG_PIN) {
    P->pinval = (P->pinval & (~keymask))|u32_setbits(keymask,keyval);
  }
  else {
    if ((idreg < 0)||(idreg >= PE4302_RG_MAXCNT)) return (-1);
    P->rgval[idreg] = (P->rgval[idreg] & (~keymask))|u32_setbits(keymask,keyval);
  }
  return(0);
}

 
int32_t pe4302_pt_sethw(PICSTRUCT *PS, int32_t port, PE4302_Reg_Obj *P, 
                          int32_t idreg, uint32_t keymask, uint32_t keyval)
{
  int32_t  rtn;
  uint32_t rgadd, rgval;
  if (keymask != 0x00000000) {
    pe4302_pt_set(P, idreg, keymask, keyval);
  }
  if (idreg == PE4302_RG_PIN) {
    rgadd = P->addr[PTCFG_PIN_BASE];
    rgval = P->pinval;
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 1);
  }
  else {
    if ((idreg < 0)||(idreg >= PE4302_RG_MAXCNT)) return (-1);
    rgadd = (P->addr[PTCFG_RG_BASE])|(P->addr[PTCFG_RG_WROFF])|(P->rgadd[idreg]);
    rgval = P->rgval[idreg];
    rtn = pic_jpmemwr(PS, port, rgadd, rgval, 1);
  }
  udelay(100);
  return(rtn);
}


int32_t pe4302_pt_addr(PE4302_Reg_Obj *P, uint32_t addr_key, uint32_t addr)
{
  int32_t rtn = -1;
  if (addr_key < PTCFG_ADDR_MAX) {
    P->addr[addr_key] = addr; 
    rtn = 0;
  }
  return (rtn);
}


int32_t pe4302_pt_init(PE4302_Reg_Obj *P, int32_t module_type)
{
  int32_t i;

  /* Configure the fields */
  P->id         = PE4302_ID;
  P->mtype      = module_type;
  P->debuglevel = 0;
  P->pinval     = 0x00000000;
  P->rgadd[0]   = 0x00000000;
  P->rgval[0]   = 0x00000000;
 
  /* Initialize base and offset addresses */ 
  for (i=0; i<PTCFG_ADDR_MAX; i++)  P->addr[i] = 0x00000000;
  
  return (0);
}


int32_t pe4302_pt_debug(PE4302_Reg_Obj *P)
{
  int32_t i;

  if (P->id != PE4302_ID) {
    printf("\n");
    printf("WARNING: PE4302 object not initialized.\n");
    printf("\n");
    return (-1);
  }
  else {
    printf("\n");
    printf("PE4302 Debug ...\n");
    printf("  id         = 0x%8.8X \n",P->id);        
    printf("  mtype      = %i      \n",P->mtype);     
    printf("  debuglevel = %i      \n",P->debuglevel);
    printf("  pinval     = 0x%8.8X \n",P->pinval);    
    printf("  addr[PTCFG_PIN_BASE] = 0x%8.8X \n",P->addr[PTCFG_PIN_BASE]);
    printf("  addr[PTCFG_RG_BASE]  = 0x%8.8X \n",P->addr[PTCFG_RG_BASE]);
    printf("  addr[PTCFG_RG_WROFF] = 0x%8.8X \n",P->addr[PTCFG_RG_WROFF]);
    printf("  addr[PTCFG_RG_RDOFF] = 0x%8.8X \n",P->addr[PTCFG_RG_RDOFF]);
    if (P->debuglevel > 1) {
      for (i=0; i<PE4302_RG_MAXCNT; i++)
        printf("  PE4302_RG  REG: %3i   ADD: 0x%8.8X   VAL: 0x%8.8X\n",
                i, P->rgadd[i], P->rgval[i]);
    }
    printf("\n");
  }
  return (0);
}

