#define  SONET_NUM_PRG_BITS                32
#define  SONET_MODRESET_REMOVE     0x00000001  /* Must Be Used For Module To Come Out Of Reset               */
#define  SONET_ENABLE_ERROR_DETECT 0x00000002  /* Enable Frame Errors Detection, Only Collect Good Frames    */
#define  SONET_DISABLE_SCRAMBLING  0x00000004  /* Turn Off Fiber Optic Descrambling (Rx) or Scrambling (Tx)  */
#define  SONET_OC12ENABLE          0x00000008  /* Use If Line Rate Is 622MBits/Sec, Don't Use For 155 MBits  */
#define  SONET_OC3FROMOC12_ENABLE  0x00000010  /* Pull Out OC3 Signal From OC12 Input, Bits Below Choose OC3 */ 
#define  SONET_OC3_0_FROMOC12      0x00000000  /* Choose OC3 #0 From OC12 For Acquisition                    */
#define  SONET_OC3_1_FROMOC12      0x00000020  /* Choose OC3 #1 From OC12 For Acquisition                    */
#define  SONET_OC3_2_FROMOC12      0x00000040  /* Choose OC3 #2 From OC12 For Acquisition                    */
#define  SONET_OC3_3_FROMOC12      0x00000060  /* Choose OC3 #3 From OC12 For Acquisition                    */
#define  SONET_INPUT_ENABLE        0x00000080  /* Enable Module For Input                                    */
#define  SONET_RXCLKFORTX          0x00000100  /* Use Rx Recovered Clock Instead Of Ref Clk For Tx Output    */
/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_reset_sonetr1 (PICSTRUCT *p, int_4 mport)
{
  /* This Puts Module Into Reset By Setting All Config Bits = 0 */
   pic_wmodreg(p,mport,0,0,SONET_NUM_PRG_BITS); 
   return 0;
}
/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_setup_sonet (PICSTRUCT *p, int_4 mport, int_4 dir,int_4 rate, int_4 gain) 
{
   int_4 i; int_u4 sonet_cfg_bits = 0;

   /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SONET_NUM_PRG_BITS); 
   
   /* To Configure the module for operation 'bit wise or' in configuration values                    */
   /* This Line Must Always Be Here, It Take the module out of reset mode and starts Data Acquistion */
   sonet_cfg_bits = sonet_cfg_bits | SONET_MODRESET_REMOVE; 

   /* Turn Off Scrambling/ Descrambling         */
    if (findflag("SCRAMOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SONET_DISABLE_SCRAMBLING;

   /* Check For OC12 Data In Open String        */
    if (findflag("OC12",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SONET_OC12ENABLE;
            
   /* Check For OC3 Remove From OC12            */
    if ((findflag("OC3FROMOC12",p->config) >= 0) || (findflag("OC3FROMOCX",p->config) >= 0))
      {
        sonet_cfg_bits = sonet_cfg_bits | SONET_OC3FROMOC12_ENABLE;
        if ((i=findflag("OC3NUM",p->config)) >= 0)
          {
            if (p->config[i+6] != '=')    sonet_cfg_bits = sonet_cfg_bits | SONET_OC3_0_FROMOC12;
            else if (p->config[i+7]=='1') sonet_cfg_bits = sonet_cfg_bits | SONET_OC3_1_FROMOC12;
            else if (p->config[i+7]=='2') sonet_cfg_bits = sonet_cfg_bits | SONET_OC3_2_FROMOC12;
            else if (p->config[i+7]=='3') sonet_cfg_bits = sonet_cfg_bits | SONET_OC3_3_FROMOC12;
            else                          sonet_cfg_bits = sonet_cfg_bits | SONET_OC3_0_FROMOC12;
          }  
      } 
 
  
   /* Set Module For Acquisition If Desired     */
    if (dir <= 0)
      {
        sonet_cfg_bits = sonet_cfg_bits | SONET_INPUT_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | SONET_RXCLKFORTX; /* Autotmatic For Rx Data To Loop To Tx */
      } 
    else if (findflag("RXCLKFORTX",p->config) >= 0)
      sonet_cfg_bits = sonet_cfg_bits | SONET_RXCLKFORTX; 

   /* Error Check Configuration Setup */
    if ((sonet_cfg_bits & SONET_OC3FROMOC12_ENABLE) != 0) {
      if ((sonet_cfg_bits & SONET_OC12ENABLE) == 0) {
         printf("icelib.c, pic_setup_sonet():OC12 Must Be Enabled For OC3 Removal");
        }
      } 
         
   vprintf2("Setup SONET port=%d cfgbits=%d\n",mport,sonet_cfg_bits);
   /* Download Bit Wise ORed Bits To Module  */
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SONET_NUM_PRG_BITS);  
   return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
/*--------------------------------------------------------------*/
#define  SR2_NUM_PRG_BITS                32
#define  SR2_MODRESET_REMOVE     0x00000001  /* Must Be Used For Module To Come Out Of Reset               */
#define  SR2_ENABLE_ERROR_DETECT 0x00000002  /* Enable Frame Errors Detection, Only Collect Good Frames    */
#define  SR2_DISABLE_SCRAMBLING  0x00000004  /* Turn Off Fiber Optic Descrambling (Rx) or Scrambling (Tx)  */
#define  SR2_INPUT_ENABLE        0x00000008  /* Enable Module For Input                                    */
#define  SR2_RXCLKFORTX          0x00000010  /* Use Rx Recovered Clock Instead Of Ref Clk For Tx Output    */
#define  SR2_OC3ENABLE           0x00000020  /* Use If Line Rate Of 155 MBits                              */
#define  SR2_OC12ENABLE          0x00000040  /* Use If Line Rate Is 622MBits/Sec                           */
#define  SR2_OC24ENABLE          0x00000080  /* Use If Line Rate Is 1244MBits/Sec                          */
#define  SR2_OC48ENABLE          0x00000100  /* Use If Line Rate Is 2488MBits/Sec                          */
#define  SR2_DEMUXENABLE         0x00000200  /* Enable Realtime Demux Of Incoming Signal                   */
#define  SR2_OC3FROMOCX_ENABLE   0x00000400  /* Pull Out OC3  Signal From OCx  Input                       */ 
#define  SR2_OC12FROMOCX_ENABLE  0x00008000  /* Pull Out OC12 Signal From OCx  Input                       */
#define  SR2_OC24FROMOCX_ENABLE  0x00040000  /* Pull Out OC24 Signal From OC48 Input                       */
#define  SR2_SLAVE_ENABLE        0x00100000  /* First Frame Of Slave Module Is Demux #0 Channel            */
#define  SR2_MUXCFG_ENABLE       0x00100000  /* Mux Cfg Info From Port 0 To Port 1 - For SlimPIC           */
#define  SR2_MUXOUT_ENABLE       0x00200000  /* Mux Ouput From Port 1 To Port 0 - For SlimPIC              */

#define  SR2_DISABLE_FRAMING     0x00200000  /* Disable Frame Detection                                    */
#define  SR2_FEC_ENABLE          0x00400000  /* Use FEC Oscillator As Reference                            */
/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_reset_sonetr2 (PICSTRUCT *p, int_4 mport)
{
  /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
   pic_wmodreg(p,mport,0,0,SR2_NUM_PRG_BITS); 
   return 0;
}
/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_setup_sonetr2 (PICSTRUCT *p, int_4 mport, int_4 dir,int_4 rate, int_4 gain) 
{
   int_4 i; int_u4 sonet_cfg_bits = 0;

   /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR2_NUM_PRG_BITS); 
   
   /* To Configure the module for operation 'bit wise or' in configuration values                    */
   /* This Line Must Always Be Here, It Take the module out of reset mode and starts Data Acquistion */
   sonet_cfg_bits = sonet_cfg_bits | SR2_MODRESET_REMOVE; 

   /* Mux Config And Output Data To Port 1 */
    if (p->type == ICESLIC3)
      sonet_cfg_bits = sonet_cfg_bits | SR2_MUXCFG_ENABLE | SR2_MUXOUT_ENABLE;

   /* Turn ON Error Detect-Only Acquire Valid Frames             */
    if (findflag("ERRDETON",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR2_ENABLE_ERROR_DETECT; 
 
    if (findflag("FRAMINGOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR2_DISABLE_SCRAMBLING | SR2_DISABLE_FRAMING; 

    if (findflag("FECENABLE",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR2_FEC_ENABLE;

   /* Turn Off Scrambling/ Descrambling         */
    if (findflag("SCRAMOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR2_DISABLE_SCRAMBLING;

   /* Setup Type Of Signal To Acquire/Playback  */
    if      (findflag("OC12",p->config) >= 0)  /* Check For OC12 Data In Open String        */
       sonet_cfg_bits = sonet_cfg_bits | SR2_OC12ENABLE;
    else if (findflag("OC24",p->config) >= 0)  /* Check For OC24 Data In Open String        */
       sonet_cfg_bits = sonet_cfg_bits | SR2_OC24ENABLE;
    else if (findflag("OC48",p->config) >= 0)  /* Check For OC48 Data In Open String        */
       sonet_cfg_bits = sonet_cfg_bits | SR2_OC48ENABLE;
    else
       sonet_cfg_bits = sonet_cfg_bits | SR2_OC3ENABLE;
 /*  
   if (mport == 1) {
       if (((sonet_cfg_bits & SR2_OC48ENABLE) != 0) || ((sonet_cfg_bits & SR2_OC24ENABLE) != 0)) {
           printf("OC48 and OC24 Supported On Module Port #2 Only!\n");
           return(-1);
         }  
     }
 */           
   /* Check For OC3 Remove From OCx            */
    if ((findflag("OC3FROMOCX",p->config) >= 0) || (findflag("OC3FROMOC12",p->config) >= 0))
       { 
        /* Enable Demux Operation              */
        sonet_cfg_bits = sonet_cfg_bits | SR2_DEMUXENABLE;  

        /* Error Check OC3 Removal Operation   */
        if ((sonet_cfg_bits & SR2_OC3ENABLE) != 0) {
            printf("\n Error-OC12 or OC24 or OC48 Must Be Enabled For OC3 Removal\n");             
            return(-1); 
          }

        i=findintflag("OC3NUM",p->config); 
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */ 
            printf("\n Warning-OC3 Number For Removal Not Specified-Assuming OC3NUM=0\n");
            i = 0; 
          }
        if ((i >= 4) && ((sonet_cfg_bits & SR2_OC12ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC12 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
        if ((i >= 8) && ((sonet_cfg_bits & SR2_OC24ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 7\n");
            return(-1); 
          }
        if ((i >= 16) && ((sonet_cfg_bits & SR2_OC48ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 15\n");
            return(-1); 
          }
        sonet_cfg_bits = sonet_cfg_bits | SR2_OC3FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 11) & 0x00007800);            
      } 
 
      /* Check For OC12 Remove From OCx            */
    if (findflag("OC12FROMOCX",p->config) >= 0) { 
        /* Error Check OC12 Removal Operation      */
        if (((sonet_cfg_bits & SR2_OC3ENABLE) != 0) || ((sonet_cfg_bits & SR2_OC12ENABLE) != 0)) {
            printf("\n Error-OC24 or OC48 Must Be Enabled For OC12 Removal\n");             
            return(-1); 
          }     
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR2_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR2_DEMUXENABLE;          

        i=findintflag("OC12NUM",p->config);
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */
            printf("\n Warning-OC12 Number For Removal Not Specified-Assuming OC12NUM=0\n");
            i = 0; 
          }

        if ((i >= 2) && ((sonet_cfg_bits & SR2_OC24ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 1\n");
            return(-1); 
          }

        if ((i >= 4) && ((sonet_cfg_bits & SR2_OC48ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
            
        sonet_cfg_bits = sonet_cfg_bits | SR2_OC12FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 16) & 0x00030000);         
      } 
  
    /* Check For OC24 Remove From OC48            */
    if (findflag("OC24FROMOCX",p->config) >= 0) {
        /* Error Check OC24 Removal Operation   */
        if ((sonet_cfg_bits & SR2_OC48ENABLE) == 0) {
            printf("\n Error-OC48 Must Be Enabled For OC24 Removal\n");             
            return(-1); 
          }     
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR2_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR2_DEMUXENABLE; 

        i=findintflag("OC24NUM",p->config);
          
        if (i < 0) { /* Flag Not Found Assume 0 */
            printf("\n Warning-OC24 Number For Removal Not Specified-Assuming OC24NUM=0\n");
            i = 0;
          }
        if (i >= 2) { 
            printf("\n Error-OC24 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 1\n");
            return(-1); 
          }            
        sonet_cfg_bits = sonet_cfg_bits | SR2_OC24FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x00080000);            
      } 

    if ((p->type==ICEPIC4) && ((sonet_cfg_bits & SR2_OC48ENABLE) != 0) && ((sonet_cfg_bits & SR2_DEMUXENABLE) == 0)) {
        printf("\n Error-Raw OC48 Acquire/Playback Rates (320 MB/Sec) Too High For ICE-PIC4T!\n");
        printf("For Raw OC48 Acquire/Playback ICE-PIC5 Series Or Higher PCI Card Needed\n");  
        return(-1);  
      }
 
    if ((findflag("SLVENABLE",p->config) >= 0) && ((sonet_cfg_bits &  SR2_DEMUXENABLE) != 0) && (dir <= 0))
      sonet_cfg_bits = sonet_cfg_bits | SR2_SLAVE_ENABLE;

   /* Set Module For Acquisition If Desired Else Playback     */
    if (dir <= 0) {      
        sonet_cfg_bits = sonet_cfg_bits | SR2_INPUT_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | SR2_RXCLKFORTX; /* Autotmatic For Rx Data To Loop To Tx */
      }  
    else if (findflag("RXCLKFORTX",p->config) >= 0)
      sonet_cfg_bits = sonet_cfg_bits | SR2_RXCLKFORTX; 
         
   vprintf2("Setup SONETR2 port=%d cfgbits=%08x\n",mport,sonet_cfg_bits);
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR2_NUM_PRG_BITS);  /* Download Bit Wise ORed Bits To Module  */
   return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
/*--------------------------------------------------------------*/
#define  SR4_NUM_PRG_BITS                32
#define  SR4_MODRESET_REMOVE     0x00000001  /* Must Be Used For Module To Come Out Of Reset               */
#define  SR4_CARDTYPE            0x00000002  /* Set Card Type To PIC5T When Ready                          */
#define  SR4_ENABLE_ERROR_DETECT 0x00000004  /* Enable Frame Errors Detection, Only Collect Good Frames    */
#define  SR4_DISABLE_SCRAMBLING  0x00000008  /* Turn Off Fiber Optic Descrambling (Rx) or Scrambling (Tx)  */
#define  SR4_INPUT_ENABLE        0x00000010  /* Enable Module For Input                                    */
#define  SR4_SLAVE_ENABLE        0x00000020  /* First Frame Of Slave Module Is Demux #0 Channel            */
#define  SR4_DISABLE_FRAMING     0x00000040  /* Disable Search For Framing Bring In Data W/Unknown Framing */
#define  SR4_REFCLK_SEL0         0x00000080  /* Reference Clock Select For FEC Type                        */
#define  SR4_REFCLK_SEL1         0x00000100  /* Reference Clock Select For FEC Type                        */
#define  SR4_DEMUXENABLE         0x00000200  /* Enable Realtime Demux Of Incoming Signal                   */
#define  SR4_OC3_96ENABLE        0x00000400  /* A Side: Use If Line Rate  5GHz B Side: Rate 155 MBits      */
#define  SR4_OC12_192ENABLE      0x00000800  /* A Side: Use If Line Rate 10GHz B Side: Rate 622 MBits      */
#define  SR4_OC24ENABLE          0x00001000  /* B Side: Use If Line Rate Is 1244MBits/Sec                  */
#define  SR4_OC48ENABLE          0x00002000  /* B Side: Use If Line Rate Is 2488MBits/Sec                  */

#define  SR4_OC3FROMOCX_ENABLE   0x00004000  /* Pull Out OC3  Signal From OCx   Input                      */ 
#define  SR4_OC12FROMOCX_ENABLE  0x00008000  /* Pull Out OC12 Signal From OCx   Input                      */
#define  SR4_OC24FROMOCX_ENABLE  0x00010000  /* Pull Out OC24 Signal From OCx   Input                      */
#define  SR4_OC48FROMOCX_ENABLE  0x00020000  /* Pull Out OC48 Signal From OC192 Input                      */
#define  SR4_OC96FROMOCX_ENABLE  0x00040000  /* Pull Out OC48 Signal From OC192 Input                      */

/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_enable_sonetr4 (PICSTRUCT *p, int_4 mport,int_4 dis_enable)
{

  /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
   if (dis_enable == 0)
     {
       pic_wmodreg(p,mport,1,0,SR4_NUM_PRG_BITS); 
       udelay(100); 
     }
   else
     {
       pic_wmodreg(p,mport,1,1,SR4_NUM_PRG_BITS); 
       udelay(100);
     }
   return 0;
}


/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_setup_sonetr4 (PICSTRUCT *p, int_4 mport, int_4 dir,int_4 rate, int_4 gain) 
{
   int_4 i; int_u4 sonet_cfg_bits = 0;
   int_u4 mjtrdval = 0;

   /* This Puts Module Into Reset By Setting All Config Bits = 0 */     
   pic_wmodreg(p,mport,1,sonet_cfg_bits,SR4_NUM_PRG_BITS); 
   udelay(100);                  
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR4_NUM_PRG_BITS); 
   udelay(100);
    
   /* To Configure the module for operation 'bit wise or' in configuration values                    */
   /* This Line Must Always Be Here, It Take the module out of reset mode and starts Data Acquistion */
    sonet_cfg_bits = sonet_cfg_bits | SR4_MODRESET_REMOVE; 

   /* Check For Full Rate OCc192                                 */  
    if ((findflag("OC192FULL",p->config) >= 0) && mport == 1)
     { 
        
       pic_rmodreg (p,mport,0,&mjtrdval,SR4_NUM_PRG_BITS);
       if (mjtrdval != 0xAA55AA55)
         {
           printf("SonetR4: Wrong Firmware Download For OC192 Signal\n");
           printf("Use Flags 'IOMFPGA=_192full' AND 'FORCE' For Correct Download\n");
           return(-1); 
         }
       if ((findflag("G709FEC",p->config) >= 0) && mport == 1)
        sonet_cfg_bits = sonet_cfg_bits | SR4_REFCLK_SEL1;
 
      /* Set Module For Acquisition If Desired Else Playback     */
       if (dir <= 0)      
         sonet_cfg_bits = sonet_cfg_bits | SR4_INPUT_ENABLE;
   
       vprintf2("Setup SONETR4 port=%d cfgbits=%08x\n",mport,sonet_cfg_bits);
       pic_wmodreg(p,mport,0,sonet_cfg_bits,SR4_NUM_PRG_BITS);  /* Download Bit Wise ORed Bits To Module  */
       return(2);
     }

   /* Set Up ITU G.709 FEC Clock Ref Source     */      
    if ((findflag("G709FEC",p->config) >= 0) && mport == 1)
       sonet_cfg_bits = sonet_cfg_bits | SR4_REFCLK_SEL1; 
    if ((findflag("G709FEC",p->config) >= 0) && mport == 2)
       sonet_cfg_bits = sonet_cfg_bits | SR4_REFCLK_SEL0; 

   /* Turn ON Error Detect-Only Acquire Valid Frames             */
    if (findflag("ERRDETON",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR4_ENABLE_ERROR_DETECT; 
 
   /* Turn Off Scrambling/ Descrambling         */
    if (findflag("SCRAMOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR4_DISABLE_SCRAMBLING;

    /* Turn Off Scrambling/ Descrambling         */
    if (findflag("FRAMINGOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR4_DISABLE_FRAMING | SR4_DISABLE_SCRAMBLING; 

   
   /* Setup Type Of Signal To Acquire/Playback  */
    if (mport == 1) 
      {
        if (findflag("OC96",p->config) >= 0)  /* Check For OC96 Data In Open String        */
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC3_96ENABLE;
        else
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC12_192ENABLE;
      }
    else 
      { 
        if      (findflag("OC12",p->config) >= 0)  /* Check For OC12 Data In Open String        */
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC12_192ENABLE;
        else if (findflag("OC24",p->config) >= 0)  /* Check For OC24 Data In Open String        */
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC24ENABLE;
        else if (findflag("OC48",p->config) >= 0)  /* Check For OC48 Data In Open String        */
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC48ENABLE;
        else
          sonet_cfg_bits = sonet_cfg_bits | SR4_OC3_96ENABLE;
      }
              

   /* Check For OC3 Remove From OCx            */
    if ((findflag("OC3FROMOCX",p->config) >= 0) || (findflag("OC3NUM",p->config) >= 0))
       { 
        /* Enable Demux Operation              */
        sonet_cfg_bits = sonet_cfg_bits | SR4_DEMUXENABLE;  

        /* Error Check OC3 Removal Operation   */
        if (((sonet_cfg_bits & SR4_OC3_96ENABLE) != 0) && (mport == 2)){
            printf("\n Error-OC12 or OC24 or OC48 Must Be Enabled For OC3 Removal\n");             
            return(-1); 
          }

        i=findintflag("OC3NUM",p->config); 
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */ 
            printf("\n Warning-OC3 Number For Removal Not Specified-Assuming OC3NUM=0\n");
            i = 0; 
          }
        if ((i >= 4) && ((sonet_cfg_bits & SR4_OC12_192ENABLE) != 0) && (mport == 2)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC12 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
        if ((i >= 8) && ((sonet_cfg_bits & SR4_OC24ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 7\n");
            return(-1); 
          }
        if ((i >= 16) && ((sonet_cfg_bits & SR4_OC48ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 15\n");
            return(-1); 
          }

        if ((i >= 64) && ((sonet_cfg_bits & SR4_OC12_192ENABLE) != 0) && (mport == 1)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 63\n");
            return(-1); 
          }
        
        sonet_cfg_bits = sonet_cfg_bits | SR4_OC3FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);            
      } 
 
      /* Check For OC12 Remove From OCx            */
    if (findflag("OC12FROMOCX",p->config) >= 0) { 
        /* Error Check OC12 Removal Operation      */
  /*      if ((((sonet_cfg_bits & SR4_OC3_96ENABLE) != 0) || ((sonet_cfg_bits & SR4_OC12_192ENABLE) != 0)) && (mport == 2)) {
            printf("\n Error-OC24, Or OC48 Must Be Enabled For OC12 Removal\n");             
            return(-1); 
          }     
  */
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR4_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR4_DEMUXENABLE;          

        i=findintflag("OC12NUM",p->config);
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */
            printf("\n Warning-OC12 Number For Removal Not Specified-Assuming OC12NUM=0\n");
            i = 0; 
          }

        if ((i >= 2) && ((sonet_cfg_bits & SR4_OC24ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 1\n");
            return(-1); 
          }

        if ((i >= 4) && ((sonet_cfg_bits & SR4_OC48ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
         
        if ((i >= 16) && ((sonet_cfg_bits & SR4_OC12_192ENABLE) != 0) && (mport == 1)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 15\n");
            return(-1); 
          }  

        sonet_cfg_bits = sonet_cfg_bits | SR4_OC12FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);         
      } 
  
    /* Check For OC24 Remove From OC48            */
    if (findflag("OC24FROMOCX",p->config) >= 0) {
        /* Error Check OC24 Removal Operation   */
      /*  if (((sonet_cfg_bits & SR4_OC48ENABLE) == 0) && (mport == 2)){
            printf("\n Error-OC48 Must Be Enabled For OC24 Removal\n");             
            return(-1); 
          }
      */     
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR4_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR4_DEMUXENABLE; 

        i=findintflag("OC24NUM",p->config);
          
        if (i < 0) { /* Flag Not Found Assume 0 */
            printf("\n Warning-OC24 Number For Removal Not Specified-Assuming OC24NUM=0\n");
            i = 0;
          }
        if (i >= 2) { 
            printf("\n Error-OC24 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 1\n");
            return(-1); 
          }            

         if ((i >= 8) && ((sonet_cfg_bits & SR4_OC12_192ENABLE) != 0) && (mport == 1)) {
            printf("\n Error-OC24 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 7\n");
            return(-1); 
          }   
        sonet_cfg_bits = sonet_cfg_bits | SR4_OC24FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);            
      } 

     if (findflag("OC48FROMOCX",p->config) >= 0) {
        /* Error Check OC24 Removal Operation   */
      /*  if (((sonet_cfg_bits & SR4_OC48ENABLE) == 0) && (mport == 2)){
            printf("\n Error-OC48 Must Be Enabled For OC24 Removal\n");             
            return(-1); 
          }
      */     
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR4_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR4_DEMUXENABLE; 

        i=findintflag("OC48NUM",p->config);
          
        if (i < 0) { /* Flag Not Found Assume 0 */
            printf("\n Warning-OC48 Number For Removal Not Specified-Assuming OC48NUM=0\n");
            i = 0;
          }
        if (i >= 4) { 
            printf("\n Error-OC48 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
            return(-1); 
          }            

     
        sonet_cfg_bits = sonet_cfg_bits | SR4_OC48FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);            
      } 

    if ((p->type==ICEPIC4) && ((sonet_cfg_bits & SR4_OC48ENABLE) != 0) && ((sonet_cfg_bits & SR4_DEMUXENABLE) == 0)) {
        printf("\n Error-Raw OC48 Acquire/Playback Rates (320 MB/Sec) Too High For ICE-PIC4T!\n");
        printf("For Raw OC48 Acquire/Playback ICE-PIC5 Series Or Higher PCI Card Needed\n");  
        return(-1);  
      }
 
    if ((findflag("SLVENABLE",p->config) >= 0) && ((sonet_cfg_bits &  SR4_DEMUXENABLE) != 0) && (dir <= 0))
      sonet_cfg_bits = sonet_cfg_bits | SR4_SLAVE_ENABLE;

   /* Set Module For Acquisition If Desired Else Playback     */
    if (dir <= 0)      
      sonet_cfg_bits = sonet_cfg_bits | SR4_INPUT_ENABLE;
   
   vprintf2("Setup SONETR4 port=%d cfgbits=%08x\n",mport,sonet_cfg_bits);
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR4_NUM_PRG_BITS);  /* Download Bit Wise ORed Bits To Module  */
   return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
/*--------------------------------------------------------------*/
#define  SR5_NUM_PRG_BITS                32
#define  SR5_MODRESET_REMOVE     0x00000001  /* Must Be Used For Module To Come Out Of Reset               */
#define  SR5_CARDTYPE            0x00000002  /* Set Card Type To PIC5T When Ready                          */
#define  SR5_ENABLE_ERROR_DETECT 0x00000004  /* Enable Frame Errors Detection, Only Collect Good Frames    */
#define  SR5_DISABLE_SCRAMBLING  0x00000008  /* Turn Off Fiber Optic Descrambling (Rx) or Scrambling (Tx)  */
#define  SR5_INPUT_ENABLE        0x00000010  /* Enable Module For Input                                    */
#define  SR5_SLAVE_ENABLE        0x00000020  /* First Frame Of Slave Module Is Demux #0 Channel            */
#define  SR5_DISABLE_FRAMING     0x00000040  /* Disable Search For Framing Bring In Data W/Unknown Framing */
#define  SR5_REFCLK_SEL0         0x00000080  /* Reference Clock Select For FEC Type                        */
#define  SR5_REFCLK_SEL1         0x00000100  /* Reference Clock Select For FEC Type                        */
#define  SR5_DEMUXENABLE         0x00000200  /* Enable Realtime Demux Of Incoming Signal                   */
#define  SR5_OC3_96ENABLE        0x00000400  /* A Side: Use If Line Rate  5GHz B Side: Rate 155 MBits      */
#define  SR5_OC12_192ENABLE      0x00000800  /* A Side: Use If Line Rate 10GHz B Side: Rate 622 MBits      */
#define  SR5_OC24ENABLE          0x00001000  /* B Side: Use If Line Rate Is 1244MBits/Sec                  */
#define  SR5_OC48ENABLE          0x00002000  /* B Side: Use If Line Rate Is 2488MBits/Sec                  */

#define  SR5_OC3FROMOCX_ENABLE   0x00004000  /* Pull Out OC3  Signal From OCx   Input                      */ 
#define  SR5_OC12FROMOCX_ENABLE  0x00008000  /* Pull Out OC12 Signal From OCx   Input                      */
#define  SR5_OC24FROMOCX_ENABLE  0x00010000  /* Pull Out OC24 Signal From OCx   Input                      */
#define  SR5_OC48FROMOCX_ENABLE  0x00020000  /* Pull Out OC48 Signal From OC192 Input                      */
#define  SR5_OC96FROMOCX_ENABLE  0x00040000  /* Pull Out OC48 Signal From OC192 Input                      */

/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_enable_sonetr5 (PICSTRUCT *p, int_4 mport,int_4 dis_enable)
{

  /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
   if (dis_enable == 0)
     {
       pic_wmodreg(p,mport,1,0,SR5_NUM_PRG_BITS); 
       udelay(100); 
     }
   else
     {
       pic_wmodreg(p,mport,1,1,SR5_NUM_PRG_BITS); 
       udelay(100);
     }
   return 0;
}
/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^*/
int_4 pic_setup_sonetr5 (PICSTRUCT *p, int_4 mport, int_4 dir,int_4 rate, int_4 gain) 
{
   int_4 i; int_u4 sonet_cfg_bits = 0;

 
   /* This Puts Module Into Reset By Setting All Config Bits = 0 */  
    /* This Puts Module Into Reset By Setting All Config Bits = 0 */     
   pic_wmodreg(p,mport,1,sonet_cfg_bits,SR5_NUM_PRG_BITS);
   udelay(100);                  
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR5_NUM_PRG_BITS); 
   udelay(100);
    
   /* To Configure the module for operation 'bit wise or' in configuration values                    */
   /* This Line Must Always Be Here, It Take the module out of reset mode and starts Data Acquistion */
    sonet_cfg_bits = sonet_cfg_bits | SR5_MODRESET_REMOVE; 

   /* Set Up ITU G.709 FEC Clock Ref Source     */      
    if (findflag("G709FEC",p->config) >= 0) 
       sonet_cfg_bits = sonet_cfg_bits | SR5_REFCLK_SEL0; 

   /* Turn ON Error Detect-Only Acquire Valid Frames             */
    if (findflag("ERRDETON",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR5_ENABLE_ERROR_DETECT; 
 
   /* Turn Off Scrambling/ Descrambling         */
    if (findflag("SCRAMOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR5_DISABLE_SCRAMBLING;

    /* Turn Off Scrambling/ Descrambling         */
    if (findflag("FRAMINGOFF",p->config) >= 0)
       sonet_cfg_bits = sonet_cfg_bits | SR5_DISABLE_FRAMING | SR5_DISABLE_SCRAMBLING; 

   
   /* Setup Type Of Signal To Acquire/Playback  */
    if      (findflag("OC12",p->config) >= 0)  /* Check For OC12 Data In Open String        */
      sonet_cfg_bits = sonet_cfg_bits | SR5_OC12_192ENABLE;
    else if (findflag("OC24",p->config) >= 0)  /* Check For OC24 Data In Open String        */
      sonet_cfg_bits = sonet_cfg_bits | SR5_OC24ENABLE;
    else if (findflag("OC48",p->config) >= 0)  /* Check For OC48 Data In Open String        */
      sonet_cfg_bits = sonet_cfg_bits | SR5_OC48ENABLE;
    else
      sonet_cfg_bits = sonet_cfg_bits | SR5_OC3_96ENABLE;
      
              

   /* Check For OC3 Remove From OCx            */
    if ((findflag("OC3FROMOCX",p->config) >= 0) || (findflag("OC3FROMOC12",p->config) >= 0))
       { 
        /* Enable Demux Operation              */
        sonet_cfg_bits = sonet_cfg_bits | SR5_DEMUXENABLE;  

        /* Error Check OC3 Removal Operation   */
        if ((sonet_cfg_bits & SR5_OC3_96ENABLE) != 0) {
            printf("\n Error-OC12 or OC24 or OC48 Must Be Enabled For OC3 Removal\n");             
            return(-1); 
          }

        i=findintflag("OC3NUM",p->config); 
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */ 
            printf("\n Warning-OC3 Number For Removal Not Specified-Assuming OC3NUM=0\n");
            i = 0; 
          }
        if ((i >= 4) && ((sonet_cfg_bits & SR5_OC12_192ENABLE) != 0))  {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC12 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
        if ((i >= 8) && ((sonet_cfg_bits & SR5_OC24ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 7\n");
            return(-1); 
          }
        if ((i >= 16) && ((sonet_cfg_bits & SR5_OC48ENABLE) != 0)) {
            printf("\n Error-OC3 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 15\n");
            return(-1); 
          }

      
        sonet_cfg_bits = sonet_cfg_bits | SR5_OC3FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);            
      } 
 
      /* Check For OC12 Remove From OCx            */
    if (findflag("OC12FROMOCX",p->config) >= 0) { 
        /* Error Check OC12 Removal Operation      */
  /*      if ((((sonet_cfg_bits & SR5_OC3_96ENABLE) != 0) || ((sonet_cfg_bits & SR5_OC12_192ENABLE) != 0)) && (mport == 2)) {
            printf("\n Error-OC24, Or OC48 Must Be Enabled For OC12 Removal\n");             
            return(-1); 
          }     
  */
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR5_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR5_DEMUXENABLE;          

        i=findintflag("OC12NUM",p->config);
          
        if (i < 0) { /* Did Not Find Flag, Assume Zero */
            printf("\n Warning-OC12 Number For Removal Not Specified-Assuming OC12NUM=0\n");
            i = 0; 
          }

        if ((i >= 2) && ((sonet_cfg_bits & SR5_OC24ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 1\n");
            return(-1); 
          }

        if ((i >= 4) && ((sonet_cfg_bits & SR5_OC48ENABLE) != 0)) {
            printf("\n Error-OC12 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
            return(-1); 
          }
         
        sonet_cfg_bits = sonet_cfg_bits | SR5_OC12FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);         
      } 
  
    /* Check For OC24 Remove From OC48            */
    if (findflag("OC24FROMOCX",p->config) >= 0) {
        /* Error Check OC24 Removal Operation   */
      /*  if (((sonet_cfg_bits & SR5_OC48ENABLE) == 0) && (mport == 2)){
            printf("\n Error-OC48 Must Be Enabled For OC24 Removal\n");             
            return(-1); 
          }
      */     
        /* Error Check For More Than 1 Demux Op   */
        if ((sonet_cfg_bits & SR5_DEMUXENABLE) != 0) {
            printf("\n Error-Only 1 Demux Operation Allowed\n");             
            return(-1);  
          }        
        else
          sonet_cfg_bits = sonet_cfg_bits | SR5_DEMUXENABLE; 

        i=findintflag("OC24NUM",p->config);
          
        if (i < 0) { /* Flag Not Found Assume 0 */
            printf("\n Warning-OC24 Number For Removal Not Specified-Assuming OC24NUM=0\n");
            i = 0;
          }
        if (i >= 2) { 
            printf("\n Error-OC24 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 1\n");
            return(-1); 
          }            

     
        sonet_cfg_bits = sonet_cfg_bits | SR5_OC24FROMOCX_ENABLE;
        sonet_cfg_bits = sonet_cfg_bits | ((i << 19) & 0x001f80000);            
      } 

    if ((p->type==ICEPIC4) && ((sonet_cfg_bits & SR5_OC48ENABLE) != 0) && ((sonet_cfg_bits & SR5_DEMUXENABLE) == 0)) {
        printf("\n Error-Raw OC48 Acquire/Playback Rates (320 MB/Sec) Too High For ICE-PIC4T!\n");
        printf("For Raw OC48 Acquire/Playback ICE-PIC5 Series Or Higher PCI Card Needed\n");  
        return(-1);  
      }
 
    if ((findflag("SLVENABLE",p->config) >= 0) && ((sonet_cfg_bits &  SR5_DEMUXENABLE) != 0) && (dir <= 0))
      sonet_cfg_bits = sonet_cfg_bits | SR5_SLAVE_ENABLE;

   /* Set Module For Acquisition If Desired Else Playback     */
    if (dir <= 0)      
      sonet_cfg_bits = sonet_cfg_bits | SR5_INPUT_ENABLE;
   
   vprintf2("Setup SONETR5 port=%d cfgbits=%08x\n",mport,sonet_cfg_bits);
   pic_wmodreg(p,mport,0,sonet_cfg_bits,SR5_NUM_PRG_BITS);  /* Download Bit Wise ORed Bits To Module  */
   
   return 0;
}
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
/*--------------------------------------------------------------*/
#define  SR6_NUM_PRG_BITS                32

/* Register Address Definitions */
#define SR6_RESETREG_ADR         0x00000000
#define SR6_ENDISREG_ADR         0x00010000
#define SR6_CFGREG_ADR           0x00020000
#define SR6_SONETREG_ADR         0x00030000
#define SR6_TRCVRREG_ADR         0x00040000
#define SR6_TCLKCFGREG_ADR       0x00050000
#define SR6_DCLKCFGREG_ADR       0x00060000
#define SR6_CBCFGREG_ADR         0x000B0000
#define SR6_CBRSTREG_ADR         0x000C0000
#define SR6_TRCVRRAM_ADR         0x000D0000
#define SR6_HOSTMEM_ADR          0x000F0000 

/* Reset Register, Register 0, Bit Definitions          */
#define SR6_RESET_BIT            0x00000001  /* Remove Reset Single Port       */       
#define SR6_GLBRESET_BIT         0x00000002  /* Remove Reset Dual   Port       */

/* Enable Register, Register 1, Bit Definitions         */
#define SR6_ENABLE_BIT           0x00000001  /* Enable Single Port Bit         */
#define SR6_GLBENABLE_BIT        0x00000002  /* Enable Dual   Port Bit         */

/* Config Register, Register 2, Bit Definitions         */
#define SR6_ACQUIRE_BIT          0x00000001  /* Enable Acquire Single Port Bit */
#define SR6_GLBACQUIRE_BIT       0x00000002  /* Enable Acquire Dual   Port Bit */
#define SR6_LVDSDBG_BIT          0x00000004  /* Enable Sing Port LVDS Rx Debug */
#define SR6_GLBLVDSDBG_BIT       0x00000008  /* Enable Dual Port LVDS Rx Debug */
#define SR6_ICEPORTLB_BIT        0x00000010  /* Loop Data From ICE O to ICE I  */ 

/* Sonet Register, Register 3, Bit Definitions          */
#define SR6_S_MODRESET_REMOVE     0x00000001  /* Must Be Used For Module To Come Out Of Reset               */
#define SR6_S_CARDTYPE            0x00000002  /* Set Card Type To PIC5T When Ready                          */
#define SR6_S_FEC_ENABLE          0x00000004  /* Enable FEC Data Processing                                 */
#define SR6_S_DISABLE_SCRAMBLING  0x00000008  /* Turn Off Fiber Optic Descrambling (Rx) or Scrambling (Tx)  */
#define SR6_S_INPUT_ENABLE        0x00000010  /* Enable Module For Input                                    */
#define SR6_S_SLAVE_ENABLE        0x00000020  /* First Frame Of Slave Module Is Demux #0 Channel            */
#define SR6_S_DISABLE_FRAMING     0x00000040  /* Disable Search For Framing Bring In Data W/Unknown Framing */
#define SR6_S_DEMUXENABLE         0x00000080  /* Enable Realtime Demux Of Incoming Signal                   */
#define SR6_S_OC96ENABLE          0x00000100  /* Use If Line Rate 4976 <Hz                                  */
#define SR6_S_OC192ENABLE         0x00000200  /* Use If Line Rate 9958 MHz                                  */
#define SR6_S_OC3ENABLE           0x00000400  /* Use If Line Rate  155 MHz                                  */
#define SR6_S_OC12ENABLE          0x00000800  /* Use If Line Rate  622 MHz                                  */
#define SR6_S_OC24ENABLE          0x00001000  /* Use If Line Rate 1244 Mhz                                  */
#define SR6_S_OC48ENABLE          0x00002000  /* Use If Line Rate 2488 MHz                                  */
#define SR6_S_OC3FROMOCX_ENABLE   0x00004000  /* Pull Out OC3  Signal From OCx   Input                      */ 
#define SR6_S_OC12FROMOCX_ENABLE  0x00008000  /* Pull Out OC12 Signal From OCx   Input                      */
#define SR6_S_OC24FROMOCX_ENABLE  0x00010000  /* Pull Out OC24 Signal From OCx   Input                      */
#define SR6_S_OC48FROMOCX_ENABLE  0x00020000  /* Pull Out OC48 Signal From OC192 Input                      */
#define SR6_S_OC96FROMOCX_ENABLE  0x00040000  /* Pull Out OC96 Signal From OC192 Input                      */

/* Transceiver Config Register, Register 4 Bit Definitions */
#define SR6_GXBPDWN_BIT           0x00000001  /* Transceiver Power Down Bit   */
#define SR6_PLLPDWN_BIT           0x00000002  /* PLL Power Down Bit           */
#define SR6_TXDREST_BIT           0x00000004  /* Tx Digital Reset             */
#define SR6_RXAREST_BIT           0x00000008  /* Rx Analog  Reset             */ 
/* Transceiver Read Bits                                   */
#define SR6_TXPLLLK_BIT           0x00000001  /* Transceiver Tx PLL Locked    */
#define SR6 RXFRQLK_BIT           0x00000002  /* Rx Locked To Data            */              

/* Crossbar Config Regiser, Register B, Bit Definitions    */
#define SR6_CBSINGPORT_BIT       0x00000001  /* Single ICE Port Op-NO Mux      */
#define SR6_CBDUALPORT_BIT       0x00000002  /* Dual ICE Port Op-Byte Mux Op   */ 

/* Host Memory, Register F, Bit Definitions                */
#define SR6_HMCLKVAL_OFFSET      0x00000000  /* Current AnyRate Freq           */
#define SR6_HMPORT_OFFSET        0x00000010  /* Offset For Each Port Config    */ 


/* Memory Read Offsets    */
#define SR6_FPGAVER_ADDR         0xF0000000  /* FPGA Version Number            */
#define SR6_TRCVRSIG_ADDR        0xE0000000  /* Transceiver Signals            */
#define SR6_MEMTEST_ADDR         0xD0000000  /* Memory Test Register           */

/* FPGA Download Versions */
#define FPGA_HIRATE_DLD  0xC3C3C3C3
#define FPGA_LORATE_DLD  0xB4B4B4B4
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
/*--------------------------------------------------------------*/
int_4 sr6_synclk_cfg (PICSTRUCT *p, int_4 mport, int_u4 maddr, int_u4 value)
{   
   
   pic_wmodreg(p,mport,maddr,((value >> 8) & 0x000000FF),SR6_NUM_PRG_BITS);         /* Set Register Address To Write */
   udelay(500);

   pic_wmodreg(p,mport,maddr,(value & 0x000000FF) | 0x00004000,SR6_NUM_PRG_BITS);   /* Write Data To Register        */ 
   udelay(500);

   return(0);    
}
/*--------------------------------------------------------------*/
int_4 pic_sonetr6_dump(PICSTRUCT *p, int_4 mport,int_4 flag)
  {
    int_u4 mbval;
    int ii;
 
   /* FPGA Version */   
    pic_rmodreg(p,1,SR6_FPGAVER_ADDR,&mbval,SR6_NUM_PRG_BITS);
    printf("\nFPGA Download Version = %x\n",mbval);
    if (mbval == FPGA_HIRATE_DLD)
      printf("High Speed Version Download\n");
    else if (mbval == FPGA_LORATE_DLD)
      printf("Low Speed Version Download\n"); 
    else 
      printf("Download Version Unrecognized\n");

   /* Memory Bank Test */
    for (ii=1;ii<3;ii++)
      {
        printf("\nModule Port = %d\n",ii);
	pic_rmodreg(p,ii,SR6_TRCVRSIG_ADDR,&mbval,SR6_NUM_PRG_BITS);
	printf("Transceiver PLL = ");
	if ((mbval & 0x00000001) == 0x00000001)
	   printf("LOCKED\n");
	else
	   printf("UNLOCKED\n");
	      
	printf("Receiver CDR Locked To ");
	if ((mbval & 0x00000002) == 0x00000002)
	   printf("LINE DATA\n");
	else
	   printf("REFERENCE\n");
	
	if (findflag("SNTMEMTEST",p->config) >= 0)
	  { 
            pic_rmodreg(p,ii,SR6_MEMTEST_ADDR,&mbval,SR6_NUM_PRG_BITS);
            printf("Port = %d Mem Test Value = %x ",ii,mbval);
            if (mbval == 0x00000003)
              printf("... PASS \n");
            else
              printf("... FAIL \n");
	  }    
      }  

    return(0);
  }
/*--------------------------------------------------------------*/
int_4 pic_enable_sonetr6 (PICSTRUCT *p, int_4 mport,int_4 dis_enable)
{
  int_u4 mbval;
  
 /* ICE Card Can't Run Mux Mode From Port 2 */
  if (findflag("SNTALTPORT",p->config) >= 0)
    mport = 2;  

 /* This Puts Module Into Reset By Setting All Config Bits = 0 */                       
  if (dis_enable == 0)
    {
     /* Clear Enable Bit In Module Memory*/ 
      pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),0,SR6_NUM_PRG_BITS); /* Clr Mem Enable Bit */
      pic_wmodreg(p,mport,SR6_RESETREG_ADR,0,SR6_NUM_PRG_BITS);                                               /*  Port Reset     */ 
      pic_wmodreg(p,mport,SR6_CBRSTREG_ADR,0,SR6_NUM_PRG_BITS);                                               /* Cross Bar Reset */  
      udelay(100); 
    }
  else
    {           
      pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),1,SR6_NUM_PRG_BITS); /* Set Mem Enable Bit */
      pic_rmodreg(p,1,(mport * SR6_HMPORT_OFFSET)|(SR6_RESETREG_ADR>>16),&mbval,SR6_NUM_PRG_BITS);            /* Read Reset Value   */
     /* printf("Port = %d Reset Val = %x\n",mport,mbval);  */
      pic_wmodreg(p,mport,SR6_CBRSTREG_ADR,1,SR6_NUM_PRG_BITS);                                               /* Remove CB Reset    */
      pic_wmodreg(p,mport,SR6_RESETREG_ADR,mbval,SR6_NUM_PRG_BITS);                                           /* Reove Port Reset   */   
      udelay(100); 
    }
  return 0;
}
/*--------------------------------------------------------------*/
#define MAX_ICEPORT_RATE  700000000    /* Maximum ICE Port Rate */
#define FEC192RATE       1338653164    /* FEC OC192 Byte Rate   */
#define OC192RATE        1244160000    /*     OC192 Byte Rate   */
#define FEC48RATE         333257143    /* FEC OC48  Byte Rate   */
#define OC48RATE          311040000    /*     OC48  Byte Rate   */
#define OC12RATE           77760000    /*     OC12  Byte Rate   */
#define OC3RATE            19440000    /*      OC3  Byte Rate   */

#define HIRATE_LOWCUT     125000000    /* Low Rate Cutoff For High Rate Download */
#define PREFX_FLAG       0x80000000    /* Set Flag For PREFX Use                 */

int_4 pic_setup_sonetr6 (PICSTRUCT *p, int_4 mport, int_4 dir,int_4 rate, int_4 gain) 
{
  Si5326_Reg_Obj Si5326_Reg; 
  int ii;    
  int i;
  int_u4 sr6_cfgbits = 0;
  float anyclkrate   = 0.0;
  int_4 altmport     = 2;
  int_u4 arclkselreg = 0x00000055;              /* AnyRate Clock Input Source 2 On-Board Oscillator,Default  */
  int_u4 mbval;
  int_u4 dldversion  = FPGA_HIRATE_DLD;  
  int_u4 cmprate     = 0;
  int_u2 sr6lrphyram[64] = { 0xF800,0x0000,0x0C30,0x4001,0x3000,0x0000,0x6055,0x2200,
	                     0x0000,0x0000,0x8000,0x0000,0x2964,0x2000,0x0000,0x0000,
	                     0x2DB4,0xBC31,0x6000,0x0000,0x0000,0x00C0,0xD05E,0x0000,
	                     0x0006,0x03E8,0x0000,0x020E,0x0016,0x6000,0x0000,0x0000,
	                     0x1800,0x0800,0x0010,0x0002,0x0400,0x0786,0x6005,0xE494,
	                     0x25C5,0x0080,0x1022,0x0000,0x0000,0x0000,0x0100,0x0000,
	                     0x0000,0x25C5,0x0080,0x1422,0x0000,0x0010,0x0400,0x0000,
                             0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000 };

  int_u2 sr6hrphyram[64] = { 0xF800,0x0000,0x0C30,0x4001,0x0000,0x0000,0x6055,0x2A00,
	                     0x0000,0x0000,0x0004,0x0000,0x1164,0x1000,0x0000,0x0000,
                             0x2DB4,0xBC31,0x6010,0x0000,0x0000,0x2000,0xD05E,0x0000,
                             0x0006,0x03E8,0x0000,0x020E,0x0016,0xC000,0x0000,0x0000,
                             0x1800,0x0800,0x0610,0x0002,0x0400,0x0006,0x6005,0xE494,
                             0x0C05,0x0080,0x1022,0x0000,0x0000,0x0000,0xC100,0x0000,
                             0x0000,0x0C05,0x0080,0x1422,0x0000,0x0610,0x0400,0x0000,
                             0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000 };

  int_u2 *sr6phyram;

 /* ICE Card Can't Run Mux Mode From Port 2 */
  if (findflag("SNTALTPORT",p->config) >= 0)
    mport = 2; 
         
 /* Set Alternate Port To Check Setup */
  if (mport == 2)
    altmport = 1;

 /* Find FPGA Download Version  */
  pic_rmodreg(p,1,SR6_FPGAVER_ADDR,&mbval,SR6_NUM_PRG_BITS);
  dldversion = mbval;     

 /* Set PHY Mem Programming     */
  if (dldversion == FPGA_LORATE_DLD)
    sr6phyram = sr6lrphyram;
  else
    sr6phyram = sr6hrphyram;

 /* Set Default Line Rate       */
  anyclkrate = ((float) rate) / 4.0; 
  
 /* Check For Pre-Defined Rates */
  if ((findflag("OC192FULL",p->config) >= 0) || (findflag("OC192",p->config) >= 0))
    {
      rate = OC192RATE;
      
      if (findflag("G709FEC",p->config) >= 0)
        rate = FEC192RATE;

      anyclkrate = ((float) rate) / 4.0;           /* Reference Clock Freq = LineRate div 32 */   
    } 

  if (findflag("OC48",p->config) >= 0)
    {
      rate = OC48RATE;
      
      if (findflag("G709FEC",p->config) >= 0)
        rate = FEC48RATE;  
       
        anyclkrate = ((float) rate) / 4.0;         /* Reference Clock Freq = LineRate div 32 */  
    }
  if (findflag("OC12",p->config) >= 0)
    {
      rate = OC12RATE;
      anyclkrate =  ((float) rate) * 2.0;          /* Reference Clock Freq = LineRate div 16 */
    }  
  if (findflag("OC3",p->config) >= 0)
    {
      rate = OC3RATE;
      anyclkrate =  ((float) rate) * 2.0;          /* Reference Clock Freq = LineRate div 16 */
    }

 /* Rate Check Firmware Vs. Data Rate */ 
  if ((rate < HIRATE_LOWCUT) && (dldversion == FPGA_HIRATE_DLD))
    {
       printf("ERR: SonetR6: Wrong Firmware Download For Low Rate Signal-Signals < %d MBy/Sec \n",HIRATE_LOWCUT/1000000 );
       printf("ERR: Reset Card And Use Flag IOMFPGA=lr For Correct Download\n");
       return(-1); 
    }
     
 /* Rate Check Firmware Vs. Data Rate */ 
  if ((rate > HIRATE_LOWCUT) && (dldversion == FPGA_LORATE_DLD))
    {
       printf("ERR: SonetR6: Wrong Firmware Download For High Rate Signal-Signals > %d MBy/Sec \n",HIRATE_LOWCUT/1000000 );
       printf("ERR: Reset Card WITHOUT THe Use Of Flag IOMFPGA=lr For Correct Download\n");
       return(-1); 
    } 
       
 /* Stop Port */   
  pic_wmodreg(p,mport,SR6_RESETREG_ADR,0,SR6_NUM_PRG_BITS);   /*  Port Reset     */ 
  pic_wmodreg(p,mport,SR6_CBRSTREG_ADR,0,SR6_NUM_PRG_BITS);   /* Cross Bar Reset */  
  pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),0,SR6_NUM_PRG_BITS); /* Record Disble  of Port */
  pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_RESETREG_ADR>>16),0,SR6_NUM_PRG_BITS); /* Record Reset   of Port */

 
 /* Read Back Current AnyRate Frequency                       */ 
  pic_rmodreg(p,1,SR6_HMCLKVAL_OFFSET,&mbval,SR6_NUM_PRG_BITS); 
 /* printf("Rate in Mem = %d\n",mbval);  */
 
 /* Set Freq Comare Value Add PREFX Flag To Rate If Necessary */
  cmprate = rate;
  if (findflag("PREFX",p->config) >= 0)
    cmprate = cmprate | PREFX_FLAG;
  
 /* AnyRate Clock Configuration */    
  if (mbval != cmprate)
    { 
      pic_rmodreg(p,1,(altmport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),&mbval,SR6_NUM_PRG_BITS);
      if (mbval != 0)
        {
          printf("Err: Alternate Port, Port = %d, Active and running at different line rate. \n",altmport);
          printf("Err: Both ports must be operated at the same line rate and with the same clock reference\n");
          printf("Err: End the use of the other port, or change the line rate for the current port.\n");
          return(-1); 
        }   

     /* Power Down Both Transceivers */
      pic_wmodreg(p,1,SR6_TRCVRREG_ADR,SR6_GXBPDWN_BIT|SR6_PLLPDWN_BIT|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS); 
      pic_wmodreg(p,2,SR6_TRCVRREG_ADR,SR6_GXBPDWN_BIT|SR6_PLLPDWN_BIT|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS);

      Si5326_Reg.debuglevel = -1;
      Si5326_Reg.fref   =  10000000.0;
      Si5326_Reg.dclk1  = anyclkrate;
      Si5326_Reg.dclk2  = anyclkrate;
      Si5326_Reg.dwght1 = 1.0;
      Si5326_Reg.dwght2 = 1.0;   
      ii = si5326_regcfg_init(&Si5326_Reg,SI5326_REGCFG_MT_SNTXDR6);
      ii = si5326_regcfg_bwsel(&Si5326_Reg,7);       
      ii = si5326_regcfg_clkin(&Si5326_Reg,SI5326_REGCFG_CLKIN_2);             
      ii = si5326_regcfg_dividers(&Si5326_Reg);
      if ((ii != 0) && (findflag("SNTWARNOFF",p->config) < 0))
        printf("WARN: SNTXDR6 reference freq of %d Hz not possible. Using %f Hz\n",(int_u4)anyclkrate,Si5326_Reg.fclk2);
      si5326_regcfg_debug(&Si5326_Reg);
      if (findflag("PREFX",p->config) >= 0)
        arclkselreg = 0x00000015;            /* Clock Input Source 1 Ext Reference */                     
      for(ii=0; ii<Si5326_Reg.numreg; ii++) 
        {    
          if ((Si5326_Reg.regval[ii] & 0x0000FF00) == 0x00000300)                   
            Si5326_Reg.regval[ii] = (Si5326_Reg.regval[ii] & 0xFFFFFF00) | arclkselreg;  /* Set AnyRate Reference Clock Source */       
 
          sr6_synclk_cfg (p,1,SR6_TCLKCFGREG_ADR,Si5326_Reg.regval[ii]);           
          udelay(1000);
        }

      pic_wmodreg(p,1,SR6_HOSTMEM_ADR|SR6_HMCLKVAL_OFFSET,cmprate,SR6_NUM_PRG_BITS);   /* Store AnyRate Clock Frequency */

      udelay(4000);  /* Allow AnyRate To Calibrate */

     /* Remove Global Power Down Of Both Transceivers */
      pic_wmodreg(p,1,SR6_TRCVRREG_ADR,SR6_PLLPDWN_BIT|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS); 
      pic_wmodreg(p,2,SR6_TRCVRREG_ADR,SR6_PLLPDWN_BIT|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS);

      udelay(4000);

     /* Remove PLL Power Down Of Both Transceivers */
      pic_wmodreg(p,1,SR6_TRCVRREG_ADR,SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS); 
      pic_wmodreg(p,2,SR6_TRCVRREG_ADR,SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS);
     
      udelay(4000);

     /* Retrieve PLL Locked Bit                    */
     /* Read PLL 0 Lock                            */
      pic_rmodreg(p,1,SR6_TRCVRSIG_ADDR,&mbval,SR6_NUM_PRG_BITS);      
      if ((mbval & SR6_TXPLLLK_BIT) == 0)
        {
           if (findflag("PREFX",p->config) >= 0)
             {
               printf("Err: Using External 10MHz Reference and PLLs are NOT Locking to External Reference\n");
               printf("Err: Check That 10MHz External Reference is Connected\n");  
               return(-1); 
             }

           printf("Err: MPORT = 1: TX PLL NOT LOCKED To Reference !!!\n");
           return(-1);
        }
     /* Read PLL 1 Lock                            */
      pic_rmodreg(p,2,SR6_TRCVRSIG_ADDR,&mbval,SR6_NUM_PRG_BITS);
      if ((mbval & SR6_TXPLLLK_BIT) == 0)
        {
          if (findflag("PREFX",p->config) >= 0)
             {
               printf("Err: Using External 10MHz Reference and PLLs are NOT Locking to External Reference\n");
               printf("Err: Check That 10MHz External Reference is Connected\n");  
               return(-1); 
             }

          printf("MPORT = 2: TX PLL NOT LOCKED To Reference !!!\n");
          return(-1);
        }  
      
     /* Clr Tx Digital Reset, Clr Rx Analog Reset  */
      pic_wmodreg(p,1,SR6_TRCVRREG_ADR,0,SR6_NUM_PRG_BITS); 
      pic_wmodreg(p,2,SR6_TRCVRREG_ADR,0,SR6_NUM_PRG_BITS);            
    }

 /* Signal Configuration Bits */
  if (dir <= 0)
    {  
      sr6_cfgbits = sr6_cfgbits | SR6_ACQUIRE_BIT;
      if (rate > MAX_ICEPORT_RATE)
        sr6_cfgbits = sr6_cfgbits | SR6_GLBACQUIRE_BIT;  /* Set Dual Port Acquire For High Data Rates */
      else
        {
          if (findflag("ICELOOPBK",p->config) >= 0)
            sr6_cfgbits = sr6_cfgbits | SR6_ICEPORTLB_BIT;
        }

      if (findflag("LVDSDBG",p->config) >= 0)
        {
          sr6_cfgbits = sr6_cfgbits | SR6_LVDSDBG_BIT;
          if (rate > MAX_ICEPORT_RATE)
           sr6_cfgbits = sr6_cfgbits | SR6_GLBLVDSDBG_BIT;
        }      
                
       sr6phyram[41] = 0x8080;     /* Turn On Loopback */ 
      /* Set Rx Gain Value  */         
       i=findintflagdef("SNTRXGAIN",p->config,gain);
       if (i == 0)
         sr6phyram[37] = 0x0006;
       else if (i == 1)
         sr6phyram[37] = 0x0086;
       else if (i==2) 
         sr6phyram[37] = 0x0186;
       else if (i==3) 
         sr6phyram[37] = 0x0386;
       else if (i==4) 
         sr6phyram[37] = 0x0786;
       else 
         { 
           printf("Err: SONETR6 Gain = %d. Valid Range 0 <= value <= 4\n",i);
           return(-1);
         }   
    } 
  else          
    sr6phyram[41] = 0x0080;       /* Turn OFF Loopback */      
     

 /* Write PHY Ram Value To Program PHY */
  for (ii=0;ii<64;ii++)
    pic_wmodreg(p,mport,SR6_TRCVRRAM_ADR+ii,sr6phyram[ii],SR6_NUM_PRG_BITS);

  pic_wmodreg(p,mport,SR6_TRCVRREG_ADR,SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS); 
  pic_wmodreg(p,mport,SR6_TRCVRREG_ADR,0x000001A0|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS);
  udelay(2000);
  pic_wmodreg(p,mport,SR6_TRCVRREG_ADR,0x000000A0|SR6_TXDREST_BIT|SR6_RXAREST_BIT,SR6_NUM_PRG_BITS);        
  udelay(1000);
  pic_wmodreg(p,mport,SR6_TRCVRREG_ADR,SR6_RXAREST_BIT,SR6_NUM_PRG_BITS); 
  udelay(1000);       
  pic_wmodreg(p,mport,SR6_TRCVRREG_ADR,0,SR6_NUM_PRG_BITS);
 /* pic_rmodreg(p,mport,SR6_TRCVRSIG_ADDR,&mbval,SR6_NUM_PRG_BITS);  */
 /* printf("Return Val = %x\n",mbval);                               */
 
 /* Set Global Sonet Cfg Bits */ 
  pic_wmodreg(p,mport,SR6_CFGREG_ADR,sr6_cfgbits,SR6_NUM_PRG_BITS);
  pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_CFGREG_ADR>>16),sr6_cfgbits,SR6_NUM_PRG_BITS); /* Record Cfg Port */

 
 /* Sonet Config Bits */
  sr6_cfgbits = 0;

 /* Turn On FEC Processing                    */
  if (findflag("G709FEC",p->config) >= 0)
    sr6_cfgbits = sr6_cfgbits | SR6_S_FEC_ENABLE; 
 
 /* Turn Off Scrambling/ Descrambling         */
  if (findflag("SCRAMOFF",p->config) >= 0)
    sr6_cfgbits = sr6_cfgbits | SR6_S_DISABLE_SCRAMBLING;

 /* Turn Off Scrambling/ Descrambling         */
  if (findflag("FRAMINGOFF",p->config) >= 0)
    sr6_cfgbits = sr6_cfgbits | SR6_S_DISABLE_FRAMING | SR6_S_DISABLE_SCRAMBLING; 
   
 /* Setup Type Of Signal To Acquire/Playback  */   
  if (findflag("OC192",p->config) >= 0)      /* Check For OC192 Data In Open String       */         
    sr6_cfgbits = sr6_cfgbits | SR6_S_OC192ENABLE; 
  else if (findflag("OC12",p->config) >= 0)  /* Check For OC12 Data In Open String        */
    sr6_cfgbits = sr6_cfgbits | SR6_S_OC12ENABLE;
  else if (findflag("OC24",p->config) >= 0)  /* Check For OC24 Data In Open String        */
    sr6_cfgbits = sr6_cfgbits | SR6_S_OC24ENABLE;
  else if (findflag("OC48",p->config) >= 0)  /* Check For OC48 Data In Open String        */
    sr6_cfgbits = sr6_cfgbits | SR6_S_OC48ENABLE;
  else
    sr6_cfgbits = sr6_cfgbits | SR6_S_OC3ENABLE;
      
              
 /* Check For OC3 Remove From OCx            */
  if ((findflag("OC3FROMOCX",p->config) >= 0) || (findflag("OC3NUM",p->config) >= 0))
    { 
     /* Enable Demux Operation              */
      sr6_cfgbits = sr6_cfgbits | SR6_S_DEMUXENABLE;  

     /* Error Check OC3 Removal Operation   */
      if (((sr6_cfgbits & SR6_S_OC3ENABLE) != 0))
        {
          printf("\n Error-OC12 or OC48 or OC192 Must Be Enabled For OC3 Removal\n");             
          return(-1); 
        }

      i=findintflag("OC3NUM",p->config); 
         
      if (i < 0) /* Did Not Find Flag, Assume Zero */
        {  
          printf("\n Warning-OC3 Number For Removal Not Specified-Assuming OC3NUM=0\n");
          i = 0; 
        }
      if ((i >= 4) && ((sr6_cfgbits & SR6_S_OC12ENABLE) != 0))
        {
          printf("\n Error-OC3 Number For Removal Out Of Range For OC12 Input; Valid Values 0 to 3\n");
          return(-1); 
        }
      if ((i >= 8) && ((sr6_cfgbits & SR6_S_OC24ENABLE) != 0))
        {
          printf("\n Error-OC3 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 7\n");
          return(-1); 
        }
      if ((i >= 16) && ((sr6_cfgbits & SR6_S_OC48ENABLE) != 0))
        {
          printf("\n Error-OC3 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 15\n");
          return(-1); 
        }

      if ((i >= 64) && ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0))
        {
          printf("\n Error-OC3 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 63\n");
          return(-1); 
        }
        
      sr6_cfgbits = sr6_cfgbits | SR6_S_OC3FROMOCX_ENABLE;
      sr6_cfgbits = sr6_cfgbits | ((i << 19) & 0x001f80000); 

      if ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0)
        rate = rate / 64;
      else if ((sr6_cfgbits & SR6_S_OC48ENABLE) != 0) 
        rate = rate / 16;
      else if ((sr6_cfgbits & SR6_S_OC24ENABLE) != 0) 
        rate = rate / 8;
      else if ((sr6_cfgbits & SR6_S_OC12ENABLE) != 0) 
        rate = rate / 4;     
    } 
 
 /* Check For OC12 Remove From OCx            */
  if ((findflag("OC12FROMOCX",p->config) >= 0) || (findflag("OC12NUM",p->config) >= 0))
    { 
     /* Error Check For More Than 1 Demux Op   */
      if ((sr6_cfgbits & SR6_S_DEMUXENABLE) != 0)
        {
          printf("\n Error-Only 1 Demux Operation Allowed\n");             
          return(-1);  
        }        
      else
        sr6_cfgbits = sr6_cfgbits | SR6_S_DEMUXENABLE;          

      i=findintflag("OC12NUM",p->config);
          
      if (i < 0) /* Did Not Find Flag, Assume Zero */
        { 
          printf("\n Warning-OC12 Number For Removal Not Specified-Assuming OC12NUM=0\n");
          i = 0; 
        }

      if ((i >= 2) && ((sr6_cfgbits & SR6_S_OC24ENABLE) != 0))
        {
           printf("\n Error-OC12 Number For Removal Out Of Range For OC24 Input; Valid Values 0 to 1\n");
           return(-1); 
        }

      if ((i >= 4) && ((sr6_cfgbits & SR6_S_OC48ENABLE) != 0))
        {
          printf("\n Error-OC12 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
          return(-1); 
        }
         
      if ((i >= 16) && ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0))
        {
          printf("\n Error-OC12 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 15\n");
          return(-1); 
        }  

      sr6_cfgbits = sr6_cfgbits | SR6_S_OC12FROMOCX_ENABLE;
      sr6_cfgbits = sr6_cfgbits | ((i << 19) & 0x001f80000);    

      if ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0)
        rate = rate / 16;
      else if ((sr6_cfgbits & SR6_S_OC48ENABLE) != 0) 
        rate = rate / 4;
      else if ((sr6_cfgbits & SR6_S_OC24ENABLE) != 0) 
        rate = rate / 2;          
    } 
  
 /* Check For OC24 Remove From OC48            */
  if ((findflag("OC24FROMOCX",p->config) >= 0) || (findflag("OC24NUM",p->config) >= 0))
    {
      
     /* Error Check For More Than 1 Demux Op   */
      if ((sr6_cfgbits & SR6_S_DEMUXENABLE) != 0)
        {
          printf("\n Error-Only 1 Demux Operation Allowed\n");             
          return(-1);  
        }        
      else
        sr6_cfgbits = sr6_cfgbits | SR6_S_DEMUXENABLE; 

      i=findintflag("OC24NUM",p->config);
          
      if (i < 0) /* Flag Not Found Assume 0 */
        { 
          printf("\n Warning-OC24 Number For Removal Not Specified-Assuming OC24NUM=0\n");
          i = 0;
        }
      if (i >= 2)
        { 
          printf("\n Error-OC24 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 1\n");
          return(-1); 
        }            

      if ((i >= 8) && ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0))
         {
           printf("\n Error-OC24 Number For Removal Out Of Range For OC192 Input; Valid Values 0 to 7\n");
           return(-1); 
         }   
      sr6_cfgbits = sr6_cfgbits | SR6_S_OC24FROMOCX_ENABLE;
      sr6_cfgbits = sr6_cfgbits | ((i << 19) & 0x001f80000);            

      if ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0)
        rate = rate / 8;
      else if ((sr6_cfgbits & SR6_S_OC48ENABLE) != 0) 
        rate = rate / 2;      
    } 

  if ((findflag("OC48FROMOCX",p->config) >= 0) || (findflag("OC48NUM",p->config) >= 0))
    {    
     /* Error Check For More Than 1 Demux Op   */
      if ((sr6_cfgbits & SR6_S_DEMUXENABLE) != 0) 
        {
          printf("\n Error-Only 1 Demux Operation Allowed\n");             
          return(-1);  
        }        
      else
        sr6_cfgbits = sr6_cfgbits | SR6_S_DEMUXENABLE; 

      i=findintflag("OC48NUM",p->config);
          
      if (i < 0) /* Flag Not Found Assume 0 */
        { 
          printf("\n Warning-OC48 Number For Removal Not Specified-Assuming OC48NUM=0\n");
          i = 0;
        }
      if (i >= 4)
        { 
          printf("\n Error-OC48 Number For Removal Out Of Range For OC48 Input; Valid Values 0 to 3\n");
          return(-1); 
        }            
 
      sr6_cfgbits = sr6_cfgbits | SR6_S_OC48FROMOCX_ENABLE;
      sr6_cfgbits = sr6_cfgbits | ((i << 19) & 0x001f80000);            
            
      if ((sr6_cfgbits & SR6_S_OC192ENABLE) != 0)
        rate = rate / 4;        
    }  

 pic_wmodreg(p,mport,SR6_SONETREG_ADR,sr6_cfgbits,SR6_NUM_PRG_BITS);                  

 /* Rate vs Number Of Ports Check */
  if (rate > MAX_ICEPORT_RATE)
    {
      pic_rmodreg(p,1,(altmport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),&mbval,SR6_NUM_PRG_BITS); /* Read Enable Of Opposite Port */
      if ((mbval & SR6_ENABLE_BIT) != 0)
        {   
          printf("Err: Line Byte Rate of %d requires use of both ICE LVDS Ports\n",rate);
          printf("Err: But, alternate port, Port %d, is already in use.\n",altmport);
          printf("Err: User needs to halt Alternate Port playback or acquisition.\n");
          return(-1);  
        }
    } 
  else
    {
      pic_rmodreg(p,1,(altmport * SR6_HMPORT_OFFSET)|(SR6_ENDISREG_ADR>>16),&mbval,SR6_NUM_PRG_BITS); /* Read Enable Of Opposite Port */
      if ((mbval & SR6_GLBENABLE_BIT) != 0)
        {   
          printf("Err: Alternate port, Port %d, line rate requires use of both ICE LVDS Ports\n",altmport);
          printf("Err: Therfore current port is already in use.\n");
          printf("Err: User needs to halt Alternate Port %d playback or acquisition before using this port %d.\n",altmport,mport);
          return(-1);  
        }
    }


 /* Crossbar Configuration Bits */
  sr6_cfgbits = SR6_CBSINGPORT_BIT; 
  if (rate > MAX_ICEPORT_RATE)
    sr6_cfgbits = SR6_CBDUALPORT_BIT;      
  pic_wmodreg(p,mport,SR6_CBCFGREG_ADR,sr6_cfgbits,SR6_NUM_PRG_BITS);   /* Cross Bar Cfg For Dual or Sing Port */    


 /* Set Reset Values For Enable */
  if (rate > MAX_ICEPORT_RATE)
    {
     /* Store Dual Port Reset */ 
      pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_RESETREG_ADR>>16),SR6_RESET_BIT|SR6_GLBRESET_BIT,SR6_NUM_PRG_BITS); 
      return(2);
    }
  else
    {
     /* Store Sing Port Reset */ 
      pic_wmodreg(p,1,SR6_HOSTMEM_ADR|(mport * SR6_HMPORT_OFFSET)|(SR6_RESETREG_ADR>>16),SR6_RESET_BIT,SR6_NUM_PRG_BITS);       
      return(0); 
    }
}
