#ifndef _TIOLIBSHARED_H_
#define _TIOLIBSHARED_H_
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <limits.h>

#define LOCK ".icelock"
/* duraction to sleep in lock function (while lock is on) */
#define SLEEP 100

/* inline functions */
inline void packX (char *buf, int64_t num) { memcpy(buf, &num, 8); }
inline void packD (char* buf, double num) { packX(buf, *(int64_t*)(&num)); }
inline void packL (char *buf, int val) { memcpy(buf, &val, 4); }
inline int64_t unpackX (char *buf) { return *(int64_t*) buf; }
inline int ceil_int (int n, int d) { return (n+d-1)/d; }
inline int floor_int (int n, int d) { return n/d; }
inline off_t ceil_double (double n, double d) { return (off_t)((n+d-1)/d); }
inline off_t floor_double (double n, double d) { return (off_t)(n/d); }
inline off_t floor_off ( off_t n, off_t d ) { return (off_t)(n/d); }
inline off_t ceil_off (off_t n, off_t d ) { return (off_t)((n+d-1)/d); }

/* parses a config file field into a character array */
int setConfField (char *dest, char *source, int off, int len, int bm) {
  len -= off;
  if ( len > 0 ) {
    strncpy (dest, &source[off], len);
    dest[len] = '\0';
    return bm;
  }
  return 0;
}
/* parse T, G, M and K out of a string and convert to number*/
off_t parseNum ( char *cur ) {
  char *str;
  char lchar;
  int len;
  off_t mult, out;

  len = strlen(cur);
  mult = 1;
  /* grab the last character */
  lchar = toupper(cur[len-1]);
  /* parse for: T, G, M and K */
  if ( lchar == 'T' ) {
    mult = 1073741824;
    mult *= 1024;
    len--;
  }
  else if ( lchar == 'G' ) {
    mult = 1073741824;
    len--;
  }
  else if ( lchar == 'M' ) {
    mult = 1048576;
    len--;
  }
  else if ( lchar == 'K' ) {
    mult = 1024;
    len--;
  }
  /* copy the input string (make sure the last char is a terminator) */
  str = malloc ( sizeof(char) * len + 1 );
  strncpy(str, cur, len);
  str[len] = '\0'; 

  /* perform conversion */
  out = (off_t) ((double) mult)*atof(str);
  /* free the string and return */
  free(str);
  return out;
}


/* new config reader */
int read_config ( char *conf, char *type, char *iceloc, char *fuseloc, char *dmfile, char *dev, off_t *extend, int *nthreads ) {
  FILE *fd;
  char cur[FILENAME_MAX];
  int len, rv;

  int bit_mask = 0;

  /* open the file for reading*/
  fd = fopen(conf, "r");
  if ( !fd ) {
    printf("Error! Error reading configuration file: %s! Make sure file exists with correct permissions!\n", conf);
    return -1;
  }

  /* parse conf file */
  while ( (rv=fscanf(fd, "%s", cur)) > 0 ) {
    len = strlen(cur);
    if ( strncmp ("TYPE=", cur, 5) == 0 ) {
      bit_mask |= setConfField (type, cur, 5, len, 0x1);
    }
    else if ( strncmp ("VERSION=", cur, 8) == 0 ) {
      bit_mask |= setConfField (iceloc, "VERSION=BOGUS", 8, len, 0x2);
    }
    else if ( strncmp ("NMPATH=", cur, 7) == 0 ) {
      bit_mask |= setConfField (iceloc, cur, 7, len, 0x2);
    }
    else if ( strncmp ("FUSEPATH=", cur, 9) == 0 ) {
      bit_mask |= setConfField (fuseloc, cur, 9, len, 0x4);
    }
    else if ( strncmp ("DMSETUP=", cur, 8) == 0 ) {
      bit_mask |= setConfField (dmfile, cur, 8, len, 0x8);
    }
    else if ( strncmp ("MAPPER=", cur, 7) == 0 ) {
      bit_mask |= setConfField (dev, cur, 7, len, 0x10);
    }
    else if ( strncmp ("EXTEND=", cur, 7) == 0 ) {
      /* parse extend */
      *extend = parseNum(&(cur[7]));
      /* make sure that the value is acceptable */
      if ( *extend > 1 ) 
        bit_mask |= 0x20;
    }
    else if ( strncmp ("THREADS=", cur, 8) == 0 ) {
      *nthreads = atoi(&cur[8]);
      if ( *nthreads > 0 && *nthreads < INT_MAX )
        bit_mask |= 0x40;
    }
  }
  /* close file */
  fclose(fd);
  /* make sure everything was filled out! */
  if ( bit_mask != 0x7f ) {
    printf("\nError! Missing field(s)! Configuration file must contain the following entries:\n");
    printf("TYPE=<RAMD|RAID>\n");
    printf("NMPATH=<path_to_midas_header_files\n");
    printf("FUSEPATH=<path_to_fuse_mount_point\n");
    printf("DMSETUP=<path_to_dmsetup_file>\n");
    printf("MAPPER=<path_to_dev_mapper_entry>\n");
    printf("EXTEND=<midas_file_extend_size>\n");
    printf("THREADS=<num_threads>\n");
    printf("NMUSER=<user_name>\n");
    printf("\n");
    return -1;
  }
  return 0;
}


int parse_dm_full ( void *ptr ) {
  s_fsprops *fs;
  FILE *fd;
  int rv, m;
  char type[512];

  fs = (s_fsprops*) ptr;

  fd = fopen(fs->dmfile, "r");
  if ( !fd ) 
    return -ENOENT;

  /* parse for the # of disks */
#if __WORDSIZE == 64
  rv = fscanf(fd, "%ld %ld %s %d %d", &(fs->off), &(fs->tsize), type, &(fs->ndisks), &(fs->ssize));
#else
  rv = fscanf(fd, "%lld %lld %s %d %d", &(fs->off), &(fs->tsize), type, &(fs->ndisks), &(fs->ssize));
#endif
  fs->off *= 512;
  fs->tsize *= 512;
  fs->ssize *= 512;
  if( fs->ndisks <= 0 )
    return -EIO;
  /* malloc */
  fs->disks = malloc(sizeof(s_diskmap)*fs->ndisks);
  /* parse each disk */
  for (m=0; m<fs->ndisks; m++) {
#if __WORDSIZE == 64
    rv = fscanf(fd, "%s %ld", fs->disks[m].path, &(fs->disks[m].off));
#else
    rv = fscanf(fd, "%s %lld", fs->disks[m].path, &(fs->disks[m].off));
#endif
    fs->disks[m].off *= 512;
  }
  fclose(fd);
  /* make sure they don't try and use more threads than disks! */
  if (fs->nthreads > fs->ndisks)
    fs->nthreads = fs->ndisks;
  return 0;
}

void *open_fs_props ( char *conf ) {
  s_fsprops *fs;

  fs = malloc(sizeof(s_fsprops));
  fs->disks = NULL;
  /* read the config file */
  read_config(conf, fs->type, fs->iceloc, fs->fuseloc, fs->dmfile, fs->dev, &(fs->extend), &(fs->nthreads));
  /* check type */
  if ( strcmp(fs->type, "RAID")==0 ) {
  /* read the dmsetup file */
    parse_dm_full(fs);
  }
  else {
    /* assume RAMD (for now) */
    fs->tsize = 36632;
    fs->tsize *= 1048576;
    fs->off = 0;
    fs->ndisks = 1;
    fs->nthreads = 1;
    fs->ssize = 1;
  }
  return fs;
}

int close_fs_props ( void *ptr ) {
  s_fsprops *fs;
  
  fs = (s_fsprops*) ptr;
  if ( fs->disks ) 
    free(fs->disks);
  free(fs);
  return 0;
} 

/* determine if a string ends with a specific extension */
int isExt ( const char *fname, const char *ext ) {
  int len1, len2;
  len1 = strlen(fname);
  len2 = strlen(ext);
  if (len1 >= len2 && strcmp(ext, &(fname[len1-len2])) == 0)
    return (len1 - len2);
  return -1;
}

off_t getFileSize ( char *path ) {
  struct stat stbuf;
  int fe;

  fe = stat(path, &stbuf);
  if ( fe < 0 )
    return -1;
  return stbuf.st_size;
}

off_t getFsUsed ( void *ptr, int len, off_t *offs, off_t *sizes ) {
  s_fsprops *fs;
  off_t accum;
  int m;

  fs = (s_fsprops*) ptr;

  accum = 0;
  /* loop through and accum */
  for ( m=0; m<len; m++ ) {
    accum += fs->ssize * ceil_off(sizes[m], fs->ssize);
  }
  return accum;
}

/* returns the number of detached header entries */
int getDetEntries ( char *path ) {
  off_t tsize;
  tsize = getFileSize(path);
  if ( tsize % 16 == 0 )
    return tsize/16;
  return 0;
}

/* count the number of entries with a specific extension */
int countDir ( char *fdir, char *ext ) {
  DIR *dir;
  struct dirent *tmp;
  char *path;
  int cntr, dlen;

  cntr = 0;
  dlen = strlen(fdir) + 3;

  /* open the directory */
  dir = opendir(fdir);
  /* loop and look for the correct extension */
  while ( (tmp = readdir(dir)) ) {
    if ( isExt(tmp->d_name, ext) > 0 ) {
      path = malloc(dlen + strlen(tmp->d_name));
      sprintf(path, "%s/%s%c", fdir, tmp->d_name, '\0');
      /* determine how many entries there currently are */
      cntr += getDetEntries(path);
      free(path);
    }
  }
  closedir(dir);
  return cntr;
}

int accumDetSize ( char *path, off_t *size ) {
  FILE *fd;
  char buf[16];
  off_t rs;
  int m, nents;

  *size = 0;
  fd = fopen(path, "rb");
  if ( !fd ) 
    return -EIO;

  /* determine how many entries there are in this det */
  nents = getDetEntries(path);
  for ( m=0; m<nents; m++ ) {
    if ( fread(buf, 1, 16, fd) != 16 ) 
      return -EIO;
    *size = *size + unpackX(&buf[8]);
  }
  fclose(fd);
  return 0;
}

int readDet ( char *path, int len, off_t *offs, off_t *sizes) {
  FILE *fd;
  char buf[16];
  int m;

  fd = fopen(path, "rb");
  if ( !fd )
    return -EIO;
  for (m=0; m<len; m++) {
    if ( fread(buf, 1, 16, fd) != 16 )
      return -EIO;
    offs[m] = unpackX(buf);
    sizes[m] = unpackX(&buf[8]);
  }
  fclose(fd);
  return 0;
}

int writeDet ( char *path, int len, off_t *offs, off_t *sizes) {
  FILE *fd;
  char buf[16];
  int m;

  fd = fopen(path, "wb");
  if ( !fd )
    return -1;
  for (m=0; m<len; m++) {
    packX(buf, offs[m]);
    packX(&buf[8], sizes[m]);
    fwrite(buf, 1, 16, fd);
  }
  fclose(fd);
  return 0;
}

/* read all the detached headers in a directory */
int getAllDets ( char *fdir, int len, off_t *offs, off_t *sizes ) {
  DIR *dir;
  struct dirent *tmp;
  char *path;
  int cntr, dlen, ce;

  cntr = 0;
  dlen = strlen(fdir) + 3;

  /* open the directory */
  dir = opendir(fdir);
  /* loop and look for .det files */
  while ( (tmp = readdir(dir)) ) {
    if ( isExt(tmp->d_name, ".det") > 0 ) {
      if ( cntr >= len )
        return -EIO;
      /* build the path name */
      path = malloc(dlen + strlen(tmp->d_name));
      sprintf(path, "%s/%s%c", fdir, tmp->d_name, '\0');
      /* determine how many entries there currently are */
      ce = getDetEntries(path);
      readDet(path, ce, &(offs[cntr]), &(sizes[cntr]));
      free(path);
      /* increment the det counter */
      cntr += ce;
    }
  }
  closedir(dir);
  return 0;
}

int sortDets ( int ndets, off_t *paddrs, off_t *psizes ) {
  off_t paddri, psizei;
  int i, j, k;

  /* sort */
  for (i=0; i<ndets; i++) {
    paddri = paddrs[i];
    psizei = psizes[i];
    for (j=0; j<i && paddri>paddrs[j]; j++) {}
    for (k=i-1; k>=j; k--) {
      paddrs[k+1]=paddrs[k];
      psizes[k+1]=psizes[k];
    }
    paddrs[j]=paddri;
    psizes[j]=psizei;
  }
  return 0;
}

/* */
int getDeviceProps ( void *ptr, off_t *tsize, off_t *asize, off_t *ssize) {
  s_fsprops *fs;

  /* TODO if ptr == NULL open it */
  fs = (s_fsprops*) ptr;

  *tsize = fs->tsize;
  *asize = floor_off(*tsize, fs->ndisks);
  *ssize = fs->ssize;

  return 0;
}

/* */
off_t getDeviceSize ( void *ptr ) {
  s_fsprops *fs;
  off_t ret;

  if ( !ptr ) 
    return -1;
  fs = (s_fsprops*) ptr;
  /* grab the total size */
  ret = fs->tsize;
  return ret;
}

/* TODO do we need this? */
off_t getStripeSize ( void *ptr ) {
  s_fsprops *fs;
  off_t ret;

  if ( !ptr ) 
    return -1;
  fs = (s_fsprops*) ptr;
  /* grab the stripe size */
  ret = fs->ssize;
  return ret;
}

/* floor function */
off_t getAdjustedSize ( void *ptr, off_t num ) {
  s_fsprops *fs;
  off_t ret;

  if ( !ptr ) 
    return -1;
  fs = (s_fsprops*) ptr;
  /* grab the total size */
  ret = floor_off(num, fs->ndisks);
  return ret;
}

/* ceil function */
off_t getRelativeSize ( void *ptr, off_t num ) {
  s_fsprops *fs;
  off_t ret;

  if ( !ptr ) 
    return -1;
  fs = (s_fsprops*) ptr;
  /* grab the total size */
  ret = ceil_off(num, fs->ndisks);
  return ret;
}

off_t stripeAlign ( off_t size, off_t stripe ) {
  /* round up to the nearest stripe */
  return stripe * ceil_off(size, stripe);
}

/*
*/
off_t determineOffset ( void *ptr, off_t length ) {
  s_fsprops *fs;
  off_t *paddrs, *psizes;
  off_t paddr, psize, minsize, delta, fend, tsize, ssize;
  int ndets, n;

  fs = (s_fsprops*) ptr;

  /* get the # of det entries */
  ndets = countDir(fs->iceloc, ".det");
  /* if there is only 1 file, return 0 */
  if ( ndets <= 1 ) 
    return 0;
  /* allocate and initilize variables */
  paddrs = malloc( sizeof(off_t)*(ndets+1) );
  psizes = malloc( sizeof(off_t)*(ndets) );
  /* get the offsets/sizes and quick sort */
  getAllDets(fs->iceloc, ndets, paddrs, psizes);
  sortDets(ndets, paddrs, psizes);
  /* get the total size and stripe size */
  tsize = getDeviceSize(fs);
  ssize = getStripeSize(fs);
  /* init the minsize and starting address */
  minsize = getAdjustedSize(fs, tsize);
  paddrs[ndets] = minsize;
  paddr = -1;
  /* get the relative size that this file will occupy */
  psize = getRelativeSize(fs, length);
  /* loop through and try to fit it in */
  for (n=ndets-1; n>=0; n--) {
    psizes[n] = getRelativeSize(fs, psizes[n]);
    /* align with the stripe size */
    fend = stripeAlign(paddrs[n]+psizes[n], ssize);
    /* calculate the space between the two addresses */
    delta = paddrs[n+1] - fend;
    /* check if this is the largest allocation */
    if ( delta>=psize && delta<minsize ) {
      /* set the address and maxsize */
      paddr = fend;
      minsize = delta;
    }
  }
  /* free the resources */
  free(paddrs);
  free(psizes);
  return paddr;
}

int __nsleep ( const struct timespec *req, struct timespec *rem ) {
  struct timespec temp_rem;
  if(nanosleep(req,rem)==-1)
     __nsleep(rem,&temp_rem);
  return 1;
}
int msleep ( unsigned long milisec ) {
  struct timespec req={0},rem={0};
  time_t sec=(int)(milisec/1000);
  milisec=milisec-(sec*1000);
  req.tv_sec=sec;
  req.tv_nsec=milisec*1000000L;
  __nsleep(&req,&rem);
   return 1;
}

int nsleep ( unsigned long nsec ) {
  struct timespec req={0},rem={0};
  req.tv_sec = 0;
  req.tv_nsec = nsec;
  __nsleep(&req,&rem);
  return 1;
}

double get_time () {
  struct timeval tim;
  gettimeofday(&tim, NULL);
  return tim.tv_sec+(tim.tv_usec/1000000.0);
}

/** 
   Try to create a lock file for this Process ID
   Returns process ID on success or a negated errno on failure
*/
int createLock ( int pid, char *lock ) {
  FILE *fd;
  struct stat perms;
  char buf[10];
  int fid;
  int rv;

  /* init file PID to zero */
  fid = 0;
  /* attempt to open as read+ */
  fd = fopen(lock, "r+b");
  if ( fd ) {
    /* read in buffer */
    rv = fread(buf, 1, 10, fd);
    if ( rv != 9 ) {
      /*  invalid file size, close file (current file will be zero'ed) */
      fclose(fd);
      fd = NULL;
    }
    else {
      /* convert buffer to number (atoi has undefined behavior on non-integers) */
      buf[9] = '\0';
      fid = (int) strtol(buf, NULL, 10);
    }
  }
  /* open for write */
  if ( !fd ) {
    /* if the file doesn't exist, create it */
    fd = fopen(lock, "wb");
    /* if we still can't open it, throw an error */
    if ( !fd ) {
      printf("Error! Cannot create lock file: %s! Make sure the folder exists and you have appropriate privileges!\n", lock);
      return -EPERM;
    }
  }
  if ( fid>0 ) {
    /* PID isn't zero, wait to give other threads a chance */
    msleep(SLEEP);
  }
  /* seek back to the beginning */
  fseek(fd, 0, SEEK_SET);
  /* write our PID, close and return */
  fprintf(fd, "%08d\n\0", pid);
  fclose(fd);
  /* get current permissions */
  stat(lock, &perms);
  /* change permissions */
  rv = chmod(lock, perms.st_mode|S_IROTH|S_IWOTH);
  return pid;
}

off_t removeLock ( int pid, off_t off, off_t len, char *lock ) {
  FILE *fd;
  int fid;

  fd = fopen(lock, "r+b");
  if ( !fd) 
    return -1;
  if ( fscanf(fd, "%d", &fid) < 0 )
    return -1;
  if ( fid==pid ) {
    /* if the lock was ours, safely return */
    fseek(fd, 0, SEEK_SET);
    fprintf(fd, "%08d", 0);
    fclose(fd);
    return off;
  }
  /* the lock wasn't ours, return an error */
  fclose(fd);
  return -len;
}

off_t getAllocation ( void *ptr, int pid, off_t len, char *lock ) {
  if ( createLock(pid, lock) < 0 ) {
    return -1;
  }
  return determineOffset(ptr, len);
}

off_t isDetContinuous ( int ndisks, off_t o1, off_t l1, off_t o2 ) {
  off_t out, e1, e2;

  e1 = floor_off(l1, ndisks);
  e2 = ceil_off(l1, ndisks);
  out = e1+o1;
  if ( !(e1==e2 && out==o2) )
    out = -1;

  return out;
}

/* get a random # */
int getRandom () {
  /* seed */
  srand( time(NULL) );
  return rand();
}

off_t writeDetLock ( void *ptr, char *path, off_t len ) {
  s_fsprops *fs;
  FILE *fd;
  char buf[16];
  char *lock;
  off_t off, final;
  int pid;

  /* direct pointer */
  fs = (s_fsprops*) ptr;

  /* malloc and generate lock file name */
  lock = malloc(strlen(fs->iceloc) + strlen(LOCK) + 3);
  sprintf(lock, "%s/%s%c", fs->iceloc, LOCK, '\0');

  /* 
   make sure offset + adjusted size == 0 (therefore, offset has to be negative)
   note: without writing this (to disk), total size could get messed up down stream
  */
  /* init offset so that -len/ndisks + len = zero */
  off = ( fs->ndisks > 0 ) ? -(len/fs->ndisks) : -len;
  /* open det file */
  fd = fopen(path, "wb");
  if ( !fd ) {
    printf("Error, cannot create '%s'. Do you have correct permissions on the directory?\n", lock);
    free(lock);
    return -1;
  }
  /* get the process id */
  pid = (int) getpid();
  /* write to disk for getAllocation function (to calc filesystem size) */
  fseek(fd, 0, SEEK_SET);
  packX(buf, off);
  packX(&buf[8], len);
  fwrite(buf, 1, 16, fd);
  fflush(fd);
  /* now get the allocation */
  off = getAllocation(ptr, pid, len, lock);
  if ( off < 0 ) {
    free(lock);
    fclose(fd);
    return off;
  }
  /* pack the real offset in and write (before lifting the lock) */
  packX(buf, off);
  fseek(fd, 0, SEEK_SET);
  fwrite(buf, 1, 8, fd);
  fclose(fd);
  /* remove the lock */
  final = removeLock(pid, off, len, lock);
  free(lock);
  /* check if our allocation was 'safe' */
  if ( off!=-1 && off != final ) {
    /* someone else got our allocation, recursively call */
    off = writeDetLock(ptr, path, len);
  }
  /* return the offset */
  return off;
}

off_t updateDet ( void *ptr, char *path, off_t len ) {
  s_fsprops *fs;
  char *rname;
  off_t *offs, *sizes, fend, off;
  int m, nents, rnum;
  /* begin update */
  fs = (s_fsprops*) ptr;
  /* get a random number */
  rnum = getRandom();
  /* create the random file's name */
  rname = malloc(strlen(fs->iceloc) + 32);
  sprintf(rname, "%s/r_%d.det%c", fs->iceloc, rnum, '\0');
  /* create a 'safe' det file with a random name */
  off = writeDetLock(ptr, rname, len);
  /* determine how many entries are in the file */
  nents = getDetEntries(path);
  if ( nents <= 0 )
    return -1;
  offs = malloc(sizeof(off_t)*(nents+1));
  sizes = malloc(sizeof(off_t)*(nents+1));
  readDet(path, nents, offs, sizes);

  /* check if the this entry is a continuation of the last entry */
  fend = isDetContinuous(fs->ndisks, offs[nents-1], sizes[nents-1], off);
  if ( fend > 0 ) {
    sizes[nents-1] += len;
  }
  else {
    /* not a continuation, write to the end */
    offs[nents] = off;
    sizes[nents] = len;
    nents++;
  }
  /* write */
  writeDet(path, nents, offs, sizes);
  /* free resources */
  free(offs);
  free(sizes);
  /* delete temp file */
  if ( remove(rname) < 0 )
    off = -1;
  free(rname);
  return off;
}

off_t finalizeDet (void *ptr, char *path, off_t len) {
  off_t *offs, *sizes;
  off_t delta, accum;
  int cnt, m;
  /* get the number of entries in the det */
  cnt = getDetEntries(path);
  /* allocate the resources */
  offs = malloc( sizeof(off_t)*cnt );
  sizes = malloc( sizeof(off_t)*cnt );
  /* read the det */
  readDet(path, cnt, offs, sizes);
  /* accum the total size */
  accum = sizes[0];
  for (m=1; m<cnt; m++) 
    accum += sizes[m];
  /* get the difference */
  delta = accum - len;
  /* modify the last size */
  sizes[cnt-1] -= delta;
  /* write the new det */
  writeDet(path, cnt, offs, sizes);
  /* free the resources */
  free(offs);
  free(sizes);
  return delta;
}

#endif

