!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This macro coordinates a collection of SNAPPER functions defined through a table file
! By default each of these macros is executed in a graphical subpane
!
startmacro

! set default config file
switch "TBL" cfg get "snapapp" 
if ".tbl" nsubs cfg then set cfg "^{cfg}.tbl"
find cfg cfgn
if cfgn eqs "" then set cfgn "nxm.ice.tbl.^cfg"

set t:cfg cfgn
switch "VERBOSE" verbose get 0 1
switch "DISPLAY" display get "FRONT" 
switch "USE"    t:usetbl get {}
switch "THEME" theme get "GEAR2" 
switch "TEST"  test get 0 1

! get remote controllers
fname jicd api jicd properties ice
switch "JICD" jicd get "NONE" jicd

! get specified table file
set l:show 1
set l:config 0
set l:appoff 1
set l:defshow 2
set l:modcntrls 3
set l:nargs this.args.size
if nargs ge 1 then
  set appn this.args.geto(1)
else
  set appn "NONE"
endif
if appn instanceof "nxm.sys.lib.Table" or ".TBL" subs appn then
  set t:cfg appn
  set l:appoff 2
endif
if cfg ninstanceof "nxm.sys.lib.Table" then
  error "Opening configuration table"
  stop
endif

! prepare default snapper table
table snapdef create {port=MODULE1,format=SB,rate=10,clock=N,length=1.0,dec=32,&
        freq=1,gain=0,mgain=0,nfft=1024,psdr=25,psda=5,expa=0,frame=512,wave=none,afname="archive"}
table snapdef merge snapdef cfg.snapset

! remove controls table until nxm320i table fix
if snapdef.controls rexists then
  set snapdef_controls snapdef.controls
  rem snapdef.controls
endif

! prepare configuration table
set apps "NONE"
if cfg.public rexists then	! specified cases listed in table
  res apps "^apps,^cfg.public"
else
 foreach key intable cfg	! build list from all cases
  if key neqss "CASE_" continue
  sedit key key trim "CASE_"
  if cfg.private rexists and ",^key," subs ",^cfg.private," continue
  res apps "^apps,^key"
 endfor
endif

! look for the APPLIST entry
set scenario "NONE"
if cfg.applist rexists then
  set scenarios cfg.applist.getlist
  if /menu then
    gcontrol menu ans "Scenarios" "^scenarios,<Cancel>"
    if ans nrexists or ans eqs "<CANCEL>" stop
    call setscene ans
  elseif appoff le nargs then
    set appn this.args.geto(appoff)
    if cfg.applist.^appn rexists then call setscene appn
  endif
else
  set scenarios "NULL"
  if /menu then error "No APPLIST entry in config table for MENU"
endif

! set global controls modes
if cfg.defshow rexists set defshow cfg.defshow
if cfg.modcntrls rexists set modcntrls cfg.modcntrls

! parse switches after scene activation
switch "APPEXT" appext get "NONE"
switch "NORESET" noreset get 0 1
switch "SERVER" server get 0 9000
switch "SHOW" defshow get defshow
if /noshow then set defshow 0
switch "FILE" t:file get "NULL"

! parse command line and calculate number of windows
set l:config 0
set app1 "None"
do ii appoff this.args.size
  set appn this.args.geto(ii)
  if appn neqs "THIS.ARGS.GETO(II)" then
    set l:config config+1
    set app^config appn
  endif
enddo

if config le 0 set l:config 1
switch "NW" l:config get config
switch "FILL" s:fill get "LRTB"
switch "ACTIONS" actions get "Monitor,Rec/PB,Stop,Cntrls,More,Exit"
if actions eqss "CFG." then set actions ^actions

! calculate layout of windows
set l:ny 1
if config gt 3 set l:ny 2
set l:nx (config+ny-1)/ny
calc l:nc cfg.getsize 8 / round 5 min 1 max

pipe on
  set l:timeout -1
  env get "SERVER" remote
  call setLayout
  panel/controls=gc/setup/grid=tgrid/theme=^theme/display=^display
  set reg.panel.eventfilter "NOMOVE"
  gcontrol button "ACTION" "" "Config,Exit" /input
  gcontrol label  "CONFIG"   "Config" "Open"
  gcontrol choice "SCENARIO" "Scenario " scenarios scenario
  gcontrol choice "NAPP"     "Windows  " "1,2,3,4,6,8,10,12,16" "^config" /nc=3
  gcontrol prompt "COMMAND"  "Run:" " " 20 
  do nn 1 config
    gcontrol choice "CASE_^nn" "App^nn" apps /nc=nc
  enddo
 if file neqs "NULL" then
  gcontrol label  "FILE"   "File" "Open"
  foreach key intable file
    sedit file.^key args TRIM "(" ")" subs "^key" "^{key}_0"
    iceutil run gcontrol ^args
  endfor
 endif
  if server gt 0 rmif/http server "q:ack"
  set gc.action.itemcolors {MONITOR=darkgreen,REC/PB=darkgreen,STOP=darkred}
  if scenario neqs "NONE" then set gc.CONFIG.action "CLOSED"
pipe run
  call doconfig 
  if jicd neqs "NONE" then 
    invoke java.lang.System.setProperty("jicd.properties.file","^jicd")
    invoke isa=nxm.ice.api.IceSensor.forMacro(THIS)
  endif
  set this.timers 1
  set this.timer(0) 1
pipe off
  if isa rexists then invoke "isa.shutdown()"

endmacro

procedure close
call undoconfig
return

procedure processMessage

if msg.name eqs "TIMER" then
  if /autoexit then
    res count 0
    do nn 1 config
      if reg.snap^nn rexists then res count count+1
    enddo
    if count eq 0 then call setAction "EXITNOW"
  endif
  if timeout ge 0 then
    if timeout eq 0 then call setAction "EXITNOW"
    set l:timeout timeout-1
  endif
  foreach key intable gc
    if "_" nsubs key continue			! dont poll local controls
    if gc.^{key}.flag(8) isfalse continue	! dont poll editable controls
    sedit key l:nn trim "_"
    sedit key name trim ,, "_"
    if nn eq 0 then set l:nn 1
    if reg.snap^nn nrexists continue
    set val reg.snap^{nn}.res.gc.^{name}.v
    if name eqs "QFREQ" then 			! special handling of QFREQ control
      ! if SNAPAPP has RF widget that is not replicated in SNAPPER, add adjustment here
      if gc.rffreq_^{nn} rexists and reg.snap^{nn}.res.gc.rffreq nrexists then set val val+gc.rffreq_^{nn}.v
    endif
    set gc.^{key}.v val
  endfor
  
elseif msg.name eqs "SCENARIO" then
  call setAction "STOP"
  call undoconfig
  call setscene msg.data
  sendto/sync panel "LOADSETUPTABLE" {L1=^tgrid}
  call doconfig

elseif msg.name eqs "NAPP" then
  if ^msg.data eq config return
  call setAction "STOP"
  call undoconfig
  set l:config ^msg.data
  call setLayout
  sendto/sync panel "LOADSETUPTABLE" {L1=^tgrid}
  call doconfig

elseif msg.name eqss "CASE_" then
  call setAction "STOP"
  sedit msg.name l:nn trim "_"
  call stopsnapper nn
  call formsnapper nn msg.data
  call runsnapper nn

elseif "_" subs msg.name then
  sedit msg.name l:nn trim "_"
  sedit msg.name name trim ,, "_"
  if appext neqs "NONE" then run ^appext "MSG" name msg
  if nn eq 0 then
    do nn 1 config
      if name eqs "AFNAME" and config gt 1 then
        set reg.snap^{nn}.res.gc.^{name}.action "^{msg.data}_^{nn}"
      elseif reg.snap^{nn}.res.gc.^{name} rexists
        set reg.snap^{nn}.res.gc.^{name}.action msg.data
        if name eqs "DEC" then call setDec msg.name 1
      endif
    enddo
  elseif reg.snap^{nn}.res.gc.^{name} rexists
    set reg.snap^{nn}.res.gc.^{name}.action msg.data
    if name eqs "DEC" then call setDec msg.name nn
  endif

elseif msg.name eqs "ACTION" then
  call setAction msg.data

elseif msg.name eqs "COMMAND" then
  sedit "^{msg.data}" mdata "GSUBS" "%27" "'" "GSUBS" "%22" "\"" "GSUBS" "%20" " " 
  run ^mdata

elseif msg.name eqs "MORE" then
  if msg.data eqs "ALLCONTROLS"
    call setShow 2
  elseif msg.data eqs "CHECKDATAAPP1V2" 
    timex now time
    if config neq 2 or reg.snap1 nrexists or reg.snap2 nrexists then
      warn "App1 & App2 must be run to apply file equality check"
    elseif reg.snap1.res.ramfile2 rexists then
      icediff reg.snap1.res.ramfile1{fs=0} reg.snap2.res.ramfile1{fs=0}
      icediff reg.snap1.res.ramfile2{fs=0} reg.snap2.res.ramfile2{fs=0}
    else
      icediff reg.snap1.res.ramfile{fs=0} reg.snap2.res.ramfile{fs=0}
    endif
  elseif msg.data eqs "CHECKDATAPORT1V2" 
    timex now time
    do nn 1 config
      if reg.snap^{nn} rexists and reg.snap^{nn}.res.ramfile2 rexists then
        icediff reg.snap^{nn}.res.ramfile1{fs=0} reg.snap^{nn}.res.ramfile2{fs=0}
      else
        warn "App^nn must be run in /DUAL mode to apply file equality check"
      endif
    enddo
  elseif msg.data eqs "TFDoApp1V2" 
    if config neq 2 or reg.snap1 nrexists or reg.snap2 nrexists then
      warn "App1 & App2 must be running to apply TFDOA check"
    else
      if reg.snap1.res.gc.MODE.v eqss "RTMEM" then
        erase/warn=no reg.snap1.res.gc.AFNAME.v reg.snap2.res.gc.AFNAME.v
        say "Snapshot to ^reg.snap1.res.gc.AFNAME.v and ^reg.snap2.res.gc.AFNAME.v"
        set atfn "^{reg.snap1.res.gc.AFNAME.v}{AUX=^reg.snap1.res.aux0}"
        invoke reg.snap1.reg.sp.startArchiver(^atfn) 
        pause 0.5
        set atfn "^{reg.snap2.res.gc.AFNAME.v}{AUX=^reg.snap2.res.aux0}"
        invoke reg.snap2.reg.sp.startArchiver(^atfn) 
        pause 2.0
        invoke reg.snap1.reg.sp.stopArchiver() 
        invoke reg.snap2.reg.sp.stopArchiver() 
      endif
      icetfdoa ^reg.snap1.res.gc.AFNAME.v ^reg.snap2.res.gc.AFNAME.v ,,1, ^reg.snap2.res.gc.FREQ.v*1e6
    endif
  elseif msg.data eqs "STATUS"
    pic status ^reg.snap1.res.gc.card.v
  elseif msg.data eqs "DMACX"
    pic dmacx ^reg.snap1.res.gc.card.v
  else
    call setAction msg.data
  endif

elseif msg.name eqs "EXITCHK" then
 if msg.data neqss "N" pipe stop
endif
return

procedure setAction s:action
if action eqss "EXIT"
  set narch 0
  do nn 1 config
    if reg.snap^nn rexists and reg.snap^{nn}.res.gc.mode.v eqss "RTARCH" then set narch narch+1
  enddo
  if narch gt 0 and action neqs "EXITNOW" then
    gc/temp/msg=y/tleft menu "EXITCHK" "Archive in progress. Are you sure?" "Yes,No" "No"
  else
    pipe stop
  endif
elseif action eqs "POWEROFF"
  set/global POWEROFF "TRUE"
  pipe stop
elseif action eqs "NULL"
elseif action eqs "NONE"
elseif action eqss "MORE"
  gc/temp/nc=1 menu "MORE" "Action" &
   "Setup,Monitor,Record,Playback,DeArchive,OneShot,Archive,OneShotArch,CheckDataApp1v2,CheckDataPort1v2,TFDoApp1v2,PowerOff,DmacX,Status"
elseif action eqs "CONFIG" 
  set gc.CONFIG.v "OPEN"
  call setShow 1
elseif action eqs "CNTRLS" 
  call setShow calc(show+1,modcntrls,mod)
elseif reg.snap1 nrexists then 
  if action neqs "STOP" gc/temp/msg=n alert ,,, "At least App1 must be configured first"
else
  if config eq 1 	! single snapper 
    message send snap1 ,, "MODE" 0 action	
  else			! multi mode startup - special sequencing
    do nn 1 config
      if reg.snap^nn nrexists continue
      message send snap^nn ,, "MODE" -1 action		! setup, no start master/slave
      if action neqs "STOP" then message sync snap^nn	! wait till setup but not for stop's archive flush
    enddo
    do nn config 1 -1
      if reg.snap^nn nrexists continue
      if action eqs "STOP" continue
      if "SHOT" subs action or "ONE" subs action
        message send snap^nn ,, "MODE" 0 "START.ONE"	! now start master/slave 
      else
        message send snap^nn ,, "MODE" 0 "START.CONT"	! now start master/slave 
      endif
      if /nosync isfalse message sync snap^nn		! wait till running
    enddo
  endif
  switch "TIMEOUT" l:timeout get -1
  if action eqss "MON"
    set gc.action.itemcolors {MONITOR=darkred,REC/PB=darkgreen,STOP=darkgreen}
  elseif action eqss "REC"
    set gc.action.itemcolors {MONITOR=darkgreen,REC/PB=darkred,STOP=darkgreen}
  else
    set gc.action.itemcolors {MONITOR=darkgreen,REC/PB=darkgreen,STOP=darkred}
  endif
endif
return

procedure doconfig
do nn 1 config
  if app^nn rexists then
    res appn app^nn
  else
    res appn "NONE"
  endif
  call formsnapper nn appn
enddo
if /autoconfig and config gt 0 then
  picd get ^snapset1.card "CONFIG" "HWALIAS_^snapset1.card"
  info "AutoConfig of ^snapset1.card = ^HWALIAS_^snapset1.card"
endif
if appext neqs "NONE" then run ^appext "CONFIG" scenario 
do nn 1 config
  call runsnapper nn
  if appext neqs "NONE" then run ^appext "RUN" app^nn snapset^nn 
enddo
if reg.snap1 rexists then call setShow defshow
if /action neqs "NULL" then call setAction ^/action
return

procedure undoconfig 
if reg.snap1 rexists and reg.snap1.res.gc.mode.v neqs "SETUP" then
  call setAction "STOP"
endif
do nn 1 config
  if appext neqs "NONE" then run ^appext "STOP" app^nn snapset^nn
  call stopsnapper nn
enddo
if appext neqs "NONE" then run ^appext "UNCONFIG" scenario 
call setShow 1
return

procedure stopsnapper l:nn
if reg.snap^{nn} nrexists return
set reg.snap^{nn}.res.gc.mode.action "EXIT"
while reg.snap^nn rexists 
  pause 0.1
endwhile
return

procedure setShow l:nshow
if ( nshow eq 2 and show neq 2 ) or ( nshow neq 2 and show eq 2 ) then
  do nn 1 config
    if reg.snap^{nn} nrexists continue
    set reg.snap^{nn}.reg.panel.toggleborders 1
  enddo
endif
if ( nshow eq 1 and show neq 1 ) or ( nshow neq 1 and show eq 1 ) then
  set reg.panel.toggleborders 1
endif
set show nshow
return

procedure formsnapper l:nn s:appcase
timex now time
set reg.win.snap^{nn}.title "App^nn : ^appcase"
if gc.CASE_^nn rexists then
  set gc.CASE_^{nn}.v appcase
else
  gcontrol/grp="CONFIG" choice "CASE_^nn" "App^nn" apps appcase /nc=nc 
endif
rem snapset^nn
if appcase eqs "NONE" return
set casename appcase
set caselist appcase
while cfg.case_^{casename}.CASE rexists
  set casename cfg.case_^{casename}.CASE
  set caselist "^casename,^caselist"
endwhile
table controls CREATE {}
if snapdef_controls rexists then table controls MERGE controls snapdef_controls
table snapdef copy snapset^nn	! copy default table
foreach casename inlist caselist
  res t:case cfg.case_^casename	! get the case table
  if case eqs "NULL"
    say "Case ^casename not found in config file"
    return
  endif
 foreach key intable case	! loop to override config
  if key eqs "CASE" then	! copy other case
  elseif key eqs "CONTROLS" then ! copy other case
    table controls MERGE controls case.controls
  else
    set value case.^key
    if "$CASE-" subs value then			! remove item
      sedif value value trim "$CASE-" ""
      sedit value snapset^{nn}.^key subs "^value" ""
    elseif "$CASE+" subs value then		! add item
      sedit value value subs "$CASE+" "^snapset^{nn}.^key"
    elseif "$CASE" subs value then		! replace item
      sedit value value subs "$CASE" "^snapset^{nn}.^key"
    endif
    set snapset^{nn}.^key value
  endif
 endfor
endfor
if /bside then
  sedit snapset^{nn}.port snapset^{nn}.port subs "1" "2"
endif
if controls.getsize gt 0 then set l:defshow 1
table snapset^{nn} MERGE snapset^{nn} usetbl
if gc.LABEL_^nn rexists then
  set gc.LABEL_^{nn}.title "App^nn : ^appcase"
else
  set clist "NULL"
  gcontrol label "LABEL_^nn" "App^nn : ^appcase" 
  foreach key intable controls
    sedit controls.^key args TRIM "(" ")" subs "^key" "^{key}_^nn"
    iceutil run gcontrol ^args
    set clist "^clist,^key"
  endfor
  sedit clist clist^nn trim "NULL,"
endif
return

procedure runsnapper l:nn
if snapset^nn nrexists return
set switches snapset^{nn}.switches
set flags    snapset^{nn}.flags
if "VERBOSE" subs switches then
  say "The VERBOSE switch is only supported on the command line of SNAPAPP.  Removed ..."
  sedit switches switches subs "VERBOSE" "XVERBOSE"
endif
if server gt 0 set switches "^switches/SERVER=^server+^nn"
if verbose gt 0 set switches "^switches/VERBOSE=^verbose"
if noreset gt 0 set switches "^switches/NORESET"
switch aflags^nn aflags get ""
call checkNoReset
if /autosync then call applyAutoSync
if test eq 1 then
  say "Case ^nn   = ^app^{nn}"
  say "Flags    = ^flags"
  say "Switches = ^switches"
  return
endif
snapper/bg/id=snap^nn/theme=^theme/^switches ,, "^aflags" snapset^nn
if show neq 2 set reg.snap^{nn}.reg.panel.toggleborders 1
if nn eq 1 then set gc.ACTION.items actions
! initialize the operator controls
foreach item inlist clist^nn
  if reg.snap^{nn}.res.gc.^{item} nrexists continue		! not a mirror
  if item eqs "AFNAME" and gc.AFNAME_^{nn}.value neqs "" then	! push to snapper
    set reg.snap^{nn}.res.gc.^{item}.value gc.^{item}_^{nn}.value
  else								! pull from snapper
    set gc.^{item}_^{nn}.value reg.snap^{nn}.res.gc.^{item}.value
    if item eqs "DEC" then call setDec DEC_^nn nn
  endif
endfor
return

procedure checkNoReset
if /reset then return
do ii 1 nn-1
  if snapset^{nn}.card eqs snapset^{ii}.card then
    if /quiet isfalse then
      say "Applying NORESET switch to App^nn using same card=^snapset^{ii}.card as App^ii"
    endif
    set switches "^switches/NORESET"
    break
  endif
enddo
return

procedure getside s:gport u:gside
sedit gport side "RANGE" gport.length()-1 gport.length()-1
calc/type=l ^gside ^side 1 - 2 mod 1 +
return

procedure applyAutoSync
call getside snapset1.port side1
call getside snapset^{nn}.port siden
if snapset^{nn}.card eqs snapset1.card then
  do ii 2 config
    if snapset^{ii} rexists and snapset^{ii}.card neqs snapset1.card then
      if /TP then
        set aflags "TPOE|^aflags"
      else
        set aflags "XSOE|^aflags"
      endif
      break
    endif
  enddo
endif
if snapset^{nn}.card neqs snapset1.card then
  if /TP then
    set aflags "XTGO|XSTP|^aflags"
  else
    set aflags "XTGO|^aflags"
  endif
elseif siden neq side1 then
  set aflags "SGO|^aflags"
elseif nn gt 1
  set aflags "RGO|^aflags"
endif
if /quiet isfalse and aflags neqs "" then
  say "Applying ^aflags flag to App^nn to autosync Port=^snapset^{nn}.port"
endif
return

procedure setScene s:name
set appargs cfg.applist.^name
if appargs eqss "(" sedit appargs appargs range 1 appargs.length()-1
sedit "SNAPAPP," appargs "APPEND" appargs
iceutil "MODARGS" appargs this.args
set l:appoff 1
set l:configl config
set l:config 0
do ii appoff this.args.size
  set l:config config+1
  set app^config this.args.geto(ii)
enddo
if gc.NAPP rexists set gc.NAPP.v config
if gc.CONFIG rexists set gc.CONFIG.action "CLOSED"
set scenario name
call setLayout
return

! calculate layout of windows
procedure setLayout 
set l:ny 1
if config gt 3 set l:ny 2
set l:nx (config+ny-1)/ny
switch "NW" l:config get config
switch "FILL" s:fill get "LRTB"
set t:tgrid {type=grid,nx=^nx,ny=^ny,fill=^fill,name=snap,title=app}
return

procedure setDec s:lname l:nn
  pause .25
  set gc.^{lname}.units reg.snap^{nn}.res.gc.DEC.units
  invoke gc.^{lname}.refresh
return
