package nxm.ice.prim;

import nxm.ice.lib.IceResource;
import nxm.sys.lib.DataFile;
import nxm.sys.lib.Primitive;
import nxm.sys.libg.GProgress;

/** Copies data from input file. This primitives separates (detaches) the 
    header and the data into separate files. It also instantiates the native
    read & write functions in the {@link nxm.sys.lib.IOResource}.

  @author xmidas

  TODO test packet handler

 */
public class icecopy extends Primitive {

  /** default transfer length (20M for now) */
  private static final int XFER = 20971520;
  /** input data file */
  private DataFile hi;
  /** output data file */
  private DataFile ho;
  /** memory pointer */
  private long ptr;
  /** input file length */
  private long length;
  /** input file offset */
  private long offset;
  /** progress */
  private GProgress gpw;

  public int open () {    
    // open the data files
    hi = MA.getDataFile("IN");
    hi.open();
    ho = MA.getDataFile("OUT", hi, 0);
    ho.setDetached( MA.getState("/ATTACH", false) ? 0 : 1);
    ho.open(DataFile.OUTPUT);
    // get user supplied transfer length
    xfer = MA.getL("/TL", XFER);    
    // allocate a buffer    
    ptr = IceResource.malloc(xfer);
    if ( ptr < 0 ) M.error("Unable to allocate buffer");
    // init/set read offset and size
    offset = 0;
    length = (long) (hi.getSize()*hi.bpa);
    // check for progress window
    if (MA.find("/GPW")) {
      String text = "ICECOPY "+hi.getURL()+" to "+ho.getURL();
      gpw = new GProgress(this,"Transfer Progress","Run,Pause,Abort",text,0,0,0,this);
      gpw.setLabel("PROGRESS");
    }
    return NORMAL;
  }

  public int process () {
    // determine the size
    int n = (int) (( (length-offset) > xfer ) ? xfer : length-offset);
    // check for completion
    if ( n <= 0 ) return FINISH;
    // read from input file
    int nr = hi.read(ptr, 0, n);
    if (gpw!=null) gpw.setProgress(getProgress());
    // write to the output file
    int nw = ho.write(ptr, 0, nr);
    if ( nr != nw ) M.error("Problem performing I/O");
    offset += nr;
    return NORMAL;
  }

  public int close () {
    if ( ptr > 0 ) IceResource.free_mem(ptr);
    hi.close();
    ho.close();
    if (gpw!=null) gpw.close();
    return FINISH;
  }

  /** get the 0->1 progress of the transfer */
  public double getProgress() {
    return ((double)offset)/length;
  }
  
}
