package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements a generic CORE library function 

  @author Jeff Schoen
  @version $Id: icecore.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icecore extends Primitive implements Chainable, Keyable {

  private int dec=1,dump,tl=4096;
  private DataFile hi,ho;
  private Data dbi,dbo;
  private Core core;
  private Table keys;

  public int open() {

    hi = MA.getDataFile ("IN","1000,2000","S#,C#",0);
    hi.open(hi.OPTIONAL);

    String iform = MA.getS("/IFORM",hi.getFormat());
    String oform = MA.getS("/OFORM",iform);

    String func = MA.getCS("FUNC","Bogus");
    core = Core.forName(func,MA);
    core.setFormats(iform,oform);

    keys = MA.getTable("/CFG"); // parameter table from command line
    if (keys==null) keys = new Table();
    keys.setMode("KV");

    for (int n=4; n<=MA.numberOf(); n++) {
      String key = MA.getKey(n);
      Data value = MA.getData(key);
      keys.put(key,value); // to show in Query
    }

    String[] list = keys.getKeys();
    for (int i=0; i<list.length; i++) {
      String key = list[i];
      Data value = keys.getData(key);
      core.set(key,value);
      if (key.indexOf("DEC")==0) dec = Convert.o2l(value);
    }
    core.open();

    dec = MA.getL("/ODEC",dec);

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setXDelta(hi.getXDelta()*dec);
    ho.setFormat(oform);
    ho.open(ho.OPTIONAL);

    if (hi.typeClass==2) {
      tl = hi.getSubsize();
      hi.setDFS(0);
      ho.setDFS(0);
    }

    tl = MA.getL("/TL", tl);
    dbi = hi.getDataBuffer(tl); 
    dbo = ho.getDataBuffer(tl);     
    
    dump = MA.getL("/DUMP",0);
    if ((dump&1)!=0) dumpKeys();
    return (NORMAL);
  }

  public int process() {
    if (!hi.isOpen) return(NOOP);
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = core.process (dbi,n, dbo,tl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    if ((dump&2)!=0) dumpKeys();
    core.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public void dumpKeys() {
    String[] list = keys.getKeys();
    for (int i=0; i<list.length; i++) {
      String key = list[i];
      byte type = (byte)'L';
      if (key.charAt(1)==':') type = (byte)key.charAt(0);
      Data value = core.get(key,type);
      MT.writeln("Key="+key+" Value="+value+" Hex="+Convert.l2x(Convert.o2l(value)));
    }
  }

  // Chainable interface
  public Object getNextLink() { return core; }
  public Object getPrevLink() { return null; }

  // use the keyable interface to expose all of the PIC library keys
  public String[] getKeys() {
    return keys.getKeys();
  }
  public Object setKey (String key, Object value) {
    Data data = (value instanceof Data)? (Data)value : new Data(value);
    core.set(key,data);
    return value;
  }
  public Object getKey (String key) {
    Object value = null;
    value = core.getL(key);
    return value;
  }


}
