package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements various demodulators on a complex baseband input.

  This primitive has multiple algorithms: Standard and LookUpTable. 
  This primitive has multiple implementations: Java, Native-C, and FPGA

  @author Jeff Schoen
  @version $Id: icedemod.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icedemod extends Primitive implements Chainable {

  /** list of currently supported modulation types */
  public String modList,algList;

  private int mode,alg,tl=4096;
  private DataFile hi,ho,hlut;
  private Data dbi,dbo;
  private Core dm;

  public int open() {

    hi = MA.getDataFile ("IN","1000","C#",0);
    hi.open();

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setFormatMode(SCALAR);
    ho.open();

    dm = Core.forName("Demod",MA);

    modList = dm.getS("MODULATIONS");
    mode = MA.getChoice("MOD",modList,0);
    dm.set("MODULATION",mode);

    hlut = MA.getDataFile ("/LUT","2000","SI",0);
    hlut.open(hlut.OPTIONAL);
    boolean lut = hlut.isOpen();
    if (lut) {
      int nlut = (int)hlut.size;
      int flags = hlut.keywords.getL("FLAGS",0);
      Data dlut = hlut.getDataBuffer(nlut);
      hlut.read(dlut);
      dm.set("LUT.FLAGS",flags);
      dm.set("LUT.DATA",dlut);
      hlut.close();
    }

    double fullScale =  1.0;
         if (ho.getFormatType()==INT) fullScale = 65536.0;
    else if (ho.getFormatType()==LONG) fullScale = 4294967296.0;
    if (MA.getState("/FREQOUT")) {
      if (hi.getFormatType()==FLOAT || hi.getFormatType()==DOUBLE) fullScale *= 1.0/hi.getXDelta();
      else M.warning("Can not use frequency scale for non-Float data types");
    }
    dm.set("FULLSCALE",fullScale);
    dm.open();

    tl = MA.getL("/TL", tl);
    byte dtype = lut? INT : FLOAT;
    dbi = hi.getDataBuffer(tl,dtype); 
    dbo = ho.getDataBuffer(tl,dtype);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = dm.process (dbi,n, dbo,tl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    dm.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return dm; }
  public Object getPrevLink() { return null; }

}
