package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.dsp.prim.firkais;

/**
  Implements FFT functions on a complex input signal.

  This primitive has multiple implementations: Java, and FPGA

  @author Jeff Schoen
  @version $Id: icefft.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icefft extends Primitive implements Chainable {

  private int nfft,nwind,navg,nmax,nexp,nout,shift,flags,tl,xl;
  private double over,factor=1;
  private boolean mag,inv,ppf,real;
  private String wname;
  private DataFile hi,ho;
  private Data dbi,dbo,dbw;
  private Core fft;

  public int open() {

    fft = Core.forName("FFT",MA);
    String flagsList = fft.getS("FLAGSLIST");

    hi = MA.getDataFile ("IN","1000","S#,C#",0);
    hi.open();

    nfft  = MA.getL("NFFT");
    wname = MA.getS("WIN");
    over  = MA.getD("OVER");
    navg  = MA.getL("NAVG");
    nmax  = MA.getL("/NMAX",1);
    nexp  = MA.getL("/NEXP",1);
    shift = MA.getL("/SHIFT",0);
    flags = MA.getOptionMask("/FLAGS",flagsList,0);

    boolean real = (hi.getSPA()==1);
    if (real) flags = Parser.mask(flagsList,"+REAL",flags);

    int nout = real? nfft/2 : nfft;
    int nrot = real? 0 : nfft/2;
    int stride = (int)(nfft*(1-over));
    double delx = hi.getXDelta();
    double delf = 1.0/(nfft*delx);
    double dely = delx*stride;
    double stry = hi.getXStart(); 
    double osize = Math.max(1.0,Math.floor((hi.getSize()-nfft)/stride/navg/nmax));

    // generate the window function
    if (wname.equals("FPPF")) {
      int nptap=nfft, nppt=4, ntap=nptap*nppt;
      float[] fbo = new float[ntap*2];
      float[] taps = new float[ntap];
      firkais.generate(1, 0.80/nptap,0.0,0.2/nptap, 70.0, fbo,ntap,false);
      for (int i=0; i<ntap; i++) taps[i]=fbo[i+i]*nfft; // complex to real
      dbw = new Data(taps);
      flags = Parser.mask(flagsList,"+PPF",flags);
    }
    else dbw = DataOp.getWindow(wname,nfft,factor,M);
    if (nrot!=0) for (int i=0; i<dbw.size; i+=2) dbw.setF(i,-dbw.getF(i));

    mag = fft.isSet(flagsList,"MAG",flags);
    inv = fft.isSet(flagsList,"INV",flags);
    ppf = fft.isSet(flagsList,"PPF",flags);

    fft.set("NFFT",nfft);
    fft.set("NOUT",nout);
    fft.set("NAVG",navg);
    fft.set("NMAX",nmax);
    fft.set("NEXP",nmax);
    fft.set("FLAGS",flags);
    fft.set("SHIFT",shift);
    if (ppf) fft.set("FPPF",dbw);
    else     fft.set("WINDOW",dbw);
    fft.open();

    byte dtype = fft.getDataType();
    String oform = (mag?"S":"C") + "F";

    ho = MA.getDataFile("OUT","2000",oform,0);
    ho.propagate(hi,"TIMECODE|KEYWORDS"); 
    ho.setSubSize (nout);
    ho.setXStart(-nrot*delf);
    ho.setXDelta(delf);
    ho.setXUnits(Units.FREQUENCY);
    ho.setYStart(stry + (nfft+(navg*nmax-1)*stride)/2 * delx);
    ho.setYDelta(dely*navg*nmax);
    ho.setYUnits(Units.TIME);
    ho.setSize(osize);
    ho.open();
    ho.setDFS(0);

    tl = MA.getL("/TL", nfft);
    xl = 2*tl*nout/nfft;
    dbi = hi.getDataBuffer(tl,dtype); 
    dbo = ho.getDataBuffer(xl);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = fft.process (dbi,n, dbo,xl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    fft.close();
    hi.close();
    ho.setSize(Math.floor(ho.seek()/nfft));
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return fft; }
  public Object getPrevLink() { return null; }

}
