package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements an FIR Filter on time domain data.

  @author Jeff Schoen
  @version $Id: icefir.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icefir extends Primitive implements Chainable {

  private int dec,gain,order,nfir,tl=4096;
  private DataFile hi,ho,hf;
  private Data dbi,dbo,dbf;
  private Core fir;

  public int open() {

    hi = MA.getDataFile ("IN","1000","S#",0);
    hi.open();

    hf = MA.getDataFile ("FIR","1000","S#",0);
    hf.open();

    dec = MA.getL("/DEC",1);
    gain = MA.getL("/GAIN",0);
    order = MA.getL("/ORDER",4);

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setXDelta(hi.getXDelta()*dec);
    ho.open();

    fir = Core.forName("FIR",MA);

    fir.set("DECIMATION",dec);
    fir.set("ORDER",order);
    fir.set("GAIN",gain);

    nfir = (int)hf.getSize();
    dbf = hf.getDataBuffer(nfir,FLOAT); 
    hf.read(dbf);
    hf.close();
    fir.set("TAPS",dbf);

    fir.open();

    byte dtype = INT;
    tl = MA.getL("/TL", tl);
    dbi = hi.getDataBuffer(tl,dtype); 
    dbo = ho.getDataBuffer(tl,dtype);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = fir.process (dbi,n, dbo,tl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    fir.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return fir; }
  public Object getPrevLink() { return null; }

}
