package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Class Loader for running a JVM on an ICE card.

  @author Jeff Schoen
  @version %I%, %G%
*/
public class icejvm extends Primitive {

  Table refTab,lnTab,bcTab,foTab;
  OpCode[] joc;
  ConstantPool[] jcp;

  byte rep=DataTypes.IEEE; 
  int classPtr=0;
  int cSlots=1,vSlots=0,runcp=0;
  int codeBase,cpLen,codeLen,maxBytes;
  byte[] dbuf,mbuf,tbuf;
  boolean v1,v2,pre;

  public int open() {
    String line;
    pre = MA.getState("/PRE");
    int v = MA.getL("/VERBOSE");
    v1 = (v>=1);
    v2 = (v>=2);

    // Open and parse the field offset file
    foTab = Convert.o2t("nxm.ice.dat.jvmfields.tbl",M);

    // Open and parse the byte code support file
    TextFile bc = new TextFile(M,"nxm.ice.dat.jvmcodes.lst");
    bc.open(bc.INPUT);
    bcTab = new Table();
    joc = new OpCode[256];
    while ((line=bc.readProper())!=null) new OpCode(line);
    bc.close();

    // Open the byte code instrucion file
    DataFile bf = new DataFile (M,"nxm.ice.dat.jvmcodes.prm","1000","SL",0);
    bf.open(bf.OUTPUT);
    Data bfd = bf.getDataBuffer(512);
    for (int i=0; i<256; i++) if (joc[i]!=null) { 
      if (joc[i].code<0 || joc[i].code>255) M.error("Bad Byte Code definition: "+joc[i]);
      int c = joc[i].code & 0xFF; 
      bfd.setL(0x000|c,joc[i].inst1); 
      bfd.setL(0x100|c,joc[i].inst2); 
    }
    bf.write (bfd);
    bf.close();

    // Prep the compiled Java ByteCode array
    maxBytes= MA.getL("/MAXCODE",0x10000);
    dbuf = new byte[maxBytes];
    mbuf = new byte[maxBytes];
    tbuf = new byte[2]; tbuf[1]=2; tbuf[0]=0;

    // Open the top class file
    jcf2m (MA.getCS("JCF"));
    
    // Open and write the Midas load file
    DataFile df = MA.getDataFile("JLF","1000","SI",0,codeLen,null);
    df.setComment("ICE XILINX JVM Load File");
    df.keywords.put("CODELEN",new Data(codeLen));
    df.setExt("prm");
    df.open(df.OUTPUT);
    for (int i=0; i<2048; i++) { df.write(bfd.buf,i,1); df.write(tbuf,1,1); }
    for (int i=0; i<codeLen; i++) { df.write(dbuf,i,1); df.write(mbuf,i,1); }
    tbuf[1]=0x6;  df.write(tbuf,0,2); /* start code */
    df.close();

    return (FINISH);
  }

  private void print (String s) {
    if (v1) System.out.print(s);
  }
  private void println (String s) {
    if (v1) System.out.println(s);
  }
  private void vprint (String s) {
    if (v2) System.out.print(s);
  }
  private void vprintln (String s) {
    if (v2) System.out.println(s);
  }

/**
  Internal elements:

  PC   - Program Counter
  FPTR - Frame Pointer
  SPTR - Stack Pointer
  CPTR - Class Pointer

  Stack Layout:

  STK0
  STK1   FPTR  LOC0
  STK2         LOC1
  STK3         LOC2 
               LOC3
               LOC4
               STK0
               STK1
               STK2
*/

  private void jcf2m (String jcf) {

    // Open and read the class file
    BaseFile bf = new BaseFile(M,jcf);
    bf.setExt("class");
    bf.open(bf.INPUT);
    int bytes = (int)bf.getSize();
    byte[] buf = new byte[bytes+16];
    bf.read (buf,0,bytes);
    bf.close();

    // parse the class file header
    int i=0,j=0,k=0; 
    int   magic = Convert.unpackL(buf,i,rep); i+=4;
    short minor = Convert.unpackI(buf,i,rep); i+=2;
    short major = Convert.unpackI(buf,i,rep); i+=2;
    println("Magic="+Convert.l2x(magic)+" Minor="+minor+" Major="+major);

    short cpcnt = Convert.unpackI(buf,i,rep); i+=2;
    println("Constant Count="+cpcnt);
    jcp = new ConstantPool[cpcnt]; 
    for (j=1; j<cpcnt; j+=jcp[j-1].slots) i=getConstant(j,buf,i);

    classPtr = 0; // must match cPtr reset value in JVM

    // setup allocation tables
    lnTab = new Table();
    refTab = new Table();
    codeBase = classPtr;
    for (j=1; j<cpcnt; j+=jcp[j-1].slots) setConstant(j);

    short access = Convert.unpackI(buf,i,rep); i+=2;
    short cthis  = Convert.unpackI(buf,i,rep); i+=2;
    short csuper = Convert.unpackI(buf,i,rep); i+=2;
    String cnthis = getC(getI(cthis));
    String cnsuper = getC(getI(csuper));
    println("Access="+getAcc(access)+" this="+cnthis+" super="+cnsuper);

    short ifcnt  = Convert.unpackI(buf,i,rep); i+=2;
    println("Interfaces count="+ifcnt); 
    for (j=0; j<ifcnt; j++) i=getInterface(j,buf,i);

    short fcnt  = Convert.unpackI(buf,i,rep); i+=2;
    println("Field count="+fcnt); 
    for (j=0; j<fcnt; j++) i=getField(j,buf,i);

    cpLen = codeLen = codeBase + cSlots*4; // mark start of code sections

    short mcnt  = Convert.unpackI(buf,i,rep); i+=2;
    println("Method count="+mcnt); 
    for (j=0; j<mcnt; j++) i=getMethod(j,buf,i);

    short acnt  = Convert.unpackI(buf,i,rep); i+=2;
    println("Attributes count="+acnt); 
    for (j=0; j<acnt; j++) i=getAttribute(j,buf,i,false);

    if (i!=bytes) println("Extra bytes="+(bytes-i));

    println("Total Load cSlots="+cSlots+" vSlots="+vSlots+" bytes="+codeLen);

    // insert bootstrap code
    for (k=0; k<classPtr; ) dbuf[k++]=0;

    // invoke run method
    //dbuf[k++]=(byte)0xB8; dbuf[k++]=0; dbuf[k++]=(byte)runcp;  dbuf[k++]=(byte)0xB1;
    dbuf[k++]=0; dbuf[k++]=(byte)0xB8; dbuf[k++]=0; dbuf[k++]=(byte)runcp;

    if (v2) refTab.dump();
    if (!pre) getCode (dbuf,0,codeBase+4,mbuf); // bootstrap
    if (!pre) getCode (dbuf,cpLen,codeLen-cpLen,mbuf);

  }

  private class ConstantPool {
    byte tag;
    int len;
    int size;
    int slot=0;
    int slots=1;
    int offset;
    String type;
    Object value;
    String name;
    public ConstantPool (byte[] buf, int i) {
      int itop=i;
      tag = buf[i++];
      switch (tag) {
        case  1: type="Utf8  "; len=Convert.unpackI(buf,i,rep); i+=2; value = new String(buf,i,len); break;
        case  3: type="Int   "; len=4; value = Data.fromBuffer(buf,rep,i,1,"SL"); break;
        case  4: type="Float "; len=4; value = Data.fromBuffer(buf,rep,i,1,"SF"); break;
        case  5: type="Long  "; len=8; value = Data.fromBuffer(buf,rep,i,1,"SX"); slots=2; break;
        case  6: type="Double"; len=8; value = Data.fromBuffer(buf,rep,i,1,"SD"); slots=2; break;
        case  7: type="Class "; len=2; value = Data.fromBuffer(buf,rep,i,1,"SI"); break;
        case  8: type="String"; len=2; value = Data.fromBuffer(buf,rep,i,1,"SI"); break;
        case  9: type="Field "; len=4; value = Data.fromBuffer(buf,rep,i,2,"SI"); break;
        case 10: type="Method"; len=4; value = Data.fromBuffer(buf,rep,i,2,"SI"); break;
        case 11: type="InterF"; len=4; value = Data.fromBuffer(buf,rep,i,2,"SI"); break;
        case 12: type="NaType"; len=4; value = Data.fromBuffer(buf,rep,i,2,"SI"); break;
        default: System.out.println("Unknown tag="+tag); len=-1;
      }
      size = i+len-itop;
    }
    public int getI (int index) {
      return ((Data)value).getI(index);
    }
    public String getS () {
      return value.toString();
    }
    public String toString() { 
      String s = " type="+type+" value="+value;
      if (name!=null) s += " name="+name;
      if (slot!=0) s += " slot="+slot;
      return s;
    }
  }

  private int getConstant (int id, byte[] buf, int i) {
    ConstantPool cp = new ConstantPool(buf,i);
    jcp[id-1]=cp; 
    return i+cp.size;
  }

  private void setConstant (int id) {
    ConstantPool cp = jcp[id-1];
    switch (cp.tag) { 
      case 3: case 4: case 5: case 6: 		// constants
        cp.slot = cSlots; cSlots += cp.slots;
        int len = cp.slots*4;
        cp.offset = codeBase + cp.slot*4; 
        System.arraycopy(((Data)cp.value).buf,0, dbuf,cp.offset, len);
        break; 
      case 10: 					// methods
        cp.slot = cSlots++;
      case 9: 					// fields
        int cid = cp.getI(0);
        int nid = cp.getI(1);
        int cidn = jcp[cid-1].getI(0);
        int nidn = jcp[nid-1].getI(0);
        int nidt = jcp[nid-1].getI(1);
        String cname = jcp[cidn-1].getS();
        String fname = jcp[nidn-1].getS();
        String ftype = jcp[nidt-1].getS();
        cp.name = cname + "." + fname + "<" + ftype + ">";
	refTab.put(fname,id);
        break; 
    }
    println("Const#"+id+" "+cp);
  }

  private String getC (int id) {
    return jcp[id-1].value.toString();
  }
  private int getI (int id) {
    return Convert.o2l(jcp[id-1].value);
  }

  private int getInterface (int id, byte[] buf, int i) {
    return i+2;
  }

  private final static int 
    A_PUBLIC=0x1,A_PRIVATE=0x2,A_PROTECTED=0x4,A_STATIC=0x8,A_FINAL=0x10,A_VOLATILE=0x20,A_TRANSIENT=0x40,A_NATIVE=0x80;

  private String getAcc (int access) {
    return Parser.mask2s("Public,Priv,Prot,Static,Final,Volatile,Transient,X,Native",access);
  }

  private int getField (int id, byte[] buf, int i) {
    short access = Convert.unpackI(buf,i,rep); i+=2;
    short name   = Convert.unpackI(buf,i,rep); i+=2;
    short descr  = Convert.unpackI(buf,i,rep); i+=2;
    short attr   = Convert.unpackI(buf,i,rep); i+=2;
    int cp = Convert.o2l(refTab.get(getC(name)));
    int slot = 0;
    String tostr = "Field#"+id+" CP="+cp+" access="+getAcc(access)+" attr="+attr+" desc="+getC(descr) +" name="+getC(name);
    if ((access&A_STATIC)!=0) {
      if ((access&A_FINAL)==0) slot = cSlots++;
    } else {
      slot = vSlots++;
    }
    if (cp>0) jcp[cp-1].slot = slot;
    println(tostr+" slot="+slot);
    for (int j=0; j<attr; j++) i=getAttribute(j,buf,i,false);
    return i;
  }

  private int getMethod (int id, byte[] buf, int i) {
    short access = Convert.unpackI(buf,i,rep); i+=2;
    short name   = Convert.unpackI(buf,i,rep); i+=2; String sname = getC(name);
    short descr  = Convert.unpackI(buf,i,rep); i+=2;
    short attr   = Convert.unpackI(buf,i,rep); i+=2;
    int cptr = (classPtr>>2)&0xFFF;
    int tag = (codeLen>>2)&0xFFF;
    int nargs = getNArgs(getC(descr));
    int cp = Convert.o2l(refTab.get(sname));
    String tostr = "Method#"+id+" CP="+cp+" access="+getAcc(access)+" attr="+attr+" desc="+getC(descr)+" name="+getC(name);
    int slot = (cp>0)? jcp[cp-1].slot : 0;
    lnTab.put(""+codeLen,tostr);
    for (int j=0; j<attr; j++) i=getAttribute(j,buf,i,true);
    int gap = methodLocals-nargs+1;
    if (nargs>15) M.error("Too many args="+nargs);
    int mgap = (gap+3)/4;
    if (mgap>15) M.error("Too many args+locals="+gap);
    println(tostr+" Args="+nargs+" Locals="+ methodLocals+" Stack="+methodStack+" slot="+slot);
    if (cp>0) Convert.packL(dbuf,codeBase+slot*4,(tag<<0)|(mgap<<12)|(cptr<<16)|(nargs<<28));
    if (sname.equals("run")) runcp=cp;
    for (;(codeLen&0x3)!=0;codeLen++); // force 4by method align for tableswitch
    return i;
  }

  private int methodStack;
  private int methodLocals;
  private int methodLength;

  private int getAttribute (int id, byte[] buf, int i, boolean code) {
    short name = Convert.unpackI(buf,i,rep); i+=2;
    int   len  = Convert.unpackL(buf,i,rep); i+=4;
    int   inext = i+len;
    vprintln("Attribute#"+id+" name="+getC(name)+" off="+codeLen+" len="+len);
    if (code) {
      methodStack = Convert.unpackI(buf,i,rep); i+=2;
      methodLocals = Convert.unpackI(buf,i,rep); i+=2;
      int length = Convert.unpackL(buf,i,rep); i+=4;
      int itop = i;
      System.arraycopy (buf,itop, dbuf,codeLen, length);
      codeLen += length;
    }
    return inext;
  }

  private int getNArgs (String desc) {
    int nargs=0;
    for (int i=0; i<desc.length(); i++) {
      char c = desc.charAt(i);
      if (c==')') break;
      nargs++;
      if (c=='(') nargs=0;
      if (c=='L') i = desc.indexOf(';',i);
    }
    return nargs;
  }

  private void getCode (byte[] buf, int itop, int length, byte[] mbuf) {
    OpCode bclast = null;
    for (int i=itop; i<itop+length; ) {
      int  ic = buf[i]&0xFF;
      OpCode bc = joc[ic];
      if (bc==null) M.error("Unhandled Java Byte Code = "+ic);
      if (bc.niy) M.warning("Uses unimplemented Java Byte Code "+ic+" = "+bc.name);
      else if (mbuf!=null && !bc.name.equals("nop")) mbuf[i]=0x1;
      OpCode bcq = (OpCode)bcTab.get(bc.name+"_quick");
      Object lns = lnTab.get(""+i);
      if (lns!=null) vprintln(lns.toString());
      int ln = (i-itop);
      String pc = Convert.l2x(i).substring(6);
      String bcx = Convert.l2x(ic).substring(8);
      vprint(" "+pc+" "+ln+" "+bcx+" "+bc.name); i++;
      short iarg = Convert.unpackI(buf,i,rep);
      if (bc.args==2) Convert.swap2(buf,i,1);
      if (bc.args==4) Convert.swap4(buf,i,1);
      if (bc.args>0) { vprint(" "+Convert.bb2hex(buf,i,bc.args)); i+= bc.args; }

      if (bc.niy);
      else if (bc.name.startsWith("goto") || bc.name.startsWith("if")) {
        vprint(" (line="+(ln+iarg)+")");
        Convert.packI(buf,i-2,(short)(itop+ln+iarg)); // absolute addr
      }
      else if (bc.name.startsWith("new")) {
        Data d = (Data) jcp[iarg-1].value;
        vprint(" ("+getC(d.getI(0))+")");
      }
      else if (bc.name.startsWith("invoke")) {
        Data d = (Data) jcp[iarg-1].value;
        Data e = (Data) jcp[d.getI(1)-1].value;
        String mname = getC(e.getI(0));
        String mtype = getC(e.getI(1));
        int cp = refTab.getL(mname);
        vprint(" ("+mtype+" "+mname+")");
        if (mname.equals("getObjectAt") && mtype.equals("(I)Ljava/lang/Object;")) {
          /* replace invokestatic with noops to leave address on stack as new object */
          buf[i-3]=buf[i-2]=buf[i-1]=0;
        } else {
          if (cp<=0) M.warning("No Constant Pool entry for method="+mname);
          else if (iarg>255) M.warning("Method offset="+iarg+" too large for method="+mname);
          else buf[i-2] = (byte)jcp[cp-1].slot; 
        }
      }
      else if (bc.name.startsWith("iinc")) {
        Convert.swap2(buf,i-2,1); // swap back
      }
      else if (bc.name.startsWith("ldc")) {
        int cp =  b2l(buf[i-1]);
        if (cp<=0) M.error("No Constant Pool entry for constant="+cp);
        vprint(" (0x"+Convert.l2x(Convert.o2l(jcp[cp-1].value))+")");
        buf[i-1] = (byte)jcp[cp-1].slot;
        buf[i-2] = bcq.bc;
      }
      else if (bc.name.endsWith("load") || bc.name.endsWith("store")) {
	int off = b2l(buf[i-1]);
	if (off>63) System.out.println(" ERR: load/store offset "+off+" out of range");
      }
      else if (bc.name.startsWith("get") || bc.name.startsWith("put")) {
        Data d = (Data) jcp[iarg-1].value;
        Data e = (Data) jcp[d.getI(1)-1].value;
        Data f = (Data) jcp[d.getI(0)-1].value;
        String fclas = getC(f.toL());
        String fname = getC(e.getI(0));
        String ftype = getC(e.getI(1));
        int ip = fclas.lastIndexOf("/");
        String foname = fclas.substring(ip+1)+"."+fname;
        if (foTab.get(foname.toUpperCase())==null) M.warning("Field="+foname+" not defined in jvmfields");
        int foff = foTab.getL(foname.toUpperCase());
        vprint(" ("+ftype+" "+foname+" off="+foff+")");
        boolean get = bc.name.startsWith("get");
        int cp = refTab.getL(fname);
        if (cp<=0) M.warning("No Constant Pool entry for field="+fname);
        else if (iarg>255) M.warning("Method offset="+iarg+" too large for field="+fname);
	byte xcmd = bcq.bc;
	byte xact = (byte)jcp[cp-1].slot;
	byte xoff = 0;
        // xact mask=(MRD,MWR,MAD,XIO,SRD,SWR,SAD,0)
        if (bc.name.indexOf("static")>0 ) {
          byte xsc = get? (byte)0xfe : (byte)0xff; // for special commands
          if (fname.equals("ioadr")) { xcmd = xsc; xact = 0x02; }
          if (fname.equals("saddr")) { xcmd = xsc; xact = 0x02; }
          if (fname.equals("sdata")) { xcmd = xsc; xact = get? (byte)0x08:(byte)0x04; }
          if (fname.equals("ioctl")) { xcmd = xsc; xact = 0x10; }
          if (fname.equals("maddr")) { xcmd = xsc; xact = 0x20; }
          if (fname.equals("mdata")) { xcmd = xsc; xact = get? (byte)0x80:(byte)0x40; }
        } else {  // class variable
	  xact = get? (byte)0xA0:(byte)0x61;
	  xoff = (byte)((foff>=128)?foff-256:foff); 
        }
	buf[i-3]=xcmd;
        buf[i-2]=xact;
        buf[i-1]=xoff;
      }
      else if (bc.name.startsWith("tableswitch")) {
        for (;((i-itop)&0x3)!=0;i++);
        Convert.swap4(buf,i,3);
        int vdf =  Convert.unpackL(buf,i); i+=4;	// default
        int vlo =  Convert.unpackL(buf,i); i+=4;
        int vhi =  Convert.unpackL(buf,i); i+=4;
        vprint(" ("+vdf+","+vlo+","+vhi+")");
        Convert.swap4(buf,i,vhi-vlo+1);
        for (int j=vlo; j<=vhi; j++) {
          int vjo =  Convert.unpackL(buf,i);		// jump offset
          vprint("\n "+(i-itop)+"   "+j+": "+vjo+" (line="+(ln+vjo)+")"); i+=4;
        }
      }
      else if (bc.name.startsWith("lookupswitch")) {
        for (;((i-itop)&0x3)!=0;i++);
        Convert.swap4(buf,i,2);
        int vdf =  Convert.unpackL(buf,i); i+=4;	// default
        int vnp =  Convert.unpackL(buf,i); i+=4;	// number of match-jump pairs
        vprint(" ("+vdf+","+vnp+")");
        Convert.swap4(buf,i,vnp*2);
        for (int j=0; j<vnp; j++) {
          int vmv =  Convert.unpackL(buf,i);
          int vjo =  Convert.unpackL(buf,i+4);		// jump offset
          vprint("\n "+(i-itop)+"   "+vmv+": "+vjo+" (line="+(ln+vjo)+")"); i+=8;
        }
      }
      bclast = bc;
      if (v2) vprintln("");
    }
  }

  private int b2l (byte b) {
    int l = b;
    if (l<0) l += 256;
    return l;
  }


  private class OpCode {
    
    String name;
    int code,args;
    int inst1,inst2;
    boolean niy;
    byte bc;

    public OpCode (String line) {
      Parser p = new Parser(line,true,false);
      code = Convert.s2l(p.get(1)) & 0xFF;
      args = Convert.s2l(p.get(3));
      name = p.get(4);
      inst1 = parse(p.get(5));
      inst2 = parse(p.get(6));
      niy = (p.get(5).length()+p.get(6).length())==0;
      if (name.endsWith("_quick")) {
	// fill in quick versions with same instructions as non-quick
        OpCode ocnq = (OpCode)bcTab.get(name.substring(0,name.length()-6));
        if (ocnq!=null) { inst1=ocnq.inst1; inst2=ocnq.inst2; bc=ocnq.bc; } 
      }
      bc = (byte)((code>=128)? code-256 : code);
      if (joc[code]!=null) M.error("Overwriting "+joc[code]+" with "+this);
      joc[code] = this;
      bcTab.put(name,this);
    }

    public String toString() {
      return "OpCode="+code+" Name="+name+" Args="+args;
    }

    private int parse (String s) {
      int i, inst=0;
      if (s==null) return(0);
      s = "|"+s+"|";
      if ((i=find(s,"PSET",0))>0)  inst |= 0x00000001;
      if ((i=find(s,"SELX",0))>0)  inst |= 0x00000002;
      if ((i=find(s,"DUAL",0))>0)  inst |= 0x00000004;
      if ((i=find(s,"HOLD",0))>0)  inst |= 0x00000008;
      if ((i=find(s,"MINV",0))>0)  inst |= 0x00000010;
      if ((i=find(s,"MRET",0))>0)  inst |= 0x00000020;
      if ((i=find(s,"XXXX",0))>0)  inst |= 0x00000040;
      if ((i=find(s,"FRMAX",0))>0) inst |= 0x00000700;
 else if ((i=find(s,"FRMA",9))!=9) inst |= ((i&0xF)<< 8);
 else if ((i=find(s,"STKA",1))!=0) inst |= ((-i&0xF)<< 8);
      if ((i=find(s,"FRMBX",0))>0) inst |= 0x00007000;
 else if ((i=find(s,"FRMB",9))!=9) inst |= ((i&0xF)<<12);
 else if ((i=find(s,"STKB",3))!=0) inst |= ((-i&0xF)<<12);
      if ((i=find(s,"STKI",0))!=0) inst |= ((i&0x3)<<16);
      if ((i=find(s,"STKI",0))==1) inst |= 0x001C0080;	// normal stack inc with store
      if ((i=find(s,"STKU",0))==1) inst |= 0x00010000;
      if ((i=find(s,"STKD",0))!=0) inst |= ((-i&0x3)<<16);
      if ((i=find(s,"STKD",0))==1) inst |= 0x000C0000;	// normal stack dec with recall
      if ((i=find(s,"WSTKA",0))>0) inst |= 0x00040000;
      if ((i=find(s,"WSTKB",0))>0) inst |= 0x00080000;
      if ((i=find(s,"WRAMA",0))>0) inst |= 0x00100000;
      if ((i=find(s,"WRAMB",0))>0) inst |= 0x00200000;
      if ((i=find(s,"WRAMC",0))>0) inst |= 0x00400000;
      if ((i=find(s,"ALT",0))>0)   inst |= 0x00800000;
      if ((i=find(s,"SELB","RAMB,STKA"))==1) inst |= 0x00000080;
      if ((i=find(s,"SELA","RAMA,RAMB,RAMC,RAMX,STKB,PUSH,RSHF,LSHF,TBD,AND,ROR,XOR,NEG,SUB,ADD,MUL"))>=0) inst |= (i<<24);
      if ((i=find(s,"OPTS",0))>0)  inst |= (i<<28);
      if ((i=find(s,"COND","H,L,EQ,EQZ,LT,LTZ,GT,GTZ,NH,NL,NE,NEZ,GE,GEZ,LE,LEZ"))>=0) inst |= (i<<28);
      if ((i=find(s,"CVT","I2L,I2F,I2D,L2I,L2F,L2D,F2I,F2L,F2D,D2I,D2L,D2F,I2B,I2C,I2S"))>=0) inst |= (i<<28);
      return inst;
    }

    private int find (String s, String key, int def) {
      key = "|"+key;
      int kl = key.length();
      int i = s.indexOf(key);
      if (i<0) return def;
      int j = s.indexOf("|",i+1);
      if (i+kl+1>=j) return 1;
      return Convert.s2l(s.substring(i+kl+1,j));
    }

    private int find (String s, String key, String list) {
      key = "|"+key;
      int kl = key.length();
      int i = s.indexOf(key);
      if (i<0) return -1;
      int j = s.indexOf("|",i+1);
      if (i+kl+1==j) return -1;
      return Parser.find(list,s.substring(i+kl+1,j),0,-1);
    }
  }
}
