package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Decifer PCI Express debug files and merge into analysys records

  @author Jeff Schoen
  @version $Id: icediff.java,v 1.1 2006/10/07 18:54:49 jgs Exp $
*/
public class icepex extends Primitive {

  private DataFile hd,h1,h2;
  private Data dd,d1,d2;
  private double offset;
  private int diffs,echo;
  private int dumped,maxShow=32;
  private int lw=8; 
  private int tstart,tend,trained=0;
  private int skps=0,ftss=0,idls=0,ts1s=0,ts2s=0,dlps=0,tlps=0,cmps=0,oths=0;

  public static String echoList = "Other,DLP,TLP,TSN,Comp";
  public static int E_OTH=0x01, E_DLP=0x02, E_TLP=0x04, E_TSN=0x08, E_CMP=0x10;

  final static int REG_DVID=0x00, REG_CODE=0x08, REG_PBAR=0x10, REG_MDAU=0x44, REG_IOCR=0x54,
                   REG_FIFO=0x40, REG_MCSR=0x48, REG_MDAR=0x58, REG_MDTC=0x5C, REG_DCSR=0xE4,
		   REG_OMB1=0x60, REG_OMB2=0x64, REG_OMB3=0x68, REG_OMB4=0x6C, REG_LCSR=0xEC;

  private boolean pcieConfig, pciePad;
  private int pcieBAR = 0xD8400000;
  private int lanes,pcilen,pcioff,pcires,pcidat,icrc,lcrc;
  private byte[] pcibuf;
  private int[] pciexp;

  public int open() {

    if (MA.getState("/GENERATE")) {
      lanes = MA.getL("/LANES",lw);
      pciePad = MA.getState("/PAD",false);
      pcigen();
      hd = MA.getDataFile ("OUT","2000","SI",0);
      hd.setFS(lw);
      hd.setSize(pcilen/2/lw);
      hd.open(hd.OUTPUT);
      hd.write(pcibuf,0,pcilen);
      hd.close();
      return (FINISH);
    }

    h1 = MA.getDataFile ("IN1","1000","SX",0);
    h1.open();

    h2 = MA.getDataFile ("IN2","1000","SX",0);
    h2.open(h2.OPTIONAL);

    hd = MA.getDataFile ("OUT");
    hd.open(hd.OUTPUT|hd.OPTIONAL);

    xfer = MA.getL("/TL",1048576);
    d1 = h1.getDataBuffer(xfer);
    d2 = h2.getDataBuffer(xfer);
    dd = hd.getDataBuffer(xfer);

    echo   = MA.getOptionMask("/ECHO",echoList,0);
    tstart = MA.getL("/TSTART",0);
    tend   = MA.getL("/TEND",0);
    maxShow = MA.getL("/SHOW",MA.getState("/ALL")?-1:maxShow);
    
    return (NORMAL);
  }

  public int process() {

   if (h1.isOpen) {
    offset = h1.seek();
    int n1 = h1.read(d1);
    if (n1<=0) return (FINISH);
    int nd1 = checkBytes (d1,n1, dd,xfer);
   }
   if (h2.isOpen) {
    offset = h2.seek();
    int n2 = h2.read(d2);
    if (n2<=0) return (FINISH);
    int nd2 = checkBytes (d2,n2, dd,xfer);
   }
    return (NORMAL);
  }

  public int close() {
    if (h1!=null) h1.close();
    if (h2!=null) h2.close();
    if (hd!=null) hd.close();
    if (trained!=0) M.info("First TS at t="+Convert.l2x(trained));
    return (NORMAL);
  }

  public static byte COMMA = (byte)0xBC;
  public static byte SKP   = (byte)0x1C;
  public static byte IDL   = (byte)0x7C;
  public static byte FTS   = (byte)0x3C;
  public static byte PAD   = (byte)0xF7;
  public static byte TS1   = (byte)0x4A;
  public static byte TS2   = (byte)0x45;
  public static byte STP   = (byte)0xFB;
  public static byte SDP   = (byte)0x5C;
  public static byte END   = (byte)0xFD;
  public static byte OTH   = (byte)0x0;
  public static byte CMP   = (byte)0x1;

  private boolean isAll (byte[] b, int off, int n, byte val) {
    for (int i=0; i<n; i++) {
      if (b[off++]!=val) return false;
    }
    return true;
  }

  public static long COMMAS = 0xBCBCBCBCBCBCBCBCL;
  public static long SKPS   = 0x1C1C1C1C1C1C1C1CL;
  public static long IDLS   = 0x7C7C7C7C7C7C7C7CL;
  public static long FTSS   = 0x3C3C3C3C3C3C3C3CL;
  public static long PADS   = 0xF7F7F7F7F7F7F7F7L;
  public static long TS1S   = 0x4A4A4A4A4A4A4A4AL;
  public static long TS2S   = 0x4545454545454545L;

  public static int RPT1   = 0x01CCBBAA;
  public static int RPT2   = 0x02CCBBAA;
  public static int RPT9   = 0x09CCBBAA;

  private byte[] buffer;
  private long ioff;
  private int boff,time;
  private int rpt,llen,loff;

  private void dump (byte type, int blen) {

    String text = null;

    if (type==0  ) { oths++; if ((echo&E_OTH)!=0) text = "OTH"; }
    if (type==SDP) { dlps++; if ((echo&E_DLP)!=0) text = "DLP"; }
    if (type==STP) { tlps++; if ((echo&E_TLP)!=0) text = "TLP"; }
    if (type==IDL) { idls++; if ((echo&E_TSN)!=0) text = "IDL"; }
    if (type==FTS) { ftss++; if ((echo&E_TSN)!=0) text = "FTS"; }
    if (type==TS1) { ts1s++; if ((echo&E_TSN)!=0) text = "TS1"; }
    if (type==TS2) { ts2s++; if ((echo&E_TSN)!=0) text = "TS2"; }
    if (type==SKP) { skps++; if ((echo&E_TSN)!=0) text = "SKP"; }
    if (type==CMP) { cmps++; if ((echo&E_CMP)!=0) text = "CMP"; }

    if (text==null) return;
    if (tend!=0 && time>tend) return;
    if (tstart!=0 && time<tstart) return;
    if (maxShow>0 && dumped>maxShow) return;

    if (llen>0 && isSame(buffer,loff,llen,boff,blen)) { rpt++; return; }
    if (rpt>0) System.out.println("  x="+rpt);
    System.out.print(text+" t="+Convert.l2x(time)+" d="+Convert.bb2hex(buffer,boff,blen)+" i="+ioff);
    if (type==SDP) {
      String xtra = " c=";
      int dtype = buffer[boff+1]&0xF0;
      switch (dtype) {
        case 0x00: xtra += "Ack    "; break;
        case 0x10: xtra += "Nak    "; break;
        case 0x20: xtra += "PowMod "; break;
        case 0x30: xtra += "Vendor "; break;
        case 0x40: xtra += "IFC1-P "; break;
        case 0x50: xtra += "IFC1-N "; break;
        case 0x60: xtra += "IFC1-C "; break;
        case 0xC0: xtra += "IFC2-P "; break;
        case 0xD0: xtra += "IFC2-N "; break;
        case 0xE0: xtra += "IFC2-C "; break;
        case 0x80: xtra += "UpFC-P "; break;
        case 0x90: xtra += "UpFC-N "; break;
        case 0xA0: xtra += "UpFC-C "; break;
      }
      System.out.print(xtra);
    }
    if (type==STP) {
      short seq = Convert.unpackI(buffer,boff+1,IEEE);
      String xtra = " s="+seq+" c=";
      int dtype = buffer[boff+3]&0x7F;
      switch (dtype) {
        case 0x00: xtra += "MRd    "; break;
        case 0x20: xtra += "MRdx   "; break;
        case 0x01: xtra += "MRdLk  "; break;
        case 0x21: xtra += "MRdLkx "; break;
        case 0x40: xtra += "MWr    "; break;
        case 0x60: xtra += "MWrx   "; break;
        case 0x02: xtra += "IORd   "; break;
        case 0x42: xtra += "IOWr   "; break;
        case 0x04: xtra += "CfgRd0 "; break;
        case 0x44: xtra += "CfgWr0 "; break;
        case 0x05: xtra += "CfgRd1 "; break;
        case 0x45: xtra += "CfgWr1 "; break;
        case 0x0A: xtra += "Cpl    "; break;
        case 0x4A: xtra += "CplD   "; break;
        case 0x0B: xtra += "CplLk  "; break;
        case 0x4B: xtra += "CplLkD "; break;
      }
      if ((dtype&0xF8)==0x30) xtra += "Msg    ";
      if ((dtype&0xF8)==0x70) xtra += "MsgD   ";
      System.out.print(xtra);
    }
    rpt=1;
    dumped++;
    loff=boff; llen=blen;
  }

  private boolean isSame (byte[] b, int off1, int len1, int off2, int len2) {
    if (len1 != len2) return false;
    len1 += off1;
    for (; off1<len1; off1+=8,off2+=8) {
      if (b[off1]==(byte)0xAA && b[off1+1]==(byte)0xBB && b[off1+2]==(byte)0xCC) continue;
      for (int i=0; i<8; i++) if (b[off1+i]!=b[off2+i]) return false;
    }
    return true;
  }

  private int getLength (byte[] b, int boff) {
    int len=0;
    if (b[boff]==SDP) {
      len = 2;
    }
    else if (b[boff]==STP) {
      int head = Convert.unpackL(b,boff+3,IEEE);
      if ((head&0x808F0C00)!=0) return 0;
      int fmt = (head>>29)&0x3;
      int dlen = head&0x3FF;
      if (fmt==0) len = 3;
      if (fmt==1) len = 4;
      if (fmt==2) len = 3+dlen;
      if (fmt==3) len = 4+dlen;
      len += 2;
    }
    return len*4;
  }

  private int checkBytes (Data d, int n, Data dd, int nd) {
    byte[] b = d.buf;
    long b0,b1,b2,b3,b4;
    int bs,be,bl,xlp,nr=0;
    skps=0;ftss=0;idls=0;ts1s=0;ts2s=0;dlps=0;tlps=0;cmps=0;oths=0;
    rpt=0; llen=0;
    buffer = b; // for dumps
    int ftime=0;
    for (int i=0; i<n-8;) {
      boff = i*8;
      ioff = (long)offset+i;
      xlp = getLength(b,boff);
      if (xlp>0) xlp = Math.min(xlp,(b.length-boff)/4);
      b0 = Convert.unpackX(b,boff); i++;
      bs = b[boff]; 
      be = (xlp<=0)? bs : b[boff+xlp-1]; 
      if (b0==COMMAS) {
        boolean hasTime=true; 
        b1 = Convert.unpackX(b,8*i); i++;  // defines special packet
        bl = Convert.unpackL(b,8*i); i++;  // should be repeat of 2
        b2 = Convert.unpackX(b,boff+6*8);  // TS id line
        be = Convert.unpackL(b,boff+7*8);  // should be repeat of 9
             if (b1==SKPS && bl==RPT2) { dump(SKP,20); }
        else if (b1==IDLS && bl==RPT2) { dump(IDL,20); }
        else if (b1==FTSS && bl==RPT2) { dump(FTS,20); }
        else if (b2==TS1S && be==RPT9) { i+=5; if (trained==0) trained=time; dump(TS1,60); }
        else if (b2==TS2S && be==RPT9) { i+=5; if (trained==0) trained=time; dump(TS2,60); }
        else hasTime=false;
        if (hasTime) time = Convert.unpackL(b,i*8-4);
        if (hasTime && ftime==0) ftime=time;
      }
      else if ((int)(b0&0xFFFFFF)==0xCCBBAA) {
        if (ftime!=0) time = Convert.unpackL(b,boff+4);
      }
      else if (bs==SDP && be==END) {
        dump(SDP,8); 
      }
      else if (bs==STP && xlp>0 && (boff+xlp+8)<b.length && b[boff+xlp+7]==END) { // handle unaligned trailing DLP case
        dump(STP,Math.min(48,xlp)); 
        i += (xlp-4)/8;
      }
      else if (bs==STP && be==END) {
        dump(STP,Math.min(48,xlp)); 
        i += (xlp-4)/8;
      }
      else if (b0==0 || b0==SKPS) {
      }
      else if (isAll(b,boff,8,b[boff])) {
        dump(CMP,16);
      }
      else if (isAll(b,boff,4,b[boff]) && isAll(b,boff+4,4,b[boff+4])) {
        dump(CMP,16);
      }
      else {
        dump(OTH,8);
      }
    }
    if (llen>0) System.out.println("  x="+rpt);
    M.info("SKP="+skps+" DLP="+dlps+" TLP="+tlps+" TS1="+ts1s+" TS2="+ts2s
         +" FTS="+ftss+" IDL="+idls+" CMP="+cmps+" OTH="+oths+" t="+Convert.l2x(ftime));
    return nr;
  }


  private void pcigen() {
    pcibuf = new byte[0x40000];
    pciexp = new int[256];
    pcidat = 0x01CE0000;
    int seq=0;
 //   pciNOOP(0,128);
    for (int i=0; i<4; i++) pciSKIP();

    // training block
    if (MA.getState("/T")) {
      for (int i=0; i<2; i++) pciTSEQ(1);  // 1st TSEQ is lost to osena delay
      for (int i=0; i<1; i++) pciTSEQ(2);
      pciNOOP(0,8);
    }
    for (int i=0; i<4; i++) {
      int j = (i>=2)? 0x80000000 : 0;  // FC2 or FC1
      pciDLP(j|0x400C007C); // P
      pciDLP(j|0x500C0004); // NP
      pciDLP(j|0x60000000); // Cpl
    }
    pciDLP(0x0);	// Ack message
    pciDLP(0x0);	// Ack message
    if (MA.getState("/T")) {
      pciDLP(0x0);	// Ack message
      pciNOOP(0,2);
      pciSKIP();
    }
    pcieConfig = true;
    pciTLPw(seq++,REG_PBAR,1,0xFFFFFFFF);
    pciTLPr(seq++,REG_PBAR,1);
    pciTLPw(seq++,REG_PBAR,1,pcieBAR);
    pcieConfig = false;

    pciTLPr(seq++,REG_PBAR,1);
    pciTLPr(seq++,REG_MCSR,1);
    pciTLPr(seq++,REG_IOCR,1);
    pciTLPr(seq++,REG_DCSR,1);
    pciTLPr(seq-1,REG_DCSR,1); // test skip of resent packet
    pciTLPr(seq-16,REG_DCSR,1); // test skip of resent packet
    pciTLPr(seq++,REG_OMB1,1);
    pciNOOP(0,5*16*lanes/4);

    // test NAK for seq and CRC errs
    if (MA.getState("/N")) {
      pciTLPr(0x123,REG_OMB1,1);
      pciTLPr(seq++,REG_DVID,1);
    }

    // configuration block
    if (MA.getState("/C")) {
      pciDLP(0x00000123);
      pciNOOP(0,8);
      pciTLPw(seq++,REG_OMB1,1,0x44444444,0x6);
      pciNOOP(0,8);
      pciTLPw(seq++,REG_OMB2,1,0x55555555);
      pciNOOP(0,8);
      pciDLP(0x00000123);
      pciDLP(0x00000456);
      pciTLPr(seq++,REG_DVID,1);
      pciTLPr(seq++,REG_OMB1,1);
      pciDLP(0x00000ABC);
      pciTLPw(seq++,REG_IOCR,1,0x30000000);
      pciTLPr(seq++,REG_IOCR,1);
      pciTLPr(seq++,REG_IOCR,1);
      pciTLPr(seq++,REG_IOCR,1);
      pciTLPr(seq++,REG_IOCR,1);
      pciNOOP(0,8);
      pciDLP(0x80120345);
      pciDLP(0x90120345);
      pciDLP(0xA0120345);
    }

    // DMA Test blocks
    int mdar = MA.getL("/MDAR",0xDC00FF00);
    int mdau = MA.getL("/MDAU",0x00000032);
    int mdtc = MA.getL("/MDTC",0x003C0);
    int mcsr = MA.getL("/MCSR",0x000000B3);
    if (MA.getState("/DAC")) mcsr |= 0x0800;

    // DMA Test write block
    if (MA.getState("/W")) {
      pciTLPw(seq++,REG_MDAR,1,mdar);
      pciTLPw(seq++,REG_MDAU,1,mdau);
      pciTLPw(seq++,REG_MDTC,1,mdtc);
      pciTLPw(seq++,REG_MCSR,1,mcsr);
      pciNOOP(0,64*mdtc/lanes);
    }

    // DMA Test read block
    if (MA.getState("/R")) {
      mcsr &= ~0x0020;
      pciTLPw(seq++,REG_MDAR,1,mdar);
      pciTLPw(seq++,REG_MDAU,1,mdau);
      pciTLPw(seq++,REG_MDTC,1,mdtc);
      pciTLPw(seq++,REG_MCSR,1,mcsr);
      pciNOOP(0,lanes*8);
      pciTLPr(seq++,REG_MCSR,1);
      pciNOOP(0,1);
      pciNOOP(0,lanes*64);
      for (int i=0; mdtc>0; i++) {
        int iper = Math.min(mdtc,(i%3==0)?64:(i%10==0)?32:128);
        if (i==1) pciDLP(0x00000222);
        if (i==3) pciNOOP(0,2);
        if (i==3) pciDLP(0x00000111);
        pciTLPc(seq++,iper>>2);
        mdtc -= iper;
      }
      pciTLPr(seq++,REG_MCSR,1);
      pciNOOP(0,lanes*10);
      pciTLPr(seq++,REG_MCSR,1);
    }
    pciNOOP(0,lanes*16);
//    pciNOOP(0,1500);
  }

  private void pciDLP (int data) {

    pciKsym(0x5C);
    pciData(data,4);
    pciDataFI(icrc,2);
    pciKsym(0xFD);

    pciExpect((icrc<<16)|0xFD5C);
    pciExpect(data);
  }

  private void pciTLPw (int seq, int addr, int ndata, int data) {
    pciTLPw(seq,addr,ndata,data,0x0F);
  }

  private void pciTLPw (int seq, int addr, int ndata, int data, int mask) {
    int head[] = {0x40000000+ndata,0x00601300|mask,pcieBAR+addr};
    if (pcieConfig) head[0] |= 0x04000000;
    if (pcieConfig) head[2] = 0x05000000+addr;

    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    for (int i=0; i<ndata; i++) pciDataLE(data,4);
    pciDataFI(lcrc,4);
    pciKsym(0xFD);
    if (pciePad && ndata%2==0) pciNOOP(0,1);

    pciExpect((seq<<16)|0xFDFB);
    for (int i=0; i<3; i++) pciExpect(head[i]);
    for (int i=0; i<ndata; i++) pciExpect(data);
    pciExpect(lcrc);
    if (ndata%2==0) pciExpect(0);
  }

  private void pciTLPr (int seq, int addr, int ndata) {
    int head[] = {0x00000000+ndata,0x0060100F,0xD8400000+addr};
    if (pcieConfig) head[0] |= 0x04000000;
    if (pcieConfig) head[2] = 0x05000000+addr;

    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    pciDataFI(lcrc,4);
    pciKsym(0xFD);
    if (pciePad) pciNOOP(0,1);

    pciExpect((seq<<16)|0xFDFB);
    for (int i=0; i<3; i++) pciExpect(head[i]);
    pciExpect(lcrc);
    if (ndata%2==0) pciExpect(0);
  }

  private void pciTLPc (int seq, int ndata) {
    int head[] = {0x4A000000+ndata,0x00600000|(ndata<<2),0x01230000};
    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    for (int i=0; i<ndata; i++) pciDataLE(pcidat++,4);
    pciDataFI(lcrc,4);
    pciKsym(0xFD);
    if (pciePad && ndata%2==0) pciNOOP(0,1);
  }

  private void pciExpect (int data) {
    pciexp[pcires++] = data;
  }

  private void pciExpectedResults() {
    MT.writeln("PCIe expected results: ");
    for (int i=0; i<pcires; i+=2)
    MT.writeln("WLoad "+Convert.l2x(pciexp[i+1])+" "+Convert.l2x(pciexp[i]));
  }

  private void pciSKIP () {
    for (int i=0; i<lanes; i++) pciKsym(0x1BC);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
  }

  private void pciTSEQ (int mode) {
    for (int i=0; i<lanes; i++) pciKsym(0x1BC);
    for (int i=0; i<lanes; i++) pciKsym(0x1F7);
    for (int i=0; i<lanes; i++) pciData(0x06,1);
    for (int i=0; i<lanes; i++) pciData(0xFF,1);
    for (int i=0; i<lanes; i++) pciData(0x02,1);
    for (int i=0; i<lanes; i++) pciData(0x00,1);
    for (int j=0; j<10; j++)
      for (int i=0; i<lanes; i++) 
        pciData( (mode==2)? 0x45:0x4A, 1);
  }

  private void pciNOOP (int data, int n) {
    byte[] buf = new byte[n*2*4];
    for (int i=0; i<buf.length; ) { buf[i++]=(byte)data; buf[i++]=0; }
    pciBuffer(buf,buf.length);
  }

  private void pciKsym (int data) {
    byte[] buf = new byte[2];
    buf[0] = (byte)(data&0xFF);
    buf[1] = 1;
    pciBuffer(buf,2);
    clrCRC();
  }

  private void pciDataLE (int data, int n) {
    byte[] buf = new byte[8];
    for (int i=0; i<n; i++) {
      buf[i*2] = (byte)((data>>(8*i))&0xFF);
      buf[i*2+1] = 0;
    }
    pciBuffer(buf,n*2);
  }

  private void pciDataFI (int data, int n) {
    pciData(getFI(data),n);
  }

  private void pciData (int data, int n) {
    byte[] buf = new byte[8];
    for (int i=0; i<n; i++) {
      buf[i*2] = (byte)((data>>(8*(n-i-1)))&0xFF);
      buf[i*2+1] = 0;
    }
    pciBuffer(buf,n*2);
  }

  private void pciBuffer (byte[] buf, int len) {
    for (int i=0; i<len; ) {
      addCRC(buf[i]);
      pcibuf[pcilen++] = buf[i++];
      pcibuf[pcilen++] = buf[i++];
      if (++pcioff == lanes) {
        for (int j=0; j<(8-lanes)*2; j++) pcibuf[pcilen++]=0;
        pcioff=0;
      }
    }
  }

  static int LPOLY = (0x04C11DB7^0x1);
  static int IPOLY = (0x0000100B^0x1);

  public void clrCRC () {
    icrc = 0xFFFF;
    lcrc = 0xFFFFFFFF;
  }
  public void addCRC (byte data) {
    for (int i=0; i<8; i++) {
      int bitval = (data>>i)&0x1;
      addICRCbit(bitval);
      addLCRCbit(bitval);
    }
  }
  public void addICRCbit (int value) {
    int val = ((icrc>>15)&0x1)^value;
    icrc = ((icrc<<1)&0xFFFE) | val;
    if (val!=0) icrc ^= IPOLY;
  }
  public void addLCRCbit (int value) {
    int val = ((lcrc>>31)&0x1)^value;
    lcrc = ((lcrc<<1)&0xFFFFFFFE) | val;
    if (val!=0) lcrc ^= LPOLY;
  }
  public int getFI (int data) {
    int fi = 0;
    for (int i=0; i<32; i++) {
      int j = i^0x7;
      if (((data>>j)&0x1)==0) fi |= (0x1<<i);
    }
    return fi;
  }

}
