package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements a Quick Tune/Filter/Decimate function for front end decimation of a high speed signal

  @author Jeff Schoen
  @version $Id: iceqtfd.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class iceqtfd extends Primitive implements Chainable {

  private int dec,gain,flags,tl=4096;
  private double freq,fnyq;
  private DataFile hi,ho,hf;
  private Data dbi,dbo,dbf;
  private Core qtfd;
  private float[] taps;

  public int open() {

    hi = MA.getDataFile ("IN","1000","SB,SI",0);
    hi.open();

    dec  = MA.getL("DEC")*2; 
    freq = MA.getD("FREQ");
    gain = MA.getL("GAIN");

    fnyq = 0.5/hi.getDelta();

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setFormat("CI");
    ho.setDelta(hi.getDelta()*dec);
    ho.open();

    qtfd = Core.forName("QTFD",MA);

    String qfn = MA.getCS("/QFN","qtfir_"+dec);
    taps = getFilterTaps("nxm.ice.dat."+qfn);

    qtfd.set("DECIMATION",dec);
    qtfd.set("FREQUENCY",freq/fnyq);
    qtfd.set("GAIN",gain);
    qtfd.set("PFIR",taps);

    qtfd.open();

    tl = MA.getL("/TL", tl);
    dbi = hi.getDataBuffer(tl); 
    dbo = ho.getDataBuffer(tl);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<0) return (FINISH);
    n = qtfd.process (dbi,n, dbo,tl);
    if (n>0) ho.write(dbo,n);
    return (NORMAL);
  }

  public int close() {
    qtfd.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return qtfd; }
  public Object getPrevLink() { return null; }

  private float[] getFilterTaps (String fname) {
    DataFile hf = new DataFile(this,fname);
    hf.open();
    int taps = (int)hf.size;
    double scale = 1.0;
    if (hf.getFormat().endsWith("I")) scale = 1.0/(64*1024);
    if (hf.getFormat().endsWith("L")) scale = 1.0/(32768*4);
    Data dbf = hf.getDataBuffer(taps,DataTypes.FLOAT);
    hf.read(dbf);
    hf.close();
    float[] coef = new float[taps];
    dbf.toArray(coef);
    float sum=0;
    for (int i=0; i<taps; i++) { coef[i]*=scale; sum += coef[i]; }
    float gain = sum;
    M.info("Read filter file="+fname+" taps="+taps+" gain="+gain);
    return coef;
  }

}

