package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libm.MMath;

import nxm.ice.lib.MDevIce;
import nxm.ice.lib.DevIce.*;
import nxm.ice.core.Demod;

/**
  Utilities for ICE development.

  FUNC,P1,P2,P3

  @author Jeff Schoen
  @version %I%, %G%
*/
public class iceutil extends Primitive {


  public int open() {

    String func = MA.getU("FUNC");

         if (func.equals("TTF2M")) ttf2m();
    else if (func.equals("BIT2M")) bit2m();
    else if (func.equals("RBF2M")) rbf2m();
    else if (func.equals("ELF2M")) elf2m();
    else if (func.equals("LUT")) lut();
    else if (func.equals("RUN")) runit();
    else if (func.equals("M2XBR")) m2xbr();
    else if (func.equals("M2VBR")) m2vbr();
    else if (func.equals("MBTST")) mbtst();
    else if (func.equals("JTAG2M")) jtag2m();
    else if (func.equals("LST2CSP")) lst2csp();
    else if (func.equals("SYM2PART")) sym2part();
    else if (func.equals("NET2UCF")) net2ucf();
    else if (func.equals("QNET2UCF")) qnet2ucf();
    else if (func.equals("GERBER")) gerber();
    else if (func.equals("CRCGEN")) crcgen();
    else if (func.equals("MODARGS")) modargs();
    else M.error ("Unsupported function: "+func);

    return (FINISH);
  }

  private void runit() {
    String cmd = MA.getRawArgs().trim().substring(4);
    Shell.runCommand(M, cmd);
  }

  private void ttf2m() {
    TextFile tf = MA.getTextFile("P1");
    tf.setExt("ttf");
    tf.open(tf.INPUT);
    DataFile df = MA.getDataFile("P2","1000","SB",0);
    df.setComment("ICEPIC IOC Load Program");
    df.open(df.OUTPUT);
    byte[] buffer = new byte[128];
    String line;
    while ( (line=tf.read()) != null) {
      int n = (line.length()+1)/4;
      if (n>128) M.error("Bad TTF line length");
      for (int i=0; i<n; i++) {
        int k = Convert.s2l(line.substring(i*4,i*4+3));
        if (k>127) buffer[i]=(byte)(k-256); else buffer[i]=(byte)k;
      }
      df.write (buffer,0,n);
    }
    tf.close();
    df.close();
  }

  private boolean chk (byte[] a, byte[] b, int n) {
    for (int i=0; i<n; i++) if (a[i]!=b[i]) return false;
    return true;
  }

  private void bit2m() {
    BaseFile bf = MA.getBaseFile("P1");
    bf.setExt("bit");
    bf.open(bf.INPUT);
    int bytes = (int)bf.getSize();
    int flash = MA.getL("/FLASH"); 
    byte[] buf = new byte[bytes+16];
    byte[] syncpattern = { -1,-1,-1,-1,-86,-103,85,102};
    // find the sync pattern
    for (int t=0; ; t++) {
      for (int i=0; i<7; i++) buf[i]=buf[i+1];
      buf[7] = bf.readByte(); bytes--;
      if (chk(buf,syncpattern,8)) break;
      if (t>256) M.error("No sync found");
    }
    bytes -= 4; // to match old code
    bf.read (buf,8,bytes);
    bytes += 8;
    bf.close();
    int quads = bytes/4;
    Convert.swap4(buf,0,quads); 	// swap byte ordering for JTAG writes
    for (;(quads%4)!=0;quads++,bytes+=4) System.arraycopy( buf,bytes-4, buf,bytes, 4); // extend to a 4 quad boundary
    int crc=0; for (int i=0; i<bytes; i+=4) crc = crc32(crc,Convert.unpackL(buf,i),32);
               for (int i=bytes; i<flash; i+=4) crc = crc32(crc,-1,32);
    DataFile df = MA.getDataFile("P2","1000","SL",0,quads,null);
    df.setComment("ICE XILINX FPGA Load Program");
    df.keywords.put("CRC",Convert.l2x(crc));
    df.open(df.OUTPUT);
    df.write (buf,0,bytes);
    df.close();
  }

  private int crc32 (int crc, int data, int bits) {
    for (int i=0; i<bits; i++) {
      int val = ((data>>i)^(crc>>31))&0x1;
      crc = ((crc<<1)&0xFFFFFFFE) | val;
      if (val!=0) crc ^= 0xEDB88320;
    }
    return crc;
  }

  private int crc32r (int crc, int data, int bits) {
    for (int i=bits; i>=0; i--) {
      int val = ((data>>i)^(crc>>31))&0x1;
      crc = ((crc<<1)&0xFFFFFFFE) | val;
      if (val!=0) crc ^= 0xEDB88320;
    }
    return crc;
  }

  private void rbf2m() {
    BaseFile bf = MA.getBaseFile("P1");
    bf.setExt("rbf");
    bf.open(bf.INPUT);
    // align to 4 byte boundary
    int ibytes = (int)bf.getSize();
    int obytes = (ibytes+3)&0xFFFFFFFC;
    int off = obytes-ibytes;
    byte[] buf = new byte[obytes];
    bf.read (buf,off,ibytes);
    bf.close();
    DataFile df = MA.getDataFile("P2","1000","SB",0,obytes,null);
    df.setComment("RUMEL ALTERA FPGA Load Program");
    df.open(df.OUTPUT);
    df.write (buf,0,obytes);
    df.close();
  }

  private void elf2m() {
    int bytes;
    BaseFile bf = MA.getBaseFile("P1");
    bf.setExt("elf");
    bf.open(bf.INPUT);
    bytes=(int)bf.getSize();
    byte[] buf = new byte[bytes];
    bf.read (buf,0,bytes);
    bf.close();
    DataFile df = MA.getDataFile("P2","1000","SB",0,bytes,null);
    df.setComment("ICE XILINX PPC Load Program");
    df.open(df.OUTPUT);
    df.write (buf,0,bytes);
    df.close();
    // create segments for DSRAM and ISRAM
    byte[] dsram = new byte[0x1000]; for (int i=0; i<0x1000; i++) dsram[i]=0;
    byte[] isram = new byte[0x1000]; for (int i=0; i<0x1000; i++) isram[i]=0;
    // analyse header
    Convert.swap4(buf,20,5);
    Convert.swap2(buf,40,6);
    String ident = Convert.unpackS(buf,0,16);
    int psecs = Convert.unpackI(buf,44);
    int hoffs = Convert.unpackL(buf,28);
    int hsize = Convert.unpackI(buf,42);
    M.info("Ident="+ident+" PSECs="+psecs+" Off="+hoffs+" Size="+hsize);
    for (int i=0; i<psecs; i++) {
      int soff = hoffs+i*hsize;
      Convert.swap4(buf,soff,8);
      int vaddr = Convert.unpackL(buf,soff+8);
      int filsz = Convert.unpackL(buf,soff+16);
      int memsz = Convert.unpackL(buf,soff+20);
      int poffs = Convert.unpackL(buf,soff+4);
      M.info("Sec="+i+" Vaddr="+Convert.l2x(vaddr)+" Offst="+Convert.l2x(poffs)+
                      " FilSz="+Convert.l2x(filsz)+" MemSz="+Convert.l2x(memsz));
      if ((vaddr&0xFFFF0000)==0xFFFF0000) System.arraycopy(buf,poffs, isram,vaddr&0xFFF, filsz);
      if ((vaddr&0xFFFF0000)==0xFD000000) System.arraycopy(buf,poffs, dsram,vaddr&0xFFF, filsz);
    }
    writeRamInit("nxm.ice.xfer.ppccode.h",isram,0,0x1000,3,2,4);
    writeRamInit("nxm.ice.xfer.ppcdata.h",dsram,0,0x1000,1,2,2);
    writeMemInit("nxm.ice.xfer.ppccode.mem",isram,0,0x1000,0xfffff000);
    writeMemInit("nxm.ice.xfer.ppcdata.mem",dsram,0,0x1000,0xfd000000);
  }

  /** Midas bit test program */
  private void mbtst() {
    DataFile df = MA.getDataFile("P1","1000","SB,SI,SL",0);
    df.open();
    int  nd = (int)df.getSize();
    Data db = df.getDataBuffer(nd);
    df.read(db);
    df.close();
    int bits = df.getBPA()*8;
    int[] tally = new int[32];
    int data=0;
    for (int i=0; i<nd; i++) {
      if (bits==8) data = db.getB(i);
      if (bits==16) data = db.getI(i);
      if (bits==32) data = db.getL(i);
      for (int b=0,m=1;b<16; b++,m<<=1) {
	if ((data&m)!=0) tally[b]++;
      }
    }
    MT.writeln("Analysis of "+nd+" samples from URL="+df.getURL());
    for (int i=0; i<bits; i++) {
      float avg = (float)tally[i]/nd;
      MT.writeln("Bit Number="+i+" Average="+avg);
    }
  }


  private void lut() {
    String smode = MA.getS("P1");
    Demod.JVM dm = new Demod.JVM();
    String modList = dm.getS("MODULATIONS");
    int mode = Parser.find(modList,smode,0);
    dm.set("MODULATION",mode);
    dm.open();
    int fs = dm.getLUTframe();
    DataFile df = MA.getDataFile("P2","2000","SI",0);
    df.setFS(fs);
    df.setSize(fs*fs*4);
    df.setComment("FPGA LUT for Demod="+smode);
    df.keywords.putMain("FLAGS",""+dm.getFlags());
    df.open(df.OUTPUT);
    Data dd = df.getDataBuffer(1);
    for (int tn=0; tn<=3; tn++) {
      for (int y=0; y<fs; y++) {
        for (int x=0; x<fs; x++) {
          dd.setI(x,dm.getLUTentry(tn,y*fs+x));
        }
        df.write(dd);
      }
    }
    df.close();
    dm.close();
  }

  /** Midas 2 Xilinx Block Ram */
  private void m2xbr() {
    DataFile df = MA.getDataFile("P1","1000","S#,C#",0);
    df.open();
    int bytes = (int)df.getDataSize();
    byte[] buf = new byte[bytes];
    df.read(buf,0,bytes);
    df.close();
    boolean auto = MA.getState("/ENUM");
    String pre = MA.getCS("/PREFIX");
    pre = (pre==null)? "" : pre+".";

    TextFile tf = MA.getTextFile("P2");
    tf.open(tf.OUTPUT);
    int banks = MA.getL("/BANKS",1);
    int bpe  = MA.getL("/BPA",df.bpa);
    int npb  = bpe*2/banks;
    int bper = banks*256/8;
    for (int i=0,l=0; i<bytes; i+=bper,l++) {
      String id = Convert.l2x(l).substring(8);
      String hdata = Convert.bb2hex(buf,i,bper);
      for (int j=0; j<banks; j++) {
        String data = ""; int u=0;
        for (int k=j*npb; k<bper*2; k+=(banks*npb)) {
	  int o = 2 + bper*2 - k - npb;
 	  if (u>0 && u%8==0) data="_"+data;
	  data = hdata.substring(o,o+npb) + data;
	  u+=npb; 
	}
        if (auto) tf.writeln("defparam "+pre+"dp["+j+"].inst.ram.INIT_"+id+" = 256'h"+data+";");
	//if (auto) tf.writeln("defparam "+pre+"ram["+(j*npb*4)+"].inst.INIT_"+id+" = 256'h"+data+";");
	else      tf.writeln("defparam "+pre+"ram"+j+".INIT_"+id+" = 256'h"+data+";");
      }
    }
    tf.close();
  }

  /** Midas 2 Verilog Block Ram */
  private void m2vbr() {
    DataFile df = MA.getDataFile("P1","1000","S#,C#",0);
    df.open();
    int bytes = (int)df.getDataSize();
    byte[] buf = new byte[bytes];
    df.read(buf,0,bytes);
    df.close();

    String ram = MA.getCS("/RAM","ram");

    TextFile tf = MA.getTextFile("P2");
    tf.open(tf.OUTPUT);
    int bpa = df.getBPA();
    int bitpa = bpa*8;
    for (int i=0,l=0; i<bytes; i+=bpa,l++) {
      String data = "  assign "+ram+"["+l+"] = "+bitpa+"'h";
      data += Convert.bb2hex(buf,i,bpa).substring(2);
      tf.writeln(data+";");
    }
    tf.close();
  }

  private void writeRamInit (String name, byte[] buf, int off, int size, 
					int bank, int banks, int banksz) {
    TextFile tf = new TextFile(name);
    tf.open(tf.OUTPUT);
    int bper = 256/8;
    Convert.swap4(buf,off,size/4);
    for (int i=0,l=0; i<size; i+=2*bper,l++) {
      String id = Convert.l2x(l).substring(8);
      for (int j=bank; j<bank+banks; j++) {
        String data = "";
        for (int k=j-bank; k<2*bper/banksz; k+=2) {
	  data = Convert.bb2hex(buf,off+i+k*banksz,banksz).substring(2) + data;
          data = "_"+data;
        }
        data = data.substring(1);
        tf.writeln("defparam ppe.ram"+j+".INIT_"+id+" = 256'h"+data+";");
      }
    }
    Convert.swap4(buf,off,size/4);
    tf.close();
  }

  private void writeMemInit (String name, byte[] buf, int off, int size, int addr) {
    TextFile tf = new TextFile(name);
    tf.open(tf.OUTPUT);
    int bper = 256/32;
    Convert.swap4(buf,off,size/4);
    tf.writeln("@"+Convert.l2x(addr).substring(2));
    for (int i=0; i<size; i+=bper*4) {
      String data = "";
      for (int j=i; j<i+bper*4; j+=4) {
	data += Convert.bb2hex(buf,off+j,4).substring(2);
        data += " ";
      }
      tf.writeln(data);
    }
    Convert.swap4(buf,off,size/4);
    tf.close();
  }

  private static String[] rowlabel = {"A","B","C","D","E","F","G","H","J","K",
    "L","M","N","P","R","T","U","V","W","Y","AA","AB","AC","AD","AE","AF","AG",
    "AH","AJ","AK","AL","AM","AN","AP","AR","AT","AU","AV","AW","AY","BA","BB"};

  private void lst2csp() {
    String line;
    int nrow = MA.getL("/ROWS");
    int ncol = MA.getL("/COLS",nrow);
    int npin = MA.getL("/PINS",ncol*nrow);
    boolean reorder = MA.getState("/REORDER");
    if (npin<0 || nrow<0) M.error("Must specify /ROWS= and /PINS=");
    TextFile tf = MA.getTextFile("P1"); tf.setExt("lst");
    if (!tf.open(tf.INPUT)) M.error("Couldnt open "+tf.getURL());
    TextFile xf = MA.getTextFile("P1"); xf.fn.setRoot("symbolpre"); xf.setExt("csa");
    if (!xf.open(xf.INPUT)) M.error("Couldnt open "+xf.getURL());
    TextFile sf = MA.getTextFile("P1"); sf.setExt("csa");
    if (!sf.open(sf.OUTPUT)) M.error("Couldnt open "+sf.getURL());
    TextFile pf = MA.getTextFile("P1"); pf.setExt("lib");
    if (!pf.open(pf.OUTPUT)) M.error("Couldnt open "+pf.getURL());
    String name = tf.fn.getRoot().toUpperCase();
    String[] spin  = new String[npin+1];
    String[] stext = new String[npin+1];
    int ip=0;
    while ( (line=tf.read()) != null) {
      int ie = line.indexOf('=');
      int i = reorder? Convert.s2l(line.substring(0,ie))-1 : ip;
      spin[i] = line.substring(0,ie);
      stext[i] = line.substring(ie+1);
      ip++;
    }
    if (ip!=npin) {
      M.warning("Number lines="+ip+" != mnumber pins="+npin);
      npin = ip;
    }
    int grid = 254000;
    int hgrid = grid/2;
    int hpin = npin/2;
    int ix1  = 10, ix2 = ix1+40, ixs = ix1+2;
    int iy1  = 10, iy2 = iy1+hpin+4;
    pf.writeln(name);
    while ( (line=xf.read()) != null && !line.equals("XXXX")) sf.writeln(line);
    sf.writeln("  (SYMDEF L0 \""+name+"\" \"\" (PT "+(ixs*grid)+" "+((iy2-1)*grid)+")");
    sf.writeln("   (FIGURE FIG0 LC25 NO_SHEET (OPENSHAPE");
    sf.writeln("    (PT "+(ix1*grid)+" "+(iy2*grid)+")");
    sf.writeln("    (PT "+(ix2*grid)+" "+(iy2*grid)+")");
    sf.writeln("    (PT "+(ix2*grid)+" "+(iy1*grid)+")");
    sf.writeln("    (PT "+(ix1*grid)+" "+(iy1*grid)+")");
    sf.writeln("    (PT "+(ix1*grid)+" "+(iy2*grid)+")");
    sf.writeln("   ))");
    sf.writeln("   (TEXTLOC SYMBOL_NAME TC2 (PT "+(ixs*grid)+" "+((iy2-2)*grid)+"))");
    sf.writeln("   (TEXTLOC PART_NAME TC2 (PT "+(ixs*grid)+" "+((iy2-3)*grid)+"))");
    for (int i=0; i<npin; i++) {
      int j = i+1;
      int ix = (i<hpin)? ix1 : ix2;
      int ixt = (i<hpin)? ix+1 : ix-12;
      int iy = (i<hpin)? iy1+(hpin-i) : iy1+1+(i-hpin);
      sf.writeln("   (TERMINAL "+j+" TC0 (PT "+(ix*grid)+" "+(iy*grid)+"))");
      sf.writeln("   (TEXT TX"+j+" \""+stext[i]+"\" TC2 NO_SHEET (PT "+(ixt*grid)+" "+(iy*grid-hgrid)+"))");
      int pn = (ncol==1)? i : getPinNumber(spin[i],nrow,ncol);
      addLine(pf,""+pn+".0");
    }
    sf.writeln("  )");
    while ( (line=xf.read()) != null) sf.writeln(line);
    flushLine(pf); 

    tf.close();
    xf.close();
    sf.close();
    pf.close();
  }

  private void sym2part() {

    Parser p; String line;
    boolean v4 = MA.getState("/V4");
    boolean v5 = MA.getState("/V5");
    TextFile tf = MA.getTextFile("P1");
    tf.setExt( v5? "pkg" : v4? "txt" : "pins" );
    if (!tf.open(tf.INPUT)) M.error("Couldnt open "+tf.getURL());
    int nrow = MA.getL("/ROWS");
    if (nrow<=0) M.error("You must specify /ROWS=n for this part up to "+rowlabel.length);
    M.info("Part Rows="+nrow);
    int npin = nrow*nrow;
    String[] rows = new String[npin+1];
    while ( (line=tf.read()) != null) {
      p = new Parser(line);
      String text = p.get( v5? 3 : v4? 6 : 2 );
      String spin = p.get( v5? 1 : v4? 4 : 3 );
      int pin = getPinNumber(spin,nrow,nrow);
      if (rows[pin]!=null) M.error("Got Lpin="+pin+" Pin="+spin+" Label="+text);
      rows[pin] = text;
    }
    tf.close();

    TextFile sf = MA.getTextFile("P1"); sf.setExt("sym");
    if (!sf.open(sf.INPUT)) M.error("Couldnt open "+sf.getURL());
    tf = MA.getTextFile("P1"); tf.setExt("lib");
    if (!tf.open(sf.OUTPUT)) M.error("Couldnt open "+tf.getURL());

    while ( (line=sf.read()) != null) {
      p = new Parser(line);
      if (p.elements()<=0) continue;
      M.info("Process "+line);
      String text = p.get(1);
      if (text.equals("GLOBAL")) { 
        String signal = p.get(2);
        flushLine(tf); 
        addLine(tf,p.get(3));
        for (int i=0; i<=npin; i++) {
          if (signal.equals(rows[i])) addLine(tf,""+i+".0");
        }
        flushLine(tf); 
      }
      else if (text.startsWith("PINS")) {
        for (int i=2; i<=p.elements(); i++) {
          String signal = p.get(i);
          int pin = findPin(signal,rows);
	  if (pin<=0) M.error("Signal "+signal+" not found");
          addLine(tf,""+pin+".0");
        }
      }
      else if (text.startsWith("IOBANK")) {
        String bank = p.get(2);
        for (int i=3; i<=p.elements(); i++) {
          String pin = p.get(i);
	  int in = findPin("IO_L"+pin+"N_"+bank,rows);
	  int ip = findPin("IO_L"+pin+"P_"+bank,rows);
	  int is = findPin("IO_L"+pin+"_"+bank,rows);
          if (is>0) addLine(tf,""+is+".0");
          if (in>0) addLine(tf,""+in+".0");
          if (ip>0) addLine(tf,""+ip+".0");
	  if (is<0 && in<0) M.error("No pin "+pin+" in bank "+bank);
	  if (is>0 && in>0) M.error("Bad pin "+pin+" in bank "+bank);
	  if (in>0 && ip<0) M.error("Mismatch pin "+pin+" in bank "+bank);
        }
      } 
    }
    tf.close();
    sf.close();
  }

  private int getPinNumber (String spin, int nrow, int ncol) {
    int irow=nrow;
    for (irow=nrow-1; irow>=0 && !spin.startsWith(rowlabel[irow]); irow--);
    if (irow<0) return -1;
    int pin = 0;
    for (int i=rowlabel[irow].length(); i<spin.length(); i++) pin = 10*pin + (spin.charAt(i)-'0');
    pin += irow*ncol;
    return pin;
  }

  private String getSPinNumber (int pin, int nrow) {
    int irow = (pin-1)/nrow;
    int icol = pin - (irow*nrow);
    return rowlabel[irow]+icol;
  }

  String outline = "";

  private void addLine (TextFile tf, String text) {
    if (outline.length()>73) {
      tf.writeln(outline+"&");
      outline = "";
    }
    outline += text+" ";
  }
  private void flushLine (TextFile tf) {
    if (outline.length()>0) tf.writeln(outline);
    outline = "";
  }

  private void net2ucf() {
    Parser p; String line;
    TextFile tf = MA.getTextFile("P1"); tf.setExt("frs");
    if (!tf.open(tf.INPUT)) M.error("Couldnt open "+tf.getURL());
    TextFile sf = MA.getTextFile("P1"); sf.setExt("ucfx");
    if (!sf.open(sf.OUTPUT)) M.error("Couldnt open "+sf.getURL());
    String part = MA.getS("/PART");
    if (part.length()<=0) M.error("You must specify /PART=name for this design");
    int nrow = MA.getL("/ROWS");
    if (nrow<=0) M.error("You must specify /ROWS=n for this part up to "+rowlabel.length);
    M.info("Part="+part+"  Rows="+nrow);
    int npin = nrow*nrow;
    String label = null;
    while ( (line=tf.read()) != null) {
      if (line.length()==0) continue;
      if (line.charAt(0) != '.') line = ".TER "+line;
      p = new Parser(line); p.clean();
      String cmd = p.get(1);
      if (cmd.equals(".ADD_TER")) { 
        label = p.get(4);
	label = label.substring(1,label.length()-1);
        int bus = 0;
	int ll = label.length();
        char c = label.charAt(label.length()-2);
        if (ll>1 && label.charAt(ll-1)>='0' && label.charAt(ll-1)<='9') bus=1;
        if (bus>0 && ll>2 && label.charAt(ll-2)>='0' && label.charAt(ll-2)<='9') bus=2;
        if (bus!=0) label = label.substring(0,ll-bus)+"<"+label.substring(ll-bus,ll)+">";
	int i = label.indexOf("<0");
	if (bus==2 && i>0) label = label.substring(0,i+1)+label.substring(i+2);
        cmd = ".TER";
       }
      if (!cmd.equals(".TER")) continue;
      String cpart = p.get(2);
      if (!cpart.equals(part)) continue;
      String cpin = p.get(3);
      String spin = getSPinNumber(Convert.s2l(cpin),nrow);
      M.info("Terminal "+cpart+" "+cpin+" "+label);
      sf.writeln("NET \""+label.toLowerCase()+"\" LOC = \""+spin.toLowerCase()+"\" ;");
    }
    tf.close();
    sf.close();
  }

  private int findPin (String text, String[] rows) {
    for (int i=1; i<rows.length; i++) {
      if (matchPin(text,rows[i])) return i;
    }
    return -1;
  }

  private boolean matchPin (String text, String label) {
    if (label==null) return false;
    return label.indexOf(text)>=0;
  }

  private void qnet2ucf() {
    Parser p; String line;
    TextFile tf = MA.getTextFile("P1"); tf.setExt("qnet");
    if (!tf.open(tf.INPUT)) M.error("Couldnt open "+tf.getURL());
    TextFile sf = MA.getTextFile("P1"); sf.setExt("ucfx");
    if (!sf.open(sf.OUTPUT)) M.error("Couldnt open "+sf.getURL());
    String part = MA.getS("/PART");
    if (part.length()<=0) M.error("You must specify /PART=name for this design");
    M.info("Part="+part);
    String label = null;
    while ( (line=tf.read()) != null) {
      if (line.length()==0) continue;
      p = new Parser(line); p.clean();
      String cmd = p.get(1);
      if (cmd.equals("NET_NAME")) { 
        label = null;
        String net = line=tf.read();
        net = net.substring(1,net.length()-1);
        if (net.indexOf(".")>0) continue;
        if (net.startsWith("VCC")) continue;
        if (net.startsWith("GND")) continue;
        if (net.equals("NC")) continue;
	label = convertNet(net);
      }
      if (label==null) continue;
      if (!cmd.equals("NODE_NAME")) continue;
      String cpart = p.get(2);
      if (!cpart.equals(part)) continue;
      String cpin = p.get(3);
      M.info("Terminal "+cpart+" "+cpin+" "+label);
      sf.writeln("NET \""+label.toLowerCase()+"\" LOC = \""+cpin.toLowerCase()+"\" ;");
    }
    tf.close();
    sf.close();
  }

  private String convertNet (String net) {
    int i,len = net.length();
    boolean tnum = net.charAt(len-2)>='0' && net.charAt(len-2)<='9';
    if (net.endsWith("+")||(net.endsWith("P")&&tnum)) 
	net = net.substring(0,len-2)+"P"+net.charAt(len-2);
    if (net.endsWith("-")||(net.endsWith("N")&&tnum)) 
	net = net.substring(0,len-2)+"N"+net.charAt(len-2);
    for (i=len; i>0; i--) {
      char c = net.charAt(i-1);
      if (c<'0' || c>'9') break;
    }
    if (i!=len && len-i<4) net = net.substring(0,i)+"<"+net.substring(i)+">";
    if (net.startsWith("DDR2_")) net = "M"+net.substring(5);
    return net;
  }

  private void jtag2m() {
    int lbuf = 32768, offset=0;
    int ibits=0,obits=0;
    int iwords=0,owords=0;
    int tdiw=0,tdow=0;
    int state = 21;
    boolean oot,tot;

    oot = MA.getState("/OOT");
    tot = MA.getState("/TOT");

    BaseFile bf = MA.getBaseFile("P1");
    bf.open(bf.INPUT);
    byte[] buf = new byte[lbuf];

    DataFile dfc = MA.getDataFile("P2","1000","5L",0);
    dfc.open(dfc.OUTPUT);
    Data dfcb = dfc.getDataBuffer(1);

    DataFile dfi = MA.getDataFile("P3","1000","SL",0);
    dfi.open(dfi.OUTPUT);
    Data dfib = dfi.getDataBuffer(lbuf);

    DataFile dfo = MA.getDataFile("P4","1000","SL",0);
    dfo.open(dfo.OUTPUT);
    Data dfob = dfo.getDataBuffer(lbuf);

    int size=(int)bf.getSize();
    for (;size>0;) {
      int nbuf = Math.min(lbuf,size);
      bf.read (buf,0,nbuf);
      for (int i=0; i<nbuf; i++) {
        int data = buf[i];
        boolean tdi = (data&0x1)!=0;
        boolean tdo = (data&0x2)!=0;
        boolean tms = (data&0x4)!=0;
        tdiw = tdiw>>>1; if (tdi) tdiw |= 0x80000000; 
        tdow = tdow>>>1; if (tdo) tdow |= 0x80000000; 
        if (++ibits==32) { dfib.setL(iwords,tdiw); iwords++; ibits=0; }
        if (++obits==32) { dfob.setL(owords,tdow); owords++; obits=0; }
	switch (state) {
	  case 0: 
	    if (tms) state++;
	    break;
	  case 1: case 2:
	    if (tms) state++;
	    else state+=10;
	    break;
          case 3: 
	    if (!tms) state = 0;
	    break;
	  case 11: case 12:
            tdiw=tdow=ibits=obits=iwords=owords=0; 
	    if (oot) obits = -1;
	    if (tot) ibits = -1;
	    state+=10; 
	    break;
          case 21: case 22:
            if (!tms) break;
	    if (ibits!=0) { int shift = 32-ibits; tdiw >>>= shift; }
	    if (obits!=0) { int shift = 32-obits; tdow >>>= shift; }
	    dfcb.setL(0,state-20);
	    dfcb.setL(1,ibits);
	    dfcb.setL(2,iwords);
	    dfcb.setL(3,tdiw);
	    dfcb.setL(4,tdow);
	    dfc.write(dfcb);
	    if (iwords>0) dfi.write(dfib,iwords);
	    if (owords>0) dfo.write(dfob,owords);
	    state = -3;
	    break;
          case -3: 
	    state = 0;
	    break;
          default:
	    M.error("Illegal State "+state+" at "+offset);
        }
	offset++;
      }
      size -= nbuf;
    }
    bf.close();
    dfc.close();
    dfi.close();
    dfo.close();
  }

  static double e2m = 0.0254;
  static String form = "%-6s%-10s%8.2f%8.3f%8.2f%8.3f%4d%8.2f%8.3f  %-6s%2d%8.2f%8.3f  D%1d";

  public void gerber() {
    TextFile tfi = MA.getTextFile("P1");
    if (!tfi.open(tfi.INPUT)) M.error("Opening input URL="+tfi.getURL());
    M.info("Opened gerber input URL="+tfi.getURL());
    TextFile tfo = MA.getTextFile("P2");
    if (!tfo.open(tfo.OUTPUT)) M.error("Opening output URL="+tfo.getURL());
    tfo.writeln("! GAPFile Version 1.0");
    tfo.writeln("! Created by: NeXtMidas "+Installation.NM_VERSION);
    tfo.writeln("GAPFile    Version 1.0    (c) 1991 CAD Solutions, Inc.");
    tfo.writeln("!     Type            Width          Length     Rotate   Offset      Swap  Tool      Size      Legend");
    tfo.writeln("!                English  Metric English  Metric     English  Metric Code  No. English  Metric");
    tfo.writeln("|-----------------------+---------------+-------------------+-------------------------+--------------");
    while (true) {
      String line = tfi.read();
      if (line==null) break;
      Parser p = new Parser (line,true);
      String pos = p.get(1);
      if (pos==null || pos.length()==0) continue;
      char c = pos.charAt(0);
      if (c<'0' || c>'9') continue;
      String shape  = p.get(2);
      String usage  = p.get(3);
      double size   = Convert.s2d(p.get(4));
      int    orient = Convert.s2l(p.get(5));
      double length = Convert.s2d(p.get(6));
      if (shape.equals("CIRCLE")) shape = "ROUND";
      if (shape.equals("ANNULUS")) { shape = "DONUT"; length=orient; orient=0; }
      if (shape.equals("FINGER")) shape = "OBLONG";
      if (shape.equals("BULLET")) shape = "OBLONG";
      if (shape.equals("DIAMOND")) shape = "SQUARE";
      if (shape.equals("RECTANGLE") || shape.equals("OBLONG")) {
        length += size;
        if (orient==0 || orient==180) { double t=size; size=length; length=t; }
      }
      String dcode = "D"+pos;
      Double dsize = new Double(size);
      Double esize = new Double(size*e2m);
      Double dlength = new Double(length);
      Double elength = new Double(length*e2m);
      Integer iorient = new Integer(0);
      Integer izero = new Integer(0);
      Double dzero = new Double(0.0);
      Object[] args = { dcode,shape,dsize,esize,dlength,elength,
                        iorient,dzero,dzero,dcode,izero,dzero,dzero,izero };
//      M.info("String.format() only supported on 1.5 JDK");
      String s = String.format(form,args);
      System.out.println(s);
      tfo.writeln(s);
    }
    tfi.close();
    tfo.close();
    M.info("Closed GAP output URL="+tfo.getURL());
  }

  int evalcrc;

  private void crcgen() {

    int size = MA.getL("P1",32);
    int poly = MA.getL("P2",0x04C11DB7);
    int seed = MA.getL("P3",0xFFFFFFFF);
    int bits = MA.getL("P4",32);
    int data = MA.getL("P5",0x00000000);
    String label = MA.getU("P6");
    boolean flip = MA.getState("/FLIP");
    boolean comp = MA.getState("/COMP");
    boolean eqn  = MA.getState("/EQN");
    boolean red  = MA.getState("/RED");

    String[] eqs = new String[size];

    int crc = seed;
    int sizem = size-1;
    int mask = (size==32)? -1 : (0x1<<size)-1;
    int maskm = mask & 0xFFFFFFFE;
    int polym = poly & 0xFFFFFFFE;

    if (eqn) for (int b=0; b<size; b++) eqs[b] = "C"+b;

    for (int i=0; i<bits; i++) {
      int val = ((data>>i) ^ (crc>>sizem)) & 0x1;
      crc = ((crc<<1)&maskm) | val;
      if (val!=0) crc ^= polym;
      if (eqn) {
        String a = eqs[sizem]+"^D"+i;
        for (int b=sizem; b>0; b--) {
          eqs[b] = eqs[b-1];
	  if ( ((poly>>b)&0x1) != 0) eqs[b] += "^"+a;
        }
        eqs[0] = a;
      }
    }
    if (flip) {
      int crcf = 0;
      for (int i=0; i<size; i++) if (((crc>>i)&0x1)!=0) crcf |= ( 0x1 << (i^0x7) );
      crc = crcf;
    }
    if (comp) crc ^= mask;

    MT.writeln("CRC size="+size+" mask="+Convert.l2x(maskm)+" polym="+Convert.l2x(polym));
    MT.writeln("CRC in="+Convert.l2x(seed)+" data="+Convert.l2x(data)+" out="+Convert.l2x(crc));

    MR.put(label,crc);

    evalcrc = 0;
    if (red) for (int b=0; b<size; b++) eqs[b] = reduceEqs(bits,data,eqs[b],b);
    if (eqn) for (int b=0; b<size; b++) MT.writeln("crc["+b+"]<="+eqs[b]);
    if (red) MT.writeln("CRC eval="+Convert.l2x(evalcrc));

  }

  private String reduceEqs (int bits, int data, String eqs, int b) {
    int n,c,ones;
    eqs = eqs+"^"; // trailer for pattern matching
    for (c=n=0; c<eqs.length(); c++) if (eqs.charAt(c)=='C') n++;
    String out = "C"+(n%2);
    ones = n;
    for (int i=0; i<bits; i++) {
      String patt = "D"+i+"^"; int plen = patt.length();
      for (c=n=0; c<eqs.length(); c++) if (eqs.regionMatches(c,patt,0,plen)) n++;
      if (n%2==1) out += "^d["+i+"]";
      if (n%2==1 && ((data>>i)&0x1)==1) ones++;
    }
    if (ones%2==1) evalcrc |= (1<<b);
    return out+";";
  }

  private void modargs () {
    Args args1 = Args.parseCommand(M,MA.getS(2));
    Args args2 = (Args) MA.getO(3);
    args2.getKV().clear();
    args2.getKV().merge(args1.getKV());
    args2.nargs = args1.nargs;
    args2.nargsx = args1.nargsx;
  }

}
