package nxm.ice.prim;

import nxm.ice.lib.Core;
import nxm.ice.lib.VRTPacket;
import nxm.sys.inc.*;
import nxm.sys.lib.*;

/**
  Implements VRT functions 

  @author Jeff Schoen
  @version $Id: icenoop.java,v 1.9 2011/05/24 13:21:12 jgs Exp $
*/
public class icevrt extends Primitive implements Chainable {

  private int fmts,flgs,xfer,yfer,mode,pktsz;
  private VRTPacket vpkh;
  private DataFile hi,ho;
  private Data dbi,dbo;
  private Core core;
  private Time time;


  public int open() {

    mode = MA.getSelectionIndex("/MODE","V2M,BYP,M2V",1,-2);

    hi = MA.getDataFile ("IN","1000","S#,C#",0);
    hi.open();

    if (mode<0) {
      hi.setPacket("NONE"); 
      if (!hi.getDataRep().equals("IEEE")) M.error("Core VRT packets must be IEEE format");
      hi.setDataRep("EEEI"); // disable conversion during the read
    }

    ho = MA.getDataFile ("OUT",hi,0);
    ho.open();

    if (mode>0) {
      ho.setPacket("VRT");
    }
    fmts = Core.getFormats(hi.getFormat(),ho.getFormat());
    flgs = MA.getOptionMask("/VRTFLAGS",VRTPacket.flagsList,0);

    vpkh = new VRTPacket(hi.getFormat(),flgs);

    int sid = MA.getL("/SID",0);
    vpkh.setSID(sid);
    int bits = MA.getL("/BITS",(int)(hi.dbpe*8));
    vpkh.setBits(bits);
    int bxfer = MA.getState("/ALT")? 1024 : 1440;
    xfer = MA.getL("/TL",(bxfer*8/bits));
    bxfer = xfer*bits/8;
    vpkh.setBytes(bxfer);
    int flags = MA.getL("/FLAGS",0);
    double dsec = xfer*hi.getXDelta();
    pktsz = 32 + bxfer;
    yfer = xfer*2;

    time = hi.getTime();
    vpkh.setTC(time.getWSec(),time.getFSec());

    core = Core.forName("VRT",MA);
    core.set("FORMATS",fmts);
    core.set("MODE",mode);
    core.set("PKTHDR",vpkh.buf);
    core.set("DSEC",dsec);
    core.set("FLAGS",flags);
    core.open();

    dbi = hi.getDataBuffer(xfer); 
    dbo = ho.getDataBuffer(yfer);     
    
    return (NORMAL);
  }

  public int process() {
    int n = hi.read(dbi);
    if (n<=0) return (FINISH);
    int m = core.process (dbi,n, dbo,yfer);
    if (m>0) ho.write(dbo,m);
    return (NORMAL);
  }

  public int close() {
    if (mode>0) ho.setDataRep("IEEE");
    core.close();
    hi.close();
    ho.close();
    return (NORMAL);
  }

  public Object getNextLink() { return core; }
  public Object getPrevLink() { return null; }

}
