package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.ice.lib.NetIO;
import nxm.ice.lib.ICEPacket;
import nxm.ice.lib.VRTPacket;
import nxm.ice.lib.SDDSPacket;

/**
  Manage playback of a Midas file/pipe to an ICE formatted network packet stream

  @author Jeff Schoen
  @version $Id: sinkice.java,v 1.8 2009/05/26 13:07:53 jgs Exp $
*/
public class sinkice extends Primitive {

  private Data data;
  private DataFile hi,ho;
  private PacketHandler pkh;
  private ICEPacket opkh,ipkh;
  private VRTPacket vpkh;
  private SDDSPacket spkh;
  private NetIO nio = new NetIO(true);
  private boolean rt,wrap,skip,net,inputIsICE;
  private double tdelta,tstart,dout;
  private int sock=0,bxfer,pt,sid;
  private long pdata=0;
  private Time time;
  private String drep;

  public static String ptList = "ICE,SDDS,VRT";
  public static int ICE=1, SDDS=2, VRT=3;

  private int lossy,loss,lcnt;

  public int open() {

    hi = MA.getDataFile ("IN","1000","S#,C#",0);
    hi.open();

    opkh = (ICEPacket) hi.getPacketHandler();
    inputIsICE = (opkh!=null);

    sid  = MA.getL("/SID",0);
    lossy= MA.getL("/LOSSY");
    wrap = MA.getState("/WRAP");
    rt   = MA.getState("/RT");
    pt   = MA.getState("/SDDS")? SDDS : ICE;
    pt   = MA.getChoice("/PT",ptList,ICE);
    drep = MA.getS("/REP",hi.getDataRep());

    if (pt==ICE) {
      pkh = ipkh = new ICEPacket();
      ipkh.setFormat(hi.getFormat());
      xfer = MA.getL("/TL",(int)(4096/hi.dbpe));
      bxfer = (int)(xfer*hi.dbpe);
    }
    else if (pt==SDDS) {
      pkh = spkh = new SDDSPacket();
      spkh.setFormat(hi.getFormat());
      xfer = (int)(1024/hi.dbpe);
      bxfer = (int)(xfer*hi.dbpe);
    }
    else if (pt==VRT) {
      int flgs = drep.equals("IEEE")? VRTPacket.FLG_BIGE : 0;
      flgs = MA.getOptionMask("/VRTFLAGS",VRTPacket.flagsList,flgs);
      pkh = vpkh = new VRTPacket(hi.getFormat(),flgs);
      vpkh.setSID(sid);
      int bits = MA.getL("/BITS",(int)(hi.dbpe*8));
      vpkh.setBits(bits);
      bxfer = MA.getState("/ALT")? 1024 : 1440;
      xfer  = MA.getL("/TL",bxfer*8/bits);
      bxfer = xfer*bits/8;
      vpkh.setBytes(bxfer);
    }

    ho = MA.getDataFile ("OUT",hi,0);
    net = ho.getURL().startsWith("udp:");
    ho.setDataRep(drep);

    if (net) {
      sock = nio.nioOpen(ho.getURL(),0,1);
      if (sock<=0) M.error("Could not open ICE formatted network stream URL="+ho.getURL());
      pdata = nio.nioAlloc(bxfer+64);
    } else {
      if (pt==ICE && inputIsICE) pkh=ipkh=opkh;	// use output packet handler
      else ho.setPacketHandler(pkh);	// add output packet handler
      ho.open();
    }

    data = hi.getDataBuffer(xfer);

    tdelta = hi.getDelta();
    tstart = dout = 0;
    time   = hi.getTime();
    if (time.getWSec()==0) time.fromCurrent();
    if (rt && tdelta>=0.1) M.warning("Realtime with dt>0.1 ?");

    return (NORMAL);
  }

  public int process() {
    if (rt) {
      if (tstart==0 && hi.avail()<xfer) return (NOOP);
      if (tstart==0) tstart = Time.current();
      if (Time.current()<tstart+dout*tdelta) return (NOOP);
    }
    int n = hi.read(data);
    if (n==xfer) {  
      if (pt==ICE && !inputIsICE) ipkh.setTC (0,0,0.0,time.getWSec(),time.getFSec());
      if (pt==VRT) vpkh.setTC (time.getWSec(),time.getFSec());
      if (lossy>0 && simdrop()) {
        // drop this on the floor
      }
      else if (net) {
        int bytes = bxfer;
        if      (pt==ICE)  { bytes += 64;
          Native.ja2p (ipkh.buf,0, pdata,0, 64);
          Native.ja2p (data.buf,0, pdata,64, bxfer);
        }
        else if (pt==SDDS) { bytes += 56;
          Native.ja2p (spkh.buf,8, pdata,0, 56);
          Native.ja2p (data.buf,0, pdata,56, bxfer);
        }
        else if (pt==VRT)  { bytes += 32;
          Native.ja2p (vpkh.buf,0, pdata,0, 32);
          Native.ja2p (data.buf,0, pdata,32, bxfer);
        }
        nio.nioSend (sock,0, pdata,bytes, 0);
      }
      else {
        ho.write(data); 
      }
      if (pt==ICE && !inputIsICE) ipkh.upCount(); 
      if (pt==SDDS) spkh.upCount();
      if (pt==VRT) vpkh.upCount();
      time.addSec(xfer*tdelta);
      dout += n;
    }
    else if (n>0) M.warning("Partial packet of "+n+" elements not output");
    else if (n==0) return (NOOP);
    else if (wrap && hi.isFile()) hi.seek(0.0);
    else return (FINISH);
    return (NORMAL);
  }

  public int close() {
    if (net) nio.nioClose(sock);
    if (pdata!=0) nio.nioFree(pdata);
    hi.close();
    ho.close();
    return (NORMAL);
  }

  private boolean simdrop() {
    if (lossy<=0) return false;
    if (++lcnt>100) {
      if (++loss>lossy) loss=1;
      lcnt = 1;
    }
    return (lcnt<=loss);
  }

}
