package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import nxm.ice.lib.ICEPacket;
import nxm.ice.lib.MDevIce;
import nxm.ice.lib.RamDiskResource;
import nxm.ice.lib.DevIce.DmaMap;
import nxm.ice.lib.DevIce.TimeCode;
import nxm.ice.inc.PicSlave;

/**
  Synchronize a PIC playback from a Midas file/pipe.

  @author Jeff Schoen
  @version %I%, %G%
*/
public final class sinkpic extends Primitive implements PicSlave,Keyable {

  private static int rFILE=-1,rSTOPPED=0,rONESHOT=1,rCONTINUOUS=2,
	rSTOPTOP=3,rSTOPNOW=4,rSPIN=5,rRELOAD=6,rRESTART=7,rABORT=8;
  public static String replayList="-NShot,File,Stopped,OneShot,Continuous,"
	+"StopTop,StopNow,Spin,Reload,Restart,Abort";
  private static int replayListOffset=-3;

  public static String algorithmList = MDevIce.algList;

  private static int sSTOP=1,sWAIT=2,sRUN=3;
  public static String syncList="Master,SlaveWait,SlaveRun";

  private static int tOFF=0,tONESHOTS=1,tCONTINUOUS=2,tTHROTTLED=3,tONDEMAND=4;  // throttle modes
  public static String throttleList="Off,OneShots,Continuous,Throttled,OnDemand";

  private static int mOFF=0,mASYNC=1,mINFO=2,mFULL=3;  // monitor async modes
  public static String monitorList="Off,Async,Info,Full";

  private long ondx,indx,rndx,bdelay,bstart,ramsz;
  private int ocyc,icyc,rcyc,status,frame,bytes,flags,dmamode,throttle,monitor;
  private int dmac,port,bits,gain,rate,trate,xfer,dec,skip,replay,lost,lostc,block;
  private int packet,pktmod,ipkt,ptype,mnbytes,tinc,tcmode,tcpp,archtl;
  private int stats,pmin,pmax,pavg,ptot;
  private boolean sgo,sss,xts,autors,reset,inmem,wrap,doneReading,isSlave,nodma,rtfile;
  private double freq,mtodo,delta,tctolr=10e-9,tcoff,poll,wait,delay,sfactor,timer,timetop;
  private String alias,slave;
  private TimeCode tc = new TimeCode();
  private Time tcs, time = new Time();
  private DataFile hr,ho,hi;
  private MDevIce pic;
  private DmaMap map;
  private ICEPacket pkt;
  private PicSlave picSlave;
  private Table keys = new Table();

  public int open() {

    // open the PIC device
    alias  = MA.getS("DEV");
    pic = new MDevIce (MA,alias);
    if (pic.open()<=0) M.error("Problem opening device: "+alias); 

    // get operational switches
    autors = MA.getState("/AUTORS");
    wrap   = MA.getState("/WRAP");
    rtfile = MA.getState("/RTFILE");
    throttle = MA.getSelectionIndex("/THROTTLE",throttleList,0,-1);

    // get PIC parameters
    ptype = pic.getKeyL(0,pic.KEY_PTYPE);
    port  = pic.getKeyL(0,pic.KEY_PINDEX);
    reset = !(ptype==0 || port==0);
    reset = MA.getState("/RESET",reset);

    // get input file
    hi = MA.getDataFile ("IN");
    hi.open(hi.NATIVE);
    tcs = hi.getTime();

    // get RAM template
    hr = MA.getDataFile ("OUT");
    hr.open(hr.INOUT);
    hr.setOutput(false);
    frame = (hr.typeClass==2)? hr.getSubSize() : 1;

    // get output reduction 
    dec = MA.getL("DEC"); 
    skip = MA.getL("/SKIP",1); 
    // adjust for backward compatibility with old syntax
    if (ptype==pic.IOPT_MODULE && dec>1 && skip==1) { skip=dec; dec=1; }
    delta = hr.getXDelta();
    if (dec==0) dec=1;

    // get the monitor output pipe
    ho = MA.getDataFile ("/MON",hr,0);
    ho.setDelta(hr.getDelta()*skip);
    ho.size = (hr.size-1)/skip + 1;
    ho.open(ho.OPTIONAL);

    // force type 1000 handling
    hi.setDFS(0);
    hr.setFS(0); hr.update();
    ho.setDFS(0);

    // get IO port parameters
    if (hr.bps>0) bits=hr.bps*8; else bits = -hr.bps;
    if (hr.spa==2) bits = -bits;
    bits = MA.getL("/BITS",bits);
    rate = (int)Math.round(dec/delta);
    rate = MA.getL("/SRATE",rate);
    trate = MA.getL("/TRATE",Math.max(10000,rate/8));
    delay = MA.getD("/PREFILL",MA.getD("/DELAY",0.2));
    freq = MA.getD("FREQ",.25*rate);
    gain = MA.getL("GAIN");

    // get transfer parameters
    xfer = frame * MA.getL("/TL", (hr.typeClass==2)? 1:4096 );
    ramsz = (long)hr.size;
    if (ramsz%xfer!=0 || ramsz==xfer) M.warning("Ram size="+ramsz+" should be an integer multiple>1 of /TL="+xfer);
    bytes = (int)(xfer*hr.dbpe);
    ramsz = (long)(hr.size*hr.dbpe);
    if (bytes*skip>ramsz/2) M.error("/TL * DEC must be less than 1/2 ram buffer");

    block = MA.getL("/BLOCK",-1);
    if (M.pipeMode==Midas.PINIT) replay=rCONTINUOUS; else replay=rFILE;
    replay = MA.getSelectionIndex("/REPLAY",replayList,replay,replayListOffset);
    wait   = MA.getD("/WAIT");

    // get status loggers
    stats  = MA.getL("/STATS");

    // get timecode parameters
    tcmode = pic.getKeyL(0,pic.KEY_TCMODE);
    if (!tcs.isZero()) time.fromTime(tcs);
    else time.fromCurrent();
    if (tcmode==pic.TCM_CPU) tcoff=0;	// CPU includes year 
    else tcoff = time.getYiS();		// add current year
    tcoff  = MA.getD("/TCOFF",tcoff);
    tctolr = MA.getD("/TCTOLR",tctolr);
    tcpp   = MA.getL("/TCPP",Math.max(1,(int)(hr.size/xfer/skip)));

    // setup packet output
    packet = MA.getLength("/PACKET");
    pktmod = MA.getL("/PKTMODE",-1);
    pkt = new ICEPacket(hi.getFormat(),xfer,pktmod);

    // master/slave modes
    slave = MA.getS("/SLAVE","NONE");
    isSlave = !slave.equals("NONE");
    sss = slave.equals("SS"); 
    xts = slave.equals("XT");
    sgo = isSlave && !(sss||xts);
    sgo = !(sss||xts||slave.equals("NONE"));
    slave = MA.getS("/MASTER",null);

    // initialize port
    if (reset) status = pic.reset(0); else status=1;
    if (status<0) M.error("Problem resetting device"); 

    // allocate and map memory buffer
    map = pic.mapFile(hr);
    if (map==null) M.error("Problem mapping DMA memory"); 

    flags = 0;
    if (sgo) flags = pic.FLG_SGO;
    if (sss) flags = pic.FLG_RGO;
    if (xts) flags = (pic.FLG_XGO|pic.FLG_TGO);
    if (throttle==tONESHOTS||throttle==tONDEMAND) flags |= pic.FLG_NCCLK;

    dmac = pic.ioPort(ptype,port,-1,1,bits,rate,2*freq/rate,dec,gain,flags);
    if (dmac<=0) M.error("Problem setting up IO port"); 

    status = pic.dmaSetup(dmac,1,map,block,0);
    if (status<0) M.error("Problem setting up DMA channel "+status); 

    // report DMA Channel number
    MR.put(MA.getU("/DMAC"),dmac);
    nodma = pic.getKeyL(dmac,pic.KEY_NODMA)>0;

    // prepare for in memory mode
    inmem = hi.getURL().equals(hr.getURL());
    inmem = MA.getState("/INMEM",inmem);

    // initialize modes
    if (replay==rSTOPPED && MA.getState("/RELEASE")) replay=rSTOPNOW; // to deallocate
    if (isSlave) replay=rSTOPPED; // special slave startup
    if (!MA.find("/POLL")) setPollTime(0.025);
    if (!tcs.isZero()) pic.setKeyD(dmac,pic.KEY_TCOFF,tcs.getWSec(),tcs.getFSec());
    dmamode = pic.DMA_STOP;
    todo = 0; 
    mtodo = map.bytes;
    ocyc = icyc = rcyc = 0; 
    ondx = indx = rndx = 0;
    if (inmem && !rtfile) doneReading = loadRamBuffer();
    archtl  = MA.getL("/ARCHTL", (int)Math.min((long)B1M,map.bytes/16));
    sfactor = 100.0/map.bytes;

    int csize = pic.getKeyL(dmac,pic.KEY_CBUFSZ);
    bdelay = (long)(delay/delta * hi.dbpe);
    bstart = Math.min( (long)(0.95*map.bytes), bdelay);
    bstart = Math.min( map.bytes, Math.max( bstart, 2*csize) );
    if (throttle==tONESHOTS) bstart = map.bytes;
    if (throttle==tONDEMAND) bstart = 0;

    // get monitor throttling mode
    monitor = MA.getSelectionIndex("/MONITOR",monitorList,mFULL,-1);

    // get extra query list entries
    setKeys(MA.getTable("/PICKEYS"));

    return (NORMAL);
  }

  public int process() {

    if (replay==rABORT) return(FINISH);
    int noop = NOOP;

    // transfer data in
    if (!doneReading && replay!=rSTOPPED) {
      boolean isStream = rtfile || hi.isStream();
      long maxavail = rtfile? availRTF() : isStream? hi.io.avail() : (long)((hi.size-hi.seek())*hi.dbpe);
      int avail = (int)Math.min(B1G,maxavail);
      int abytes = (int)Math.min(archtl,map.bytes-indx);
      if (avail>0) abytes = Math.min(abytes,avail);		// trim to whats available
      else if (isStream);					// piped wait for it
      else if (wrap) hi.seek(0.0);				// go back to beginning
      else doneReading = true;					// mark finished
      if (avail>0 && avail<abytes) abytes = avail; 		// end of file | all in pipe
      if ( (icyc>(rcyc-lost) && (indx+abytes)>rndx) 		// wait for RAM pointer
	|| (icyc>ocyc && (indx+abytes)>ondx) ) avail=0;		// wait for output pointer
      if (packet>=0 && avail>pkt.headerLength) {
	hi.read (pkt.buf,0,pkt.headerLength);
	avail = abytes = pkt.getBytes();
      }
      if (avail>=abytes) {
        if (!inmem) {
	  long vaddr = map.getVirtualAddress(indx,abytes);	// map this space
	  abytes = hi.read(vaddr,0,abytes);
        }
	if (abytes>0) { indx += abytes;
	  if (indx>=map.bytes) { icyc++; indx=0; }
 	  if (throttle==tONDEMAND) pic.setKeyL(dmac,pic.KEY_INBYTE_W64,(int)(indx>>6));
	  if (throttle==tONESHOTS && indx==0 && icyc>1) pic.dmaFunc(dmac,pic.DMA_RESHOT);
	  if (wrap && !isStream && abytes==maxavail) hi.seek(0.0); // prevent pause on short wrapped files
        }
	noop = NORMAL;	// we did work here, dont pause if nothing to monitor
      }
      if (throttle==tCONTINUOUS || throttle==tTHROTTLED) {	// check throttle modes
	int index = pic.dmaFunc(dmac,pic.DMA_STATUS);
	double full = doneReading? 1.0 : ((double)(indx-index)/map.bytes);
	if (full<=0) full += 1; 				// handle buffer wrap around
	if (throttle==tCONTINUOUS && full<0.25) { pic.setKeyL(dmac,pic.KEY_RATE,trate); throttle=tTHROTTLED; }
	if (throttle==tTHROTTLED && full>0.75) { pic.setKeyL(dmac,pic.KEY_RATE,rate); throttle=tCONTINUOUS; }
      }
    }

    // start a playback
    if (todo==0) {
      if (replay==rSTOPTOP||replay==rSTOPNOW) {
	pic.dmaFunc(dmac,pic.DMA_RESET);
	replay=rSTOPPED;
      }
      if (replay==rSTOPPED) return(noop);
      ocyc=rcyc=lost=lostc=0; ondx=rndx=0; pkt.setCount(0);
      if (icyc==0 && indx<bstart) return (noop);		// wait for input data delay
      int newmode = nodma? pic.DMA_ENABLE : pic.DMA_CONTINUOUS;
      if (replay==rONESHOT) newmode = pic.DMA_ONESHOT;
      if (throttle==tONESHOTS) newmode = pic.DMA_RESHOT;
      if (throttle==tONDEMAND) newmode = pic.DMA_ONDEMAND;
      if (slave!=null) setSlaveReplay(replay,newmode);		// wait for slave startup
      if (wait>0) Time.sleep(wait);
      pic.dmaFunc(dmac,newmode);
      if (stats>0) M.info("Sinkpic START");
      dmamode=newmode;					// dont change dmamode until dmaFunc completes
      todo=mtodo; 
      timer = timetop = Time.current();
    }

    if (replay==rSPIN) return(noop);
    if (replay==rSTOPNOW) todo=0; 

    // transfer data out
    if (todo>0) {
      long index = ((long)pic.dmaFunc(dmac,pic.DMA_STATUS_W64))<<6;
      if (index<rndx) rcyc = pic.dmaFunc(dmac,pic.DMA_CYCLE);		// low overhead cycle/index counter
      rndx = index;
      if (inmem && !rtfile) { indx=rndx; icyc=rcyc+1; }
      if (rcyc<ocyc+lost) return (noop);				// data not ready
      if (doneReading && !inmem && (ocyc>icyc || ondx>=indx)) setReplay(rABORT); // finish playback
      int behind = rcyc - (icyc+lost);					// check for falling behind
      if (index>indx) behind++;
      if (behind>0) { 
        //M.warning("Falling behind "+behind+" buffers"); 
        M.warning("Falling behind "+behind+" buffers"+" "+rcyc+" "+rndx+" "+icyc+" "+indx); 
        lost += behind; 
      }
      int tbytes = bytes;						// total bytes
      if (ondx+tbytes>map.bytes) tbytes=(int)(map.bytes-ondx);		// trim bytes
      if (index<ondx+tbytes && rcyc<=ocyc+lost) return (noop);		// not ready
      int count = pkt.getCount();
      if (tcmode!=pic.TCM_OFF && count%tcpp==0) {			// handle timecode every tcpp packets
        int tcstat = pic.tc (dmac,-1.0,delta,tc,0); tc.wsec+=tcoff;	// get current timecode
        pkt.setTC (tcmode,tcstat,tc.offset,tc.wsec,tc.fsec);		// set in packet header
        if (count==0) ho.setTime(tc.wsec,tc.fsec-tc.offset*delta); 	// record 1st TC to output header
      } 
      if (ho.isOpen) {
	long vaddr;							// map this space
        if (ondx+bytes>map.bytes) {
	  M.warning("Partial buffer out "+ondx+" "+map.bytes);
	  vaddr = map.getVirtualAddress(ondx-bytes,bytes);		// map this space
        } else {
	  vaddr = map.getVirtualAddress(ondx,bytes);			// map this space
        }
	if (monitor==mOFF || (monitor!=mFULL && ho.io.avail()<tbytes)){ // monitor pipe not ready
          if (monitor==mINFO) M.info("Dropping monitor frame");
        }
	else {
	  ho.write(vaddr,0,bytes);					// output data
	}
      }
      ondx+=bytes*skip; todo=mtodo-ondx; pkt.upCount();			// advance pointers
      if (tcmode==pic.TCM_CPU) time.fromCurrent();
      else if (tcmode!=pic.TCM_OFF) {
	double dtc = (ocyc*map.bytes + ondx)/hr.dbpe;
	if (wrap && dtc>hi.size) dtc = dtc % hi.size;
	time.fromJ1950(hi.getTimeAt(dtc));
      }
    }

    // stop a playback 
    if (todo<=0) {
      ocyc++; todo=0; ondx-=map.bytes;
      long gap = (rndx-indx) + (rcyc-icyc-lost+1)*map.bytes;
      int pf = (doneReading)? 0 : (int)Math.round(sfactor*gap);
      if (stats>0 && replay!=rSTOPNOW) {
	timer = Time.current();
        if (pf<pmin) pmin=pf; if (pf>pmax) pmax=pf; ptot+=pf;
	float mbps = (float)(1e-6*map.bytes/Math.max(0.01,timer-timetop));
        if (ocyc%stats==0) { pavg = ptot/stats;
          M.info("Sinkpic STATS={CYC="+ocyc+",LOST="+lost+",FAVG="+pavg+",FMIN="+pmin+",FMAX="+pmax+",MBPS="+mbps+"}");
          pmin=100; pmax=0; ptot=0;
        }
	timetop = timer;
      }
      int lostd = pic.dmaFunc(dmac,pic.DMA_LOST);
      if (lostd>lostc) { M.warning("Lost "+(lostd-lostc)+" card buffers"); lostc=lostd; }
      if (replay==rCONTINUOUS) { todo=mtodo; return(noop); }
      if (replay==rONESHOT) { pic.dmaFunc(dmac,pic.DMA_WAIT); rcyc=1; }
      pic.dmaFunc(dmac,pic.DMA_STOP);
      pic.dmaFunc(dmac,pic.DMA_RESET);
      dmamode=pic.DMA_STOP;
      if (slave!=null) setSlaveReplay(replay,dmamode);                  // wait for slave shutdown
      if (stats>0) M.info("Sinkpic STOP");
      if (replay==rFILE) return(FINISH); 
      if (replay==rRESTART) replay=rCONTINUOUS;
      else replay=rSTOPPED;
    }

    return (NORMAL);
  }

  public int close() {
    if (dmac>0) pic.dmaFunc(dmac,pic.DMA_CANCEL);
    if (dmamode!=pic.DMA_STOP && stats>0) M.info("Sinkpic CANCEL");
    if (hr!=null) hr.close();
    if (ho!=null) ho.close();
    if (hi!=null) hi.close();
    if (map!=null) map.close();
    if (pic!=null) pic.close();
    replay = rSTOPPED;
    dmamode = pic.DMA_STOP;
    return (NORMAL);
  }

  private boolean loadRamBuffer() {
    if (!(hi.io instanceof RamDiskResource)) {
      int tbytes = (int)(hr.size*hr.dbpe);
      long vaddr = map.getVirtualAddress(0L,tbytes);	// map this space
      hi.read(vaddr,0,tbytes);
    }
    icyc=1; indx=0;
    return true;
  }

  private long availRTF() {
    long ibyte = map.bytes*icyc + indx;
    long inbyte = (long)hi.getInByte();
    long avail = inbyte - ibyte;
    if (throttle==tONDEMAND) avail -= bdelay;
    if (avail<archtl) hi.ioh.read(hi.hb,0,512,0L); // update inByte field
    return avail;
  }

  public void setReplay (String value) {
    if (value.equalsIgnoreCase("STOP")) value = "STOPNOW";
    if (value.equalsIgnoreCase("START")) value = "CONTINUOUS";
    int nreplay = Parser.find(replayList,value,replay,replayListOffset);
    if (nreplay>replayListOffset) setReplay(nreplay);
    else M.warning("Illegal replay mode: "+value);
  }
  private void setReplay (int nreplay) {
    if (nreplay==0 && replay!=0) replay = rSTOPNOW;
    else if (nreplay==rRESTART && replay==0) replay = rCONTINUOUS;
    else if (nreplay==rONESHOT && replay==rCONTINUOUS) replay = rSTOPTOP;
    else replay=nreplay;
    if (replay==rSTOPNOW) process(); // process now
    if (replay>=rABORT && slave!=null) setSlaveReplay(replay,pic.DMA_STOP);
    if (replay==rCONTINUOUS && nodma) replay = rSPIN;
  }
  private void setSlaveReplay (int value, int mode) {
    if (picSlave==null) picSlave = (PicSlave)M.registry.get(slave);
    if (picSlave==null) { M.warning("Slave="+slave+" does not exist"); return; }
    if (mode==pic.DMA_STOP && value<rABORT) value = rSTOPNOW;
    if (nodma && value==rSPIN) value = rCONTINUOUS;
    picSlave.setReplay(getReplay(value));
    while (picSlave.getDmaMode()!=mode) Time.sleep(pollTime);
  }
  public void setMonitor (String value) {
    monitor = Parser.find(monitorList,value,monitor,-1);
  }
  public void setRate (int value) { 
    rate=value; pic.setKeyL(dmac,pic.KEY_RATE,rate); 
  }
  public void setFreq (double value) { 
    freq=value; pic.setKeyD(dmac,pic.KEY_FREQ,2*freq/rate); 
  }
  public void setDec (int value) { 
    dec=value; 
  }
  public void setGain (int value) { 
    gain=value; pic.setKeyL(dmac,pic.KEY_GAIN,gain); 
  }
  public void setSkip (int value) { 
    skip=value; 
  }
  public void setStats (int value) {
    stats = value;
  }


  private String getReplay (int replay) {
    return Parser.get(replayList,replay-replayListOffset);
  }
  public String getReplay() { return getReplay(replay); }
  public int getRate() { return rate; }
  public double getFreq() { return freq; }
  public int getDec() { return dec; }
  public int getGain() { return gain; }
  public int getSkip() { return skip; }
  public int getCycle() { return rcyc; }
  public long getIndex() { return rndx; }
  public double getProgress() { return ((double)rndx)/map.bytes; }
  public String getMonitor() { return Parser.get(monitorList,monitor,-1); }
  public TimeCode getTimeCode() { return tc; }
  public Time getTime() { return time; }
  public int getHBLost() { return lost; }
  public int getCBLost() { return lostc; }
  public int getPFull() { return pavg; }
  public int getDmaMode() { return dmamode; }
  public double getOffset() { return hi.getOffset(); }
  public MDevIce getDevIce() { return pic; }


  // use the keyable interface to expose all of the PIC library keys

  private void setKeys (Table kt) {
    if (kt==null || kt.size()==0) return;
    for (Table.Iterator ti=kt.iterator(); ti.getNext(); ) {
      String sval = (ti.value==null)? "QUERY" : ti.value.toString();
      if (sval.length()==0 || sval.equals("QUERY")) keys.put(ti.key,ti.value); // to show in Query
      else setKey(ti.key,ti.value); // actually set
    }
  }

  public String[] getKeys() {
    return keys.getKeys();
  }

  public Object setKey (String key, Object value) {
    if (!thisIsMe()) MQ.put("SET."+key,0,value); // force all sets to be performed from this thread
    else if (key.equals("REPLAY")) setReplay(Convert.o2s(value));
    else if (key.equals("RATE"))   setRate(Convert.o2l(value));
    else if (key.equals("FREQ"))   setFreq(Convert.o2d(value));
    else if (key.equals("DEC"))    setDec(Convert.o2l(value));
    else if (key.equals("GAIN"))   setGain(Convert.o2l(value));
    else {  // now expose all PIC library keys
      int ikey = pic.name2key(key);
      if (ikey<=0) return null;
      int ityp = pic.getKeyType(ikey);
      if (ityp==DOUBLE) pic.setKeyD(dmac,ikey,Convert.o2d(value));
      else              pic.setKeyL(dmac,ikey,Convert.o2l(value));
      keys.put(key,value); // to show in Query
    }
    return value;
  }

  public Object getKey (String key) {
    Object value = null;
         if (key.equals("REPLAY")) value=            getReplay();
    else if (key.equals("RATE"))   value=Convert.l2o(getRate());
    else if (key.equals("FREQ"))   value=Convert.d2o(getFreq());
    else if (key.equals("DEC"))    value=Convert.l2o(getDec());
    else if (key.equals("GAIN"))   value=Convert.l2o(getGain());
    else {  // now expose all PIC library keys
      int ikey = pic.name2key(key);
      if (ikey<=0) return null;
      int ityp = pic.getKeyType(ikey);
      if (ityp==DOUBLE) value=Convert.d2o(pic.getKeyD(dmac,ikey));
      else              value=Convert.l2o(pic.getKeyL(dmac,ikey));
    }
    return value;
  }

}
