package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.ice.lib.NetIO;
import nxm.ice.lib.ICEPacket;
import nxm.ice.lib.VRTPacket;
import nxm.ice.lib.SDDSPacket;

/**
  Copies data from an ICE formatted network packet stream to an ordinary Midas file.

  @author Jeff Schoen
  @version $Id: sourceice.java,v 1.4 2009/05/26 13:07:53 jgs Exp $

*/
public class sourceice extends Primitive implements Keyable {

  private static int rFILE=-1,rSTOPPED=0,rONESHOT=1,rCONTINUOUS=2,
        rSTOPTOP=3,rSTOPNOW=4,rSPIN=5,rARCHIVE=6,rRESTART=7,rABORT=8;
  public static String replayList="-NShot,File,Stopped,OneShot,Continuous,"
        +"StopTop,StopNow,Spin,Archive,Restart,Abort";
  private static int replayListOffset=-3;

  public static String ptList = "ICE,SDDS,VRT";
  public static int ICE=1, SDDS=2, VRT=3;

  public static String pkthdrList = "Strip,Copy,Attach,Detach";
  public static int STRIP=1,COPY=2,ATTACH=3,DETACH=4;

  private NetIO nio;
  private Data data;
  private DataFile hi,ho;
  private ICEPacket  ipkh = new ICEPacket();
  private VRTPacket  vpkh = new VRTPacket();
  private SDDSPacket spkh = new SDDSPacket();
  private boolean swap,net,usenat,check,attach,detach,skipframe;
  private int replay,pt,bxoff,bxall,bxfer,errors,pkthdr,skip,iskip,frame,nper;
  private double done,todo,drate;
  private String format;
  private byte[] bdata;
  private Table keys;
  private Time time = new Time();

  public int open() {

    hi = MA.getDataFile ("IN","1000","S#,C#",0);
    net = hi.getURL().startsWith("udp:");

    pt     = MA.getState("/SDDS")? SDDS : ICE;
    pt     = MA.getChoice("/PT",ptList,pt);
    check  = MA.getState("/CHECK");
    usenat = MA.getState("/NATIVE",false);

    todo   = MA.getD("/MAXOUT");
    skip   = MA.getL("/SKIP",1);
    frame  = MA.getL("/FRAME",0);
    pkthdr = MA.getSelectionIndex("/PKTHDR",pkthdrList,STRIP);
    format = MA.getFormat("FORMAT","S#,C#");
    drate  = MA.getD("RATE");

    nio = new NetIO(usenat);

    if (net) {
      if (nio.open(hi.getURL(),0,-1)<=0) M.error("Problem opening ICE formatted network stream URL="+hi.getURL());
      ho = MA.getDataFile ("OUT","1000",format,0);
      ho.setXDelta(1.0/drate);
      ho.setXUnits(Units.TIME);
      ho.setSize(Math.max(todo,4096/ho.bpa));
      xfer = (4096/ho.bpa);
    } else {
      hi.open();
      if (hi.getPacketHandler()==null) M.error("Input is not an ICE formatted file");
      ho = MA.getDataFile ("OUT",hi,0);
      PacketHandler pkh = ho.getPacketHandler();
      String packet = ho.getPacket(); 
      int idet = packet.indexOf("/DET");
      if (pkthdr==STRIP) ho.setPacketHandler(null);
      else if (pkthdr==ATTACH && idet>=0) ho.setPacket(packet.substring(0,idet));
      else if (pkthdr==DETACH && idet<0) ho.setPacket("/DET");
      if (pt==ICE) xfer = ((ICEPacket)pkh).getSize();
      if (pt==VRT) xfer = ((VRTPacket)pkh).getSize();
      if (pt==SDDS) xfer = ((SDDSPacket)pkh).getSize();
    }
    if (frame>0) {
      ho.setFS(frame);
      ho.setYUnits(Units.TIME);
      ho.setYDelta(skip*frame/drate);
    }
    ho.open();
    ho.setDFS(1);

    if (pt==SDDS) {
      xfer = (1024/ho.bpa);
      bxoff = 8;
    } else if (pt==VRT) {
      xfer = (1440/ho.bpa);
      bxoff = 0;
    } else {
      xfer = MA.getL("/TL",xfer);
      bxoff = 0;
    }
    bxall = 64 + (xfer*ho.bpa);
    bxfer = bxall-bxoff;
    data  = ho.getDataBuffer(bxall/ho.bpa);
    bdata = data.buf;
    swap  = MA.getState("/SWAP",(pt==SDDS) && (ho.bpa==2));
    nper  = (frame>0)? frame/xfer : 1;
    skip *= nper;
    iskip = -1;
    skipframe = (skip>1) || (frame>0);

    if (M.pipeMode==Midas.PINIT) replay=rCONTINUOUS; else replay=rFILE;
    replay = MA.getSelectionIndex("/REPLAY",replayList,replay,replayListOffset);
    time.fromCurrent();

    return (NORMAL);
  }

  public int process() {
    if (replay==rABORT) return(FINISH);
    if (replay==rSTOPPED) return(NOOP);
    int n=0;
    if (net) {
      int i = (pkthdr==STRIP)? 64:bxoff;
      n = nio.recv(0,bdata,bxoff,bxfer,0);
      if (n<=0) return ((n<0)? FINISH : NOOP);
      if (++iskip>=skip) iskip=0;
      if (skipframe && iskip>=nper) return (NORMAL);
      if (swap) Convert.swap2(bdata,64,512);
      ho.write(bdata,i,n+bxoff-i); 
    } else {
      n = hi.read(data,xfer);
      if (n<=0) return ((n<0)? FINISH : NOOP);
      if (++iskip>=skip) iskip=0;
      if (skipframe && iskip>=nper) return (NORMAL);
      ho.write(data,n); 
    }
    if (check) checkPacket();
    done += xfer;
    if (todo>0 && done>todo) return (FINISH);
    return (NORMAL);
  }

  public int close() {
    nio.close();
    hi.close();
    ho.close();
    if (check) M.info("Transfer addr="+hi.getURL()+" size="+done+" errors="+errors);
    return (NORMAL);
  }

  public void setReplay (String value) {
    if (value.equalsIgnoreCase("STOP")) value = "STOPNOW";
    if (value.equalsIgnoreCase("START")) value = "CONTINUOUS";
    int nreplay = Parser.find(replayList,value,replay,replayListOffset);
    if (nreplay>replayListOffset) setReplay(nreplay);
    else M.warning("Illegal replay mode: "+value);
  }
  private void setReplay (int nreplay) {
         if (nreplay==0 && replay!=0) replay = rSTOPNOW;
    else if (nreplay==rRESTART && replay==0) replay = rCONTINUOUS;
    else if (nreplay==rONESHOT && replay==rCONTINUOUS) replay = rSTOPTOP;
    else replay=nreplay;
    if (replay==rSTOPNOW) process(); // process now
  }
  private String getReplay (int replay) {
    return Parser.get(replayList,replay-replayListOffset);
  }

  public int getCycle() { return 0; }
  public double getOffset() { return ho.getOffset(); }

  int total=0, nextseq=-1;
  private void checkPacket() {
    int seq;
    if (pt==SDDS) {
      spkh.setBuffer(bdata,0);
      seq = spkh.getCount();
    }
    else if (pt==VRT) {
      vpkh.setBuffer(bdata,0);
      seq = vpkh.getCount();
    } else {
      ipkh.setBuffer(bdata,0);
      seq = ipkh.getCount();
    }
    if (nextseq>0) {
      if (seq != nextseq) { M.warning("Sequence gap curr="+seq+" next="+nextseq); errors++; }
    }
    if (++total%100==0) M.info("Packet total="+total+" samples="+(total*xfer));
    nextseq = seq+1;
    if (pt==SDDS) {
      if ((nextseq&0x1F)==0x1F) nextseq++;
      nextseq &= 0xFFFF;
    } else if (pt==VRT) {
      nextseq &= 0xF;
    }
  }

  // use the keyable interface to expose all of the PIC library keys
  public String[] getKeys() {
    return keys.getKeys();
  }
  public Object setKey (String key, Object value) {
    if (!thisIsMe()) MQ.put("SET."+key,0,value); // force all sets to be performed from this thread
    else if (key.equals("REPLAY")) setReplay(Convert.o2s(value));
    else if (key.equals("VLAN")||key.equals("IPVLAN")) ;
    else if (key.equals("JOIN")||key.equals("IPCONN")) ;
    else System.out.println("Unsupported Set "+key+" to "+value);
    return value;
  }
  public Object getKey (String key) {
    Object value = null;
         if (key.equals("TIME")) return time;
    else if (key.equals("CYCLE")) return new Data(0);
    else System.out.println("Unsupported Get "+key);
    return value;
  }

}
