package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import nxm.ice.lib.Archiver;
import nxm.ice.lib.ICEPacket;
import nxm.ice.lib.MDevIce;
import nxm.ice.lib.RamDiskResource;
import nxm.ice.lib.DevIce.DmaMap;
import nxm.ice.lib.DevIce.TimeCode;
import nxm.ice.inc.PicSlave;

/**
  Synchronize a PIC acquisition into a Midas file/pipe.

  If /MULTI=n is used, this primitive controls multiple ports at once.

  @author Jeff Schoen
  @version %I%, %G%
*/
public final class sourcepic extends Primitive implements PicSlave,Keyable {

  private static int rFILE=-1,rSTOPPED=0,rONESHOT=1,rCONTINUOUS=2, // replay modes
	rSTOPTOP=3,rSTOPNOW=4,rSPIN=5,rARCHIVE=6,rRESTART=7,rABORT=8,
	rNEWDEVICE=9, rREOPEN=10, rRECONNECT=11, rRELEASE=12, rRESTART_WAIT=99;
  public static String replayList="-NShot,File,Stopped,OneShot,Continuous,"
	+"StopTop,StopNow,Spin,Archive,Restart,Abort,NewDevice,Reopen,Reconnect,Release";
  private static int replayListOffset=-3;

  public static String algorithmList = MDevIce.algList;

  private static int sMASTER=1,sWAIT=2,sRUN=3;	// master/slave modes
  public static String syncList="Master,SlaveWait,SlaveRun";

  private static int aOFF=0,aOPEN=1,aSNAP=2,aRT=3;  // archive modes
  public static String archList="Off,Open,SnapShot,RealTime";

  private static int pOFF=0,pICE=1,pICET=2,pSDDS=3;  // packet modes
  public static String pktmodeList="Off,Ice,IceT,SDDS";

  private static int mOFF=0,mASYNC=1,mINFO=2,mFULL=3;  // monitor async modes
  public static String monitorList="Off,Async,Info,Full";

  private long rndx,ondx,andx,ramsz;
  private int rcyc,ocyc,stats,pmin,pmax,pavg,ptot,snaptc,autors,renum;
  private int status,frame,bytes,dmamode,skip,nchan,nchano,multi,chan=0;
  private int dmac,port,bits,rate,xfer,dec,gain,replay,block,lost,lostc;
  private int packet,apacket,pktmod,apktmod,ptype,mnbytes,tinc,tcmode,tcpp,tccnt;
  private int archmode=aOFF,archtl,archtc,archcn,alg,nargs,ipflag,archsf,monitor;
  private boolean module,tuner,core,host,fdec,flush,autoss,sgo,sss,xts,reset,isSlave,sdds,nodma,rtfile;
  private double freq,afreq,dfreq,mtodo,delta,ratio,wait,sfactor,maxout;
  private double tctolr,tcmaxtolr,tcoff,timer,timetop,timeout,archto,archdur,archoff,timearch,timeRTF;
  private String alias,slave,archfn,pmss,apmss;
  private TimeCode tc = new TimeCode(), tclast = new TimeCode();
  private Time time = new Time();
  private DataFile hr,ho;
  private Archiver ha;
  private DmaMap map,smap;
  private ICEPacket pkt,apkt;
  private MDevIce pic;
  private int[] dmacs,ports;
  private PicSlave picSlave;
  private Table keys = new Table();

  public int open() {

    // open the PIC device
    alias  = MA.getS("DEV");
    pic = new MDevIce (MA,alias);
    if (pic.open()<=0) M.error("Problem opening device: "+alias); 

    // get operational switches
    autors = MA.getL("/AUTORS");
    autoss = MA.getState("/AUTOSS");
    renum  = MA.getL("/RENUM");
    sdds   = MA.getState("/SDDS");
    rtfile = MA.getState("/RTFILE");
    multi  = MA.getL("/MULTI",1);
    dmacs  = new int[multi];
    ports  = new int[multi];

    // get PIC parameters
    ptype = pic.getKeyL(0,pic.KEY_PTYPE);
    port  = pic.getKeyL(0,pic.KEY_PINDEX);
    reset = !(ptype==0 || port==0 || (multi>1 && !autoss));
    reset = MA.getState("/RESET",reset);
    tinc  = pic.getKeyL(port,pic.KEY_TINC);
    tuner = (ptype == pic.IOPT_TUNER) || (ptype == pic.IOPT_TBANK);
    core  = (ptype == pic.IOPT_CORE) || (ptype == pic.IOPT_MCORE);
    module = (ptype == pic.IOPT_MODULE);
    host  = MA.getState("/HOST",module);
    nchan = core? 1 : pic.getKeyL(port,pic.KEY_CHNS);
    nchan = (ptype==pic.IOPT_TBANK||ptype==pic.IOPT_CORE)? MA.getL("/NCHN",nchan) : 1;
    nchano = MA.getL("/NCHNOUT",nchan);
    fdec = MA.getState("/FRAMEDEC");

    // parse packet modes
    pktmod = pOFF;
    apacket = MA.getLength("/APACKET");
    if (apacket>=0) apmss = parsePacketMode(MA.getS("/APACKET")); 
    apktmod = pktmod; 
    if (apktmod==pSDDS) apacket=-1;
    pktmod = pOFF;
    packet = MA.getLength("/PACKET");
    if (packet>=0) pmss = parsePacketMode(MA.getS("/PACKET"));
    if (pktmod==pOFF && apktmod==pSDDS) { pktmod=apktmod; pmss=apmss; }
    if (pktmod==pSDDS) packet=-1;
    ipflag = ICEPacket.FIXED | (pktmod==pICET? ICEPacket.ABSC:0);

    // get input RAM template
    hr = MA.getDataFile ("IN");
    hr.open(hr.INOUT);
    hr.setOutput(false);
    frame = (hr.typeClass==2)? hr.getSubSize() : 1;
    if (pktmod==pSDDS) frame = frame*1088/1024;

    // get reduction 
    dec = MA.getL("DEC"); 	// happens on card
    skip = MA.getL("/SKIP", 1);	// happens in host
    if (module && !host && dec==1 && skip>1) { dec=skip; skip=1; }
    if (module && host && dec>1) { skip*=dec; dec=1; }

    // get resampling ratio
    ratio = MA.getD("/RATIO"); 
    delta = hr.getXDelta();
    if (ratio>0) pic.setKeyD(0,pic.KEY_RATIO,ratio);

    // get transfer parameters
    xfer = frame * MA.getL("/TL", (frame>1)? 1:4096 );
    if (M.pipeMode==Midas.PINIT) replay=rCONTINUOUS; else replay=rFILE;
    replay = MA.getSelectionIndex("/REPLAY",replayList,replay,replayListOffset);
    flush  = MA.getState("/FLUSH");
    wait   = MA.getD("/WAIT");
    timeout= MA.getD("/TIMEOUT",2.0);
    maxout = MA.getD("/MAXOUT");

    // create ICE packet 
    pkt = new ICEPacket(hr.getFormat(),xfer,ipflag);
    pkt.setChannel( (renum>0)? renum:port );

    // get the output pipe
    ho = MA.getDataFile ("OUT",hr,0);
    ho.setYDelta(delta*frame*skip*(module? dec:1));
    int oskip = (module && host)? 1 : skip;
    ho.size = (hr.size-1)/oskip + 1;

    // open output
    if (pktmod==pICE || pktmod==pICET) {
      ho.setPacketHandler(pkt);
      if (packet>0) ho.setPacket(ho.getPacket()+pmss); 
    }
    ho.open(ho.NATIVE);
    if (pktmod==pSDDS) ho.keywords.putMain("PACKET",pmss);

    // force type 1000 handling
    hr.setFS(0); hr.update();
    ho.setDFS(0);
    ramsz = (long)hr.size;
    if (ramsz%xfer!=0 || ramsz==xfer) M.info("Ram size="+ramsz+" should be an integer multiple>1 of /TL="+xfer);
    bytes = (int)(xfer*hr.dbpe);
    block = MA.getL("/BLOCK",bytes);
    if (skip>1 || dec>1) block=-1;

    // get IO port parameters
    if (hr.bps>0) bits=hr.bps*8; else bits = -hr.bps;
    if (hr.spa==2) bits = -bits;
    bits = MA.getL("/BITS",bits);
    rate = (int)Math.round(1/delta);
    if (tuner) rate = (int)Math.round(hr.spa*dec/delta);
    if (core && !fdec) rate = (int)Math.round(dec/delta);
    if (ratio>0) rate = (int)Math.round(hr.spa*dec/(delta*ratio));
    rate = MA.getL("/SRATE",rate);
    freq = MA.getD("FREQ",rate*.25);
    dfreq = MA.getD("/DFREQ",0.0);
    gain = MA.getL("GAIN");

    // get timecode parameters
    time.fromCurrent(); 
    tcmode = pic.getKeyL(0,pic.KEY_TCMODE);
    if (tcmode==pic.TCM_CPU) tcoff=0; // CPU includes year 
    else tcoff = time.getYiS(); // add current year
    tcoff  = MA.getD("/TCOFF",tcoff);
    tctolr = MA.getD("/TCTOLR");
    tcmaxtolr = MA.getD("/TCMAXTOLR");
    double tcps = MA.getD("/TCPS");	// time codes per second
    double sptc = (tcps>0)? 1.0/(tcps*delta) : hr.size/Math.max(multi,nchan);	// samples per tc
    tcpp = Math.max(1,(int)Math.round(sptc/(multi*xfer*skip)));
    tcpp = MA.getL("/TCPP",tcpp);
    tcpp *= skip;

    // master/slave modes
    slave = MA.getS("/SLAVE","NONE");
    isSlave = !slave.equals("NONE");
    sss = slave.equals("SS");
    xts = slave.equals("XT");
    sgo = isSlave && !(sss||xts);
    slave = MA.getS("/MASTER",null);

    // get the optional algorithm
    alg = MA.getSelectionIndex("/ALG",algorithmList,0,0);
    if (alg>0) pic.setKeyL(0,pic.KEY_ALG,alg);

    // initialize port
    if (reset) status = pic.reset(0); else status=1;
    if (status<0) M.error("Problem resetting device"); 

    // allocate and map memory buffer
    map = pic.mapFile(hr);
    if (map==null) M.error("Problem mapping DMA memory"); 
    if (multi>1) map.setPaged(false);

    // open the PIC card IO ports
    openPorts();

    // report DMA Channel number
    MR.put(MA.getU("/DMAC"),dmac);
    nodma = pic.getKeyL(dmac,pic.KEY_NODMA)>0;

    // setup framed decimation in IOC chip
    if (module && frame>1 && !host) {
      pic.setKeyL(dmac,pic.KEY_FRAME,frame);
      pic.setKeyL(dmac,pic.KEY_DEC,dec);
    }

    // setup tuner resampling 
    if (ratio>0) setRatio(ratio);

    // set algorithm arguments
    nargs = MA.getL("/NARG");
    for (int n=0; n<nargs; n++) pic.setKeyL(dmac,pic.KEY_ARGS+n,MA.getL("/ARG"+n));

    // initialize modes
    if (isSlave) replay=rSTOPPED; // special slave startup
    if (!MA.find("/POLL")) setPollTime(0.025);
    dmamode = pic.DMA_STOP;
    mtodo = smap.bytes;
    todo = 0; ocyc = 0; ondx = andx = 0;
    sfactor = 100.0/(smap.bytes>>6); // in 64by host words
    if (bytes*skip>smap.bytes) M.error("Skip cannot be greater than one buffer");
    if (rtfile) updateRTF(true);

    // setup archiving
    archtl = MA.getL("/ARCHTL",128*1024);
    archto = MA.getL("/ARCHTO",-1);
    archsf = MA.getL("/ARCHSF");
    archfn = MA.getS("/ARCH");
    archcn = MA.getL("/ARCHCHN");
    if (nchano>1 && (apacket>=0 || archcn>=0)) archtl = xfer;
    apkt = new ICEPacket(ho.getFormat(),archtl,ipflag);
    if (apktmod==pSDDS) archtl = archtl*1088/1024;;
    startArchiver (archfn,0.0,-1.0);

    // get status logger
    stats  = MA.getL("/STATS");

    // get monitor throttling mode
    monitor = MA.getSelectionIndex("/MONITOR",monitorList,mFULL,-1);

    // get extra query list entries
    setKeys(MA.getTable("/PICKEYS"));

    return (NORMAL);
  }


  public int process() {

    if (rtfile) updateRTF(false);
    if (replay==rABORT) return(FINISH);

    if (replay==rRELEASE) {
      for (int n=0; n<multi; n++) pic.dmaFunc(dmacs[n],pic.DMA_RESET);
      replay=rSTOPPED;
    }

    // start an acquisition
    if (todo==0) {
      if (replay==rSTOPTOP || replay==rSTOPNOW) replay=rSTOPPED;
      if (replay==rSTOPPED) return(NOOP);
      if (replay==rRESTART && autors>=2) { 
	Time.sleep(0.2); pic.dmaFunc(dmac,pic.DMA_KILL);
	replay=rRESTART_WAIT; timer=Time.current();
      }
      if (replay==rRESTART_WAIT) {
	if (pic.dmaFunc(dmac,pic.DMA_ACTIVE)==0) Time.sleep(0.2);	// ready to start
	else if (Time.current()-timer < timeout+1) return (NOOP);	// wait some more
	else {
	  if (autors>=3) pic.dmaFunc(dmac,pic.DMA_BURY);
	  M.warning("All DMA channels on this input did not shut down");
	  Time.sleep(0.5);
	}
	replay=rCONTINUOUS;
      }
      rcyc=ocyc=lost=lostc=0; rndx=ondx=andx=0;				// init counters
      pmin=100; pmax=0; ptot=0; pkt.setCount(0); apkt.setCount(0);	// init more counters
      int newmode = (nodma)? pic.DMA_ENABLE : (replay==rONESHOT)? pic.DMA_ONESHOT : pic.DMA_CONTINUOUS;	
      if (slave!=null) setSlaveReplay(replay,newmode);			// wait for slave startup
      if (wait>0) Time.sleep(wait);					// specified startup wait
      for (int n=multi-1; n>=0; n--) pic.dmaFunc(dmacs[n],newmode);
      if (stats>0) M.info("Sourcepic START");
      dmamode=newmode;	// dont change dmamode until dmaFunc completes
      todo=mtodo; 
      snaptc=1; tc.initMetrics(); tclast.initMetrics();
      timer = Time.current(); timetop = timearch = timer;
    }

    if (ha.isOpen && archto>0 && (Time.current()-timearch)>archto) {
      ha.checkChannels(); 
      ha.flush();
      timearch=Time.current();
    }
    if (archmode==aOPEN) startArchiver(archfn,0.0,archdur);

    if (replay==rSPIN) return(NOOP);
    if (replay==rSTOPNOW || replay==rRESTART) todo=0; 

    // transfer data
    if (todo>0) {
      long index = ((long)pic.dmaFunc(dmac,pic.DMA_STATUS_W64))<<6;
      if (index<rndx) rcyc = pic.dmaFunc(dmac,pic.DMA_CYCLE);		// low overhead cycle/index counter
      rndx = index;
      int behind = rcyc - (ocyc+lost);					// check for falling behind
      if (behind>1 || (behind==1 && rndx>ondx)) { 
	M.warning("Falling behind "+behind+" buffers. Rcyc="+rcyc+" Ocyc="+ocyc+" Lost="+lost); 
	lost=rcyc-ocyc; 
      }
      int tbytes = bytes*nchano;					// total bytes
      int sbytes = tbytes;						// total bytes with skip
      if (ha.isOpen) sbytes *= skip;					// must include skip
      if (ondx+sbytes>smap.bytes) sbytes=(int)(smap.bytes-ondx);	// trim bytes to end of buffer
      if ((rcyc<ocyc+lost) || (rndx<ondx+sbytes && rcyc<=ocyc+lost)) {	// data not ready
	if (Time.current()-timer<timeout) return (NOOP);		// under timeout
	M.warning("Detected DMA stall > timeout. Rcyc="+rcyc+" Ocyc="+ocyc+" Lost="+lost);
	if (autors>0) setReplay(rRESTART);				// auto restart
	timer=Time.current(); return (NOOP);				// reinit stall timer
      }
      for (int n=1; n<multi; n++) { 					// now check others for multi
	index = ((long)pic.dmaFunc(dmacs[n],pic.DMA_STATUS_W64))<<6;
	if (index<ondx+sbytes && rcyc<=ocyc+lost) return (NOOP);	// not ready
      }
      if (tcmode!=pic.TCM_OFF) processTimeCode();			// handle TC 

      if (ho.isOpen) {
	long vaddr;
	if (ondx+tbytes>smap.bytes) {					// output data
	  M.info("Partial buffer out "+ondx+" "+smap.bytes);
	  vaddr = map.getVirtualAddress(ondx-tbytes,tbytes);		// map this space
	} else {
	  vaddr = map.getVirtualAddress(ondx,tbytes);			// map this space
	}
	if (monitor==mOFF || (monitor!=mFULL && ho.io.avail()<tbytes)){	// monitor pipe not ready
	  if (monitor==mINFO) M.info("Dropping monitor frame");
        }
	else if (nchano>1) {						// channelized data
	  if (packet<0) 
	    ho.write(vaddr,0,tbytes);					// output single data
	  else for (int n=0; n<nchano; n++) {
	    pkt.setChannel(n+1);					// set pkt channel
	    ho.write(vaddr,0+n*bytes,bytes);				// output banked data
	  }
	}
	else if (multi==1) 
	  ho.write(vaddr,0,bytes);					// output single data
	else for (int n=0; n<multi; n++) { 
	  pkt.setChannel( (renum>0)? renum+n : ports[n] );		// set pkt channel
	  ho.write(vaddr,0+n*(int)smap.bytes,bytes);			// output multi data
	}
	if (maxout>0 && ho.seek()>=maxout) return(FINISH);		// all output done
      }
      if (ha.isOpen) {							// archive data
	index = ondx+sbytes;						// last valid index
	tbytes = (int)(archtl*ha.dbpe);					// total bytes per atransfer
	if (nchano>1 && (apacket>=0 || archcn>=0)) tbytes *= nchano;
	for (; andx+tbytes<=index; andx+=tbytes) { 			// archive output
 	  long vaddr = map.getVirtualAddress(andx,tbytes);		// map this space
	  if (nchano>1) {						// channelized data
	    if (apacket<0 && archcn<0) 
	      ha.write(vaddr,0,tbytes,-1);				// output single data
	    else for (int n=0; n<nchano; n++) {
	      apkt.setChannel(n+1);					// set pkt channel
	      if (archcn<0 || n==archcn)
	      ha.write(vaddr,0+n*bytes,bytes,n);			// output banked data
	    }
	  }
	  else if (multi==1) ha.write(vaddr,0,tbytes,-1);		// archive single data
	  else for (int n=0; n<multi; n++) {
	    apkt.setChannel( (renum>0)? renum+n : ports[n] );		// set apkt channel
	    ha.write(vaddr,0+n*(int)smap.bytes,tbytes,n);		// archive multi data
	  }
	  if (apacket>=0) { apkt.upCount(); apkt.decTCO(archtl); }	// increment counters
	} 
        if (archmode==aSNAP && ha.seek()*delta>archdur) archmode=aOFF;
        if (archmode==aOFF) stopArchiver();
      }
      ondx+=(bytes*nchano*skip); todo=(mtodo-ondx); 			// advance pointers
      pkt.decTCO(xfer*skip); pkt.upCount(skip);				// update TC and count
      time.addSec( delta*xfer*skip );					// update time
      timer = Time.current();
    }

    // stop an acquisition or recycle to top of buffer
    if (todo<=0) {
      // wait for raw DMA to complete, even though decimated pipe out is done
      if (replay==rONESHOT && pic.dmaFunc(dmac,pic.DMA_POLL)!=0) { todo=-1; return(NOOP); }
      ocyc++; andx=0; todo=0; ondx-=smap.bytes;
      if (stats>0 && replay!=rSTOPNOW && replay!=rRESTART) {
	int pf=0; if (pic.dmaFunc(dmac,pic.DMA_CYCLE)==(ocyc+lost)) 
	  pf = (int)Math.round(sfactor*pic.dmaFunc(dmac,pic.DMA_STATUS_W64));
	if (pf==100) pf=0; if (pf<pmin) pmin=pf; if (pf>pmax) pmax=pf; ptot+=pf;
	float mbps = (float)(1.e-6 * smap.bytes / Math.max(0.01,timer-timetop));
	if (ocyc%stats==0) { pavg = ptot/stats;
	  String s = "CYC="+ocyc+",LOST="+lost+",FAVG="+pavg;
	  if (sdds) { int gap=pic.getKeyL(dmac,pic.KEY_GAP); s += ",GMIN="+(gap&0xFFFF)+",GMAX="+((gap>>16)&0xFFFF); }
          else s += ",FMIN="+pmin+",FMAX="+pmax;
	  M.info("Sourcepic STATS={"+s+",MBPS="+mbps+"}");
          pmin=100; pmax=0; ptot=0;
        }
	timetop=timer;        
      }
      if (flush && ha.isOpen) ha.flush();
      int lostd = pic.dmaFunc(dmac,pic.DMA_LOST);
      if (lostd>lostc) { M.warning("Sourcepic LOST="+(lostd-lostc)+" card buffers"); lostc=lostd; }
      if (replay==rCONTINUOUS) { todo=mtodo; return(NORMAL); }
      // wait .2 sec for graceful exit of other channels, then force a DMA stall
      if (replay==rRESTART && autors>=2) { Time.sleep(0.2); pic.dmaFunc(dmac,pic.DMA_KILL); }
      for (int n=0; n<multi; n++) {
        pic.dmaFunc(dmacs[n],pic.DMA_STOP);
        pic.dmaFunc(dmacs[n],pic.DMA_RESET);
      }
      dmamode=pic.DMA_STOP;
      if (slave!=null) setSlaveReplay(replay,dmamode);			// wait for slave shutdown
      if (stats>0) M.info("Sourcepic STOP");
      if (replay==rFILE) return(FINISH); 
      if ( replay==rONESHOT) {
        if (snaptc<0) hr.update();
        if (flush) flushRamBuffer();
        rcyc = 1;
      }
      if (replay!=rRESTART) replay=rSTOPPED;
      else if (autors>=2) replay=rRESTART_WAIT; 
      else replay=rCONTINUOUS;
      timer=Time.current();
      if (ha.isOpen) ha.checkChannels();
    }

    return (NORMAL);
  }

  public int close() {
    if (dmamode!=pic.DMA_STOP && stats>0) M.info("Sourcepic CANCEL");
    for (int n=0; n<multi; n++) {
      if (dmacs!=null && dmacs[n]>0) pic.dmaFunc(dmacs[n],pic.DMA_CANCEL);
    }
    stopArchiver();
    if (hr!=null) hr.close();
    if (ho!=null) ho.close();
    if (map!=null) map.close();
    if (pic!=null) pic.close();
    replay = rSTOPPED;
    dmamode = pic.DMA_STOP;
    return (NORMAL);
  }

  private void flushRamBuffer() {
    if (hr.io instanceof RamDiskResource) return;
    int tbytes = (int)(hr.size*hr.dbpe);
    hr.setOutput(true);
    hr.seek(0.0);
    long vaddr = map.getVirtualAddress(0L,tbytes);
    hr.write(vaddr,0,tbytes);
    hr.setOutput(false);
  }

  private void updateRTF (boolean init) {
    if (init) timer = Time.current();
    else if (timer-timeRTF<0.2) return;
    timeRTF = timer;
    hr.setInByte(getBytes());
    hr.update();
  }

  private void openPorts() {
    if (ptype==pic.IOPT_TBANK) {
      int pktlen = MA.getL("/PKTLEN",bytes);
      pic.setKeyL(port,pic.KEY_CHNS,nchan);
      pic.setKeyL(port,pic.KEY_PKTLEN,pktlen);
      pic.setKeyD(port,pic.KEY_DFREQ,2*dfreq/rate);
    }
    for (int n=0; n<multi; n++) {
      // loop through ports - treat normal mode as special case of multi=1
      int flags = 0;
      if (n==0) { 
        if (sgo) flags = pic.FLG_SGO;
        if (sss) flags = pic.FLG_RGO;
        if (xts) flags = (pic.FLG_XGO|pic.FLG_TGO);
      }
      else if (n==1 && !autoss) flags = pic.FLG_SGO;
      else flags = pic.FLG_RGO;

      double nfreq = (freq+n*dfreq)*2/rate; // normalized frequency
      dmac = pic.ioPort(ptype,port,-1,-1,bits,rate,nfreq,dec,gain,flags);
      if (dmac<=0) M.error("Problem setting up IO port"); 

      smap = map.getSubMap(n,multi); // get block n of multi
      status = pic.dmaSetup(dmac,-1,smap,block,0);
      if (status<0) M.error("Problem setting up DMA channel "+status); 

      dmacs[n]=dmac;
      ports[n]=port;

      if (autoss) port += tinc;  // same side increments by 2 unless CPC!=default
      else if (n%2==0) port++;	 // B is always A+1
      else port = port-1 + tinc; // next is on same side as last
    }

    dmac = dmacs[0];
    port = ports[0];

    afreq = pic.getKeyD(dmac,pic.KEY_FREQ)*rate/2; // report actual freq used
  }

  private String parsePacketMode (String ss) {
    String pmss;
    if (ss==null || ss.length()==0) pmss = "ICE";
    else if (ss.startsWith("DET")) pmss = "ICE/"+ss;
    else if (ss.equals("SDDS")) pmss = "SDDS/ICE";
    else pmss = ss;
         if (pmss.startsWith("SDDS")) { pktmod = pSDDS; }
    else if (pmss.startsWith("ICET")) { pktmod = pICET; pmss = pmss.substring(4); }
    else if (pmss.startsWith("ICE"))  { pktmod = pICE; pmss = pmss.substring(3); }
    else M.error("Illegal Packet Mode Switch String: "+ss);
    return pmss;
  }

  private void processTimeCode() {
    if (snaptc>0 || archtc>0);						// snap|archive exceptions
    else { 
      tccnt += skip;
      if (tccnt<tcpp) return; 						// only handle TC every tcpp packets
      tccnt -= tcpp;
    }
    int tcstat = pic.tc (dmac,-1.0,delta,tc,0); 			// get current timecode
    if (tcstat>0 && tc.wsec<100) { 					// check for ray day 1st 100 sec of year
      int days = (int)Math.round(-(tc.diff(tclast)+tcoff)/86400);
      if (days==365 || days==366) tcoff += days*86400; 			// bump to next year
    }
    tc.wsec += tcoff;							// add year offset
    double tboff = (ocyc+lost)*mtodo; 					// byte offset at top of ram buffer
    double toff = tc.offset - (tboff+ondx)/hr.bpa/nchano;		// sample offset relative to packet start
    pkt.setTC (tcmode,tcstat,toff,tc.wsec,tc.fsec);			// set in packet header
    if (tcstat == pic.TC_NOCLOCK) {
      M.warning("Time code NOCLOCK err detected");
      if (autors>0) setReplay(rRESTART);
    }
    if (tcstat<0 && tcstat!=-9) return;					// done for now
    if (tctolr>0 && tclast.delta>=0) {					// TC tolerance check
      double tcerr = tc.err(tclast);
      if (tcmaxtolr>0 && Math.abs(tcerr)>tcmaxtolr) {
	M.warning("Time code slip="+tcerr+" > maxtolr="+tcmaxtolr+". Assumed bogus and ignored");
        tclast.update(tc); return;					// update last TC for differential metrics
      }
      if (Math.abs(tcerr)>tctolr) {
        M.warning("Time code slip="+tcerr+" > tolr="+tctolr+" delta="+tc.delta);
        if (autors>0) setReplay(rRESTART);
      }
    }
    tc.computeMetrics(tclast,delta);	 				// calulate TC metrics
    if (snaptc>0) { snaptc=-1;
      hr.setTime(tc.wsec,tc.fsec-tc.offset*delta); hr.update();	 	// record 1st TC to ramdisk header
      ho.setTime(tc.wsec,tc.fsec-tc.offset*delta); 			// record 1st TC to output header
    }
    ho.setTimeAt(tc.wsec+(tc.fsec-toff*delta)); 			// record all TC to output header
    if (ha.isOpen) {
      toff = tc.offset - ((tboff+andx)/hr.bpa)/nchano;			// sample offset relative to packet start
      if (archtc!=0) { archtc=0;  					// to archive header
        ha.setTime(tc.wsec,tc.fsec-(tc.offset-archoff)*delta); }
      if (apacket>=0) apkt.setTC (tcmode,tcstat,toff,tc.wsec,tc.fsec);	// set in arch packet header
      ha.setTimeAt(tc.wsec+(tc.fsec-toff*delta)); 			// record all TC to output header
    }
    pkt.getTC (time,0.0,delta);						// update time
    tclast.update(tc); 							// update last TC for differential metrics
  } 

  public void stopArchiver() {
    archmode = aOFF;
    if (ha!=null && ha.isOpen && thisIsMe()) ha.close();
  }
  public void startArchiver (String fname) {
    archfn = fname;
    archdur = -1;
    archmode = aOPEN;
  }
  private void startArchiver (String fname, double time, double dur) {
    stopArchiver();
    ha = new Archiver(this);
    if (fname==null||fname.length()==0||fname.startsWith("NULL")) return;
    if (ramsz<=archtl) M.error("Ram buffer length="+ramsz+" must be increased to at least 2x the /ARCHTL="+archtl);
    if (ramsz%archtl!=0) M.error("Ram size="+ramsz+" must be an integer multiple>1 of /ARCHTL="+archtl);
    if (MA.getL("/SDDSMULTI")>0) M.error("Switch /SDDSMULTI=n deprecated. Replace with file qualifier {FUNC=SDDSUNPACK,MULTI=n}");
    ha.init (this,fname,hr,0);
    if (multi>1) ha.setMulti(multi);
    if (nchan>1) ha.setNChan(nchan);
    if (archsf>0) { ha.setMulti(1); ha.setMaxLineSize(hr.size*hr.dbpe*archsf); }
    if (apacket>=0) { apkt.setCount(0); ha.setPacketHandler(apkt); }
    if (apacket>0) ha.setPacket( ha.getPacket()+apmss ); 
    ha.setFS(0);
    ha.open(ha.NATIVE|ha.OPTIONAL);
    if (!ha.isOpen) return;
    if (apktmod==pSDDS) ha.keywords.putMain("PACKET",apmss);
    if (ha.io instanceof FileResource && !(ha.io instanceof NFileResource)) 
	M.warning("UhOh. Using slower non-native file access.  Check installation.");
    archmode = (dur>0)? aSNAP:aRT;
    archtc=1; // need time code
    int tbytes = (int)(archtl*ha.dbpe);
    andx = (ondx/tbytes)*tbytes;
    archoff = (ocyc*mtodo+andx)/hr.bpa;
  }

  public void setReplay (String value) { 
    if (value.equalsIgnoreCase("STOP")) value = "STOPNOW";
    if (value.equalsIgnoreCase("START")) value = "CONTINUOUS";
    int nreplay = Parser.find(replayList,value,replay,replayListOffset); 
    if (nreplay>replayListOffset) setReplay(nreplay);
    else M.warning("Illegal replay mode: "+value);
  }
  private void setReplay (int nreplay) {
         if (nreplay==0 && replay!=0) replay = rSTOPNOW; 
    else if (nreplay==rRESTART && replay==0) replay = rCONTINUOUS; 
    else if (nreplay==rONESHOT && replay==rCONTINUOUS) replay = rSTOPTOP; 
    else if (nreplay>=rNEWDEVICE && replay!=0) M.warning ("Port must be Stopped to perform NewDevice|Reopen|Reconnect");
    else replay=nreplay;
    if (replay==rSTOPNOW) process(); // process now
    if (replay==rRECONNECT) { openPorts(); replay=0; }
    if (replay>=rABORT && slave!=null) setSlaveReplay(replay,pic.DMA_STOP);
    if (replay==rCONTINUOUS && nodma) replay = rSPIN;
  }
  private void setSlaveReplay (int value, int mode) {
    if (picSlave==null) picSlave = (PicSlave)M.registry.get(slave);
    if (picSlave==null) { M.warning("Slave="+slave+" does not exist"); return; }
    if (mode==pic.DMA_STOP && value<rABORT) value = rSTOPNOW;
    if (nodma && value==rSPIN) value = rCONTINUOUS;
    picSlave.setReplay(getReplay(value));
    int smode = picSlave.getDmaMode();
    while (smode!=mode) {
      if (smode==pic.DMA_CONTINUOUS && mode==pic.DMA_ENABLE) break;
      if (mode==pic.DMA_CONTINUOUS && smode==pic.DMA_ENABLE) break;
      Time.sleep(pollTime);
      smode = picSlave.getDmaMode();
    }
  }
  public void setRate (int value) { 
    if (value==rate) return;
    if (pic.setKeyL(dmac,pic.KEY_RATE,rate)<0) {	// can't set realtime
      MA.put("/SRATE",""+value);			// impose desired rate
      setState(RESTART);				// perform a restart
    } else {
      rate=value;
    }
  }
  public void setFreq (double value) { 
    freq=value;
    if (ptype==pic.IOPT_TBANK) pic.setKeyD(dmac,pic.KEY_FREQ,2*value/rate);
    else if (chan>0) pic.setKeyD(dmacs[chan-1],pic.KEY_FREQ,2*value/rate); 
    else for (int n=0; n<multi; n++) pic.setKeyD(dmacs[n],pic.KEY_FREQ,2*(value+n*dfreq)/rate); 
    afreq = pic.getKeyD(dmac,pic.KEY_FREQ)*rate/2; // report actual freq used
    // if (afreq!=freq) M.info("Frequency "+freq+" rounded to "+afreq);
  }
  public void setDec (int value) { 
    for (int n=0; n<multi; n++) pic.setKeyL(dmacs[n],pic.KEY_DEC,value); 
    dec = pic.getKeyL(dmac,pic.KEY_DEC);
    if (dec!=value) M.info("Decimation "+value+" rounded to "+dec);
    if (dmamode==pic.DMA_STOP);
    else if (autors>0) setReplay(rRESTART);
    else if (tuner && tcmode!=pic.TCM_OFF) setReplay(rRESTART);
  }
  public void setGain (int value) { 
    gain=value; 
    if (ptype==pic.IOPT_TBANK) pic.setKeyL(dmac,pic.KEY_GAIN,gain);
    else if (chan>0) pic.setKeyL(dmacs[chan-1],pic.KEY_GAIN,gain); 
    else for (int n=0; n<multi; n++) pic.setKeyL(dmacs[n],pic.KEY_GAIN,gain); 
  }
  public void setRatio (double value) { 
    if (ptype==pic.IOPT_TBANK) pic.setKeyD(dmac,pic.KEY_RATIO,value);
    else if (chan>0) pic.setKeyD(dmacs[chan-1],pic.KEY_RATIO,value); 
    else for (int n=0; n<multi; n++) pic.setKeyD(dmacs[n],pic.KEY_RATIO,value); 
    ratio = pic.getKeyD(dmac,pic.KEY_RATIO);	// report actual ratio used
  }
  public void setChannel (int value) { 
    if (ptype==pic.IOPT_TBANK) {
      chan = value;
      afreq = getChannelFreq(chan);
      pic.setKeyL(dmac,pic.KEY_CHAN,chan);
    } else {
      chan = Math.max(1,Math.min(multi,value)); 
      gain = pic.getKeyL(dmacs[chan-1],pic.KEY_GAIN);
      afreq = pic.getKeyD(dmacs[chan-1],pic.KEY_FREQ)*rate/2;
    }
  }
  public void setAlgorithm (String value) { 
    alg = Parser.find(algorithmList,value,alg,0); 
    pic.setKeyL(dmac,pic.KEY_ALG,alg);
  }
  public void setMonitor (String value) { 
    monitor = Parser.find(monitorList,value,monitor,-1); 
  }
  public void setArgument (int n, int value) { 
    pic.setKeyL(dmac,pic.KEY_ARGS+n,value);
  }
  public void setStats (int value) {
    stats = value;
  }

  private String getReplay (int replay) { 
    return Parser.get(replayList,replay-replayListOffset);
  }
  public String getReplay() { return getReplay(replay); }
  public int getDec() { return dec; }
  public int getGain() { return gain; }
  public int getRate() { return rate; }
  public double getRatio() { return ratio; }
  public double getFreq() { return afreq; }
  public int getChannel() { return chan; }
  public int getCycle() { return rcyc; }
  public long getIndex() { return rndx; }
  public double getProgress() { return ((double)rndx)/map.bytes; }
  public double getBytes() { return (((double)rcyc)*map.bytes + rndx); }
  public int getSkip() { return skip; }
  public String getAlgorithm() { return Parser.get(algorithmList,alg); }
  public String getMonitor() { return Parser.get(monitorList,monitor,-1); }
  public int getArgument(int n) { return pic.getKeyL(dmac,pic.KEY_ARGS+n); }
  public Time getTime() { return time; }
  public TimeCode getTimeCode() { return tclast; }
  public String getFormat() { return hr.getFormat(); }
  public int getHBLost() { return lost; }
  public int getPFull() { return pavg; }
  public int getCBLost() { return lostc; }
  public int getDmaMode() { return dmamode; }
  public double getOffset() { return ha.isOpen? ha.getOffset() : ho.getOffset(); }
  public MDevIce getDevIce() { return pic; }
  public Archiver getArchiver() { return ha; }
  public ICEPacket getPkt() { return pkt; }
  public ICEPacket getAPkt() { return apkt; }

  public double getDelta() { 
    double delta = 1.0/rate;
    if (core) delta *= dec;
    if (tuner) delta *= (hr.spa*dec);
    if (ratio>0) delta /= ratio;
    return delta;
  }

  public double getChannelFreq (int chn) { 
    double fcny = 0.0;
    if (ptype==pic.IOPT_TBANK) {
      pic.setKeyL(dmac,pic.KEY_CHAN,chn);
      Data data = new Data(freq*2/rate);
      pic.getKey(dmac,pic.KEY_NFREQ,data);
      pic.setKeyL(dmac,pic.KEY_CHAN,chan);
      fcny = data.getD(0);
    } else {
      int n = Math.max(1,Math.min(multi,chn))-1;
      fcny = pic.getKeyD(dmacs[n],pic.KEY_FREQ); 
    }
    return fcny*rate/2;
  }

  // use the keyable interface to expose all of the PIC library keys

  private void setKeys (Table kt) {
    if (kt==null || kt.size()==0) return;
    for (Table.Iterator ti=kt.iterator(); ti.getNext(); ) {
      String sval = (ti.value==null)? "QUERY" : ti.value.toString();
      if (sval.length()==0 || sval.equals("QUERY")) keys.put(ti.key,ti.value); // to show in Query
      else setKey(ti.key,ti.value); // actually set
    }
  }

  public String[] getKeys() {
    return keys.getKeys();
  }

  public Object setKey (String key, Object value) {
    if (!thisIsMe()) MQ.put("SET."+key,0,value); // force all sets to be performed from this thread
    else if (key.equals("REPLAY")) setReplay(Convert.o2s(value));
    else if (key.equals("RATE"))   setRate(Convert.o2l(value));
    else if (key.equals("FREQ"))   setFreq(Convert.o2d(value));
    else if (key.equals("DEC"))    setDec(Convert.o2l(value));
    else if (key.equals("GAIN"))   setGain(Convert.o2l(value));
    else if (key.equals("RATIO"))  setRatio(Convert.o2d(value));
    else if (key.startsWith("ALG")) setAlgorithm(Convert.o2s(value));
    else if (key.startsWith("CHAN")) setChannel(Convert.o2l(value));
    else {  // now expose all PIC library keys
      int ikey = pic.name2key(key);
      if (ikey<=0 && key.equals("VLAN")) ikey = pic.KEY_IPVLAN;
      if (ikey<=0 && key.equals("JOIN")) ikey = pic.KEY_IPCONN;
      if (ikey<=0 && key.equals("LEAVE")) ikey = pic.KEY_IPDISC;
      if (ikey<=0) return null;
      int ityp = pic.getKeyType(ikey);
      if (ikey==pic.KEY_IPCONN || ikey==pic.KEY_IPDISC) value = new Integer( pic.str2ip(Convert.o2s(value)) );
      if (ityp==DOUBLE) pic.setKeyD(dmac,ikey,Convert.o2d(value));
      else              pic.setKeyL(dmac,ikey,Convert.o2l(value));
      keys.put(key,value); // to show in Query
    }
    return value;
  }

  public Object getKey (String key) {
    Object value = null;
         if (key.equals("REPLAY")) value=            getReplay();
    else if (key.equals("RATE"))   value=Convert.l2o(getRate());
    else if (key.equals("FREQ"))   value=Convert.d2o(getFreq());
    else if (key.equals("DEC"))    value=Convert.l2o(getDec());
    else if (key.equals("GAIN"))   value=Convert.l2o(getGain());
    else if (key.equals("RATIO"))  value=Convert.d2o(getRatio());
    else {  // now expose all PIC library keys
      int ikey = pic.name2key(key);
      if (ikey<=0) return null;
      int ityp = pic.getKeyType(ikey);
      if (ityp==DOUBLE) value=Convert.d2o(pic.getKeyD(dmac,ikey));
      else              value=Convert.l2o(pic.getKeyL(dmac,ikey));
    }
    return value;
  }

}
