package nxm.ice.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.libm.MMath;
import nxm.ice.lib.VCore;
import nxm.ice.lib.DevIce;

/**
  Streams data through an ICE FPGA Verilog Core Simulation.

  @author Jeff Schoen
  @version $Id: vcore.java,v 1.13 2004/11/09 21:36:25 jgs Exp $
*/
public class vcore extends Primitive {

  private DataFile hi,ho;
  private Data dbi,dbo;
  private VCore core;
  private String sig,iform,oform;
  int dec,order,bpei,bpeo,xfes,sysreg;
  int[] coef,pciexp;
  byte[] pcibuf;
  int pcilen,pcioff,pcires;
  double xdr = 1;
  int lanes = -1;
  boolean pm;

  private final static int CORE_ENA=0x01, CORE_PLY=0x02, CORE_I8B=0x100, CORE_O8B=0x1000;

  public int open() {

    hi = MA.getDataFile ("IN");
    hi.open();

    iform = hi.getFormat();
    oform = MA.getS("/FORM",iform);

    ho = MA.getDataFile ("OUT",hi,0);
    ho.setFormat(oform);

    sig = MA.getS("SIG");
    dec = 1;

    core = new VCore();
    core.open();
    sysreg = 0x0;
    core.syswr(0x0,sysreg);

    lanes = MA.getL("/LANES",8);
    pm = MA.getState("/PM");

    if (sig.equals("NOOP")) {
      dec = MA.getL("P2",1);
      sysreg = 0x4; 		// use decimation
      sysreg |= (getFmt(iform)<<8);
      sysreg |= (getFmt(oform)<<12);
      core.syswr(1<<2,(dec-1));	// ratio
    }
    else if (sig.equals("DMOD")) {
      sysreg = 0x08;
      sysreg |= (getFmt(iform)<<8);
      sysreg |= (getFmt(oform)<<12);
      loadLUT( MA.getCS("P1"), 0x4 , sysreg);	// load LUT entries
    }
    else if (sig.equals("FIR")) {
      dec = MA.getL("P2",1);
      boolean by2 = MA.getState("/BY2");
      coef = getFilterTaps(MA.getCS("P1"));
      int nc = coef.length;
      if (by2) { 
        nc = (nc+1)/2; iform = "C"+iform.substring(1);
        for (int i=0; i<nc; i++) coef[i] = coef[i+i];
      }
      int order = MA.getL("/ORDER",4);
      symfir (order,nc,coef,dec, 12<<2,13<<2);
      if (by2) dec*=2;
      ho.setXDelta(hi.getXDelta()*dec);
      sysreg = 0;
      sysreg |= getFmt(iform)<<8;
      sysreg |= getFmt(oform)<<12;
      double gain = MA.getD("/GAIN",0.0);
      int scale = (int)(0x1000 * Math.pow(2.0,gain/6));
      core.syswr(9<<2,scale);			// scaling
      M.info("Gain="+gain+" Scale="+Convert.l2x(scale));
    }
    else if (sig.equals("QFIR")) {
      dec = MA.getL("P2",1);
      coef = getFilterTaps(MA.getCS("P1"));
      int nc = coef.length;
      int order = MA.getL("/ORDER",5);
      int width = MA.getL("/WIDTH",8);
      int ndiv  = MA.getL("/NDIV",256);
      double gain = MA.getD("/GAIN",0.0);
      int scale = (int)(0x2000 * Math.pow(2.0,gain/6));
      int freq  = -(int)(MA.getD("/FREQ",0)*0x80000000L);
      qfir (order,nc,coef,ndiv,scale, 12<<2,13<<2);
      core.syswr(1<<2,dec-1);
      core.syswr(6<<2,freq);	
      ho.setXDelta(hi.getXDelta()*dec);
      ho.setXStart(-hi.getXDelta()*(nc/2-dec));
      sysreg = 0;
      sysreg |= getFmt(iform)<<8;
      sysreg |= getFmt(oform)<<12;
      M.info("Taps="+nc);
      M.info("Freq="+Convert.l2x(core.sysrd(6<<2)));
    }
    else if (sig.equals("CIC")) {
      dec = MA.getL("P1",1);
      order = MA.getL("P2",5);
      ho.setXDelta(hi.getXDelta()*dec);
      int scale = (int)(order*log2((double)dec)+1);
      scale = MA.getL("/SHIFT",scale);
      core.syswr(4,dec-1);
      core.syswr(8,scale);
      M.info("CIC dec="+dec+" shift="+scale);
    }
    else if (sig.equals("RES")) {
      coef = getFilterTaps(MA.getCS("P1"));
    }
    else if (sig.equals("TFDD")) {
      dec = MA.getL("P2",2);
      int ovsr = MA.getL("/OVSR",1);
      int movsr = MA.getL("/MOVSR",1);
      int freq = (int)(MA.getD("/FREQ",0.25)*(0x80000000L/ovsr));
      int odec = dec*ovsr;
      int cdec = Math.max(1,odec/2);
      double ratio = MA.getD("/RATIO");
      boolean resamp = (ratio>0);
      if (!resamp) ratio=1;
      int iratio = (int)(0x10000000/ratio);
      int repeat = MA.getL("/REPEAT",ratioMoN(ratio));
      int gain = MA.getL("/GAIN",0);
      if (ovsr>1) gain += (int)(log2((double)ovsr)*6+.5);
      int scale = MA.getL("/SHIFT",getScale(gain,cdec));
      int order = MA.getL("/ORDER",pm?1:3);
      int mdec = (order==1)? 4 : 1;
      boolean round = MA.getState("/ROUND",true);
      boolean tuner = MA.getState("/TUNER",true);
      ovsr = Math.max(ovsr,movsr);
      xdr = ho.spa/ratio;
      core.syswr(1<<2,cdec-1);	// deci
      core.syswr(2<<2,scale);	// scaling
      core.syswr(3<<2,iratio);	// ratio
      core.syswr(4<<2,repeat);	// repeat
      core.syswr(5<<2,ovsr);	// ovsr
      core.syswr(6<<2,-freq);	// freq
      core.syswr(7<<2,0);	// poff
      M.info("Cdec="+Convert.l2x(cdec)+" Freq="+Convert.l2x(freq)+" Scale="+Convert.l2x(scale));
      boolean od1 = (odec==1);
      boolean htf = MA.getState("/HTF",(odec==mdec));
      String cfn = htf?  "cfir_80f" :  resamp? "dfir_25" : "cfir_80";
      cfn = MA.getCS("/CFN",cfn);
      coef = getFilterTaps("nxm.ice.dat."+cfn);
      symfir (order,coef.length,coef,od1?1:2, 10<<2,11<<2);
      cfn = htf?  "pfir_80f" : "pfir_80";
      cfn = MA.getCS("/PFN",cfn);
      coef = getFilterTaps("nxm.ice.dat."+cfn);
      symfir (order,coef.length,coef,2, 12<<2,13<<2);
      if (resamp) {
        coef = getFilterTaps(pm?"nxm.ice.dat.res_8x256_90":"nxm.ice.dat.res_10x2048_90");
        M.info("Ratio "+Convert.l2x(iratio)+" Repeat="+Convert.l2x(repeat)+" Coef size "+coef.length);
        core.syswr(14<<2,0);
        for (int i=0; i<coef.length/2; i++) core.syswr(15<<2,coef[i]);
      }
      String demod = MA.getS("/DEMOD",null);
      if (demod!=null) {
        sysreg |= 0x80; // DEMOD
        loadLUT("nxm.ice.dat.LUT_"+demod+".prm",8<<2,sysreg);
      }
      if (tuner)     sysreg |= 0x10; // TUNER
      if (movsr>1)   sysreg |= 0x20; // MOVSR
      if (resamp)    sysreg |= 0x40; // RESAMP
      sysreg |= getFmt(iform)<<8;
      sysreg |= getFmt(oform)<<12;
      core.syswr(0x0,sysreg);
    }
    else if (sig.equals("PCIE")) {
      sysreg = 0x30 | (getLaneMask()<<8);
      core.syswr(0x0,sysreg);
      pcigen();
    }
    else if (sig.equals("PPC")) {
      sysreg = 0x0;
      core.syswr(0x0,sysreg);
    }
    else if (sig.equals("JVM")) {
      String fname = MA.getCS("/JF","nxm.ice.lib.IcePM");
      byte[] bc = getByteCodes(fname+".prm");
      core.syswr(0x0,0x0); 
      sysreg = 0x1;
      if (MA.getState("/V")) sysreg |= 0x100;
      core.syswr(0x0,sysreg);
      for (int i=0; i<bc.length; i+=2) {
        // if code translate section done - reset pointer
        if (i==4096) { core.syswr(0x0,0); core.syswr(0x0,sysreg); }
        core.syswr(0x4, ((bc[i+1]<<8)&0xF00) | (bc[i]&0xFF) );
      }
    }
    else if (sig.equals("RAM")) {
      sysreg = 0x0;
    }
    else if (sig.equals("DMA")) {
      sysreg = 0x0;
    }
    else if (sig.equals("IOM")) {
      int flags = 0;
      sysreg = 0;
      core.syswr(0x0,sysreg);
      if (MA.getState("/PLAY")) {
        flags = 2;
        sysreg = 0x13;
        for (int i=0; i<4; i++) {
          core.syswr(20,0x000e0358);
          core.syswr(20,0x00040358);
          core.syswr(20,0x000e0348);
          core.syswr(20,0x0004034c);
        }
      } else {
        flags = 1;
        sysreg = 0x03;
      }
      flags = MA.getL("/FLAGS",flags);
      sysreg = MA.getOptionMask("/MCFG",DevIce.mcfgYList,sysreg);
      core.syswr(0x10,flags);
    }
    else if (sig.equals("TC")) {
      int off = MA.getL("/OFF",0);
      int count = (int)( 1e-3/hi.getXDelta());
      int vft = (int)((2.0+off)*hi.getXDelta()*1e9/5);
      core.syswr(0x4,0x030F);	// barker
      core.syswr(0x4,0x0304);
      core.syswr(0x4,0x1396);	// sec
      core.syswr(0x4,0x1255);
      core.syswr(0x4,0x1394);	// min
      core.syswr(0x4,0x1253);
      core.syswr(0x4,0x1392);	// hrs
      core.syswr(0x4,0x1121);
      core.syswr(0x4,0x1393);	// day
      core.syswr(0x4,0x1392);
      core.syswr(0x4,0x5131);
      core.syswr(0x4,0x2397);	// msec
      core.syswr(0x4,0x2398);
      core.syswr(0x4,0x2399);
      core.syswr(0x4,0x0300 | getBitsRev(vft,7,10));	// vft
      core.syswr(0x4,0x0300 | getBitsRev(vft,3,6));
      core.syswr(0x4,0x8308 | getBitsRev(vft,0,2));
      core.syswr(0x4,off);	// start count
      core.syswr(0x4,count-1);	// repeat count-1
      sysreg = 0x80000000;
    }
    else M.error("Unsupported Core Sig="+sig);

    sysreg |= CORE_ENA; // universal core enable bit
    sysreg = MA.getL("/SYSREG",sysreg);
    if (!sig.equals("JVM")) core.syswr(0x0,sysreg);
    int stat = core.sysrd(0x0);
    M.info("SysReg="+Convert.l2x(sysreg)+"  SysStat="+Convert.l2x(stat));

    ho.setXDelta(hi.getXDelta()*dec*xdr);
    ho.open();
    bpei = (int)hi.dbpe;
    bpeo = (int)ho.dbpe;
    xfer = MA.getL("/TL",1024);
    dbi  = hi.getDataBuffer(xfer);
    xfes = xfer*4;
    dbo  = ho.getDataBuffer(xfes);
    
    return (NORMAL);
  }

  public int process() {
    int nbw,nin,nout; byte[] bufi;
    if (sig.equals("PPC")) {
      ppcgen();
      return (FINISH);
    }
    if (sig.equals("JVM")) {
      core.syswr(0x0,sysreg);
      int ticks = MA.getL("/TICKS",6);
      for (int i=0; i<20; i++) core.sysrd(0x0);
      if (MA.find("/MSG")) sendMsg(MA.getS("/MSG"));
      if (MA.find("/MSG2")) sendMsg(MA.getS("/MSG2"));
      for (int i=0; i<ticks; i++) core.sysrd(0x0);
      return (FINISH);
    }
    if (sig.equals("RAM")) {
      ramgen();
      return (FINISH);
    }
    if (pcibuf!=null) {
      nin = pcilen;
      bufi = pcibuf;
      nbw = 128;
      for (int i=16; i>0; i--) core.sysrd(0x0);
      core.syswr(0x58,0x1234567);
    } else {
      nin = hi.read(dbi);
      bufi = dbi.buf;
      nbw = 64;
    }
    nout = core.process (bufi,nin*bpei, dbo.buf,xfes*bpeo, nbw);
    if (nout>0) ho.write(dbo,nout/bpeo);
    if (verbose) M.info("Process Nin="+nin+" Nout="+nout);
    if (nin<=0 || pcibuf!=null) return FINISH;
    return (NORMAL);
  }

  public int close() {
    hi.close();
    ho.close();
    if (MA.getState("/EXP")) pciExpectedResults();
    core.close();
    return (NORMAL);
  }

  private int getFmt (String format) {
    int fmt=0;
    if (format.charAt(1)=='B') fmt |= 0x1;
    if (format.charAt(1)=='L') fmt |= 0x4;
    if (format.charAt(0)=='C') fmt |= 0x8;
    return fmt;
  }

  private byte[] pciePrep (byte[] bufi, int nin, int lanes) {
    int nout = nin;
    for (int i=lanes; i<3; i++) nout *= 2;
    byte[] buf = new byte[nout*2];
    for (int i=0; i<nout*2; i++) buf[i]=0;
    if (lanes==0) for (int i=0; i<nin/1; i++) System.arraycopy(bufi,i*2, buf,i*16, 2);
    if (lanes==1) for (int i=0; i<nin/2; i++) System.arraycopy(bufi,i*4, buf,i*16, 4);
    if (lanes==2) for (int i=0; i<nin/4; i++) System.arraycopy(bufi,i*8, buf,i*16, 8);
    return buf;
  }

  private int getBitsRev (int bits, int bit1, int bit2) {
    int nbit = bit2-bit1+1;
    int bitr = 0;
    for (int i=0; i<nbit; i++) {
      bitr |= ((bits>>(bit2-i))&0x1) << i;
    }
    return bitr;
  }

  private void loadLUT (String dfn, int addr, int sysreg) {
    DataFile df = new DataFile(M,dfn);
    core.syswr(0x0,sysreg|CORE_ENA);		// enable tap counter
    df.open();
    Data db = df.getDataBuffer((int)df.size);
    int flags = df.keywords.getL("FLAGS",0);
    df.read(db);
    df.close();
    int n = db.getBufLength()/4;
    for (int i=0; i<n; i++) core.syswr(addr,db.getL(i)); // load taps
    core.syswr(addr,flags);	// load flags always right after lut
    core.syswr(0x0,sysreg);			// reset tap counter
  }

  private int[] getFilterTaps (String fname) {
    DataFile hf = new DataFile(this,fname);
    hf.open();
    int taps = (int)hf.size; 
    Data dbf = hf.getDataBuffer(taps,DataTypes.LONG);
    hf.read(dbf);
    hf.close();
    int[] coef = new int[taps];
    dbf.toArray(coef);
    int sum=0;
    for (int i=0; i<taps; i++) sum += coef[i];
    float gain = ((float)sum)/65536;
    M.info("Read filter file="+fname+" taps="+taps+" gain="+gain);
    return coef;
  }

  private byte[] getByteCodes (String fname) {
    DataFile hbc = new DataFile(this,fname);
    hbc.open();
    int bytes = (int)(hbc.size*hbc.getBPE()); 
    byte[] buffer = new byte[bytes];
    hbc.read(buffer,0,bytes);
    hbc.close();
    return buffer;
  }

  private int symfir (int order, int taps, int[] coef, int dec, int addrs, int addrc) {

    boolean odd  = (taps%2 == 1);
    boolean sym  = true;
    boolean anti = true;
    boolean nosym = MA.getState("/NOSYM");

    // parse taps
    for (int i=0; i<taps/2; i++) {
      if (Math.abs(coef[i]-coef[taps-1-i])>1) sym = false;
      if (Math.abs(coef[i]+coef[taps-1-i])>1) anti = false;
    }
    sym |= anti;
    if (nosym) sym = false;
    odd &= sym;
    int staps = taps;
    if (sym) staps/=2; 			// symmetric
    if (odd) staps++;
    int otap = order;			// order taps
    for (int i=1; i<dec; i*=2) otap++;	// disallow degenerate case of short decimating filters
    int ttaps = 1<<otap;		// total taps
    for (; ttaps<staps; ttaps+=ttaps,otap++);
    int tpc = 1<<(otap-order);		// taps per cell
    int cps = (tpc+dec-1)/dec;		// cycles per sample
    int htaps = sym? ttaps : ttaps/2;
    int otaps = htaps / dec;
    int o1 = (order==1)?1:0;

    M.info("dec="+dec+" tpc="+tpc+" cps="+cps+" ttaps="+ttaps+" staps="+staps+" sym="+sym+" odd="+odd+" anti="+anti);

    if (tpc>32) M.error("Max TapsPerCell=32 exceeded");
    int odec = dec;
    int otpc = tpc;
    tpc--;
    cps--;
    dec--;

    // form sequence
    int aadr=tpc, ashft=0;
    int badr=dec, bshft=0;
    int cadr=tpc, cshft=0;
    int radr=1+(1-o1)*tpc, rshft=0; 
    int seq,idec=0;
    if (odd) radr--;
    int[] sequence = new int[32];
    for (int tcnt=0; tcnt<=tpc; tcnt++) {
      int ccnt = tcnt&cps;
      boolean aload = (ccnt==0);
      boolean bload = (ccnt==cps);
      boolean acc = (tcnt!=0);
      boolean aco = (tcnt==tpc);
      if (bload) {
        if (idec==0) rshft=tpc-dec-o1;
        else rshft+=2;
        if (++idec>dec) idec=0;
      }
      seq = ((radr+rshft)<<18) | ((cadr+cshft)<<12) | ((badr+bshft)<<6) | ((aadr+ashft));
      if (aload) seq |= 0x01000000;
      if (bload) seq |= 0x02000000;
      if (acc)   seq |= 0x04000000;
      if (aco)   seq |= 0x08000000;
      M.info ("Seq "+tcnt+" = "+Convert.l2x(seq));
      sequence[tcnt] = seq;
      if (aload) { ashft++; radr++; }
      if (bload) { bshft++; radr--; }
      aadr-=odec; if (aadr<0)   aadr+=tpc;
      badr+=odec; if (badr>tpc) badr-=(tpc+2);
      cadr-=odec; if (cadr<0)   cadr+=tpc;
    }
    // load sequence
    seq = 0x80000000;
    if (anti)  seq |= 0x10000000;
    if (!sym)  seq |= 0x20000000;
    core.syswr(addrs,seq|(tpc<<16)|otaps);
    for (int tcnt=tpc; tcnt>=0; tcnt--) core.syswr(addrs,sequence[tcnt]);

    // load taps
    if (odd) coef[staps-1] /= 2;
    if (sym) {
      for (int i=staps-1; i>=0; i--) core.syswr(addrc,coef[i]); // actual taps
      for (int i=staps; i<ttaps; i++) core.syswr(addrc,0);	// pad taps
    } else if (nosym) {
      int fill = ttaps - staps;
      for (int i=0; i<=fill/2; i++) core.syswr(addrc,0);	// pad taps
      for (int i=taps-1; i>=0; i--) core.syswr(addrc,coef[i]); 	// actual taps
      for (int i=fill/2+1; i<fill; i++) core.syswr(addrc,0);	// pad taps
    } else {
      for (int i=staps; i<ttaps; i++) core.syswr(addrc,0);	// pad taps
      for (int i=taps-1; i>=0; i--) core.syswr(addrc,coef[i]); 	// actual taps
    }

    return 0;
  }

  private int qfir (int order, int taps, int[] coef, int ndiv, int scale, int addrs, int addrc) {

    int ttaps = (int)Math.pow(2,order)*2;
    M.info(" taps="+taps+" ttaps="+ttaps);

    // load taps
    core.syswr(addrc,0x80000000); // bit=31 is tap enable
    for (int i=0; i<ttaps/2; i++) core.syswr(addrc,coef[i]&0xFFFF);
    core.syswr(addrc,0x0);	// flush tap enable

    // load sincos lut
    core.syswr(0<<2,0x00000001);
    int amp = Math.min(scale,0x7FFF);
    for (int i=0; i<ndiv; i++) {
      double phase = (i*TWOPI/ndiv);
      int tapsin = (int)(amp*Math.sin(phase));
      int tapcos = (int)(amp*Math.cos(phase));
      int tapcx  = ((tapsin<<16)&0xFFFF0000) | (tapcos&0xFFFF);
      core.syswr(addrs,tapcx); 
    }
    core.syswr(0<<2,0x00000000);
    
    return 0;
  }

  private int oqfir (int order, int width, int taps, int[] coef, int dec, int freq, int ndiv, int addrs, int addrc) {

    int steps = ndiv;
    int ttaps = (int)Math.pow(2,order);
    int lc = width-1;		// last column
    int lr = ttaps/width-1;	// last row
    M.info("dec="+dec+" taps="+taps+" ttaps="+ttaps);

    freq |= ((ndiv-1)<<8);
    int seqfreq = (freq<<16)|freq;
    core.syswr(addrs,seqfreq);

    // load taps
    for (int i=ttaps-1; i>=0; i--) {		// tap index
      int ic = i%width;				// column index
      int ir = i/width;				// row index
      for (int k=0; k<steps; k++) {		// freq or pinc block
        core.syswr(addrs,(k<<16)|k);
        for (int j=0; j<steps; j++) {		// phase
          int jj = ((j+1)*dec)%width;
          int ii = (i+jj)%ttaps;
          int tap = (ii<taps)? coef[ii] : 0;
          double phase = (-TWOPI/steps)*(j*dec-ii)*k;
          int tapsin = (int)(tap*Math.sin(phase));
          int tapcos = (int)(tap*Math.cos(phase));
          int tapcx  = ((tapsin<<16)&0xFFFF0000) | (tapcos&0xFFFF);
          core.syswr(addrc,tapcx); 
    } } }

    // load sequencer
    int nper = width/dec;
    for (int i=steps-1; i>=0; i--) {
      int seq = (int)Math.pow(2,dec)-1;
      seq <<= (nper-(i%nper)-1)*dec;
      seq |= 0x8000;
      seq = (seq<<16)|seq;
      core.syswr(addrs,seq);
      M.info("Sequence "+i+"  "+Convert.l2x(seq));
    }

    // reset
    core.syswr(addrs,seqfreq);

    return 0;
  }

  private int ratioMoN (double ratio) {
    double tolr = 0.0000001;
    for (int m=1; m<0xFFFF; m++) {
      double d = m/ratio;
      int n = (int)(d+.5);
      if (Math.abs(d-n)<tolr) return (m<<16)|n;
    }
    return 0;
  }

  private final double GCFIR_GAIN = 0.9899;
  private final double GPFIR_GAIN = 1.1494;

  private int getScale (int gain, int decratio) {
    int coarse,scale,gaincomp,cdec;
    double dgain,sgain,fgain;
    cdec = decratio*4;
    coarse = (gain>=0)? (gain/6) : (gain-5)/6;
    scale = (int)(5.0*log2((double)cdec/4)+0.98) - coarse;
    dgain = powi(cdec/4,5) * powi(2,-scale);
    scale = Math.max(0,scale+5); /* from 20b to 18b data */
    sgain = 1.0;
    sgain *= GCFIR_GAIN; /* compensate for CFIR gain */
    sgain *= GPFIR_GAIN; /* compensate for PFIR gain */
    fgain = Math.pow( 2.0, ((double)gain)/6.0 ) / (sgain*dgain);
    if (decratio>1 && fgain>=2.0 && scale>0) { fgain/=2; scale--; }
    gaincomp = (int)(0x8000 * fgain);
    if (gaincomp>=0x10000) gaincomp = 0xffff;
    return (scale<<16)|gaincomp;
  }

  private double log2 (double value) {
    double val;
    val = Math.log(value) * 1.44269504088896;
    return val;
  }
  private double powi (int base, int power) {
    double val;
    if (power==2) val = (double)base*(double)base;
    else val = Math.pow( (double)base, (double)power );
    return val;
  }

  final static int REG_DVID=0x00, REG_CODE=0x08, REG_PBAR=0x10, REG_MDAU=0x44, REG_IOC=0x54,
                   REG_FIFO=0x40, REG_MCSR=0x48, REG_MDAR=0x58, REG_MDTC=0x5C, REG_DCSR=0xE4,
		   REG_OMB1=0x60, REG_OMB2=0x64, REG_OMB3=0x68, REG_OMB4=0x6C, REG_LCSR=0xEC;

  private boolean pcieConfig = false;
  private int pcieBAR = 0xD8400000;

  private void pcigen() {
    pcibuf = new byte[0x40000];
    pciexp = new int[256];
    int seq=20;
    pciNOOP(0,128);
    for (int i=0; i<4; i++) pciSKIP();
    if (MA.getState("/T")) {
      for (int i=0; i<2; i++) pciTSEQ(1);  // 1st TSEQ is lost to osena delay
      for (int i=0; i<1; i++) pciTSEQ(2);
      pciNOOP(0,8);
    }
    for (int i=0; i<4; i++) {
      int j = (i>=2)? 0x80000000 : 0;  // FC2 or FC1
      pciDLP(j|0x400C007C);
      pciDLP(j|0x500C0004);
      pciDLP(j|0x60000000);
    }
    pciDLP(0x0);	// Ack message
    pciDLP(0x0);	// Ack message
    if (MA.getState("/T")) {
      pciDLP(0x0);	// Ack message
      pciNOOP(0,2);
      pciSKIP();
    }
    pciTLPr(seq++,REG_CODE,1);
    pcieConfig = true;
    pciTLPw(seq++,REG_PBAR,1,0xFFFFFFFF);
    pciTLPr(seq++,REG_PBAR,1);
    pciTLPw(seq++,REG_PBAR,1,pcieBAR);
    pcieConfig = false;
    if (MA.getState("/C")) {
      pciDLP(0x123);
      pciNOOP(-1,8);
      pciTLPw(seq++,REG_OMB1,1,0x44444444,0x6);
      pciNOOP(-1,8);
      pciTLPw(seq++,REG_OMB2,1,0x55555555);
      pciNOOP(-1,8);
      pciDLP(0x123);
      pciDLP(0x456);
      pciTLPr(seq++,REG_DVID,1);
      pciTLPr(seq++,REG_OMB1,1);
      pciDLP(0xABC);
      pciTLPw(seq++,REG_IOC,1,0x30000000);
      pciTLPr(seq++,REG_IOC,1);
      pciTLPr(seq++,REG_IOC,1);
      pciTLPr(seq++,REG_IOC,1);
      pciTLPr(seq++,REG_IOC,1);
      pciNOOP(-1,8);
      pciDLP(0x80120345);
      pciDLP(0x90120345);
      pciDLP(0xA0120345);
    }
    if (MA.getState("/R")) {
      int rlen = MA.getL("/RLEN",16);
      pciTLPc(seq++,rlen,0);
      pciTLPr(seq++,REG_MCSR,1,0);
      pciTLPc(seq++,rlen,1);
      pciTLPr(seq++,REG_MCSR,1);
      pciTLPr(seq++,REG_LCSR,1);
      pciTLPw(seq++,REG_LCSR,1,0x00000008);
      pciTLPr(seq++,REG_DCSR,1);
      pciTLPr(seq++,REG_DCSR,1);
      pciTLPw(seq++,REG_DCSR,1,0xFFFFFFFF);
      pciTLPr(seq++,REG_DCSR,1);
    }
    if (MA.getState("/W")) {
      int mdar = MA.getL("/MDAR",0xDC00FF00);
      int mdau = MA.getL("/MDAU",0x00000032);
      int mdtc = MA.getL("/MDTC",0x003C0);
      int pay  = MA.getL("/PAYLOAD",256);
      int mcsr = MA.getL("/MCSR",0x000000B3);
      if (pay==256) mcsr |= 0x1000;
      if (pay==512) mcsr |= 0x2000;
      if (pay==1024) mcsr |= 0x3000;
      if (MA.getState("/DAC")) mcsr |= 0x0800;
      if (MA.getState("/READ")) mcsr &= ~0x0020;
      int ccm = MA.getL("/CCM",0);
      pciTLPw(seq++,REG_MDAR,1,mdar);
      pciTLPw(seq++,REG_MDAU,1,mdau);
      pciTLPw(seq++,REG_MDTC,1,mdtc);
      pciTLPw(seq++,REG_MCSR,1,mcsr);
      while (ccm-- >0) pciTLPc(seq++,16);
      pciNOOP(0,lanes*15);
      pciTLPr(seq++,REG_MCSR,1);
      pciNOOP(0,lanes*10);
      pciTLPr(seq++,REG_MCSR,1);
      pciNOOP(0,lanes*30);
      //pciSKIP();
      pciNOOP(0,mdtc/lanes);
      pciTLPr(seq++,REG_MCSR,1);
    }
    pciNOOP(0,1500);
  }

  private void pciDLP (int data) {
    short[] sbuf = new short[8];
    int crc = 0xDCBA;
    pciKsym(0x5C);
    pciData(data,4);
    pciData(crc,2);
    pciKsym(0xFD);
    pciExpect((crc<<16)|0xFD5C);
    pciExpect(data);
  }

  private void pciTLPw (int seq, int addr, int ndata, int data) {
    pciTLPw(seq,addr,ndata,data,0x0F);
  }

  private void pciTLPw (int seq, int addr, int ndata, int data, int mask) {
    short[] sbuf = new short[24+ndata*4];
    int head[] = {0x40000000+ndata,0x00601300|mask,pcieBAR+addr};
    int crc = 0xDCBADCBA;
    if (pcieConfig) head[0] |= 0x04000000;
    if (pcieConfig) head[2] = 0x05000000+addr;
    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    for (int i=0; i<ndata; i++) pciDataLE(data,4);
    pciData(crc,4);
    pciKsym(0xFD);
    pciExpect((seq<<16)|0xFDFB);
    for (int i=0; i<3; i++) pciExpect(head[i]);
    for (int i=0; i<ndata; i++) pciExpect(data);
    pciExpect(crc);
    if (ndata%2==0) pciExpect(0);
    if (ndata%2==0) pciNOOP(-1,1);
  }

  private void pciTLPr (int seq, int addr, int ndata) {
    pciTLPr(seq,addr,ndata,1);
  }
  private void pciTLPr (int seq, int addr, int ndata, int pad) {
    short[] sbuf = new short[24+ndata*4];
    int head[] = {0x00000000+ndata,0x0060100F,0xD8400000+addr};
    int crc = 0xDCBADCBA;
    if (pcieConfig) head[0] |= 0x04000000;
    if (pcieConfig) head[2] = 0x05000000+addr;
    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    pciData(crc,4);
    pciKsym(0xFD);
    pciExpect((seq<<16)|0xFDFB);
    for (int i=0; i<3; i++) pciExpect(head[i]);
    pciExpect(crc);
    if (ndata%2==0) pciExpect(0);
    pciNOOP(-1,pad);
  }

  private void pciTLPc (int seq, int ndata) {
    pciTLPc(seq,ndata,1);
  }
  private void pciTLPc (int seq, int ndata, int pad) {
    int head[] = {0x4A000000+ndata,0x00600000|(ndata<<2),0x01230000};
    int crc = 0xDCBADCBA;
    pciKsym(0xFB);
    pciData(seq,2);
    pciData(head[0],4);
    pciData(head[1],4);
    pciData(head[2],4);
    for (int i=0; i<ndata; i++) pciDataLE(0x01CE0000|i,4);
    pciData(crc,4);
    pciKsym(0xFD);
    pciNOOP(-1,pad);
  }

  private void pciExpect (int data) {
    pciexp[pcires++] = data;
  }

  private void pciExpectedResults() {
    MT.writeln("PCIe expected results: ");
    for (int i=0; i<pcires; i+=2)
    MT.writeln("WLoad "+Convert.l2x(pciexp[i+1])+" "+Convert.l2x(pciexp[i]));
  }

  private void pciSKIP () {
    for (int i=0; i<lanes; i++) pciKsym(0x1BC);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
    for (int i=0; i<lanes; i++) pciKsym(0x11C);
  }

  private void pciTSEQ (int mode) {
    for (int i=0; i<lanes; i++) pciKsym(0x1BC);
    for (int i=0; i<lanes; i++) pciKsym(0x1F7);
    for (int i=0; i<lanes; i++) pciData(0x06,1);
    for (int i=0; i<lanes; i++) pciData(0xFF,1);
    for (int i=0; i<lanes; i++) pciData(0x02,1);
    for (int i=0; i<lanes; i++) pciData(0x00,1);
    for (int j=0; j<10; j++)
      for (int i=0; i<lanes; i++) 
        pciData( (mode==2)? 0x45:0x4A, 1);
  }

  private void pciNOOP (int data, int n) {
    byte[] buf = new byte[n*2*4];
    for (int i=0; i<buf.length; ) { buf[i++]=(byte)data; buf[i++]=0; }
    pciBuffer(buf,buf.length);
  }

  private void pciKsym (int data) {
    byte[] buf = new byte[2];
    buf[0] = (byte)(data&0xFF);
    buf[1] = 1;
    pciBuffer(buf,2);
  }

  private void pciDataLE (int data, int n) {
    byte[] buf = new byte[8];
    for (int i=0; i<n; i++) {
      buf[i*2] = (byte)((data>>(8*i))&0xFF);
      buf[i*2+1] = 0;
    }
    pciBuffer(buf,n*2);
  }

  private void pciData (int data, int n) {
    byte[] buf = new byte[8];
    for (int i=0; i<n; i++) {
      buf[i*2] = (byte)((data>>(8*(n-i-1)))&0xFF);
      buf[i*2+1] = 0;
    }
    pciBuffer(buf,n*2);
  }

  private void pciBuffer (byte[] buf, int len) {
    for (int i=0; i<len; ) {
      pcibuf[pcilen++] = buf[i++];
      pcibuf[pcilen++] = buf[i++];
      if (++pcioff == lanes) {
        for (int j=0; j<(8-lanes)*2; j++) pcibuf[pcilen++]=0;
        pcioff=0;
      }
    }
  }

  private int getLaneMask() {
    if (lanes==8) return 3;
    if (lanes==4) return 2;
    if (lanes==2) return 1;
    if (lanes==1) return 0;
    M.error("Illegal lane count "+lanes);
    return -1;
  }

  private int BUSR=1, BUSW=2, MEMR=3, MEMW=4, MEMRWM=10;

  private void ppcgen() {

    ppc_func (BUSW, 0x10000008, 0x11223344, 4, 0);
    for (int i=0; i<16; i++) core.sysrd(0x0); 
    ppc_func (BUSR, 0x10000008, 0x11223344, 0, 4);
    for (int i=0; i<16; i++) core.sysrd(0x0); 
    ppc_rmsg (8);

    ppc_func (MEMW, 0xFD001108, 0x12345678, 12, 0);
    for (int i=0; i<16; i++) core.sysrd(0x0); 
    ppc_func (MEMR, 0xFD001108, 0x0, 0, 12);
    for (int i=0; i<16; i++) core.sysrd(0x0); 
    ppc_rmsg (8);

    int sz = 32;
    ppc_func (MEMW, 0x10001234, 0xFEFE0000, sz, 0);
    for (int i=0; i<32; i++) core.sysrd(0x0); 
    ppc_func (MEMR, 0x10001234, 0x0, 0, sz);
    for (int i=0; i<32; i++) core.sysrd(0x0); 
    ppc_rmsg (24);

  }

  private int ppc_func (int func, int addr, int data, int nwrite, int nread) {
    M.info("func="+func+" addr="+Convert.l2x(addr)+
                        " data="+Convert.l2x(data)+" nw="+nwrite+" nr="+nread);
    ppcad(0);
    ppcwr(0x01010101);		// key
    ppcwr(0x123);		// node
    ppcwr((nread<<16)|func);	// head0
    ppcwr(addr);		// head1
    for (int i=0; i<nwrite; i+=4) ppcwr(data++);		// data
    ppcad(0x0);
    ppcwr(0x55000000|nwrite);		// key
    return data;
  }

  private void ppc_rmsg (int nread) {
    ppcad(0);
    for (int i=0; i<nread; i++) {
      int data = ppcrd();
      M.info("rMSG "+i+" "+Convert.l2x(data));
    }
  }

  private void ppcad (int addr) {
    core.syswr(0x4,addr); 
    core.syswr(0x4,addr>>8); 
  }

  private void ppcwr (int data) {
    for (int i=0; i<4; i++) {
      core.syswr(0x0,data&0xff); 
      data >>= 8;
    }
  }
 
  private int ppcrd () {
    int data,word=0;
    for (int i=0; i<4; i++) {
      data = core.sysrd(0x0); 
      word |= (data&0xFF)<<(i*8);
    }
    return word;
  }
 
  private void ramgen() {
    int n = MA.getL("/BLOCKS",10);
    int sysreg = 0x00220001;
    core.syswrn(0x0,sysreg,32); 
    for (int i=0; i<n; i++) {
      int aw = 0x00123400 + i*64;
      int ar = 0x42789A00 + i*64;
      core.syswrn(0x1,aw,4); 
      core.syswrn(0x1,ar,4); 
    }
    core.syswrn(0x0,sysreg,256); 
  }

  private void sendMsg (String test) {
    int addr=0x000001000;
    int kaddr=addr;
    int key=0x55000000;
    core.syswr(addr++,0x00000000); // key
    core.syswr(addr++,0x00000001); // node
    if (test.equals("RB")) {
      core.syswr(addr++,0x00080001); 
      core.syswr(addr++,0x04001234); 
    }
    else if (test.equals("WB")) {
      core.syswr(addr++,0x00100002); 
      core.syswr(addr++,0x04000400); 
      core.syswr(addr++,0x00001234); 
      core.syswr(addr++,0x00012345); 
      core.syswr(addr++,0x00123456); 
      core.syswr(addr++,0x01234567); 
    }
    else if (test.equals("RM")) {
      core.syswr(addr++,0x00080003); 
      core.syswr(addr++,0xFD000ff8); 
    }
    else if (test.equals("WM")) {
      core.syswr(addr++,0x00080004); 
      core.syswr(addr++,0x00000400); 
      core.syswr(addr++,0x00001234); 
      core.syswr(addr++,0x00005678); 
    }
    else if (test.equals("RIO")) {
      core.syswr(addr++,0x00200003); 
      core.syswr(addr++,0x01101200); 
    }
    else if (test.equals("RIOX")) {
      core.syswr(addr++,0x00800003); 
      core.syswr(addr++,0x01101200); 
    }
    else if (test.equals("WIO")) {
      core.syswr(addr++,0x00400004); 
      core.syswr(addr++,0x01101200); 
      for (int i=0; i<16; i++)
      core.syswr(addr++,0x000F0000+i); 
    }
    else if (test.equals("WIOX")) {
      core.syswr(addr++,0x00800004); 
      core.syswr(addr++,0x01101200); 
      for (int i=0; i<32; i++)
      core.syswr(addr++,0x000F0000+i); 
    }
    else if (test.equals("RW")) {
      core.syswr(addr++,0x0014000E); 
      core.syswr(addr++,0xFD001234); 
      core.syswr(addr++,0x1000AAAA); 
      core.syswr(addr++,0xBB101010); 
    }
    else if (test.equals("IA")) {
      core.syswr(addr++,0x00000006); 
      core.syswr(addr++,0x01101200); 
    }
    else { key |= 4;
      core.syswr(addr++,0x00040002); 
      core.syswr(addr++,0xFD001234); 
      core.syswr(addr++,0x12345678); 
    }
    core.syswr(kaddr,key); 
    // wait for message response
    while ( (core.sysrd(kaddr)>>>24) != 0xAA);

  }
}
