/*
   Non-Midas acquisition routine for ICE card on the network

   Author: Jeff Schoen

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "icedefs.h"
#include "icelib.h"
#include "packet.h"

#include "../inc/netlib.h"
#include "../lib/netlib.c"

main (int argc, char *argv[])
{
  char *func,*addr,*flags, *fname, naddr[80];
  int_4 i,size,loop,sock,status;

  if (argc < 6) {
    BAIL:
    printf("Syntax: testnio <func> <addr> <size> <flags>\n");
    printf("          <func>    = function = ACQ|CHK          local IF             MCast addr     \n");
    printf("          <addr>    = network address (ex. udp:192.168.201.105:7777/224.1.40.%d:7777) \n");
    printf("          <size>    = length in bytes (ex. 1e6 is roughly 1000 SDDS packets) \n");
    printf("          <loop>    = loop count (index can be %d in the <addr> string) (ex. 10) \n");
    printf("          <flags>   = packet type = ICE|SDDS\n");
    printf("          <file>    = output file name <optional>\n");
    printf("\n");
    exit(0);
  }

  func = argv[1]; 
  addr = argv[2]; 
  size = atoi(argv[3]); 
  loop = atoi(argv[4]); 
  flags = argv[5];
  fname = argv[6];

  status = 0;

  if (strcmp(func,"ACQ")==0) {
    for (i=0; i<loop; i++) {
      sprintf(naddr,addr,i);
      status += acquire (naddr,size,flags,fname);
    }
  }    
  else { 
    printf("Function=%s not coded yet\n",func); 
  }
  if (loop>1) printf("Total errors=%d\n",status);

  exit (status);
}

#define PKTSIZE (8192+64)

int_4 acquire (char *addr, int_4 size, char *flags, char *fname) {

  int_4 sock,xfer,total,seq,seqNext=-1,errs=0,sdds,status;
  ICEPACKETSTRUCT *ipkt;
  SDDSPACKETSTRUCT *spkt;
  unsigned char *ucdata;
  double tdb,t1,t2;
  char *pdata;
  FILE *fp = NULL;

  sock = nioOpen (addr,0,-1);
  if (sock<=0) { printf("Err opening socket=%s status=%d\n",addr,status); return -1; }

  if (fname != NULL) fp = fopen(fname,"w");

  printf("Transfer addr=%s ",addr);

  xfer  = PKTSIZE;
  pdata = nioAlloc(xfer);
  ipkt  = (ICEPACKETSTRUCT*)pdata;
  spkt  = (SDDSPACKETSTRUCT*)(pdata-8);  // Not an ICE/SDDS packet
  ucdata = (unsigned char *)&spkt->frame;
  sdds  = strcmp(flags,"SDDS")==0;

  t1 = pic_time();

  for (total=0; total<size; ) {

    status = nioRecv (sock, 0, pdata,xfer, 0);

    if (status==0)
      usleep(10000);
    else if (status<0) { 
      printf("nioRecv error\n");
      break;
    }
    else if (sdds) {
      seq = 256*ucdata[0] + ucdata[1];
      if (seqNext<0);
      else if (seq==seqNext);
      else printf("\n Seq errs=%d actual=%d expect=%d",errs++,seq,seqNext);
      seqNext = seq+1;
      if ((seqNext&0x1F)==0x1F) seqNext++;
      seqNext &= 0xFFFF;
      total += (status-56);
    }
    else {
      seq = ipkt->count;
      if (seqNext<0);
      else if (seq==seqNext);
      else printf("\n Seq errs=%d actual=%d expect=%d",errs++,seq,seqNext);
      seqNext = seq+1;
      total += (status-64);
    }
    if (fp!=NULL && status>0) fwrite(pdata,1,status,fp);
  }

  t2 = pic_time();

  nioFree(pdata);
  nioClose(sock);
  if (fp!=NULL) fclose(fp);

  tdb = 1.e-6*total; t2-=t1;
  if (errs>0)  printf("\nTransfer addr=%s ",addr);
  printf(" size=%lfMby elapse=%lfs rate=%lfMby/s errs=%d\n",tdb,t2,tdb/t2,errs);

  return errs;
}
